# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""The imgupload manage extension."""

import webob
import six

from cinder.api import extensions
from cinder.api.openstack import wsgi

from powervc_cinder.scohack import scohack_location as location
from powervc_cinder.scohack import scohack_utils as utils

from oslo_utils import excutils

from cinder import rpc

from powervc_cinder.k2aclient import exceptions as k2aexc

import logging
from powervc_cinder import _
import eventlet
import datetime

LOG = logging.getLogger(__name__)

from webob.exc import (HTTPBadRequest)


authorize = extensions.extension_authorizer('volume', 'scohack_image_upload')


def _get_store_from_uri(context, uri, loc=None):
    """
    Given a URI, return the store object that would handle
    operations on the URI.

    :param uri: URI to analyze
    """
    scheme = uri[0:uri.find('/') - 1]
    store = _get_store_from_scheme(context, scheme, loc)
    return store


def _get_store_from_scheme(context, scheme, loc=None):
    """
    Given a scheme, return the appropriate store object
    for handling that scheme.
    """
    if scheme not in location.SCHEME_TO_CLS_MAP:
        msg = (_("scohack: unknownScheme: scheme: >%(scm)s<") %
               dict(scm=scheme))
        raise Exception(msg)
    scheme_info = location.SCHEME_TO_CLS_MAP[scheme]
    store = scheme_info['store_class'](context, loc)
    return store


def _get_from_backend(context, uri, **kwargs):
    """Yields chunks of data from backend specified by uri"""

    loc = location.get_location_from_uri(uri)
    store = _get_store_from_uri(context, uri, loc)
    return store.get(loc)


def _getReader(image_meta, image_data):
    if image_meta.get('copy_from'):
        return utils.CooperativeReader(image_data)
    else:
        return utils.ChunkReader(image_data)


class ScohackImageUploadController(wsgi.Controller):
    """The SCO image upload API controller for the OpenStack API."""

    def __init__(self):
        self.pool = eventlet.GreenPool(size=1024)

    @staticmethod
    def _notify_imgupload_error(context, method, payload):
        anotifier = rpc.get_notifier(publisher_id='imgupload')
        anotifier.error(context, method, payload)

    @staticmethod
    def _get_from_store(context, where):
        image_data, image_size = _get_from_backend(context, where)
        image_size = int(image_size) if image_size else None
        return image_data, image_size

    def create(self, req):
        """
        Adds a new image to K2. Two scenarios exist when creating an
        image:

        1. If the image data is available directly for upload, create can be
           passed the image data as the request body and the metadata as the
           request headers.

        2. If the image data exists somewhere else, you can upload indirectly
           from the external source using the x-imgupload-copy-from header.
           The request body *must* be encoded as application/octet-stream,
           otherwise an HTTPBadRequest is returned.

        Upon a successful save of the image data and metadata, a response
        containing metadata about the image is returned.

        :param req: The WSGI/Webob Request object

        :raises HTTPBadRequest if x-imgupload-copy-from is missing
                and the request body is not application/octet-stream
                image data.
        """

        try:

            # parse metadata
            image_meta = self._parse_meta(req)

            context = req.environ['cinder.context']

            copy_from = image_meta.get('copy_from')
            if copy_from:
                image_data, image_size = self._get_from_store(context,
                                                              copy_from)

                msg = (_("scohack: from copy_from http server content-length, "
                         "size: >%(size)s<")
                       % dict(size=image_size))
                LOG.info(msg)

                if not image_size > 0:
                    msg = _("scohack: content-length not specified by "
                            "http server.")
                    LOG.warn(msg)

                image_meta['size'] = image_size
            else:
                image_data = req.body_file

            # get back end store
            store = _get_store_from_scheme(context, 'scohack')

            context = req.environ['cinder.context']

            store.add(_getReader(image_meta, image_data),
                      image_meta,
                      context)
        except webob.exc.HTTPError:
            msg = _("scohack: HTTP error while uploading file")
            LOG.debug(msg)
            notifier_err = dict(image_meta=image_meta, error_message=msg)
            self._notify_imgupload_error(context,
                                         'imageupload.create',
                                         notifier_err)
            with excutils.save_and_reraise_exception():
                LOG.exception(msg)

        except (ValueError) as e:
            msg = _("scohack: upload failed: ValueError")
            notifier_err = dict(image_meta=image_meta, error_message=msg)
            self._notify_imgupload_error(context,
                                         'imageupload.create',
                                         notifier_err)
            LOG.exception(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg),
                                           content_type="text/plain",
                                           request=req)

        except k2aexc.K2aK2Error as e:
            msg = _("scohack: K2aK2Error error while uploading file")
            LOG.debug(msg)
            notifier_err = dict(image_meta=image_meta, error_message=msg)
            self._notify_imgupload_error(context,
                                         'imageupload.create',
                                         notifier_err)
            if e.status in webob.exc.status_map:
                k2msg = six.text_type(e.k2msg)
                raise webob.exc.status_map[e.status](explanation=k2msg,
                                                     content_type="text/plain",
                                                     request=req)
            else:
                ct = 'text/plain'
                emsg = six.text_type(e.msg)
                raise webob.exc.HTTPInternalServerError(explanation=emsg,
                                                        request=req,
                                                        content_type=ct)

        except Exception as e:
            scohack_httpcode = 500
            if hasattr(e, 'scohack_httpcode'):
                scohack_httpcode = e.scohack_httpcode
            msg = (_("scohack: upload failed: exception: "
                     ">%(ex)s<, code: >%(code)d<") %
                   dict(ex=e, code=scohack_httpcode))
            notifier_err = dict(image_meta=image_meta, error_message=msg)
            self._notify_imgupload_error(context,
                                         'imageupload.create',
                                         notifier_err)
            LOG.exception(msg)
            x = "text/plain"
            emsg = six.text_type(msg)
            raise webob.exc.status_map[scohack_httpcode](explanation=emsg,
                                                         content_type=x,
                                                         request=req)

        response = self._to_json_response(image_meta)

        notifier_info = dict(image_meta=image_meta, response=response)
        anotifier = rpc.get_notifier(publisher_id='imgupload')
        anotifier.info(context, 'imgupload.create', notifier_info)
        return response

    def _to_json_response(self, image_meta):
        x = {'cinder_volume_id': image_meta.get('volume_id'),
             'cinder_volume_name': image_meta.get('volume_name'),
             'vios_id': image_meta.get('vios_id'),
             'hmc_id': image_meta.get('hmc_id'),
             'size': image_meta.get('size'),
             'checksum': image_meta.get('checksum', None),
             'created_at': str(datetime.datetime.now()), }
        return {'imgupload': x}

    def _parse_meta(self, req):

        # all comparisons lower case
        lheaders = {}
        for k, v in req.headers.iteritems():
            lheaders[k.lower()] = v

        content_type = lheaders.get('content-type')
        if not content_type or content_type != 'application/octet-stream':
            msg = _("scohack: Content-Type must be application/octet-stream")
            LOG.debug(msg)
            raise HTTPBadRequest(explanation=msg)

        x = {"copy_from": lheaders.get('x-imgupload-copy-from'),
             "ms_id": lheaders.get("x-imgupload-ms-id"),
             "vios_id": lheaders.get("x-imgupload-vios-id"),
             "hmc_id": lheaders.get("x-imgupload-hmc-id"),
             "volume_id": lheaders.get("x-imgupload-cinder-volume-id"),
             "checksum": lheaders.get("x-imgupload-file-checksum"),
             "scg_vios_id": lheaders.get("x-imgupload-scg-vios-id"),
             "size": lheaders.get("x-imgupload-file-size", "0"), }

        msg = (_("scohack: request summary: "
                 "hmc_id: >%(hmc_id)s<, "
                 "volume_id: >%(volume_id)s<, "
                 "copy_from: >%(copy_from)s<, "
                 "size: >%(size)s<, "
                 "checksum: >%(checksum)s<,"
                 "scg_vios_id: >%(scg_vios_id)s<, "
                 "ms_id: >%(ms_id)s<, "
                 "vios_id: >%(vios_id)s< ") %
               dict(hmc_id=x.get('hmc_id'),
                    volume_id=x.get('volume_id'),
                    copy_from=x.get('copy_from'),
                    size=x.get('size'),
                    checksum=x.get('checksum'),
                    scg_vios_id=x.get('scg_vios_id'),
                    ms_id=x.get('ms_id'),
                    vios_id=x.get('vios_id')))
        LOG.info(msg)

        if x.get('size') != "0" and x.get('copy_from'):
            msg = _("scohack: copy_from set, x-imgupload-file-size header "
                    "specified on request will be ignored and, instead, "
                    "will be obtained from copy_from")
            LOG.info(msg)

        if not x.get('copy_from') and x.get('size') == "0":
            msg = _("scohack: x-imgupload-file-size not specified.")
            LOG.warn(msg)

        if not x.get('copy_from'):
            transfer_encoding = lheaders.get('transfer-encoding')
            if transfer_encoding is None or transfer_encoding != "chunked":
                msg = _("scohack: when transmitting a file via the "
                        "request body the transfer-encoding header "
                        "must be set to chunked")
                LOG.debug(msg)
                raise HTTPBadRequest(explanation=msg)

        if not x.get('checksum'):
            msg = _("scohack: x-imgupload-file-checksum not specified.")
            LOG.warn(msg)

        if not x.get('volume_id'):
            msg = _("scohack: volume_id must be specified")
            LOG.debug(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg),
                                           content_type="text/plain",
                                           request=req)

        if not x.get('hmc_id'):
            msg = _("scohack: hmc_id must be specified")
            LOG.debug(msg)
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg),
                                           content_type="text/plain",
                                           request=req)

        x['size'] = int(x.get('size'))

        return x


class Scohack_image_upload(extensions.ExtensionDescriptor):
    """Scohack image upload support."""

    name = "ScohackImageUpload"
    alias = "scohack"
    namespace = "http://www.ibm.com/openstack/cinder/contrib/scohack/v1.0"
    updated = "2013-09-24T00:00:00+00:00"

    def get_resources(self):
        resources = []

        ca = {'create': 'POST'}
        res = extensions.ResourceExtension('scohack',
                                           ScohackImageUploadController(),
                                           collection_actions=ca)
        resources.append(res)

        ca = {'create': 'POST'}
        res = extensions.ResourceExtension('imgupload',
                                           ScohackImageUploadController(),
                                           collection_actions=ca)
        resources.append(res)

        return resources
