# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from datetime import datetime
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, V999

CINDER_CONF = '/etc/cinder/cinder.conf'
FABRIC_CONF = '/etc/cinder/fabrics.conf'
PASTE_CONF = '/etc/cinder/api-paste.ini'
POWERVC_MANAGE = '/opt/ibm/powervc/bin/powervc-manage'


######################################################
########  Cinder Data Manager Definition  ############
######################################################
class CinderDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'cinder'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db2', 'rabbitmq', 'httpd', 'keystone', 'glance']

    ###############################################
    #####  Data Manager - Upgrade Interface  ######
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        super(CinderDataManager, self).upgrade_conf_files(version)
        #If this was ever 1.2.1 we need to upgrade the fabrics.conf file
        self.__upgrade_fabrics_conf()

    ###############################################
    ######  Data Manager - Backup Interface  ######
    ###############################################
    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config = utils.get_config_data()
        #Generate a unique time-stamp to put in the existing cinder.conf files
        timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
        config['cinder_restore_timestamp'] = timestamp
        #Loop through each CONF file adding a time-stamp to determine changes
        for host in self.__get_storage_providers():
            conf_file = '/etc/cinder/cinder-%s.conf' % host
            utils.set_openstack_config_value(
                conf_file, 'restore_timestamp', timestamp)

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""
        config = utils.get_config_data()
        #Loop through each CONF file seeing if the host is new or old
        for host in self.__get_storage_providers():
            conf_file = '/etc/cinder/cinder-%s.conf' % host
            val = utils.get_openstack_config_value(
                conf_file, 'restore_timestamp')
            #If file came from the restore, we need to verify service exists
            if val is None or val != config['cinder_restore_timestamp']:
                self.__update_new_provider(host)
            #If file is left from before the restore, need to delete it
            else:
                self.__remove_old_provider(host)

    ###############################################
    ##  Data Manager - Subclass Implementation  ###
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'cinder'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        services = ['api', 'scheduler']
        services.extend(self.__get_volume_services())
        return ['openstack-cinder-%s' % svc for svc in services]

    def _get_disabled_services(self):
        """Provides a list of the Disabled Services for the Component"""
        services = ['backup', 'volume']
        return ['openstack-cinder-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['api', 'scheduler', 'cinder-manage',
                'httpd-error', 'powervc-dbsync']
        return ['/var/log/cinder/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('CINDER', 'cinder')

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/cinder-manage db sync',
                '/opt/ibm/powervc/bin/powervc-dbsync cinder']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {CINDER_CONF: dict()}
        #If this is the Mgmt System, also need to update the Cinder Conf Files
        for provider in self.__get_storage_providers():
            config['/etc/cinder/cinder-%s.conf' % provider] = dict()
        #Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._added_auth_conf_properties(config[key])
            self._added_messaging_conf_properties(config[key])
            self.__added_cinder_mgmtsystem_conf_properties(config[key])
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {CINDER_CONF: dict()}
        #If this is the Mgmt System, also need to update the Cinder Conf Files
        for provider in self.__get_storage_providers():
            config['/etc/cinder/cinder-%s.conf' % provider] = dict()
        #Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._deleted_auth_conf_properties(config[key])
            self._deleted_messaging_conf_properties(config[key])
            config[key]['DEFAULT']['enabled_apis'] = V122
            config[key]['DEFAULT']['sql_connection'] = V123
            config[key]['DEFAULT']['lock_path'] = V123
        return config

    def _get_audit_conf_properties(self):
        """Provides a tuple of configuration properties for auditing"""
        sects = ['composite:openstack_volume_api_v1',
                 'composite:openstack_volume_api_v2']
        return (PASTE_CONF, sects, ['keystone', 'keystone_nolimit'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/cinder']

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return ['/etc/cinder/volumes']

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('cinder', CINDER_CONF,
                 'keystone_authtoken', 'admin_password')]

    def _temporary_openstack_patches(self):
        """Perform any temporary openstack patching for the Component"""
        pass

    ###############################################
    #### Internal Data Manager Helper Methods  ####
    ###############################################
    def __get_storage_providers(self):
        """Helper method to get a list of registered storage providers"""
        files = utils.get_matching_files('/etc/cinder', 'cinder-', '.conf')
        return [f[7:-5] for f in files]

    def __get_volume_services(self):
        """Helper method to get a list of cinder-volume-xxx services"""
        files1 = utils.get_matching_files(
            '/etc/init.d', 'openstack-cinder-volume-', '')
        files2 = utils.get_matching_files(
            '/usr/lib/systemd/system', 'openstack-cinder-volume-', '')
        if files1:
            return [f[17:] for f in files1]
        return [f[17:-8] for f in files2]

    def __added_cinder_mgmtsystem_conf_properties(self, config):
        """Add in the Management System cinder.conf Configuration Properties"""
        config.setdefault('database', dict())
        #Add in the properties for the Database Section
        config['database'].update(dict(
            pool_timeout=('60', V123),
            connection=('aes:ibm_db_sa://cinder:%(cinder_'
                        'password)s@%(local_ip)s:50110/cinder', V123)))
        #Add in the properties for the Oslo Currency Section
        config['oslo_concurrency'] = \
            dict(lock_path=('/var/lib/cinder/tmp', V123))
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            osapi_volume_listen_port=('9000', V120),
            default_log_levels=('urllib3=WARN', V120),
            strict_ssh_host_key_policy=('True', V122),
            ibmpowervc_policy_granular=('True', V120),
            osapi_volume_listen=('%(local_ip)s', V121),
            osapi_max_request_body_size=('1099511627776', V120),
            api_paste_config=('/etc/cinder/api-paste.ini', V120),
            ssh_hosts_key_file=('/opt/ibm/powervc/data/known_hosts', V122),
            service_down_time=('120', V122), osapi_max_limit=('8192', V123),
            glance_host=('%(local_ip)s', V122), report_interval=('30', V122),
            osapi_volume_workers=('4', V122), quota_volumes=('100000', V121),
            quota_gigabytes=('100000', V121), quota_snapshots=('100000', V121),
            volume_group=('cinder-volumes', V120), monkey_patch=('True', V121),
            volume_name_template=('<volume-%s>', V120), verbose=('True', V120),
            scheduler_manager=('powervc_cinder.scheduler.manager.'
                               'PowerVCSchedulerManager', V120),
            scheduler_driver=('powervc_cinder.scheduler.sched_drv.'
                              'PowerVCSchedulerDriver', V120),
            scheduler_host_manager=('powervc_cinder.scheduler.host_'
                                    'manager.PowerVCHostManager', V120),
            osapi_volume_extension=(
                'powervc_cinder.api.plugins.api_extensions_v2', V121),
            volume_manager=(
                'powervc_cinder.volume.manager.PowerVCVolumeManager', V120),
            monkey_patch_modules=('cinder.volume.volume_types:powervc_cinder.'
                                  'volume.volume_type_decorator,cinder.volume.'
                                  'api:powervc_cinder.volume.volume_create_'
                                  'decorator', V121)))
        #Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_cinder_powervm_conf_properties(config)
        else:
            self.__added_cinder_powerkvm_conf_properties(config)

    def __added_cinder_powervm_conf_properties(self, config):
        """Add in the PowerVM-specific cinder.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            supported_host_types=(
                'svc,ssp,xiv,emc_vmax,emc_vnx', V123),
            storwize_svc_connection_protocol=('fc', V122),
            storwize_svc_multihostmap_enabled=('True', V123)))

    def __added_cinder_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific cinder.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            supported_host_types=('svc', V122),
            storwize_svc_connection_protocol=('iscsi', V122),
            storwize_svc_multihostmap_enabled=('True', V120)))

    def __update_new_provider(self, provider):
        """Helper method to make sure the service exists for the provider"""
        service = 'openstack-cinder-volume-%s' % provider
        conf_file = '/etc/cinder/cinder-%s.conf' % provider
        #We only want to create the service if this is a PowerVM Mgmt System
        if not self.__service_file_exists(provider):
            utils.move_file(conf_file, conf_file + '_backup')
            utils.run_command(POWERVC_MANAGE + ' -o addhost -s ' +
                              service + ' -f ' + conf_file + '_backup')
            utils.remove_file(conf_file + '_backup')

    def __remove_old_provider(self, provider):
        """Helper method to make sure service is removed for old provider"""
        service = 'openstack-cinder-volume-%s' % provider
        conf_file = '/etc/cinder/cinder-%s.conf' % provider
        #We only want to delete the service if this is a PowerVM Mgmt System
        if self.__service_file_exists(provider):
            utils.run_command(POWERVC_MANAGE + ' -o removehost -s ' + service)
        #Make sure the CONF file is gone even it the svc-manage didn't
        utils.remove_file(conf_file)

    def __service_file_exists(self, provider):
        """Helper method to see if the service file exists for the provider"""
        svc = 'openstack-cinder-volume-%s' % provider
        svc1 = '/etc/rc.d/init.d/%s' % svc
        svc2 = '/usr/lib/systemd/system/%s.service' % svc
        return utils.file_exists(svc1) or utils.file_exists(svc2)

    def __upgrade_fabrics_conf(self):
        """Helper method to upgrade the fabrics.conf file"""
        props = ['address', 'user', 'password', 'uniqueid', 'display_name']
        fabs = utils.get_openstack_config_value(FABRIC_CONF, 'fc_fabric_names')
        #If there aren't any Fabrics set, then nothing to do
        if not fabs:
            return
        #Loop through each of the Fabrics, adding in the new properties
        for fab in fabs.split(','):
            for prop in props:
                newprop = 'fc_fabric_%s' % prop
                oldprop = 'fc_fabric_%s_%s' % (prop, fab)
                oldval = utils.get_openstack_config_value(FABRIC_CONF, oldprop)
                #If the value is set for the property, than add the new one
                if oldval:
                    #If this is the password property we need to encrypt it
                    if prop == 'password':
                        oldval = utils.encrypt_data(oldval)
                    #Now we can set the new property in the fabric's section
                    utils.set_openstack_config_value(
                        FABRIC_CONF, newprop, oldval, fab)
            #Since 1.2.1 only supported Brocade, we will hard-code to brocade
            utils.set_openstack_config_value(
                FABRIC_CONF, 'fc_fabric_type', 'brocade')
            utils.set_openstack_config_value(
                FABRIC_CONF, 'fabric_type', 'brocade', fab)
        #Also need to duplicate the fabric names in fc-zone-manager section
        utils.set_openstack_config_value(
            FABRIC_CONF, 'fc_fabric_names', fabs, 'fc-zone-manager')
