#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import cinder.context
from cinder import exception
from oslo_log import log as logging
from oslo_utils import importutils
from powervc_cinder.db import api as pvc_db
from cinder.volume import api as base_api
from powervc_cinder.volume import rpcapi_product as volume_rpc
import inspect
from powervc_oslo import override_context_serialize
from powervc_cinder import _
import six

LOG = logging.getLogger(__name__)

#Override the Context Serialization to add Locale
override_context_serialize(cinder.context.RequestContext)


class restricted_metadata_keys():
    VDISK_ID_KEY = "vdisk_id"
    VDISK_UID_KEY = "vdisk_uid"
    VDISK_NAME_KEY = "vdisk_name"
    BYTE_SIZE_KEY = "byte_size"


class metada_keys():
    VOLUME_WWN = "volume_wwn"


def volume_create_decorator(name, fn):
    """ Decorator for cinder.volume.api.create(). It will patch the
    default volume create api to bypass the volume type check when
    source_volume's volume type doesn't match the volume type passed
    in. This is the monkey patch to address the volume clone issue
    introduced by community fix: # 1276787
    This patch can be removed once PowerVC switches to the retype and
    volume migration approach to handle the volume type change during
    volume clone.

        :param name: name of the function
        :param function: - object of the function
        :returns: function -- decorated function

    """
    def wrapped_func(*args, **kwarg):
        try:
            if fn.__name__ == 'create':
                LOG = logging.getLogger(fn.__module__)
                try:
                    idx_svol =\
                        inspect.getargspec(fn).args.index("source_volume")
                    idx_voltp =\
                        inspect.getargspec(fn).args.index("volume_type")
                except Exception as e:
                    LOG.error(_("cinder volume_create interface has been "
                                "changed. PowerVC monkey patch for volume "
                                "clone won't work. function: %(fx)s, "
                                "arguments: %(args)s") %
                              dict(fx=fn.__module__,
                                   args=inspect.getargspec(fn).args))
                    raise e

                svol = kwarg.get('source_volume')
                voltp = kwarg.get('volume_type')

                if svol and voltp:
                    if svol['volume_type_id'] != voltp['id']:
                        # this is the condition that will trigger the
                        # volume clone to fail. Patch it here.
                        svol['volume_type_id'] = voltp['id']
                        LOG.info(_("Monkey patched volume clone by powervc "
                                   "volume_create_decorator(). "
                                   "Source Volume ID: %(svol)s, "
                                   "volume type name: %(voltpnm)s, "
                                   "volume type ID: %(voltpid)s") %
                                 dict(svol=svol['id'],
                                      voltpnm=voltp['name'],
                                      voltpid=voltp['id']))
                        return fn(*args, **kwarg)

            return fn(*args, **kwarg)
        # Catching exception VolumeSizeExceedsAvailableQuota.
        # Re-sending exception with with removed false storage quotas(VMAX)
        except exception.VolumeSizeExceedsAvailableQuota as e:
            LOG.error(six.text_type(e))
            msg = (_("Requested volume exceeds "
                     "allowed Gigabytes quota."))
            raise exception.VolumeSizeExceedsAvailableQuota(message=msg)
    return wrapped_func


def volume_type_decorator(name, fn):
    """Decorator for cinder.volume.volume_types.create() which is called
    by utils.monkey_patch(). It will set the default quota for the
    created volume type.

        :param name: name of the function
        :param function: - object of the function
        :returns: function -- decorated function

    """
    def wrapped_func(*args, **kwarg):
        if fn.__name__ == 'create':
            # cinder.volume.volume_types.create() decorator
            r = fn(*args, **kwarg)
            LOG = logging.getLogger(fn.__module__)
            try:
                idx_specs = inspect.getargspec(fn).args.index('extra_specs')
                idx_name = inspect.getargspec(fn).args.index('name')
                idx_ctxt = inspect.getargspec(fn).args.index('context')
            except Exception as e:
                LOG.warn(_("Failed to get the parameters from function "
                           "cinder.volume.volume_types.create(). Default "
                           "quota didn't set for the storage template. "
                           "Error: %(err)s") % dict(err=e))
                # Just return. Don't set the storage template default quota.
                return r

            volume_type = args[idx_name]
            extra_specs = args[idx_specs]
            ctxt = args[idx_ctxt]
            volume_host = None
            if extra_specs and isinstance(extra_specs, dict):
                volume_host = extra_specs.get(
                    "capabilities:volume_backend_name", None)

            if volume_host and volume_type and ctxt:
                volume_rpcapi = volume_rpc.VolumeAPIProduct()
                try:
                    volume_rpcapi.set_volume_type_quota(
                        ctxt, volume_host, volume_type)
                    LOG.info(_("Successfully set default quota for storage "
                               "template %(vol_type)s") %
                             dict(vol_type=volume_type))
                except Exception as e:
                    LOG.warn(_("Failed to set default quota for storage "
                               "template %(vol_type)s, error: %(err)s") %
                             dict(vol_type=volume_type, err=e))
            else:
                LOG.warn(_("Cannot set default quota for storage template "
                           "%(vol_type)s due to invalid Parameters from "
                           "volume type create.")
                         % dict(vol_type=volume_type))
            return r
        else:
            return fn(*args, **kwarg)
    return wrapped_func


def API(*args, **kwargs):
    class_name = 'powervc_cinder.volume.rpcapi_product.VolumeAPIProduct'
    return importutils.import_object(class_name, *args, **kwargs)


def HostAPI(*args, **kwargs):
    class_name = 'powervc_cinder.volume.rpcapi_product.VolumeAPIProduct'
    return importutils.import_object(class_name, *args, **kwargs)


def monkey_patch_get_all():
    """
    Patch the cinder.volume.api.API class get_all method
    in order to implement a new filter.  This new filter,
    'data_volumes' will filter out all boot volumes from the
    response.  This filter will be used for both the cinder
    volumes and volumes/detail API.  Code calls a db method
    which checks for the volume_metadata attribute
    is_boot_volume.
    If the 'data_volumes' filter is not used, the base cinder
    get_all api is called.
    """
    #maintain the original api
    cndr_base_api_get_all = base_api.API.get_all

    base_api.API.powervc_mod_get_all = \
        base_api.API.get_all

    def powervc_mod_get_all(self, context, marker=None,
                            limit=None,
                            sort_keys=['created_at'],
                            sort_dirs=['desc'],
                            filters={},
                            viewable_admin_meta=False):
        # Avoid missing volume_metadata table join() exception if no
        # volumes were ever managed in the past.
        if not hasattr(monkey_patch_get_all, 'first_vol'):
            if "all_tenants" not in filters:
                volumes = cndr_base_api_get_all(self, context, marker,
                                                1, sort_keys, sort_dirs,
                                                {}, viewable_admin_meta)
            else:
                volumes = cndr_base_api_get_all(self, context, marker,
                                                1, sort_keys, sort_dirs,
                                                {"all_tenants":
                                                 filters["all_tenants"]},
                                                viewable_admin_meta)
            if not volumes:
                LOG.info(_("Special case for empty volume table. Return %s")
                         % volumes)
                return volumes
            LOG.debug("There is at least one volume. First=%s" % volumes)
            # set the cached var to avoid the above query again
            monkey_patch_get_all.first_vol = volumes
        # Normal processing
        volumes = []
        if ('data_volumes' in filters):
            try:
                volumes = pvc_db.ibm_volume_get_all_except_bootable(context)
            except exception.NotFound:
                #ignore this exception
                LOG.debug("No data volumes found.")
                pass
            except Exception as ex:
                LOG.error(_("Error getting data volumes: %s ") % ex)
                raise ex
        else:
            volumes = cndr_base_api_get_all(self, context, marker,
                                            limit, sort_keys, sort_dirs,
                                            filters,
                                            viewable_admin_meta)
        return volumes

    base_api.API.get_all = \
        powervc_mod_get_all


def monkey_patch_get_default_volume_type():
    """Override the get_default_volume_type to handle UTF-8 conversion"""
    from oslo_config import cfg
    from cinder import context
    from cinder.volume import volume_types

    def powervc_get_default_volume_type():
        # If the default volume type is set, then look it up based on the name
        if cfg.CONF.default_volume_type is not None:
            try:
                ctxt = context.get_admin_context()
                name = cfg.CONF.default_volume_type.decode('utf-8')
                return volume_types.get_volume_type_by_name(ctxt, name)
            except exception.VolumeTypeNotFoundByName as e:
                LOG.warn(_('Default volume type is not found, please'
                           'check default_volume_type config: %s'), e)
        return dict()
    #Override the get_default_volume_type with our version
    volume_types.get_default_volume_type = powervc_get_default_volume_type

monkey_patch_get_all()
monkey_patch_get_default_volume_type()


def monkey_patch_reserve_volume():
    """ Override reserve_volume to support the attaching state in case
    multiattach is enabled
    """

    def powervc_reserve_volume(self, context, volume):
        volume = self.db.volume_get(context, volume['id'])
        if volume['status'] == 'available':
            self.update(context, volume, {"status": "attaching"})
        elif volume['status'] in ('in-use', 'attaching'):
            if volume['multiattach']:
                self.update(context, volume, {"status": "attaching"})
            else:
                msg = _("Volume must be multiattachable to reserve again.")
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)
        else:
            msg = _("Volume status must be available to reserve.")
            LOG.error(msg)
            raise exception.InvalidVolume(reason=msg)

    base_api.API.reserve_volume = \
        powervc_reserve_volume

monkey_patch_reserve_volume()
