# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import ast
from cinder import context, exception
from cinder.i18n import _, _LE, _LI, _LW
from cinder.openstack.common import loopingcall
from cinder.volume.drivers.emc import emc_vmax_common, emc_vmax_masking, \
    emc_vmax_fast
from cinder.volume.drivers.emc import emc_vmax_utils
from cinder.volume.drivers.emc.emc_vmax_common import CINDER_EMC_CONFIG_FILE, \
    CINDER_EMC_CONFIG_FILE_PREFIX, CINDER_EMC_CONFIG_FILE_POSTFIX, \
    FASTPOLICY
import copy
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import six

from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import restricted_metadata_keys as rmk
from powervc_oslo.wbem import pywbem


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

vmax_cfg_options = [
    # Set default wait period to be longer than the 60 retries by
    # default for the community driver.
    # Note that this default is also set separately by the registration code
    # in the cinder_emc_config_<id>.xml file. So the first two CONF options
    # below only apply when the community driver code is not using the
    # extra_specs from the XML config file.
    cfg.IntOpt('vmax_job_retries',
               default=250,
               help="Maximum number of retries to check for Job "
               "operation to complete. SMI-Provider polling option."
               ),
    cfg.IntOpt('vmax_job_interval',
               default=10,
               help="Delay interval in seconds to wait before the next "
               "attempt. SMI-Provider polling option."
               ),
    cfg.BoolOpt('vmax_allow_manage_inuse',
                default=False,
                help="Allow volumes to be managed that are marked as in-use "
                "by the VMAX provider. VMAX volumes may be marked as in-use "
                "because they participate in a masking group, but not be "
                "mapped to a VM."),
    cfg.BoolOpt('vmax_allow_cached_cim_inst',
                default=True,
                help="Allow volume CIM instances to be cached only during "
                "a particular request. This avoids volume lookup multiple "
                "times during a single request."),
    cfg.BoolOpt('vmax_wait_for_sync',
                default=True,
                help="Controls whether the VMAX cinder driver waits for "
                "replication synchronization to complete before returning "
                "from a clone operation API."),
    cfg.BoolOpt('vmax_restrict_pool_to_xml_conf',
                default=False,
                help="If True, then the VMAX cinder driver will only report "
                "a pool or pools matching what is specified in the XML "
                "configuration file. Storage template creation will be "
                "restricted as well."),
    cfg.BoolOpt('vmax_use_thin_capacity',
                default=False,
                help="If True, then storage pool free capacity reports as"
                "the remaining actual/thin capacity rather than the "
                "unsubscribed capacity.")
]

CONF.register_opts(vmax_cfg_options)


# *****************************************************
# **            VMAX Common Extension                **
# *****************************************************
class EMCVMAXCommonExtension(emc_vmax_common.EMCVMAXCommon):
    """ PowerVC cinder volume extension to the EMC VMAX Common utility class
    """
    def __init__(self, prtcl, configuration=None):
        super(EMCVMAXCommonExtension, self).__init__(prtcl, configuration)
        # set the following overrides that hang off the 'common' class.
        self.masking = EMCVMAXMaskingExtension(prtcl)
        self.utils = EMCVMAXUtilsExtension(prtcl)
        # TODO: Can remove powervc FAST extension after community fix
        self.fast = EMCVMAXFastExtension(prtcl)
        # Just reference the one created instead of constructing new utils.
        self.provision.utils = self.utils
        self.masking.utils = self.utils
        self.masking.provision.utils = self.utils
        self.masking.fast = self.fast
        self.fast.utils = self.utils
        self.fast.provision = self.provision
        self.fast.provision.utils = self.utils
        self.sync_jobs = {}
        LOG.debug("EMCVMAXCommonExtension initialized. CONF.vmax_job_retries="
                  "%d, CONF.vmax_job_interval=%d, CONF.vmax_allow_cached_cim"
                  "_inst=%s, CONF.vmax_wait_for_sync=%s, "
                  "CONF.vmax_restrict_pool_to_xml_conf=%s, "
                  "CONF.vmax_use_thin_capacity=%s." %
                  (CONF.vmax_job_retries, CONF.vmax_job_interval,
                   CONF.vmax_allow_cached_cim_inst, CONF.vmax_wait_for_sync,
                   CONF.vmax_restrict_pool_to_xml_conf,
                   CONF.vmax_use_thin_capacity))

    def _register_config_file_from_config_group(self, configGroupName):
        """Override this method on EMCVMAXCommon.
           Community has rejected bug 1364235. Their direction is to use
           cinder-multi-backend support instead of fixing this case.

        :param configGroupName: the config group name
        :returns: string configurationFile
        """
        LOG.debug("PowerVC Override.")
        if hasattr(self.configuration, 'cinder_emc_config_file'):
            configurationFile = self.configuration.cinder_emc_config_file
        elif configGroupName is None:
            self._set_ecom_credentials(CINDER_EMC_CONFIG_FILE)
            return CINDER_EMC_CONFIG_FILE
        else:
            configurationFile = (
                CINDER_EMC_CONFIG_FILE_PREFIX + configGroupName +
                CINDER_EMC_CONFIG_FILE_POSTFIX)

        # The file saved in self.configuration may not be the correct one,
        # double check
        if configGroupName and configGroupName not in configurationFile:
            configurationFile = (
                CINDER_EMC_CONFIG_FILE_PREFIX + configGroupName +
                CINDER_EMC_CONFIG_FILE_POSTFIX)

        self._set_ecom_credentials(configurationFile)
        LOG.debug("Config_file name = %s." % configurationFile)
        return configurationFile

    def _initial_setup(self, volume, volumeTypeId=None):
        """Necessary setup to accumulate the relevant information.

        This method overrides the method in:
            cinder.volume.drivers.emc.emc_vmax_common

        Currently we override to allow the volume's "type" to specify
        the pool name and the fast policy name if present, otherwise
        obtain the same way the VMAX driver does.
        Community will not likely accept a similar fix since their
        direction is to use cinder-multi-backend support instead.

        TODO: Per https://bugs.launchpad.net/cinder/+bug/1425641,
              community extra specs are broken in Kilo. Need that fixed
              before this override will work again properly

        :param volume: the volume Object
        :returns: The extra specs dictionary
        """
        LOG.debug("PowerVC Override.")
        extraSpecs = super(EMCVMAXCommonExtension, self)._initial_setup(
            volume, volumeTypeId)
        # Volume type fields that override xml config file field are
        # handled here. If the following two things happen, then this
        # override method can go away:
        # 1. PowerVC switches to allow the 'storagetype' scope (not just
        #    'drivers') in build_default_opts() flow.
        # 2. EMC changes their driver to have the volume type extra_specs
        #    override the extra specs coming from the XML config file,
        #    rather than the other way around.
        driver_scope = 'storagetype:'
        pvc_scope = 'drivers:'
        for key in self._get_default_opts().keys():
            pvc_key = pvc_scope + key
            if pvc_key in extraSpecs:
                # The pool property is a special case for PVC.
                if key == 'storage_pool':
                    emc_key = driver_scope + 'pool'
                else:
                    emc_key = driver_scope + key
                # Map PowerVC key scope to cinder VMAX scope
                extraSpecs[emc_key] = extraSpecs[pvc_key]
        if 'storagetype:stripecount' in extraSpecs and\
                int(extraSpecs['storagetype:stripecount']) == 1 and\
                'storagetype:compositetype' in extraSpecs and\
                extraSpecs['storagetype:compositetype'] == 'striped':
            LOG.debug("Converting compositetype to 'concatenated' since "
                      "stripecount is 1.")
            extraSpecs['storagetype:compositetype'] = 'concatenated'

        LOG.debug("Aggregate extra_specs: %s " % extraSpecs)
        if self.sync_jobs:
            # Query for and log some info about synchronization jobs tracked.
            # Note that a special config option is required to get into this
            # code block and if we end up supporting this, then we would
            # need a periodic task to handle this work rather than doing
            # it inline here.
            self._update_sync_job_data()
        return extraSpecs

    @lockutils.synchronized('syncJobs', 'vmax-')
    def _update_sync_job_data(self):
        """ Run through sync jobs, update percent complete info, and
            break clone relationships if jobs are done. This is similar
            to flash copy management.
        """
        to_remove = []
        for vol_id, sync_data in self.sync_jobs.iteritems():
            percentSynced = 0
            try:
                syncInstance = self.conn.GetInstance(
                    sync_data['syncName'], LocalOnly=False)
                percentSynced = syncInstance['PercentSynced']
            except Exception as ex:
                LOG.exception(ex)
                LOG.info("Exception taken, so presume the sync job is "
                         "gone and the operation is complete for %s" %
                         vol_id)
                percentSynced = 100
            LOG.debug("Clone name '%s' with ID %s goes from %d to %d "
                      "percent synced." % (sync_data['cloneName'],
                                           vol_id, sync_data['percentSynced'],
                                           percentSynced))
            sync_data['percentSynced'] = percentSynced
            if percentSynced >= 100:
                to_remove.append(vol_id)
                # Remove the Clone relationship so it can be used as a
                # regular lun.
                try:
                    LOG.debug("Break clone relationship for vol %s." % vol_id)
                    rc, job = self.provision.delete_clone_relationship(
                        self.conn, sync_data['repServiceInstanceName'],
                        sync_data['syncName'], sync_data['extraSpecs'])
                    LOG.debug("delete_clone_relationship for sync data %s "
                              "returns rc=%d, job=%s." % (sync_data, rc, job))
                except Exception as ex:
                    LOG.exception(ex)
                    LOG.debug("Continue after exception with: %s" % sync_data)
        # Any item that reports as complete will be removed from dict even
        # if sync rel did not get successfully removed.
        for vol_id in to_remove:
            LOG.debug("Removing completed sync job: %s" %
                      self.sync_jobs[vol_id])
            del self.sync_jobs[vol_id]

    def _get_default_opts(self):
        """
        Define the default options for the default volume type extra_specs
        in the "drivers:" scope. The drivers scope is a PowerVC convention
        for extra_specs options that the cinder driver honors. So we
        provide this information to integrate the EMC driver into powerVC,
        along with the prior method override. Note that the EMC VMAX driver
        only supports 'thin' pools.
        """
        default_opts = {'storage_pool': self.pool_info['pool_name'],
                        'fastpolicy': None,
                        'stripecount': '1',
                        'slo': None,
                        'workload': None,
                        'provision_type': 'thin'
                        }
        return default_opts

    def update_volume_stats(self):
        """Override the Volume Stats to return all of the Pools"""
        total, free = (0, 0)
        # Call the Parent Class to retrieve most of the Statistics
        stats = super(EMCVMAXCommonExtension, self).update_volume_stats()
        if 'stg_sys_inst_name' in self.pool_info:
            stg_sys_inst_name = self.pool_info['stg_sys_inst_name']
        else:
            stg_sys_inst_name = self.utils.find_storageSystem(
                self.conn, self.pool_info['serial_number'])
            self.pool_info['stg_sys_inst_name'] = stg_sys_inst_name
            self.pool_info['stg_sys_name'] = stg_sys_inst_name['Name']
        # Retrieve the Capacity from each of the Pools and use the total
        # Note that in 1.2.3, for the FAST policy support, it was decided
        # that we would not accumulate the total for the pools associated
        # with a fast policy on the volume-pools API - because it would
        # be to confusing for customers using our UI. However, this method
        # is different from volume-pools API flow and the pool capacities
        # coming from the EMC driver code here may not match the capacities
        # reported for the volume-pools API. This is expected.
        pools = []
        try:
            pools = self.conn.Associators(
                stg_sys_inst_name, resultClass='EMC_VirtualProvisioningPool')
        except Exception as ex:
            LOG.debug("Could not get EMC_VirtualProvisioningPool: %s" %
                      six.text_type(ex))
        for p in pools:
            ptotal = self.utils.convert_bits_to_gbs(p['TotalManagedSpace'])
            palloc = self.utils.convert_bits_to_gbs(p['EMCSubscribedCapacity'])
            pfree = self.utils.convert_bits_to_gbs(p['RemainingManagedSpace'])
            ptotal = max(ptotal, 0)
            unsubscribed = ptotal - palloc
            if unsubscribed < 0:
                LOG.debug("Capping negative unsubscribed of %d to 0." %
                          unsubscribed)
                unsubscribed = 0
            total += ptotal
            if CONF.vmax_use_thin_capacity:
                pfree = max(pfree, 0)
                free += pfree
            else:
                # Normal case consistent with EMC driver.
                free += unsubscribed
        # Don't override values if no pools or if restricting info to what
        # community driver processes.
        if total > 0 and not CONF.vmax_restrict_pool_to_xml_conf:
            # Now we can actually override the Free/Total Capacity values
            stats['total_capacity_gb'] = total
            stats['free_capacity_gb'] = free
            stats['storage_protocol'] = 'FC'  # override 'unknown'
            LOG.debug("VMAX update stats refreshed: %s." % stats)
        else:
            LOG.debug("Don't overwrite status since may be V3 and we only "
                      "support a single V3 pool.")
        return stats

    def find_device_number(self, volume):
        """
        PowerVC version until community driver is fixed to honor finding
        the device number with respect to a set of initiator wwpns.

        TODO: The LP# 1382641 community fix is suspected to NOT be a
              complete solution. Leave this here until we can collect proper
              data to open new bug and have community resolve if needed.
        """
        if 'initiator_wwpns' not in volume:
            return super(EMCVMAXCommonExtension, self).find_device_number(
                volume)
        # otherwise, use our override
        wwpns = [wwpn.lower() for wwpn in volume['initiator_wwpns']]
        foundNumDeviceNumber = None
        lunMaskingView = None
        volumeInstance = self._find_lun(volume)
        LOG.debug("PowerVC override. volume_devid=%s, wwpns=%s" %
                  (volumeInstance['DeviceID'], wwpns))
        storageSystemName = volumeInstance['SystemName']
        unitnames = self.conn.ReferenceNames(
            volumeInstance.path,
            ResultClass='CIM_ProtocolControllerForUnit')

        for unitname in unitnames:
            controller = unitname['Antecedent']
            classname = controller['CreationClassName']
            index = classname.find('Symm_LunMaskingView')
            if index > -1:
                LOG.debug("PowerVC check initiator ports for %s." % controller)
                hwid_inst_names = self.conn.AssociatorNames(
                    controller, ResultClass='EMC_StorageHardwareID')
                hwid_wwpns = self.utils.hwid_names_to_wwpns(hwid_inst_names)
                found_match = None
                for hwid in hwid_wwpns:
                    if hwid in wwpns:
                        found_match = hwid
                        break
                # do we need to special case if connector or wwpns is not
                # passed in?
                if found_match:
                    LOG.debug("First matching wwpn: %s" % found_match)
                    unitinstance = self.conn.GetInstance(unitname,
                                                         LocalOnly=False)
                    foundNumDeviceNumber = int(unitinstance['DeviceNumber'],
                                               16)
                    lunMaskingView = controller
                    break
                else:
                    LOG.debug("Did not find matching initiator wwpn in: "
                              "%s" % hwid_wwpns)

        if foundNumDeviceNumber is None:
            LOG.debug(
                "Device number not found for volume "
                "%(volumeName)s %(volumeInstance)s."
                % {'volumeName': volume['name'],
                   'volumeInstance': volumeInstance.path})

        data = {'hostlunid': foundNumDeviceNumber,
                'storagesystem': storageSystemName,
                'maskingview': lunMaskingView
                }

        LOG.debug("Device info: %(data)s." % {'data': data})
        return data

    def _set_ecom_credentials(self, configurationFile):
        """
        Override the Set Credentials method to handle backwards
        compatibility with 1.2.2.x registrations.
        """
        super(EMCVMAXCommonExtension, self).\
            _set_ecom_credentials(configurationFile)
        if not self.ecomCACert:
            # 'Cert' is the old xml key for 1.2.2.x
            self.ecomCACert = self.utils._parse_from_file(configurationFile,
                                                          'Cert')
            if self.ecomCACert:
                LOG.debug('Picked up Cert instead of EcomCACert.')
            else:
                LOG.warn(_LW("Could not find certificate reference in "
                             "configuration file for secure https."))

    def _get_ecom_connection(self):
        """
        Override the Get Connection method to not pass x509.
        The community flow still does not connect successfully.
        """
        # Continue to replace 'http:' to force a failure in case secure
        # options were manually removed from the XML config file.
        self.url = self.url.replace('http://', 'https://')
        conn = pywbem.WBEMConnection(self.url, (self.user, self.passwd),
                                     default_namespace='root/emc', x509=None,
                                     verify_callback=None,
                                     ca_certs=self.ecomCACert,
                                     no_verification=self.ecomNoVerification)
        # If we can't connect to the SMI-S Proxy, throw an exception
        if conn is None:
            exception_message = (_LE("Cannot connect to ECOM server"))
            raise exception.VolumeBackendAPIException(data=exception_message)
        conn.debug = True
        return conn

    def _find_lun(self, volume, skip_verify=False, skip_cache=False):
        """
        PowerVC override to avoid multiple lookups per request and to do
        validation that the original volume was not deleted out-of-band
        and the deviceID is re-used. The community handles this in a way
        not compatible with our current implementation.
        :param skip_verify: Optional param. If True, then will not check and
                    raise exception if backend instance does not match
                    correlation restricted_metadata we track. Otherwise
                    the verification check is done.
        :param skip_cache: Optional param. If True, then skip the cache and
                    always perform the lookup.
        """
        if CONF.vmax_allow_cached_cim_inst and '_retrieved_inst' in volume\
                and not skip_cache:
            LOG.debug("PowerVC override to returned already retrieved "
                      "instance: %s" % (volume['_retrieved_inst']))
            return volume['_retrieved_inst']
        else:
            lu = super(EMCVMAXCommonExtension, self)._find_lun(volume)
            if lu:
                # Handle the case where the original volume was deleted and
                # the device ID was re-used for a different volume.
                emc_name = lu['ElementName']
                if emc_name and not skip_verify:
                    metad = None
                    if 'restricted_metadata' in volume:
                        metad = volume['restricted_metadata']
                    else:
                        try:
                            LOG.debug("Looking up restricted metadata for vol")
                            grm = powervc_db_api.volume_restricted_metadata_get
                            metad = grm(context.get_admin_context(),
                                        volume['id'])
                        except Exception as ex:
                            LOG.exception(ex)
                    LOG.debug("restricted_metadata=%s" % metad)
                    if not metad or rmk.VDISK_NAME_KEY not in metad:
                        # This can happen if volume just created and metadata
                        # is not persisted yet.
                        LOG.info(_LI("PowerVC metadata looked up is "
                                     "incomplete for VMAX volume: %(data)s ") %
                                 {'data': metad})
                    elif metad[rmk.VDISK_NAME_KEY] != emc_name:
                        LOG.info(
                            _LI("Volume %(vid)s EMC name '%(mname)s' has "
                                "changed on the backend to %(emc_name)s. "
                                "Lookup fails.") %
                            dict(vid=volume['id'],
                                 mname=metad[rmk.VDISK_NAME_KEY],
                                 emc_name=emc_name))
                        # The lun_status here is temporary - not persisted.
                        # The health monitor will periodically call
                        # check_for_deleted_volumes() and persist the actual
                        # state/status properties accordingly.
                        volume['lun_status'] = _LI("Deleted out-of-band "
                                                   "and DeviceID reused")
                        return None
                volume['_retrieved_inst'] = lu
            else:
                # The status here is temporary - not persisted.
                volume['lun_status'] = _LI("Not found")
            return lu

    def _create_v2_replica_and_delete_clone_relationship(
            self, repServiceInstanceName, cloneVolume, sourceVolume,
            sourceInstance, targetInstance, extraSpecs, isSnapshot=False):
        """
        Create a replica and delete the clone relationship.

        PowerVC Override:
            If targetInstance is None, we need to create it before
            calling create_element_replica() such that the target has
            the extraSpecs from the desired storage template plus the
            disk geometry of the source volume.

        :param repServiceInstanceName: the replication service
        :param cloneVolume: the clone volume object
        :param sourceVolume: the source volume object
        :param sourceInstance: the source volume instance
        :param targetInstance: the target volume instance
        :param extraSpecs: extra specifications
        :param isSnapshot: check to see if it is a snapshot
        :returns: int -- return code
        :returns: dict -- cloneDict
        """
        sourceName = sourceVolume['name']
        cloneName = cloneVolume['name']

        # Check for targetInstance and create if needed.
        # If the source is a striped volume however, we cannot pre-create
        # the target. Allow target to be created by the CIM method
        # invocation - same pool as source.
        if not targetInstance and\
                extraSpecs['storagetype:compositetype'] != 'striped':
            # Set-up for calling the lower-level create_composite_volume()
            # method in the same fashion that create_volume() does,
            # except that we hard code the stripecount to be 1 and the
            # composite type to be 'concatenated'. The reason we do this is
            # because in our override flow (1) we know that targetInstance
            # does not get passed only when the source is a simple volume
            # of this nature and (2) we need to override those two extra
            # specs since they must get ignored from the target volume type.
            # If they are not ignored, then volume copy that happens later
            # can fail.
            # First, determine the target volume size from the source volume
            # size in bytes, not converted from the cinder 'size' in GiB,
            # which may have a rounding error for on-boarded volumes.
            volumeSize = int(
                self.utils.get_volume_size(self.conn, sourceInstance))
            if (volumeSize != (sourceVolume['size'] * (1024 ** 3))):
                LOG.info(_("Source volume '%(name)s' size of %(ssb)d bytes "
                           "does not match cinder volume size of %(gbs)s GiB."
                           " Bytes difference=%(diff)d.")
                         % {'name': sourceName, 'ssb': volumeSize,
                            'gbs': sourceVolume['size'], 'diff':
                            volumeSize - (int(sourceVolume['size']) *
                                          (2 ** 30))})
            volumeName = cloneVolume['id']
            compositeType = self.utils.get_composite_type('concatenated')
            poolInst, storageSystemName = (
                self._get_pool_and_storage_system(extraSpecs))
            elementCompositionService = (
                self.utils.find_element_composition_service(self.conn,
                                                            storageSystemName))
            LOG.debug("PowerVC: Create concatenated clone volume target. "
                      "Volume: %(cloneName)s, Source Volume: %(sourceName)s, "
                      "SourceElement: %(sourceelement)s, "
                      "extra_specs: %(extraspecs)s, poolInst=%(poolInst)s, "
                      "Stripecount=1."
                      % {'cloneName': cloneName, 'sourceName': sourceName,
                         'sourceelement': six.text_type(sourceInstance.path),
                         'extraspecs': extraSpecs, 'poolInst': poolInst})

            # Create the target volume shell thin device.
            target_voldict, rc = self.provision.create_composite_volume(
                self.conn, elementCompositionService, volumeSize, volumeName,
                poolInst, compositeType, 1, extraSpecs)
            # rc is guaranteed to be 0 at this point.
            LOG.debug("Created thin concatenated volume for cinder id %s, "
                      "RC=%d." % (volumeName, rc))

            # Add provider location so that we can look up the CIM inst.
            cloneVolume['provider_location'] = six.text_type(target_voldict)
            targetInstance = self._find_lun(cloneVolume, skip_verify=True)
            if not targetInstance:
                message = (_("Could not locate the clone volume target "
                             "on the array for provider_location: %(loc)s.")
                           % {'loc': cloneVolume['provider_location']})
                raise exception.VolumeDriverException(message=message)
            LOG.debug("Target thin volume created per extra specs. Now make "
                      "an ElementReplica into: %s" %
                      six.text_type(targetInstance.path))
        else:
            LOG.debug("Target instance already created or source is striped. "
                      "Meta-config=%s. Continue with ElementReplica into: %s" %
                      (extraSpecs['storagetype:compositetype'],
                       six.text_type(targetInstance.path) if
                       targetInstance else 'none.'))

        # TODO: Call the super method. For now we keep the super processing
        #       inline to investigate not waiting for synchronization
        #       (aka flash copy) to complete.
        try:
            rc, job = self.provision.create_element_replica(
                self.conn, repServiceInstanceName, cloneName, sourceName,
                sourceInstance, targetInstance, extraSpecs)
        except Exception:
            LOG.warning(_LW(
                "Clone failed. Cleaning up the target volume. "
                "Clone name: %(cloneName)s "
                "Source name: %(sourceName)s "
                "Extra specs: %(extraSpecs)s "),
                {'cloneName': cloneName,
                 'sourceName': sourceName,
                 'extraSpecs': extraSpecs})
            if targetInstance is not None:
                # Check if the copy session exists.
                storageSystem = targetInstance['SystemName']
                syncInstanceName = self.utils.find_sync_sv_by_target(
                    self.conn, storageSystem, targetInstance, False)
                if syncInstanceName is not None:
                    # Remove the Clone relationship.
                    rc, job = self.provision.delete_clone_relationship(
                        self.conn, repServiceInstanceName, syncInstanceName,
                        extraSpecs, True)
                storageConfigService = (
                    self.utils.find_storage_configuration_service(
                        self.conn, storageSystem))
                self._delete_target_volume_v2(
                    storageConfigService, targetInstance, extraSpecs)
                if 'provider_location' in cloneVolume:
                    # clear it since rollback already performed.
                    cloneVolume['provider_location'] = ""
            # Re-throw the exception.
            raise
        cloneDict = self.provision.get_volume_dict_from_job(
            self.conn, job['Job'])

        fastPolicyName = extraSpecs[FASTPOLICY]
        if isSnapshot:
            if fastPolicyName is not None:
                storageSystemName = sourceInstance['SystemName']
                self._add_clone_to_default_storage_group(
                    fastPolicyName, storageSystemName, cloneDict, cloneName)
            LOG.info(_LI("Snapshot creation %(cloneName)s completed. "
                     "Source Volume: %(sourceName)s."),
                     {'cloneName': cloneName,
                      'sourceName': sourceName})

            return rc, cloneDict

        cloneVolume['provider_location'] = six.text_type(cloneDict)
        # Note - this point, the cloneVolume will be cleaned up at the
        # top emc_vmax_product driver level if an exception is raised later
        # in this method. By Liberty, there will be community fix that
        # correctly handles the cleanup within this method in a parallel
        # fashion to vmax V3 cleanup. TODO: We will re-evaluate then.
        syncInstanceName, storageSystemName = (
            self._find_storage_sync_sv_sv(cloneVolume, sourceVolume,
                                          waitforsync=CONF.vmax_wait_for_sync))

        if CONF.vmax_wait_for_sync:
            # Remove the Clone relationship so it can be used as a regular lun.
            rc, job = self.provision.delete_clone_relationship(
                self.conn, repServiceInstanceName, syncInstanceName,
                extraSpecs)
        else:
            # Don't break the clone relationship, but keep track of it
            # so that it can be managed and removed later.
            sync_data = {'syncName': syncInstanceName,
                         'cloneLoc': cloneVolume['provider_location'],
                         'repServiceInstanceName': repServiceInstanceName,
                         'extraSpecs': copy.deepcopy(extraSpecs),
                         'cloneName': cloneName,
                         'percentSynced': 0}
            self.sync_jobs[cloneVolume['id']] = sync_data
            LOG.debug("Added synchronization entry for cloned ID %s: %s" %
                      (cloneVolume['id'], sync_data))

        # I wonder if the vmax driver is incorrect in adding the volume
        # to the default FAST policy group after the synchronization is
        # complete?
        LOG.debug("FAST policy is: %s." % fastPolicyName)
        if fastPolicyName is not None:
            self._add_clone_to_default_storage_group(
                fastPolicyName, storageSystemName, cloneDict, cloneName,
                extraSpecs)
        return rc, cloneDict

    def _create_clone_v2(self, repServiceInstanceName, cloneVolume,
                         sourceVolume, sourceInstance, isSnapshot, extraSpecs):
        """Create a clone (v2).
           PowerVC override. TODO: This is a large method and all we are
           doing is changing a few lines. We should dialog with
           the EMC community team about either restructuring this so we
           can override with less tech debt or always creating the target
           for us using extra_specs rather than ignoring the extra specs
           by letting CreateElementReplica make the clone target atomically
           with the replication.

        :param repServiceInstanceName: the replication service
        :param cloneVolume: the clone volume object
        :param sourceVolume: the source volume object
        :param sourceInstance: the device ID of the volume
        :param extraSpecs: extra specifications
        :param isSnapshot: check to see if it is a snapshot
        :returns: int -- return code
        :raises: VolumeBackendAPIException
        """
        # Check if the source volume contains any meta devices.
        metaHeadInstanceName = self.utils.get_volume_meta_head(
            self.conn, sourceInstance.path)

        if metaHeadInstanceName is None:  # Simple volume.
            LOG.debug("No meta-head. Call PowerVC override for simple "
                      "volume cloning.")
            return self._create_v2_replica_and_delete_clone_relationship(
                repServiceInstanceName, cloneVolume, sourceVolume,
                sourceInstance, None, extraSpecs, isSnapshot=isSnapshot)
        else:
            # Composite volume with meta device members.
            # Check the meta members capacity info.
            metaMemberInstanceNames = (
                self.utils.get_meta_members_of_composite_volume(
                    self.conn, metaHeadInstanceName))
            volumeCapacities = self.utils.get_meta_members_capacity_in_bit(
                self.conn, metaMemberInstanceNames)
            # Override extra specs such that target vol matches source for
            # the meta configuration (geometry)
            cloneExtraSpecs = copy.deepcopy(extraSpecs)
            metaHeadInst = self.conn.GetInstance(metaHeadInstanceName)
            propList = metaHeadInst.properties.items()
            LOG.debug("metaHead propList: %s " % propList)
            isConcatenated = 'True'
            for properties in propList:
                if properties[0] == 'IsConcatenated':
                    isConcatenated = six.text_type(properties[1].value)
            sourceType = 'concatenated' if isConcatenated == 'True'\
                else 'striped'
            cloneExtraSpecs['storagetype:compositetype'] = sourceType
            cloneExtraSpecs['storagetype:membercount'] =\
                cloneExtraSpecs['storagetype:stripecount'] =\
                str(len(metaMemberInstanceNames))
            # If striped, the fastpolicy needs to inherit from the source
            # volume as well.
            if sourceType == 'striped':
                cloneExtraSpecs[FASTPOLICY] =\
                    self._get_fast_policy_for_vol(sourceInstance)
            LOG.debug("%d source Meta volumes, %d different sizes. "
                      "CloneExtraSpecs: %s" % (len(volumeCapacities),
                                               len(set(volumeCapacities)),
                                               cloneExtraSpecs))

            # If only one meta or striped, do the same as above. But for
            # multiple metas, even if all the same size, fall through to loop
            # processing below. (PowerVC change). If source volume is
            # striped, then we can assume all the metas are the same size.
            if len(metaMemberInstanceNames) == 1 or sourceType == 'striped':
                LOG.debug("Only one meta or striped: %s" %
                          sourceVolume['name'])
                return self._create_v2_replica_and_delete_clone_relationship(
                    repServiceInstanceName, cloneVolume, sourceVolume,
                    sourceInstance, None, cloneExtraSpecs, isSnapshot)

            baseTargetVolumeInstance = None
            # Force member count to be 1 for a base concatenated vol.
            cloneExtraSpecs['storagetype:membercount'] = '1'
            LOG.debug("Volume capacities:  %s. With membercount=1, "
                      "continue on to creating the target meta volume for %s."
                      % (volumeCapacities, sourceVolume['name']))
            for volumeSizeInbits in volumeCapacities:
                if baseTargetVolumeInstance is None:  # Create base volume.
                    baseVolumeName = "TargetBaseVol"
                    volume = {'size': int(self.utils.convert_bits_to_gbs(
                        volumeSizeInbits))}
                    __, baseVolumeDict, storageSystemName = (
                        self._create_composite_volume(volume, baseVolumeName,
                                                      volumeSizeInbits,
                                                      cloneExtraSpecs))
                    baseTargetVolumeInstance = self.utils.find_volume_instance(
                        self.conn, baseVolumeDict, baseVolumeName)
                    LOG.debug("Base target volume %(targetVol)s created. "
                              "capacity requested: %(capInBits)lu.",
                              {'capInBits': volumeSizeInbits,
                               'targetVol': baseTargetVolumeInstance.path})
                    ret_base_size = int(self.utils.get_volume_size(
                        self.conn, baseTargetVolumeInstance))
                    if ret_base_size != volumeSizeInbits:
                        LOG.warn("Source=%s, Target=base. The created target "
                                 "size of %d does not match requested size of "
                                 "%d." % (sourceVolume['name'], ret_base_size,
                                          volumeSizeInbits))
                else:  # Create append volume
                    targetVolumeName = "MetaVol"
                    volume = {'size': int(self.utils.convert_bits_to_gbs(
                        volumeSizeInbits))}
                    storageConfigService = (
                        self.utils.find_storage_configuration_service(
                            self.conn, storageSystemName))
                    unboundVolumeInstance = (
                        self._create_and_get_unbound_volume(
                            self.conn, storageConfigService,
                            baseTargetVolumeInstance.path, volumeSizeInbits,
                            cloneExtraSpecs))
                    if unboundVolumeInstance is None:
                        exceptionMessage = (_LE(
                            "Error Creating unbound volume."))
                        LOG.error(exceptionMessage)
                        # Remove target volume (cleanup)
                        self._delete_target_volume_v2(storageConfigService,
                                                      baseTargetVolumeInstance,
                                                      cloneExtraSpecs)
                        raise exception.VolumeBackendAPIException(
                            data=exceptionMessage)
                    LOG.debug("New unbound volume member %s to append: %s " %
                              (targetVolumeName,
                               six.text_type(unboundVolumeInstance.path)))
                    ret_meta_size = int(self.utils.get_volume_size(
                        self.conn, unboundVolumeInstance))
                    if ret_meta_size != volumeSizeInbits:
                        LOG.warn("Source=%s, Target=meta. The created target "
                                 "size of %d does not match requested size of "
                                 "%d." % (sourceVolume['name'], ret_meta_size,
                                          volumeSizeInbits))

                    # Append the new unbound volume to the
                    # base target composite volume.
                    baseTargetVolumeInstance = self.utils.find_volume_instance(
                        self.conn, baseVolumeDict, baseVolumeName)
                    try:
                        elementCompositionService = (
                            self.utils.find_element_composition_service(
                                self.conn, storageSystemName))
                        compositeType = self.utils.get_composite_type(
                            cloneExtraSpecs['storagetype:compositetype'])
                        modifiedVolumeDict = None
                        __, modifiedVolumeDict = (
                            self._modify_and_get_composite_volume_instance(
                                self.conn,
                                elementCompositionService,
                                baseTargetVolumeInstance,
                                unboundVolumeInstance.path,
                                targetVolumeName, compositeType,
                                cloneExtraSpecs))
                        if modifiedVolumeDict is None:
                            exceptionMessage = (_LE(
                                "Error appending volume %(volumename)s to "
                                "target base volume.")
                                % {'volumename': targetVolumeName})
                            LOG.error(exceptionMessage)
                            raise exception.VolumeBackendAPIException(
                                data=exceptionMessage)
                    except Exception:
                        exceptionMessage = (_LE(
                            "Exception appending meta volume to target volume "
                            "%(volumename)s.")
                            % {'volumename': baseVolumeName})
                        LOG.error(exceptionMessage)
                        # Remove append volume and target base volume
                        self._delete_target_volume_v2(
                            storageConfigService, unboundVolumeInstance,
                            cloneExtraSpecs)
                        self._delete_target_volume_v2(
                            storageConfigService, baseTargetVolumeInstance,
                            cloneExtraSpecs)
                        LOG.info(_LI("Rollback happened for base and append "
                                     "volume."))
                        raise

            LOG.debug("Create V2 replica for multiple meta members.")
            return self._create_v2_replica_and_delete_clone_relationship(
                repServiceInstanceName, cloneVolume, sourceVolume,
                sourceInstance, baseTargetVolumeInstance, cloneExtraSpecs,
                isSnapshot)

    def _delete_target_volume_v2(self, storageConfigService,
                                 targetVolumeInstance, extraSpecs):
        """Helper function to delete the clone target volume instance.
        TODO: Remove when community fix goes in!
        :param storageConfigService: storage configuration service instance
        :param targetVolumeInstance: clone target volume instance
        :param extraSpecs: extra specifications
        """
        deviceId = targetVolumeInstance['DeviceID']
        volumeName = targetVolumeInstance['Name']
        rc = self._delete_from_pool(storageConfigService,
                                    targetVolumeInstance,
                                    volumeName, deviceId,
                                    extraSpecs[FASTPOLICY],
                                    extraSpecs)
        return rc

    def _get_fast_policy_for_vol(self, volumeInstance):
        """ Return fast policy controlling volume if one, else None """
        sg_instNames = self.conn.AssociatorNames(
            volumeInstance.path, ResultClass='CIM_DeviceMaskingGroup')
        if sg_instNames:
            prex = emc_vmax_fast.DEFAULT_SG_PREFIX
            postx = emc_vmax_fast.DEFAULT_SG_POSTFIX
            for sg_instName in sg_instNames:
                if prex in sg_instName['InstanceID']:
                    # found a fast policy
                    tail = sg_instName['InstanceID'].split(prex)[1]
                    fastpolicy = tail.split(postx)[0]
                    LOG.debug("Volume %s is under fast control of: %s " %
                              (volumeInstance.path['DeviceID'], fastpolicy))
                    return fastpolicy
        return None

    def _get_volume_dict_from_path(self, volpath):
        """Given the volume path, determine the volume dictionary.
        This is also known as the provider_location.
        Compare to emc_vmax_provision->get_volume_dict_from_job().
        """
        volumeDict = {}
        volumeDict['classname'] = volpath.classname
        keys = {}
        keys['CreationClassName'] = volpath['CreationClassName']
        keys['SystemName'] = volpath['SystemName']
        keys['DeviceID'] = volpath['DeviceID']
        keys['SystemCreationClassName'] = volpath['SystemCreationClassName']
        volumeDict['keybindings'] = keys
        return volumeDict

    def _check_and_remove_sync_rels(self, volume):
        """ Helper to do work of checking for and removing any
        SE_StorageSynchronized_SV_SV clone relationships.
        TODO: I don't know if this fix makes sense for the community(?)
              This can probably be removed with a more robust driver in place
              for Liberty.
        """
        vol_dict = ast.literal_eval(volume['provider_location'])
        vol_path = self.utils.get_instance_name(
            vol_dict['classname'], vol_dict['keybindings'])
        storageSystem = vol_dict['keybindings']['SystemName']
        sv_sync_ans = self.conn.AssociatorNames(
            vol_path, AssocClass='SE_StorageSynchronized_SV_SV',
            ResultClass='Symm_StorageVolume')
        LOG.debug("SE_StorageSynchronized_SV_SV association targets of "
                  "%s: %s." % (vol_dict['keybindings']['DeviceID'],
                               sv_sync_ans))
        if not sv_sync_ans:
            return
        repSvc = self.utils.find_replication_service(
            self.conn, storageSystem)
        num_removed = 0
        for targ in sv_sync_ans:
            # Try finding the actual association by using one direction...
            sync_path = self.utils.get_instance_name(
                'SE_StorageSynchronized_SV_SV',
                {'SyncedElement': vol_path, 'SystemElement': targ})
            try:
                sync_inst = self.conn.GetInstance(sync_path)
            except Exception:
                sync_inst = None
            LOG.debug("sync_path=%s.  GetInsance=%s" % (sync_path, sync_inst))
            if not sync_inst:
                # ...and now try the other direction...
                sync_path = self.utils.get_instance_name(
                    'SE_StorageSynchronized_SV_SV',
                    {'SyncedElement': targ, 'SystemElement': vol_path})
                sync_inst = self.conn.GetInstance(sync_path)
                LOG.debug("#2 sync_path=%s.  GetInsance=%s" % (sync_path,
                                                               sync_inst))
            if not sync_inst:
                LOG.info(_LI("Cannot get SE_StorageSynchronized_SV_SV for: %s")
                         % six.text_type(targ))
                continue
            LOG.info(_LI("Break clone relationship for "
                         "synchronization instance: %s") % sync_path)
            extraSpecs = self._initial_setup(volume)
            rc, job = self.provision.delete_clone_relationship(
                self.conn, repSvc, sync_path, extraSpecs, force=True)
            LOG.debug("Remove clone relationship returns: RC=%d, job=%s" %
                      (rc, six.text_type(job)))
            if rc == 0:
                num_removed += 1
        LOG.debug("Return. Num_removed=%d." % num_removed)


# *****************************************************
# **            VMAX Masking Extension               **
# *****************************************************
class EMCVMAXMaskingExtension(emc_vmax_masking.EMCVMAXMasking):
    """ PowerVC Override to EMC VMAX masking class """
    def __init__(self, prtcl):
        super(EMCVMAXMaskingExtension, self).__init__(prtcl)
        self.hwidService = None

    def _get_storage_hardware_id_instance_names(
            self, conn, initiatorNames, storageSystemName):
        """Given a list of initiator names find CIM_StorageHardwareID instance.
           This is one of the worst offenders in the VMAX driver so
           overriding here.
        :param conn: the connection to the ecom server
        :param initiatorName: the list of initiator names
        :param storageSystemName: the storage system name
        :returns: foundHardwardIDsInstanceNames
        """
        LOG.debug("Find SE_StorageHardwareID CIM paths for wwpns: %s" %
                  initiatorNames)
        foundHWIDNames = []
        if not self.hwidService:
            self.hwidService = self.utils.find_storage_hardwareid_service(
                conn, storageSystemName)
        i_wwpns = set([i.lower() for i in initiatorNames])
        hwInstNames = conn.AssociatorNames(
            self.hwidService, ResultClass='SE_StorageHardwareID')

        for hwInstName in hwInstNames:
            # Note: thread-safety fix not needed because our override
            #       does not look up the instance for performance.
            storageId = six.text_type(hwInstName).strip('\"')[-16:].lower()
            if storageId in i_wwpns:
                foundHWIDNames.append(hwInstName)
                i_wwpns.remove(storageId)
            if not i_wwpns:
                break
        LOG.debug("The found hardware IDs are: %s. WWPNs not found: %s" %
                  (foundHWIDNames, i_wwpns))
        return foundHWIDNames

    def _is_volume_in_storage_group(
            self, conn, storageGroupInstanceName, volumeInstance):
        """Override: Check if the volume is already part of the storage group.

        Check if the volume is already part of the storage group,
        if it is no need to re-add it.

        :param conn: the connection to  ecom
        :param storageGroupInstanceName: the storage group instance name
        :param volumeInstance: the volume instance
        :returns: boolean
        """
        # TODO: csky override - remove method when fixed in community.
        #       The gtype param is specified as a masking storage group as
        #       opposed to a default FAST SG. Also, we pass in the specific
        #       name to check for. I believe this is required to support
        #       multi-attach (shared disk).
        foundStorageGroupInstanceName = (
            self.utils.get_storage_group_from_volume(
                conn, volumeInstance.path, gtype='mask',
                sg_inst_name=storageGroupInstanceName))

        if foundStorageGroupInstanceName is not None:
            storageGroupInstance = conn.GetInstance(
                storageGroupInstanceName, LocalOnly=False)
            LOG.debug(
                "The existing storage group instance element name is: "
                "%(existingElement)s.",
                {'existingElement': storageGroupInstance['ElementName']})
            foundStorageGroupInstance = conn.GetInstance(
                foundStorageGroupInstanceName, LocalOnly=False)
            LOG.debug(
                "The found storage group instance element name is: "
                "%(foundElement)s.",
                {'foundElement': foundStorageGroupInstance['ElementName']})
            if (foundStorageGroupInstance['ElementName'] == (
                    storageGroupInstance['ElementName'])):
                return True
        return False

    def _get_sg_or_mv_associated_with_initiator(
            self, conn, controllerConfigService, volumeInstanceName,
            volumeName, connector, getSG=True):
        """Get storage group or masking view associated with connector.

        If the connector gets passed then extra logic required to
        get storage group.

        :param conn: the ecom connection
        :param controllerConfigService: storage system instance name
        :param volumeInstanceName: volume instance name
        :param volumeName: volume element name
        :param connector: the connector object
        :param getSG: True if to get storage group; otherwise get masking
        :returns: foundInstanceName(can be None)
        """
        foundInstanceName = None
        initiatorNames = self._find_initiator_names(conn, connector)
        igInstanceNameFromConnector = self._find_initiator_masking_group(
            conn, controllerConfigService, initiatorNames)
        # Device can be shared by multi-SGs in a multi-host attach case.
        storageGroupInstanceNames = (
            self.get_associated_masking_groups_from_device(
                conn, volumeInstanceName))
        LOG.debug("Found storage groups volume "
                  "%(volumeName)s is in: %(storageGroups)s",
                  {'volumeName': volumeName,
                   'storageGroups': storageGroupInstanceNames})
        if storageGroupInstanceNames:  # not empty
            # Get the SG by IGs.
            for sgInstanceName in storageGroupInstanceNames:
                # Get maskingview from storage group.
                mvInstanceName = self.get_masking_view_from_storage_group(
                    conn, sgInstanceName)
                # TODO: csky override - remove method when fixed in community
                if not mvInstanceName:
                    LOG.debug("Skip storage group %s." %
                              six.text_type(sgInstanceName))
                    continue
                LOG.debug("Found masking view associated with SG "
                          "%(storageGroup)s: %(maskingview)s",
                          {'maskingview': mvInstanceName,
                           'storageGroup': sgInstanceName})
                # Get initiator group from masking view.
                igInstanceName = (
                    self.get_initiator_group_from_masking_view(
                        conn, mvInstanceName))
                LOG.debug("Initiator Group in masking view %(ig)s: "
                          "IG associated with connector%(igFromConnector)s",
                          {'ig': igInstanceName,
                           'igFromConnector': igInstanceNameFromConnector})
                if igInstanceName == igInstanceNameFromConnector:
                    if getSG is True:
                        foundInstanceName = sgInstanceName
                        LOG.debug("Found the storage group associated with "
                                  "initiator %(initiator)s: %(storageGroup)s",
                                  {'initiator': initiatorNames,
                                   'storageGroup': foundInstanceName})
                    else:
                        foundInstanceName = mvInstanceName
                        LOG.debug("Found the masking view associated with "
                                  "initiator %(initiator)s: %(maskingview)s.",
                                  {'initiator': initiatorNames,
                                   'maskingview': foundInstanceName})
                    break
        return foundInstanceName


# *****************************************************
# **            VMAX Utils Extension                 **
# *****************************************************
class EMCVMAXUtilsExtension(emc_vmax_utils.EMCVMAXUtils):
    """Utility class for SMI-S based EMC volume drivers.

    This Utility class is for EMC volume drivers based on SMI-S.
    It supports VMAX arrays.
    """

    def __init__(self, prtcl):
        super(EMCVMAXUtilsExtension, self).__init__(prtcl)
        self.in_flight = []
        self.total_polls = 0
        self.total_jobs = 0

    def hwid_names_to_wwpns(self, hwid_name_list, toupper=False):
        """ Utility method to map HWID inst names to WWPNs """
        if toupper:
            return [six.text_type(hn).strip('\"')[-16:].upper()
                    for hn in hwid_name_list]
        else:
            return [six.text_type(hn).strip('\"')[-16:].lower()
                    for hn in hwid_name_list]

    def _wait_for_job_complete(self, conn, job, extraSpecs=None,
                               syncName=None):
        """Given the job wait for it to complete.
        Overriding community to fix thread-safety issue.
        TODO: Remove after community fix? LP# 1393568, LP# 1401279, etc.

        :param conn: connection the ecom server
        :param job: the job dict
        :param extraSpecs: the extraSpecs dict. Defaults to None
        :param syncName: if passed then we are waiting for a sync request
                         (volume copy) to complete. The job is ignored in
                         this case.
        """
        maxJobRetries = self._get_max_job_retries(extraSpecs)
        intervalInSecs = self._get_interval_in_secs(extraSpecs)

        def _wait_for_job_nested(dict_in):
            """Called at an interval until the job is finished"""
            retry_attempt = maxJobRetries - dict_in['retry_countdown']
            jobData = dict_in['job']
            # Refactored method to handle both synchronization requests and
            # job requests since the wait structure should be the same.
            if jobData:
                waitLogObj = jobData['Job']
            else:
                waitLogObj = dict_in['sync_name']
            LOG.debug("PowerVC: Job or sync object "
                      "%(job)s, attempt %(retry)d. In-flight=%(num)d."
                      % {'job': waitLogObj,
                         'retry': retry_attempt,
                         'num': len(self.in_flight)})
            if dict_in['retry_countdown'] <= 0:
                seconds = maxJobRetries * intervalInSecs
                LOG.error(_LE("PowerVC: Issuing a "
                              "failure after %(retries)d tries for job "
                              "%(job)s.") %
                          {'retries': maxJobRetries,
                           'job': waitLogObj})
                err = (_LE("Could not tell that job or synchronization request"
                           " completed after waiting %(seconds)d seconds for "
                           "'%(job)s.'") %
                       {'seconds': seconds, 'job': waitLogObj})
                # Bubble up error so don't raise LoopingCallDone() here.
                raise exception.VolumeBackendAPIException(err)
            is_job_finished = False
            try:
                if jobData:
                    is_job_finished = self._is_job_finished(conn, jobData)
                else:
                    is_job_finished = self._is_sync_complete(
                        conn, dict_in['sync_name'])
            except Exception as e:
                LOG.exception(e)
                exceptionMessage = (_LE("Issue encountered waiting "
                                        "for job or synchronization request "
                                        "'%(job)s'. Check the cinder logs for "
                                        "details.") % {'job': waitLogObj})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)
            if is_job_finished:
                self.total_jobs += 1
                self.total_polls += retry_attempt
                average = float(self.total_polls) / float(self.total_jobs)
                LOG.debug("Retry %d satisfied (average_polls=%.1f). "
                          "Job/sync is done: %s" % (retry_attempt,
                                                    average, waitLogObj))
                raise loopingcall.LoopingCallDone()
            dict_in['retry_countdown'] -= 1

        loop_param = {'retry_countdown': maxJobRetries,
                      'job': job, 'sync_name': syncName}
        LOG.debug("Begin FixedIntervalLoopingCall: Interval=%d, loop_param="
                  "%s, extraSpecs=%s." % (intervalInSecs, loop_param,
                                          extraSpecs))
        self.in_flight.append(job)
        try:
            timer = loopingcall.FixedIntervalLoopingCall(_wait_for_job_nested,
                                                         loop_param)
            event_ret = timer.start(interval=intervalInSecs).wait()
        finally:
            self.in_flight.remove(job)
        LOG.debug("FixedIntervalLoopingCall returns: %s" % event_ret)

    def wait_for_sync(self, conn, syncName):
        """Given the sync name wait for it to fully synchronize.
        Overriding community to fix thread-safety issue.
        TODO: remove after community fix?

        :param conn: connection the ecom server
        :param syncName: the syncName
        """
        # Call our other overridden method.
        self._wait_for_job_complete(conn, None, syncName=syncName)

    def _get_max_job_retries(self, extraSpecs):
        """Get max job retries either default or user defined
        """
        if extraSpecs and 'storagetype:retries' in extraSpecs:
            jobRetries = super(
                EMCVMAXUtilsExtension, self)._get_max_job_retries(extraSpecs)
        else:
            jobRetries = CONF.vmax_job_retries  # fall-back to old option
        return int(jobRetries)

    def _get_interval_in_secs(self, extraSpecs):
        """Get interval in secs, either default or user defined
        """
        if extraSpecs and 'storagetype:interval' in extraSpecs:
            intervalInSecs = super(
                EMCVMAXUtilsExtension, self)._get_interval_in_secs(extraSpecs)
        else:
            intervalInSecs = CONF.vmax_job_interval  # fall-back to old option
        return int(intervalInSecs)

    def get_storage_group_from_volume(self, conn, volumeInstanceName,
                                      gtype=None, sg_inst_name=None):
        """Override: Returns the storage group for a particular volume.

        Given the volume instance name get the associated storage group if it
        is belong to one.

        :param conn: connection to the ecom server
        :param volumeInstanceName: the volume instance name
        :param gtype: The storage group type. If None, then the first
                      associated SG is returned. If 'mask', then the first
                      default FAST policy group for the vol is returned.
                      If sg_inst_name is not None, this param is ignored.
        :param sg_inst_name: If not None, then loop through associated
                      SG instances and return the matching one, ignorning
                      gtype. If no match, return None.
        :returns: foundStorageGroupInstanceName
        """
        foundStorageGroupInstanceName = None

        storageGroupInstanceNames = conn.AssociatorNames(
            volumeInstanceName,
            ResultClass='CIM_DeviceMaskingGroup')

        for sg in storageGroupInstanceNames:
            # TODO: csky override - remove method when fixed in community
            if sg_inst_name:
                if sg_inst_name == sg:
                    foundStorageGroupInstanceName = sg
                    break
            elif gtype is None:
                foundStorageGroupInstanceName = sg
                break
            elif gtype == 'fast' and 'OS_default_' in sg['InstanceID']:
                foundStorageGroupInstanceName = sg
                break
            elif gtype == 'mask' and 'OS_default_' not in sg['InstanceID']:
                foundStorageGroupInstanceName = sg
                break
            LOG.debug("Skipping storage group %s." % sg['InstanceID'])
        found_name = foundStorageGroupInstanceName['InstanceID'] if\
            foundStorageGroupInstanceName else None
        LOG.debug("Returning storage group '%s'. gtype=%s, sg_inst_name="
                  "%s.  Number associated with vol: %d." %
                  (found_name, gtype, sg_inst_name,
                   len(storageGroupInstanceNames)))
        return foundStorageGroupInstanceName

    def parse_pool_instance_id(self, poolInstanceId):
        """ First try V3 formatting, then V2. """
        pool, system = super(EMCVMAXUtilsExtension,
                             self).parse_pool_instance_id_v3(poolInstanceId)
        LOG.debug("PowerVC insert v3 attempt: %s, %s " % (pool, system))
        if pool is None or system is None:
            return super(EMCVMAXUtilsExtension,
                         self).parse_pool_instance_id(poolInstanceId)
        return pool, system

    def get_pool_capacities(self, conn, poolName, storageSystemName):
        """PowerVC override to support config option of thin free capacity
        """
        if CONF.vmax_restrict_pool_to_xml_conf and CONF.vmax_use_thin_capacity:
            poolInstanceName = self.get_pool_by_name(
                conn, poolName, storageSystemName)
            if poolInstanceName is None:
                LOG.error(_LE(
                    "Unable to retrieve pool instance of %(poolName)s on "
                    "array %(array)s."),
                    {'poolName': poolName, 'array': storageSystemName})
                return (0, 0)
            storagePoolInstance = conn.GetInstance(
                poolInstanceName, LocalOnly=False)
            total_capacity_gb = self.convert_bits_to_gbs(
                storagePoolInstance['TotalManagedSpace'])
            free_capacity_gb = self.convert_bits_to_gbs(
                storagePoolInstance['RemainingManagedSpace'])
            free_capacity_gb = max(free_capacity_gb, 0)
            LOG.debug("PowerVC thin config option. Free cap for pool %s "
                      "is %d" % (poolName, free_capacity_gb))
            return (total_capacity_gb, free_capacity_gb)
        else:
            return super(EMCVMAXUtilsExtension, self).get_pool_capacities(
                conn, poolName, storageSystemName)


# *****************************************************
# **            VMAX FAST Extension                  **
# *****************************************************
class EMCVMAXFastExtension(emc_vmax_fast.EMCVMAXFast):
    """ Override for EMC VMAX FAST utilities """
    # TODO: csky override - remove method when fixed in community
    #                       LP 1435069
    def __init__(self, prtcl):
        super(EMCVMAXFastExtension, self).__init__(prtcl)

    def get_and_verify_default_storage_group(
            self, conn, controllerConfigService, volumeInstanceName,
            volumeName, fastPolicyName):
        """Retrieves and verifies the default storage group for a volume.

        Given the volumeInstanceName get any associated storage group and
        check that it is the default storage group. The default storage group
        should have been already created. If not found error is logged.

        :param conn: the connection to the ecom server
        :param controllerConfigService: the controller config service
        :param volumeInstanceName: the volume instance name
        :param volumeName: the volume name (String)
        :param fastPolicyName: the fast policy name (String)
        :returns: foundDefaultStorageGroupInstanceName
        """
        foundDefaultStorageGroupInstanceName = None
        storageSystemInstanceName = self.utils.find_storage_system(
            conn, controllerConfigService)

        if not self._check_if_fast_supported(conn, storageSystemInstanceName):
            LOG.error(_LE(
                "FAST is not supported on this array."))
            raise
        # csky - use gtype in override for FAST only
        #        We assume it can be a member of only a single default FAST.
        assocStorageGroupInstanceName = (
            self.utils.get_storage_group_from_volume(conn, volumeInstanceName,
                                                     gtype='fast'))
        defaultSgGroupName = ('OS_default_' + fastPolicyName + '_SG')
        defaultStorageGroupInstanceName = (
            self.utils.find_storage_masking_group(conn,
                                                  controllerConfigService,
                                                  defaultSgGroupName))
        if defaultStorageGroupInstanceName is None:
            LOG.error(_LE(
                "Unable to find default storage group "
                "for FAST policy : %(fastPolicyName)s."),
                {'fastPolicyName': fastPolicyName})
            raise

        if assocStorageGroupInstanceName == defaultStorageGroupInstanceName:
            foundDefaultStorageGroupInstanceName = (
                assocStorageGroupInstanceName)
        else:
            LOG.warn(_LW(
                "Volume: %(volumeName)s Does not belong "
                "to storage storage group %(defaultSgGroupName)s."),
                {'volumeName': volumeName,
                 'defaultSgGroupName': defaultSgGroupName})
        return foundDefaultStorageGroupInstanceName

    def add_volume_to_default_storage_group_for_fast_policy(
            self, conn, controllerConfigService, volumeInstance,
            volumeName, fastPolicyName, extraSpecs):
        """Add a volume to the default storage group for FAST policy.

        The storage group must pre-exist.  Once added to the storage group,
        check the association to make sure it has been successfully added.

        :param conn: the ecom connection
        :param controllerConfigService: the controller configuration service
        :param volumeInstance: the volume instance
        :param volumeName: the volume name (String)
        :param fastPolicyName: the fast policy name (String)
        :param extraSpecs: additional info
        :returns: assocStorageGroupInstanceName - the storage group
            associated with the volume
        """
        failedRet = None
        defaultSgGroupName = ('OS_default_' + fastPolicyName + '_SG')
        storageGroupInstanceName = self.utils.find_storage_masking_group(
            conn, controllerConfigService, defaultSgGroupName)
        if storageGroupInstanceName is None:
            LOG.error(_LE(
                "Unable to create default storage group for "
                "FAST policy : %(fastPolicyName)s."),
                {'fastPolicyName': fastPolicyName})
            return failedRet
        # csky - use try/except to tolerate if vol is already in the group.
        try:
            self.provision.add_members_to_masking_group(
                conn, controllerConfigService, storageGroupInstanceName,
                volumeInstance.path, volumeName, extraSpecs)
        except Exception as ex:
            LOG.debug("Exception may be due to volume already being a member"
                      ": %s " % six.text_type(ex))
        # Check to see if the volume is in the storage group.
        # csky - use gtype in override
        assocStorageGroupInstanceName = (
            self.utils.get_storage_group_from_volume(conn,
                                                     volumeInstance.path,
                                                     gtype='fast'))
        LOG.debug("Returning volumes FAST group: %s." %
                  assocStorageGroupInstanceName['InstanceID'])
        return assocStorageGroupInstanceName
