# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import sys
import base64

from cinder import context
from cinder import exception
from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import discovery_driver
from cinder.volume.drivers.ibm import xiv_ds8k
from cinder.volume import volume_types
from powervc_cinder import _
from oslo_log import log as logging
from oslo_utils import excutils

from oslo_utils import importutils
from powervc_cinder import exception as powervc_exception
from cinder import db as db_api

from powervc_cinder.zonemanager.powervc_fc_zone_manager import \
    PowerVCZoneManager

LOG = logging.getLogger(__name__)

RESTRICTED_METADATA_VDISK_ID_KEY = "vdisk_id"
RESTRICTED_METADATA_VDISK_UID_KEY = "vdisk_uid"
RESTRICTED_METADATA_VDISK_NAME_KEY = "vdisk_name"
RESTRICTED_METADATA_BYTE_SIZE_KEY = "byte_size"
RESTRICTED_METADATA_NAA = "wwn"
METADATA_VOLUME_WWN = "volume_wwn"

from xiv_ds8k_openstack.utils.powervc_helper import XIVPowerVCHelpers


###############################################
#####  Temporary Workaround for XIV Bug  ######
###############################################
def _get_xiv_exception_class(self):
    """Temporary Workaround for XIV bug loading exception class"""
    return exception.CinderException

XIVPowerVCHelpers._get_exception = _get_xiv_exception_class


class XIVProductDriver(
        discovery_driver.VolumeDiscoveryDriver,
        xiv_ds8k.XIVDS8KDriver):

    def __init__(self, *args, **kwargs):
        super(XIVProductDriver, self).__init__(*args, **kwargs)
        self.endpoint_desc = "XIVProductDriver provider"

    def _build_default_opts(self):
        default_opts = {'storage_pool': self.configuration.san_clustername}
        return default_opts

    def create_volume(self, volume):
        vol = super(XIVProductDriver, self).create_volume(volume)

        ctxt = context.get_admin_context()
        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                #Update db to preserve volume_type
                LOG.info(_('adding volume_type_id \
                    to volume=%s') % volume)
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})

        #Add restricted metadata to volume
        wwn = self._get_wwn(volume)
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: volume['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: volume['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: volume['name'],
                    RESTRICTED_METADATA_NAA: wwn
                    }
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, volume['id'],
                                                        metadata)
        if vol is None:
            vol = {}
        meta = db_api.volume_metadata_get(ctxt, volume['id'])
        meta = meta if meta else dict()
        meta[METADATA_VOLUME_WWN] = wwn
        vol['metadata'] = meta
        return vol

    def _get_wwn(self, volume):
        volume_type_id = volume.get('volume_type_id')
        pool = volume_types.get_volume_type_extra_specs(volume_type_id,
                                                        "drivers:storage_pool")
        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        proxy = importutils.import_class(
            'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
        helpers = proxy()
        context = {'user': x_userid, 'password': x_password,
                   'address': x_ip, 'pool': pool, 'volume': volume['name']}
        specific_volume = helpers.get_volume(context)
        wwn = specific_volume['wwn']
        LOG.info(_('getting wwn: %s') % wwn)
        return wwn

    def create_cloned_volume(self, tgt_volume, src_volume):
        """
        Overrides the superclass create_cloned_volume. Adds
        restricted metadata information to the cloned volume
        and sets default volume type if one is not specified
        """
        s = super(XIVProductDriver, self)

        ctxt = context.get_admin_context()
        volume_type_id = tgt_volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                tgt_volume['volume_type_id'] = volume_type_id
                #Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % tgt_volume)
                self.db.volume_update(ctxt, tgt_volume['id'],
                                      {'volume_type_id': volume_type_id})

        vol = s.create_cloned_volume(tgt_volume, src_volume)
        #Add restricted metadata to volume
        wwn = self._get_wwn(tgt_volume)
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: tgt_volume['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: tgt_volume['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: tgt_volume['name'],
                    RESTRICTED_METADATA_NAA: wwn
                    }

        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, tgt_volume['id'],
                                                        metadata)
        if vol is None:
            vol = {}
        meta = db_api.volume_metadata_get(ctxt, tgt_volume['id'])
        meta = meta if meta else dict()
        meta[METADATA_VOLUME_WWN] = wwn
        vol['metadata'] = meta
        return vol

    def discover_volumes(self, context, filters=None):
        """Method to Retrieve a list of volumes for a host
        Currently supported filters:

         'host_refs': [ WWPN, ... ]
           Returns only volumes that are mapped to at least one of the WWPNS in
           the specified list.  Matching is case-insensitive.
        """
        LOG.debug("discover_volumes filters %s" % filters)
        # If we are filtering, create a set from the WWPN list to make
        # filtering easier.  Convert the WWPNs to upper-case for consistency
        filter_set = None
        upper_refs = None
        if filters is not None and 'host_refs' in filters:
            upper_refs = [ref.upper() for ref in filters['host_refs']]
            filter_set = set(upper_refs)

        # If we are filtering get the volume references to get the k2udid
        vol_refs = None
        if filters is not None and 'vol_refs' in filters:
            vol_refs = filters['vol_refs']

        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        proxy = importutils.import_class(
            'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
        helpers, volumes = (proxy(), dict())
        context = {'user': x_userid, 'password': x_password, 'address': x_ip}
        for pool in helpers.get_pools_list(context).values():
            context2 = dict(context, pool=pool['name'])
            volumes.update(helpers.get_volume_list(context2))
        connection_info = helpers.get_hosts_list(context)
        return_volumes = []
        for key, vol in volumes.items():
            volume = {}
            volume['name'] = key
            volume['wwn'] = vol['wwn']
            volume['storage_pool'] = vol['pool']
            volume['status'] = 'available'
            volume['size'] = int(vol['size_in_bytes']) / 1000000000
            vdisk_uid = vol['name']
            volume['restricted_metadata'] = {
                RESTRICTED_METADATA_NAA: vol['wwn'],
                RESTRICTED_METADATA_VDISK_ID_KEY: vol['name'],
                RESTRICTED_METADATA_VDISK_UID_KEY: vdisk_uid,
                RESTRICTED_METADATA_VDISK_NAME_KEY: vol['name'],
                RESTRICTED_METADATA_BYTE_SIZE_KEY:
                int(vol['size_in_bytes']) / 1000000000
            }
            volume['metadata'] = {METADATA_VOLUME_WWN: vol['wwn']}
            volume['support'] = {"status": "supported"}
            volume['mapped_wwpns'] = vol['wwpns']
            if filter_set is not None:
                    rmt_wwpns = set([wwpn.upper() for wwpn in vol['wwpns']])
            if len(vol['wwpns']) > 0:
                volume['status'] = 'in-use'
                volume['support'] = {
                    "status": "not_supported",
                    "reasons": [
                    'This volume is not a candidate for management '
                    'because it is already attached to a virtual '
                    'machine.  To manage this volume with PowerVC, '
                    'you must bring the virtual machine under '
                    'management.  Select to manage the virtual '
                    'machine that has the volume attached.  The '
                    'attached volume will be automatically included '
                    'for management.'
                    ]
                }
                conninfo = {}
                for k, v in connection_info.items():
                    hwwpns = v['host_wwpns']
                    for w in vol['wwpns']:
                        if w in hwwpns:
                            conninfo[k] = v
                            volume['connection_info'] = conninfo

            # Store the volume k2udid
            if vol_refs:
                for vol_idx in vol_refs:
                    wwn = vol['wwn'].upper()
                    if wwn in vol_idx['uid']:
                        volume['k2udid'] = vol_idx['k2udid']
                        break
            # If we are not filtering on WWPNs or UID's, or we ARE filtering
            # and there is a match, add this disk to the list to be returned.
            if ((filter_set is None or len(filter_set & rmt_wwpns) > 0 or
                 vdisk_uid in filter_set)):
                return_volumes.append(volume)

            # This is to support onboarding of vSCSI VM.
            # get wwn from vol and match it in host_ref
            # for vSCSI filter looks like :
            # filters={'host_refs':[u'261120017380075B012A7072810XIV03IBMfcp',
            #                       u'261120017380075B01458072810XIV03IBMfcp']}
            # vol['wwn'] looks like '0017380075B01458'
            if not upper_refs is None:
                for ref in upper_refs:
                    # upper_refs are upper case, convert wwn too
                    wwn = vol['wwn'].upper()
                    if wwn in ref:
                        if not volume in return_volumes:
                            return_volumes.append(volume)
                            break
        LOG.debug("discover_volumes return %s" % return_volumes)
        return return_volumes

    def query_volumes(self, context, volumes, server_info={}, mark_boot=True):
        LOG.debug("Enter: query_volumes %(vol)s %(srv_info)s" %
                  dict(vol=volumes, srv_info=server_info))
        uid_map, wwpn_map = (dict(), dict())
        returned_volumes, boot_map = (list(), dict())

        for server_id, server in server_info.iteritems():
            mark_boot = server.get('boot') is not True
            for volume in server.get('volumes', []):
                uid = volume.get('unique_device_id')
                wwpns = volume.get('wwpns', [])
                vios_wwpn = volume.get('vios_wwpns', None)
                #Build maps of the UID and WWPN depending on which is set
                if uid is not None:
                    if uid.upper() not in uid_map:
                        server_dict = {server_id: (mark_boot, vios_wwpn)}
                    else:
                        server_dict = uid_map[uid.upper()]
                        server_dict[server_id] = (mark_boot, vios_wwpn)
                    uid_map[uid.upper()] = server_dict
                for wwpn in wwpns:
                    wwpn_map[wwpn.upper()] = (server_id, mark_boot)
        LOG.info("uid_map %s wwpn_map %s" % (uid_map, wwpn_map))
        for volume in volumes:
            volume_copy = volume.copy()
            server_id, mark_boot = (None, False)
            #Parse the UID out of the MetaData to use for matching Servers
            metadata = volume_copy.get('restricted_metadata', {})
            uid = metadata.get(RESTRICTED_METADATA_VDISK_UID_KEY)
            vdisk_id = metadata.get(RESTRICTED_METADATA_VDISK_ID_KEY)
            udid = None
            wwn = volume_copy['wwn']
            if 'k2udid' in volume_copy.keys():
                udid = volume_copy['k2udid']
            volume_copy['attach_info'] = {}
            #Parse the correct Connection Info based on the WWPN's
            connect_infos = volume_copy.get('connection_info', {})
            if len(connect_infos) > 0:
                for cinfo in connect_infos.values():
                    wwpns = cinfo.get('host_wwpns', [])
                    #Loop through each of the WWPN's looking for the Server
                    for wwpn in wwpns:
                        #If we found a Server for the WWPN, use this Info
                        if wwpn_map.get(wwpn.upper()) is not None:
                            server_id = wwpn_map[wwpn.upper()][0]
                            if server_id not in volume_copy['attach_info']:
                                mark_boot = wwpn_map[wwpn.upper()][1]
                                connection_info = None
                                #sanitize connection info for npiv
                                if cinfo is not None:
                                    cinfo['target_wwn'] = cinfo.pop(
                                        'storage_wwpns', None)
                                    conn_type = \
                                        discovery_driver.CONNECTION_TYPE_NPIV
                                    connection_info =\
                                        self._get_connection_info(
                                            uid, cinfo, conn_type)
                                    self._add_attach_info(
                                        volume_copy, boot_map,
                                        connection_info, server_id,
                                        mark_boot=mark_boot)

            wwnFound = None
            for uidkey in uid_map.keys():
                if wwn in uidkey:
                    wwnFound = uidkey

            if wwnFound:
                server_dict = uid_map[wwnFound]
                for server_id in server_dict.keys():
                    if server_id not in volume_copy['attach_info']:
                        mark_boot = server_dict[server_id][0]
                        udid = volume.get('k2udid')
                        connection_info = None
                        server_wwpn = server_dict[server_id][1]
                        #get correct connection info and itl list
                        valid_itls = self._get_vscsi_itl_info(server_wwpn,
                                                              volume_copy)
                        if valid_itls:
                            LOG.info(_("XIV Query Volume Valid ITLS=%s") %
                                     valid_itls)
                            #If given the VDisk UID from the VM, then
                            #it isn't NPIV
                            #pass the first itl to extract the target_wwn
                            cinfo = valid_itls[0]
                            cinfo['target_wwn'] = cinfo.get('target_wwpn',
                                                            None)
                            conn_type = discovery_driver.CONNECTION_TYPE_VSCSI
                            connection_info = self._get_connection_info(
                                uid, cinfo, conn_type, udid=udid)
                            connection_info['vscsi_target_lun'] =\
                                valid_itls
                            connection_info['data']['vendor'] =\
                                "IBM"
                            self._add_attach_info(volume_copy, boot_map,
                                                  connection_info,
                                                  server_id,
                                                  mark_boot=mark_boot)
            #Add the Copy of the Volume to the list to be returned
            volume_copy.pop('connection_info', None)
            returned_volumes.append(volume_copy)
        LOG.debug("exit: query_volumes %s" % returned_volumes)
        return returned_volumes

    def _get_vscsi_itl_info(self, server_infos, volume):
        """
        Return the vsci connection info for a given volume
        and given instance
        """
        valid_itls = []
        itl_list = volume.get('connection_info', [])
        if itl_list:
            for cinfo in itl_list.values():
                itl = {}
                itl['vios_wwpn'] = cinfo.get('host_wwpns', None)
                itl['target_wwpn'] = cinfo.get('storage_wwpns', None)
                if 'luns' in cinfo:
                    itl['target_lun'] = cinfo.get('luns')
                for server_wwpns in server_infos.values():
                    #make the wwpn to upper
                    vol_wwpns = itl.get('vios_wwpn', [])
                    if vol_wwpns:
                        vol_wwpns = [wwpn.upper() for wwpn in vol_wwpns]
                        server_wwpns = [wwpn.upper() for wwpn in server_wwpns]
                        common_wwpn = list(set(vol_wwpns) & set(server_wwpns))
                        if common_wwpn and len(common_wwpn) > 0:
                            valid_itls.append(itl)
            #END FOR-Loop
        return valid_itls

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        # Run the vol_list command to query the list of volumes on the XIV
        existvols = self.xiv_ds8k_proxy._call_xiv_ds8k_cli("vol_list").as_list
        volnames = dict([(v['name'], '') for v in existvols])
        # Now loop through the passed in volumes seeing which don't exist
        return [v for v in volumes if v['name'] not in volnames]

    def _decode_password(self, password):
        """Shared Helper Method to specify the deCryption method to use"""
        return base64.b64decode(password)

    def _encode_password(self, password):
        """Shared Helper Method to specify the encryption method to use"""
        return base64.b64encode(password)

    def get_storage_metadata(self):
        """Returns the additional information that PowerVC uses to populate
        various UI elements.
        """
        compression = False

        pools = self.xiv_ds8k_proxy._call_xiv_ds8k_cli("pool_list")

        def transform_pool_structure(pool):
            return {'name': pool.name,
                    'capacity': str(pool.soft_size) + " GB",
                    'free_capacity': str(pool.empty_space_soft) + " GB"}
        volume_pools = map(transform_pool_structure, pools)

        return {'volume_pools': volume_pools,
                'compression_supported': compression}

    def get_default_vol_type(self):
        """Returns the default volume type from the .conf file"""
        vtn = self.configuration.default_volume_type
        return vtn.decode('utf-8') if vtn else vtn

    def initialize_connection(self, volume, connector):
        """
        Calls the parent class to mask the volume on the storage controller,
        and here performs additional processing to work out which zones should
        be created on the switches.

        For NPIV, the virtual fibre channel ports will not yet be logged in,
        but we need to know which fabrics which ports will appear on, so that
        we know which fabrics to create the zones on.  For this reason, the
        connector must also include a data structure that maps physical WWPNs
        (which will be logged in) to the virtual WWPNs that they will provide.
        To support live partition mobility, we expect each physical WWPN to be
        providing two virtual WWPNs.

        The expected structure of a fibre channel connector is:

        {
            'host': hostname,
            'wwpns': [ list of virtual WWPNs ]  <-- this is used for masking
            'phy_to_virt_initiators': {phy_wwpn_X: [virt_wwpn_a, virt_wwpn_b],
                                       phy_wwpn_Y: [virt_wwpn_c, virt_wwpn_d]
                                      }
        }

        The expected structure of an iSCSI connector is:

        {
          'host': hostname,
          'initiator': initiator  <-- IQN format
        }
        """

        # Ensure that the passed-in data structure is correct.
        expected_fc_keys = set(['host', 'wwpns'])
        expected_iscsi_keys = set(['host', 'initiator'])
        if not (expected_fc_keys.issubset(connector) |
                expected_iscsi_keys.issubset(connector)):
            err_msg = (_('The connector %(connector)s does not contain the '
                         'required keys, which are %(expected_fc_keys)s.'
                         'or %(expected_iscsi_keys)s') %
                       {'connector': connector,
                        'expected_fc_keys': expected_fc_keys,
                        'expected_iscsi_keys': expected_iscsi_keys})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        # Need to sanitize the host, since XIV doesn't handle many characters
        if connector.get('host'):
            connector['host'] = sanitize_name(connector['host'])
        # Perform the masking on the storage controller
        s = super(XIVProductDriver, self)
        info = s.initialize_connection(volume, connector)
        info['data']['vendor'] = "IBM"
        if connector['connection_type']:
            if 'npiv' in connector['connection_type']:
                info['data']['connector_type'] = 'npiv'
            elif 'pv_vscsi' in connector['connection_type']:
                info['data']['connector_type'] = 'pv_vscsi'

        #A new field has been added to info data to return all
        #target wwpns.  We use that to poplulate target_wwn if it
        #returns empty from initialize_connection()
        if info['data']['all_storage_wwpns'] and \
           not info['data']['target_wwn']:
            info['data']['target_wwn'] = \
                info['data']['all_storage_wwpns']

        # LUN ID SAM format standard uses 8 bytes, if the lun id is greater
        # than 255(1 byte) then format needs to be changed for the top 2 bits
        # of lun id. Top two bits should be set to 0x4000 (16384) followed by
        # 8 byte Lun id
        # for eg if lun=256, will be passed 0x4100 = (0x4000 + 0x100) down
        # we will update the value to use the SAM format for setting the
        # So the lun id value will be set as lun id = 16384 + lun_id
        if info['data']['target_lun'] > 255:
            info['data']['target_lun'] += 16384

        # Perform zoning if the required data was specified
        # No zoning will be performed for vSCSI
        if not 'pv_vscsi' in connector['connection_type']:
            if 'phy_to_virt_initiators' in connector:
                try:
                    self._do_zoning(connector,
                                    info['data']['initiator_target_map'])
                    # Remove the initator_target_map so that the community
                    # zoning code won't be triggered
                    del info['data']['initiator_target_map']

                except Exception as e:
                    # Something went wrong during zoning, make sure that the
                    # disk is unmapped and any host that we created is
                    #  deleted before we return
                    LOG.exception(e)
                    with excutils.save_and_reraise_exception():
                        s.terminate_connection(volume, connector)

        return {'driver_volume_type': 'fibre_channel', 'data': info['data']}

    def _do_zoning(self, connector, initiator_target_map):

        # Our objective is to build a data structure like this, which (for each
        # fabric) maps initiators to targets.  This allows us to perform zoning
        # operations in parallel across both fabrics.
        #
        #
        # { 'a':  { 'initiator_A': [ target_W, target_X ],
        #           'initiator_B': [ target_W, target_X ]
        #         }
        #   'b':  { 'initiator_C': [ target_Y, target_Z ],
        #           'initiator_D': [ target_Y, target_Z ]
        #         }
        # }
        #
        # This means that we must work out which initiators and targets
        # are on which fabrics.  This process is further complicated because
        # NPIV means that the virtual initiator ports (which are the things
        # that we want to zone) aren't yet logged into the fabric.  What we do
        # have is the phy_to_virt_initiators map, which tells us which physical
        # ports are going to support which virtual ports.  We can therefore
        # see which physical initiator ports are on the fabric, and use that to
        # know which virtual initiator ports will end up on the same fabric.

        # Construct a Zone manager to interact with the fabrics
        zonemanager = PowerVCZoneManager(self.configuration)

        # Sanitize the initiator_targe_map to be lower case, no colons
        new_it_map = {}
        for initiator, targets in initiator_target_map.items():
            new_it_map[strip_colons(initiator)] = map(strip_colons, targets)
        initiator_target_map = new_it_map

        # Extract the map of physical-to-virtual WWPNs, sanitising to all
        # lowercase with no colons, and also build a list of physical WWPNs.
        phy_to_virt_initiators = connector['phy_to_virt_initiators']

        phy_to_virt_map = {}
        for phy, virt in phy_to_virt_initiators.items():
            phy_to_virt_map[strip_colons(phy)] = map(strip_colons, virt)
        phy_initiators = phy_to_virt_map.keys()

        target_wwpns = set()
        for targets in initiator_target_map.values():
            for target in targets:
                target_wwpns.add(strip_colons(target))

        mapping = self._get_fabric_mapping(zonemanager, phy_initiators,
                                           list(target_wwpns))

        # We now know which physical initiators and targets are on which fabric
        # We need to filter the initiator_target_map

        zones = {}
        LOG.debug("GMN:\n"
                  " mapping is %(mapping)s\n"
                  " phy_to_virt_map is %(phy_to_virt_map)s\n"
                  " initiator_target_mapis%(initiator_target_map)s" % locals())

        for fabric_name, wwpn_lists in mapping.iteritems():

            # Make a set of all the target ports on this fabric.
            target_fabric_set = set(wwpn_lists['target_port_wwn_list'])

            # For every physical initiator found on the fabric...
            for phy_initiator in wwpn_lists['initiator_port_wwn_list']:

                # ...for every virtual initiator that will be hosted by that
                # physical initiator...
                for virt_initiator in phy_to_virt_map[phy_initiator]:

                    # Get the list of target ports that we might want to zoneto
                    if virt_initiator in initiator_target_map:
                        target_zone_set = set(initiator_target_map
                                              [virt_initiator])

                        # Filter the ports to only those on the fabric
                        filtered_target_ports = (target_zone_set &
                                                 target_fabric_set)

                        # If there is at least one target port in thesamefabric
                        # as the initiator
                        if filtered_target_ports:
                            if not fabric_name in zones:
                                zones[fabric_name] = {}

                            zones[fabric_name][virt_initiator] = \
                                list(filtered_target_ports)

        # Now that we have a data structure describing the zones, zone it.
        LOG.info(_("Creating zones: %(zones)s") % {'zones': zones})
        host = connector['host']
        zonemanager.add_connection(host, zones)
        zonemanager.close()

    # Given a list of physical initiator wwpns, and a list of target wwpns,
    # asks a passed-in SAN lookup service to tell is which of those wwpns are
    # visible on which fabrics.
    def _get_fabric_mapping(self, zone_manager, phy_initiator_wwpns,
                            target_wwpns):
        LOG.info(_('Requested mapping for initiators  %(initiators)s'
                   ' and targets %(targets)s')
                 % {'initiators': phy_initiator_wwpns,
                    'targets': target_wwpns})

        # The brocade code only works if we add colons to the incoming WWPNs.
        initiator_colons = map(add_colons, phy_initiator_wwpns)
        target_colons = map(add_colons, target_wwpns)

        # Produce a mapping that shows which of the specified initiator and
        # target wwpns are logged into which switches.
        #               'a': [...]}
        #               'b': [...]}
        wwpns = {'initiator_port_wwn_list':  initiator_colons,
                 'target_port_wwn_list': target_colons}

        mapping = zone_manager.get_san_context(wwpns)

        LOG.info(_("Returned mapping is %(mapping)s")
                 % {'mapping': mapping})

        # Remove any colons that might be in the returned structure
        for fabric_name, wwpn_lists in mapping.iteritems():
            for name, wwpns in wwpn_lists.iteritems():
                wwpn_lists[name] = \
                    map(strip_colons, wwpn_lists[name])

        LOG.info(_("Returned mapping is %(mapping)s")
                 % {'mapping': mapping})

        return mapping

    def terminate_connection(self, volume, connector, **kwargs):

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        LOG.info(_("Terminating connection with parameters %(connector)s" %
                   {'connector': connector}))

        # Perform the masking on the storage controller
        s = super(XIVProductDriver, self)
        s.terminate_connection(volume, connector, **kwargs)

        # Test to see if we should remove zones on the fabrics
        if 'phy_to_virt_initiators' in connector:
            if connector.get('clean_zones', None) == "all" and \
               self.configuration.zoning_mode == 'fabric':
                LOG.info(_("Removing zones for initiators:"
                           "%(initiator_list)s") % {'initiator_list': ',\
                           '.join(connector['wwpns'])})
                # Create a zonemanager and ask it to remove zones for these
                # initiators
                try:
                    zonemanager = PowerVCZoneManager(self.configuration)
                    zonemanager.delete_connection(map(add_colons,
                                                      connector['wwpns']))
                except Exception as ex:
                    LOG.exception(ex)
                    LOG.error(_("There was a failure when the zone manager "
                                "attempted to remove zones created for WWPNs "
                                "%(wwpns)s during detach of volume %(vol)s.") %
                              (connector['wwpns'], volume['id']))

    def check_volume_health(self, volume):
        """
        Throws exceptions if something is wrong with the passed-in volume.

        Checks for existence of the underlying vdisk
        """
        #The list of volumes from the call to get_volume_list()
        #is used a couple of times.  Collect the info here to
        #avoid redundant calls.
        #Also create a new internal helper method to simplify the call.
        volumes = self._get_volume_list()

        vdisk_id = self._incorporate_restricted_metadata(volume, volumes)

        if not self._check_vdisk_exists(vdisk_id, volumes):
            raise powervc_exception.XIVVdiskNotFoundException(
                self.endpoint_desc,
                volume['id'], vdisk_id=vdisk_id)

    def _check_vdisk_exists(self, vdisk_id, volumes):
        for key, vol in volumes.items():
            if vol['existing_ref'] == vdisk_id:
                return True
        return False

    def _incorporate_restricted_metadata(self, volume_ref, volumes):
        """Query the Cinder database to retrieve any access-restricted
           metadata and attach it to the volume data.
           Returns the vdisk id, which most of the time is what we were after.
           If there is no restricted metadata, tries to find the disk and
           creates it.
        """
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume_ref['id'])
        # If the vdisk ID is not in the metadata, this could be because of an
        # upgrade from a version that didn't use it.  We try to find it by
        # looking for a volume on the SAN with a matching name
        if RESTRICTED_METADATA_VDISK_ID_KEY not in metadata:
            # Previous versions would have used this as the name on the SAN
            vdisk = self._find_vdisk_by_name(volume_ref['name'], volumes)
            if vdisk is not None:
                self._update_restricted_metadata(volume_ref['id'], vdisk)
                metadata = powervc_db_api.\
                    volume_restricted_metadata_get(ctxt, volume_ref['id'])
            else:
                return None

        volume_ref['restricted_metadata'] = metadata
        return metadata[RESTRICTED_METADATA_VDISK_ID_KEY]

    def _find_vdisk_by_name(self, vdisk_name, volumes):
        """ Look through the list of volumes for a match.
            For XIV, the only attribute needed is the name which
            will be used for the metadata update.
        """
        for key, vol in volumes.items():
            if vol['existing_ref'] == vdisk_name:
                attributes = {}
                attributes['name'] = vdisk_name
                return attributes
        return None

    def _update_restricted_metadata(self, volume_id, vdisk):
        """Perform a database update to associate the specified information
           with the specified volume.
        """
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: vdisk['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: vdisk['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: vdisk['name']
                    }
        ctxt = context.get_admin_context()
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, volume_id,
                                                        metadata)

    def _get_volume_list(self):
        """ Helper method to consolidate info for the
            XIV helper method call to get_volume_list()
            Returns the list of volumes.
        """
        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        x_pool = self.configuration.san_clustername
        proxy = importutils.import_class(
            'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
        helpers = proxy()
        context = {'user': x_userid, 'password': x_password,
                   'address': x_ip, 'pool': x_pool}
        volumes = helpers.get_volume_list(context)
        return volumes


def add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in range(0, len(wwpn), 2)]).lower()


def strip_colons(wwpn):
    wwpn = wwpn.lower()
    if len(wwpn) > 16:
        return wwpn.replace(":", "")
    else:
        return wwpn


###################################################
######  Override Create Volume on XIV Proxy  ######
###################################################
#Since the XIV Driver doesn't currently provide a way to specify a
#different pool on create_volume than what is the default specified
#when registered, we need to override the proxy to allow that pool
#to be specified until the XIV team provides support in the driver.
__import__('powervc_oslo.xiv_xsf')
gigabytes_to_blocks = importutils.import_class(
    'xiv_ds8k_openstack.gigabytes_to_blocks')
_trace_time = importutils.import_class(
    'xiv_ds8k_openstack.nova_proxy._trace_time')
XIVNovaProxy = importutils.import_class(
    'xiv_ds8k_openstack.xiv_nova_proxy.XIVNovaProxy')
SystemOutOfSpaceError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.SystemOutOfSpaceError')
PoolOutOfSpaceError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.PoolOutOfSpaceError')
XCLIError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.XCLIError')
#Load the various constant string literals that we need to use from XIV
CREATE_VOLUME_BASE_ERROR = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_BASE_ERROR')
CREATE_VOLUME_SYSTEM_OUT_OF_SPACE = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_SYSTEM_OUT_OF_SPACE')
CREATE_VOLUME_POOL_OUT_OF_SPACE = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_POOL_OUT_OF_SPACE')
#We need to import the sanitize name to use for replacing the host
sanitize_name = importutils.import_class(
    'powervc_cinder.db.sqlalchemy.models._sanitize_volume_name')


class XIVProductProxy(XIVNovaProxy):

    @_trace_time
    def create_volume(self, volume):
        """Override the create_volume to specify the pool name"""
        size = gigabytes_to_blocks(float(volume['size']))
        pool = self._get_pool_from_extra_specs(volume)
        #If a Pool override wasn't specified, use the default configured
        if not pool:
            pool = self.xiv_ds8k_info.get('xiv_ds8k_vol_pool')
        try:
            self._call_xiv_ds8k_cli('vol_create', vol=volume['name'],
                                    size_blocks=size, pool=pool)
        #If the System is Out-of-Space then log and raise that error
        except SystemOutOfSpaceError:
            self._log('info', CREATE_VOLUME_BASE_ERROR,
                      details=CREATE_VOLUME_SYSTEM_OUT_OF_SPACE)
            raise self._get_exception()(
                message=CREATE_VOLUME_BASE_ERROR %
                {'details': CREATE_VOLUME_SYSTEM_OUT_OF_SPACE})
        #If the Pool is Out-of-Space then log and raise that error
        except PoolOutOfSpaceError:
            err = CREATE_VOLUME_POOL_OUT_OF_SPACE % {'pool': pool}
            self._log('info', CREATE_VOLUME_BASE_ERROR, details=err)
            raise self._get_exception()(
                message=CREATE_VOLUME_BASE_ERROR % {'details': err})
        #If it is a Generic CLI Error, then just log and raise general
        except (XCLIError) as e:
            details = self._get_code_and_status_or_message(e)
            self._log('info', CREATE_VOLUME_BASE_ERROR, details=details)
            raise self.meta.exception.VolumeBackendAPIException(
                data=CREATE_VOLUME_BASE_ERROR % {'details': details})
        return {}

    def _update_stats(self):
        """Override the Volume Stats to return all of the Pools"""
        total, free = (0, 0)
        #Call the Parent Class to retrieve most of the Statistics
        super(XIVProductProxy, self)._update_stats()
        #Retrieve the Capacity from each of the Pools and use the total
        x_userid = self.driver.configuration.safe_get('san_login')
        x_password = self.driver.configuration.safe_get('san_password')
        x_password = base64.b64decode(x_password)
        x_ip = self.driver.configuration.safe_get('san_ip')
        context = {'user': x_userid, 'password': x_password, 'address': x_ip}
        proxy = importutils.import_class(
            'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
        helpers = proxy()

        for pool in helpers.get_pools_list(context).values():
            total += int(pool['soft_size'])
            free += int(pool['free_size'])
        #Now we can actually override the Free/Total Capacity values
        self.meta.stat['total_capacity_gb'] = total
        self.meta.stat['free_capacity_gb'] = free

    def _get_pool_from_extra_specs(self, volume):
        """Helper method to get the Pool Name from the Extra Specs"""
        if volume and volume.get('volume_type_id') is not None:
            specs = volume_types.\
                get_volume_type_extra_specs(volume['volume_type_id'])
            return specs.get('drivers:storage_pool')
        return None
