# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for Host Aggregates."""

import six

import webob.exc

from oslo_config import cfg

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova.compute import api as compute_api
from oslo_log import log as logging
from nova.ibm.ego.hapolicy import constants as ego_ha_consts
from nova.ibm.ego import ego_constants as ego_consts
from nova import exception

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova import _

AUTHORIZE = extensions.extension_authorizer('compute', 'aggregates')
LOG = logging.getLogger(__name__)


class HostAggregatesControllerExtension(wsgi.Controller):
    def __init__(self):
        self.api = compute_api.AggregateAPI()

    @wsgi.extends
    def index(self, req, resp_obj):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, action='index')
        for agg in resp_obj.obj['aggregates']:
            if self._is_default_agg(agg):
                agg['powervc_default'] = True
            else:
                agg['powervc_default'] = False

    @wsgi.extends
    def create(self, req, body):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, action='create')
        yield

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, action='index')
        if self._is_default_agg(resp_obj.obj['aggregate']):
            resp_obj.obj['aggregate']['powervc_default'] = True
        else:
            resp_obj.obj['aggregate']['powervc_default'] = False

    @wsgi.extends
    def update(self, req, id, body):
        """Check specific action permissions since OpenStack doesn't."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, action='update')
        try:
            agg = self.api.get_aggregate(ctxt, id)
            if self._is_default_agg(agg):
                # Make sure they aren't trying to change the default
                # aggregate's name.
                if len(body) != 1:
                    raise webob.exc.HTTPBadRequest()
                try:
                    updates = body['aggregate']
                except KeyError:
                    raise webob.exc.HTTPBadRequest()
                if len(updates) < 1:
                    raise webob.exc.HTTPBadRequest()

                agg_name = (updates['name'] if "name" in updates.keys()
                            else cfg.CONF.default_host_aggregate_name)
                if agg_name != cfg.CONF.default_host_aggregate_name:
                    msg = _("The default host aggregate's name cannot be "
                            "updated.")
                    raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.AggregateNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        yield

    @wsgi.extends
    def delete(self, req, id):
        """Prevent the default aggregate from being deleted."""
        ctxt = req.environ['nova.context']
        self._validate_authorization(ctxt, action='delete')
        try:
            agg = self.api.get_aggregate(ctxt, id)
            if self._is_default_agg(agg):
                msg = _("The default host aggregate cannot be deleted.")
                raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.AggregateNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        yield

    def _validate_add_host(self, context, agg_id, host):
        """Prevent the host from belonging to multiple host aggregates."""
        if cfg.CONF.allow_host_in_multiple_host_aggregates or host is None:
            # Just let the call proceed since we don't have to check.
            return
        if len(agg_utils.get_aggregate_ids_for_host(context, host)) >= 1:
            msg = _("The host cannot be added to more than one aggregate.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

    @wsgi.extends
    def action(self, req, id, body):
        """Check specific action permissions since OpenStack doesn't."""
        actions = ('add_host', 'remove_host', 'set_metadata')
        ctxt = req.environ['nova.context']
        for action, data in six.iteritems(body):
            if action in actions:
                if action == 'set_metadata':
                    # Because PRS uses aggregate metadata to control various
                    # behaviors, we need to traverse the keys to determine
                    # which authorizations to check.
                    for key in data.get('metadata', {}).keys():
                        if key.startswith(ego_consts.INITIAL_POLICY):
                            self._validate_authorization(
                                ctxt, 'initial_placement')
                        elif key.startswith(ego_consts.RUNTIME_POLICY):
                            self._validate_authorization(
                                ctxt, 'runtime_optimization')
                        elif key.startswith(ego_ha_consts.KEY_HAPOLICY):
                            self._validate_authorization(
                                ctxt, 'high_availability')
                        else:
                            self._validate_authorization(ctxt, action=action)
                else:
                    self._validate_authorization(ctxt, action=action)
                    if action == 'add_host':
                        # Prevent a single host from being in multiple
                        # host aggregates.
                        self._validate_add_host(ctxt, id, data.get('host'))
        yield

    @staticmethod
    def _validate_authorization(context, action):
        """Checks the user's access to invoke the action."""
        AUTHORIZE(context, action=action)

    @staticmethod
    def _is_default_agg(agg):
        """Checks whether the aggregate is the default aggregate."""
        return agg['name'] == cfg.CONF.default_host_aggregate_name


class Host_aggregates(extensions.ExtensionDescriptor):
    """Extend the Host Aggregates API."""

    name = "IBM PowerVC Host Aggregates Extension"
    alias = "ibm-powervc-os-aggregates"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_aggregates/api/v1.1"
    updated = "2014-09-17T21:00:00-06:00"

    def get_controller_extensions(self):
        controller = HostAggregatesControllerExtension()
        extension = extensions.ControllerExtension(self, 'os-aggregates',
                                                   controller)
        return [extension]
