#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import httplib
import json
import webob
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging
from oslo_config import cfg

from powervc_nova.network.ibmpowervm import host_sriov

authorize_show = extensions.extension_authorizer(
    'compute', 'host_sriov:show')
authorize_index = extensions.extension_authorizer(
    'compute', 'host_sriov:index')

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class HostSriovController(wsgi.Controller):
    """
    This Controller handles GET requests for SRIOV config on hosts.
    """

    def __init__(self):
        self.sriov_task = host_sriov.HostSriovTask()

    @wsgi.extends
    def index(self, req):
        """
        This method handles GET requests to query SRIOV config.
        :param req: The HTTP request object.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(
            req.environ['nova.context'], authorize_method=authorize_index)

        # Run the task and generate the HTTP response
        host_sriov_dict = self.sriov_task.get_host_sriov(context, req)
        http_code = self._get_http_code(host_sriov_dict)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_sriov_dict))

    @wsgi.extends
    def show(self, req, id):
        """
        This method handles GET requests to query SRIOV config for a specific
        host.
        :param req: The HTTP request object.
        :param id: The host to query from.
        :returns: An HTTP response object.
        """
        context = self._validate_authorization(req.environ['nova.context'],
                                               authorize_method=authorize_show)

        # Run the task and generate the HTTP response
        host_sriov_dict = self.sriov_task.get_host_sriov(context, req,
                                                         host_name=id)
        http_code = self._get_http_code(host_sriov_dict)
        return webob.Response(request=req,
                              status=http_code,
                              content_type='application/json',
                              body=json.dumps(host_sriov_dict))

    @staticmethod
    def _validate_authorization(context, authorize_method):
        """
        This method authorizes the request.
        :param context: The HTTP request context.
        :returns: The authorized HTTP request context.
        """
        authorize_method(context)
        return context.elevated()

    def _get_http_code(self, response_dict):
        """
        Inspect the response dictionary to determine if all hosts had an error.
        If all hosts had an error, return 404.  If at least one host was
        successful, return 200.

        :param response_dict: A JSON response body.
        :returns: An httplib return code.
        """
        if not response_dict or 'host-sriov' not in response_dict:
            # Some sort of test or other problem
            return httplib.OK

        successHostFound = False
        for host_dict in response_dict['host-sriov']:
            if('errors' not in host_dict or
               ('errors' in host_dict and len(host_dict['errors']) == 0)):
                successHostFound = True
                break
        if len(response_dict['host-sriov']) == 0:
            successHostFound = True

        if not successHostFound:
            return httplib.NOT_FOUND
        return httplib.OK


class Host_sriov(extensions.ExtensionDescriptor):
    """
    Capability to manage SRIOV config on all hosts.
    """
    name = "Compute Host SRIOV"
    alias = "host-sriov"
    namespace = "http://docs.openstack.org/compute/ext/ibm_host_sriov/api/v1.1"
    updated = "2015-02-03T11:00:00-06:00"

    def get_resources(self):
        """
        Provide a Resource Extension for host-sriov
        """
        return [extensions.ResourceExtension('host-sriov',
                                             HostSriovController())]
