# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from eventlet import greenthread
from nova.objects import service as service_obj
from oslo_utils import importutils, encodeutils
from nova.compute import api as base_api
from nova.compute import instance_actions
from nova.compute import task_states
from nova.compute import vm_states
import six
from eventlet import timeout
from functools import wraps
from oslo_config import cfg
from nova import context as ctx
from nova import rpc
from nova import exception as nova_exception
from nova.ibm.ego import hapolicy
from nova.ibm.ego.hapolicy import constants as ha_constants
from nova.network.neutronv2 import api as neutronv2_api
from nova.objects import instance as instance_obj
from nova.objects import compute_node as comp_node_obj
from oslo_utils import excutils, timeutils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from powervc_nova.compute import exception as pvc_exception
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_oslo import override_context_serialize
from powervc_nova import _

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

EVENT_TYPE_ERROR = 'ERROR'
EVENT_TYPE_WARN = 'WARN'
datacache = dict()

#Override the Context Serialization to add Locale
override_context_serialize(ctx.RequestContext)


def API(*args, **kwargs):
    class_name = 'powervc_nova.compute.api.PowerVCComputeAPI'
    return importutils.import_object(class_name, *args, **kwargs)


def HostAPI(*args, **kwargs):
    class_name = 'powervc_nova.compute.api.PowerVCHostAPI'
    return importutils.import_object(class_name, *args, **kwargs)


def monkey_patch_evacuate():
    """
    We need to monkey patch the nova.compute.api.API class'
    evacuate method to avoid checking the source compute service state.
    """
    base_api.API.powervc_mod_evacuate = \
        base_api.API.evacuate

    @base_api.check_instance_state(vm_state=[vm_states.ACTIVE,
                                             vm_states.STOPPED,
                                             vm_states.ERROR])
    def powervc_mod_evacuate(self, context, instance, host, on_shared_storage,
                             admin_password=None):
        service = service_obj.Service.get_by_compute_host(context.elevated(),
                                                          instance['host'])

        if service and service['compute_node']:
            hypervisor_type = service['compute_node']['hypervisor_type']
        if hypervisor_type and hypervisor_type == 'powervm':
            if self.servicegroup_api.service_is_up(service):
                context = context.elevated()
                nodes = comp_node_obj.ComputeNodeList.get_by_hypervisor(
                    context, instance['host'])
                node = nodes[0]
                hypervisor_type = node.hypervisor_type
                stats = node.stats or {}
                hyp_state = stats.get(pvmc.HYPERVISOR_STATE)
                if (hyp_state not in {'error', 'error - dump in progress',
                                      'power off'}):
                    LOG.error(_('Cannot remote restart %s because the source '
                                'host %s is expected to be down, but it was '
                                'up.'),
                               (instance['name'], instance['host']))
                    msg_fmt = _('Remote restart operation cannot be performed '
                                'when source host state is %s') % hyp_state
                    raise nova_exception.ComputeServiceInUse(message=msg_fmt)

            LOG.debug('vm evacuation scheduled', instance=instance)
            instance.task_state = task_states.REBUILDING
            instance.save(expected_task_state=[None])
            self._record_action_start(context, instance,
                                      instance_actions.EVACUATE)

            return self.compute_task_api.\
                rebuild_instance(context,
                                 instance=instance,
                                 new_pass=admin_password,
                                 injected_files=None,
                                 image_ref=None,
                                 orig_image_ref=None,
                                 orig_sys_metadata=None,
                                 bdms=None,
                                 recreate=True,
                                 on_shared_storage=on_shared_storage,
                                 host=host)

        else:
            return self.powervc_mod_evacuate(context, instance, host,
                                             on_shared_storage,
                                             admin_password=None)

    base_api.API.evacuate = powervc_mod_evacuate

monkey_patch_evacuate()


def monkey_patch_is_volume_backed_instance():
    """
    We need to monkey patch the nova.compute.api.API class'
    is_volume_backed_instance method to allow the PowerVM capture path to
    go through the compute manager and the virt-drivers for IVM/HMC.  The
    existing method in OpenStack drives capture of our PowerVM VMs down
    a volume backed / Cinder snapshot path that is not supported on our VMs.
    """
    base_api.API.powervc_mod_is_volume_backed_instance = \
        base_api.API.is_volume_backed_instance

    def powervc_mod_is_volume_backed_instance(self, context, instance,
                                              bdms=None):
        service = service_obj.Service.get_by_compute_host(context.elevated(),
                                                          instance['host'])

        if service and service['compute_node']:
            hypervisor_type = service['compute_node']['hypervisor_type']
        if hypervisor_type and hypervisor_type == 'powervm':
            base_api.LOG.debug("Instance %s is on a PowerVM host, "
                               "returning False" % instance['uuid'])
            return False
        return self.powervc_mod_is_volume_backed_instance(context,
                                                          instance, bdms)

    base_api.API.is_volume_backed_instance = \
        powervc_mod_is_volume_backed_instance

monkey_patch_is_volume_backed_instance()


def monkey_patch__delete():
    """
    We need to monkey patch the nova.compute.api.API class'
    _delete method to prevent delete of VM when compute service is down.
    See LP bug for detail: https://bugs.launchpad.net/bugs/1389899
    """
    base_api.API.powervc_mod__delete = \
        base_api.API._delete

    def powervc_mod__delete(self, context, instance, delete_type, cb,
                            **instance_attrs):
        host = instance['host']
        is_up = False
        if host:
            try:
                service = service_obj.Service.get_by_compute_host(
                    context.elevated(), instance.host)
                if self.servicegroup_api.service_is_up(service):
                    is_up = True
            except Exception as e:
                LOG.exception(e)
        else:
            # if host is None, it's result from failed deploy.
            # We should allow the delete to go through
            is_up = True
        if not is_up:
            raise nova_exception.ComputeServiceUnavailable(host=host)
        else:
            return self.powervc_mod__delete(context, instance, delete_type,
                                            cb, **instance_attrs)

    base_api.API._delete = \
        powervc_mod__delete

monkey_patch__delete()


def monkey_patch__check_and_transform_bdm():
    """
    We need to monkey patch the nova.compute.api.API class'
    _check_and_transform_bdm method to allow the multi-deploy of PowerVC
    volume backed images.
    """
    base_api.API.powervc_mod__check_and_transform_bdm = \
        base_api.API._check_and_transform_bdm

    def override_check_and_transform_bdm(self, context, base_options,
                                         instance_type,
                                         image_meta, min_count, max_count,
                                         block_device_mapping, legacy_bdm):
        return self.powervc_mod__check_and_transform_bdm(context, base_options,
                                                         instance_type,
                                                         image_meta, 1, 1,
                                                         block_device_mapping,
                                                         legacy_bdm)
    base_api.API._check_and_transform_bdm = \
        override_check_and_transform_bdm

monkey_patch__check_and_transform_bdm()


def monkey_patch_get_image_defined_bdms():
    """
    We need to monkey patch the nova.compute.api.API class'
    _get_image_defined_bdms method to prevent the deletion of image volumes in
    the following flow:
        1. deploy fails
        2. image is deleted but volumes are kept to be re-used
        3. failed deploy VM is deleted, deleting the image volumes.
    """
    base_api.API.powervc_mod__get_image_defined_bdms = \
        base_api.API._get_image_defined_bdms

    def override_get_image_defined_bdms(self, base_options, instance_type,
                                        image_meta, root_device_name):
        image_bdms = self.powervc_mod__get_image_defined_bdms(base_options,
                                                              instance_type,
                                                              image_meta,
                                                              root_device_name)
        if image_bdms:
            for bdm in image_bdms:
                vol_id = bdm.get('volume_id')
                delete_on_termination = bdm.get('delete_on_termination')
                if vol_id and delete_on_termination:
                    bdm['delete_on_termination'] = False
        return image_bdms
    base_api.API._get_image_defined_bdms = override_get_image_defined_bdms


monkey_patch_get_image_defined_bdms()


def monkey_patch_detach():
    """
    We need to monkey patch the community code till the nova code
    for multi attach goes in.
    """
    base_api.API.powervc_mod_detach = \
        base_api.API.detach_volume

    def powervc_mod_detach(self, context, instance, volume):
        if volume['attach_status'] == 'detached':
            msg = _("Volume must be attached in order to detach.")
            raise base_api.exception.InvalidVolume(reason=msg)
        # Check if the volume has attachments before trying to detach
        if volume.get('attachments') is not None:
            attachments = volume.get('attachments')
            attached = False
            for att in attachments:
                if att['instance_uuid'] == instance['uuid']:
                    attached = True
            if attached is False:
                raise base_api.exception.VolumeUnattached(
                    volume_id=volume['id'])
        self._detach_volume(context, instance, volume)

    base_api.API.detach_volume = \
        powervc_mod_detach

monkey_patch_detach()


def log_operation(op=None, instance=None, msg=None):
    """
    Log entry point of user initiated operation to aid debugging
    """
    if op and instance:
        LOG.info(_('Operation: %(op)s. Virtual machine display name: '
                   '%(display_name)s, name: %(name)s, UUID: %(uuid)s') %
                 {'op': op, 'display_name': instance['display_name'],
                  'name': instance['name'], 'uuid': instance['uuid']})
    elif msg:
        LOG.info(msg)


def check_host_maintenance_mode(func_type):
    """ Decorator to check if the target host of operation
        is in maintenance mode.
        If host status is OK
            then all operations are allowed
        If host status is Maintenance error/on/entering
            then delete/stop
        live_migrate as source host is the only operation allowed
        all the time,so this decorator is not used for live_migration.
        as source.
    """
    def func_parms(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            host_state = {'ego_stat': None}
            time_out = CONF.rpc_response_timeout

            def _get_host_status(context):
                try:
                    with timeout.Timeout(time_out, False):
                        host_state['ego_stat'] = hapolicy.API().\
                            get_hv_status(context)
                except Exception as e:
                    # If for some reason HA service is down
                    # or not able to retrieve host state
                    # then allow the operation on VM
                    LOG.info(_("Exception while retrieving host state "
                               "from HA service %s") % six.text_type(e))

            def _get_node_name(host):
                if datacache is not None and host in datacache.keys():
                    return datacache[host]

                try:
                    nodes = comp_node_obj.ComputeNodeList.get_all_by_host(
                        context.elevated(), host)
                    node_name = nodes[0].hypervisor_hostname
                    datacache[host] = node_name
                except nova_exception.NotFound:
                    raise nova_exception.NotFound(_("Host %s not found")
                                                  % host)
                return datacache[host]

            host = CONF.host
            context = ctx.get_admin_context()
            node_name = _get_node_name(host)
            # note (ymadhavi):
            # To retrieve host status from HA service
            # green thread is spawned separately to avoid
            # API call hanging if HA service is down
            gt = greenthread.spawn(_get_host_status, context)
            gt.wait()

            mainttag = None
            maint_stat = ""

            if host_state['ego_stat']:
                if node_name in host_state['ego_stat']['maintenance'].keys():
                    mainttag = host_state['ego_stat']['maintenance'][node_name]
                    maint_stat = mainttag['state']

            block_operation = False

            if maint_stat in (ha_constants.STATE_MT_ERROR,
                              ha_constants.STATE_MT_ON,
                              ha_constants.STATE_MT_ENTERING):
                block_operation = True
            if block_operation:
                exc_args = {'state': maint_stat}
                raise pvc_exception.IBMPowerVCInvalidOperation(**exc_args)
            else:
                return function(*args, **kwargs)

        return wrapper
    return func_parms


def send_ui_notification(error_message, success_message,
                         error_event_type=EVENT_TYPE_ERROR,
                         migrate_op=False, log_exception=False):
    """
    Send a notification to the GUI. If the decorated method throws an
    exception, an error notification will be sent, else a successful
    info notification will be sent. lpar name and uuid will be pulled from
    the instance object or dict parameter passed to the decorated method.

    Note: message string parameters (error_message and success_message) must be
          translated (wrappered with _() ). Replacement text variables
          must be surrounded with brackets.

          Replacement variables must be from the following list:'
                        {instance_id}
                        {instance_name}
                        {host_name}
                        {volume_id}
                        {source_host_name}
                        {target_host_name}
                        {error}

          Example: Stop of virtual machine {instance_name} on host {host_name}
                   failed with exception: {error}

    :param error_message: message string for exception notifications
                          (see note above)
    :param success_message: message string for success notifications
                            (see note above)
    :param error_event_type: either ERROR (EVENT_TYPE_ERROR)
                             or WARN (EVENT_TYPE_WARN)
    :param migrate_op: set to true for migation operations. Tells the decorator
                       to look for the migrate data dict. Only use when you
                       have replacement variables for source and target host.
    :param log_exception: if true in and error case, the event message is
                          also logged (LOG.exception)
    """
    def func_parms(f):
        @wraps(f)
        def wrapper(*args, **kwds):

            instance_dict_keys = ['uuid', 'display_name',
                                  'vm_state', 'task_state']
            migrate_dict = ['migrate_data']
            migrate_dict_source = ['src_sys_name']
            migrate_dict_target = ['dest_sys_name']

            def check_keys(arg, inst_keys):
                for key in inst_keys:
                    if key not in arg:
                        return False
                # Found them all!
                return True

            def check_inst(arg):
                # check if this is a Nova Instance object
                if isinstance(arg, instance_obj.Instance):
                    return True
                # check for a dict with all the right keys to match instance
                if isinstance(arg, dict) and check_keys(arg,
                                                        instance_dict_keys):
                    return True
                return False

            def check_migrate_source(arg):
                # check for a dict and it has the source migration host
                if isinstance(arg, dict) and check_keys(arg,
                                                        migrate_dict_source):
                    return True
                return False

            def check_migrate_target(arg):
                # check for a dict and it has the target migration host
                if isinstance(arg, dict) and check_keys(arg,
                                                        migrate_dict_target):
                    return True
                return False

            def check_migrate_target_embedded(arg):
                # check for an embedded migrate dict
                if isinstance(arg, dict) and check_keys(arg,
                                                        migrate_dict):
                    if isinstance(arg, dict) and (
                            check_keys(arg['migrate_data'],
                                       migrate_dict_target)):
                        return True
                return False

            def _notify(exception_msg=None):
                payload = {}
                payload['msg'] = ''
                payload['instance_id'] = lpar_uuid
                payload['instance_name'] = lpar_name
                payload['host_name'] = encodeutils.safe_decode(
                    CONF.host_display_name, 'utf-8')
                if volume_id is not None:
                    payload['volume_id'] = volume_id
                notifier = rpc.get_notifier(service='compute',
                                            host=CONF.host)
                try:
                    if migrate_op:
                        payload['source_host_name'] = source_host
                        payload['target_host_name'] = target_host
                    if exception_msg is not None:
                        if error_message is not None:
                            payload['msg'] = error_message
                            payload['error'] = exception_msg
                            if log_exception:
                                try:
                                    # Convert the brackets to replacement
                                    # variables
                                    log_exception_msg = error_message.replace(
                                        "{", "%(").replace("}", ")s")
                                    LOG.exception(log_exception_msg % payload)
                                except Exception as exc:
                                    LOG.debug('LOG.exception failed with '
                                              'exception: %s' %
                                              six.text_type(exc))
                            if error_event_type is EVENT_TYPE_ERROR:
                                notifier.error(context,
                                               'compute.instance.log',
                                               payload)
                            elif error_event_type is EVENT_TYPE_WARN:
                                notifier.warn(context,
                                              'compute.instance.log',
                                              payload)
                            else:
                                LOG.debug('Exception was handled, but the '
                                          'error_event_type parameter was not '
                                          'set to a supported value. No GUI '
                                          'notification sent.')
                        else:
                            LOG.debug('Operation failed, but no error_message '
                                      'was provided. No GUI notification '
                                      'sent.')
                    else:
                        if success_message is not None:
                            payload['msg'] = success_message
                            notifier.info(context,
                                          'compute.instance.log',
                                          payload)
                        else:
                            LOG.debug('Operation completed successfully, but '
                                      'no success_message was provided. No GUI'
                                      ' notification sent.')
                except Exception as exc:
                    LOG.debug('Notify failed with exception: %s' %
                              six.text_type(exc))

            lpar_uuid = None
            lpar_name = None
            source_host = None
            target_host = None
            volume_id = None
            context = None
            for arg in args:
                if check_inst(arg):
                    lpar_uuid = arg['uuid']
                    lpar_name = arg['display_name']
                if migrate_op:
                    if check_migrate_source(arg):
                        source_host = arg['src_sys_name']
                    if check_migrate_target(arg):
                        target_host = arg['dest_sys_name']
                    elif check_migrate_target_embedded(arg):
                        migrate_data = arg['migrate_data']
                        target_host = migrate_data['dest_sys_name']
                # TODO: need to see what type of context object is passed
                # with the new osee build installed
                # if isinstance(arg, RpcContext):
                #    context = arg
                if (context is not None and
                    lpar_uuid is not None and
                        (not migrate_op or target_host is not None)):
                    break

            if 'volume_id' in kwds:
                volume_id = kwds['volume_id']
            if 'bdm' in kwds and volume_id is None:
                if kwds['bdm'] is not None:
                    bdm = kwds['bdm']
                    volume_id = bdm['volume_id']
            if 'instance' in kwds:
                if check_inst(kwds['instance']):
                    if lpar_uuid is None:
                        lpar_uuid = kwds['instance']['uuid']
                    if lpar_name is None:
                        lpar_name = kwds['instance']['display_name']

            if lpar_uuid is None:
                LOG.debug('Did not find the Instance object in '
                          'the decorated method arguments')
                lpar_uuid = ''
                lpar_name = ''
            if migrate_op:
                if source_host is None:
                    source_host = ''
                if target_host is None:
                    LOG.debug('Did not find the migrate dict in '
                              'the decorated method arguments')
                    target_host = ''
            if not context:
                context = ctx.get_admin_context()
            try:
                r = f(*args, **kwds)
                # notify success
                _notify()
                return r
            except Exception as exc:
                with excutils.save_and_reraise_exception():
                    # notify failure
                    exception_msg = six.text_type(exc)
                    if not exception_msg:
                        exception_msg = exc.__class__.__name__
                    _notify(exception_msg)

        return wrapper
    return func_parms


def ui_notifier(notify_message, instance, except_to_raise=None,
                error_event_type=EVENT_TYPE_ERROR,
                log_exception=False):

    error_msg = None
    success_msg = None
    if except_to_raise is not None:
        error_msg = notify_message
    else:
        success_msg = notify_message

    @send_ui_notification(error_msg, success_msg,
                          error_event_type=error_event_type,
                          log_exception=log_exception)
    def _ui_notifier(instance, except_to_raise=None):
        if except_to_raise is not None:
            raise except_to_raise

    _ui_notifier(instance, except_to_raise=except_to_raise)


def _get_token_info(hclient):
    """Helper method to check if the token going to use are expiring"""
    is_expiring, catalog = True, None
    header = {'X-Auth-Token': hclient.auth_token, 'Accept': 'content/json'}
    token_url = cfg.CONF.neutron.admin_auth_url
    token_url += '/tokens/' + str(hclient.auth_token)
    try:
        resp, body = hclient.request(token_url, 'GET', headers=header)
        #If the Token Request was successful, check the expiration date
        if body and (resp.status_code == 200 or resp.status_code == 203):
            body = jsonutils.loads(body)
            expiration = body['access']['token']['expires']
            expiration = timeutils.parse_isotime(expiration)
            #Determine if the Token is within 15 minutes from Expiring
            is_expiring = not timeutils.is_newer_than(expiration, 900)
            if is_expiring:
                LOG.info(_("The periodic task token is about to expire"))
            catalog = body['access']['serviceCatalog']
        else:
            LOG.info(_("The periodic task token is invalid or expired"))
    except Exception as exc:
        LOG.warn(_("Error trying to determine if token expired: %s") % exc)
        LOG.exception(exc)
    return is_expiring, catalog


def populate_admin_context(context=None, context_service_catalog=None):
    """Helper method to populate the Token/ServiceCatalog on the Context"""
    if context is None:
        context = ctx.get_admin_context()
    #We don't need to do anything if the Token/ServiceCatalog are populated
    if context.auth_token and context.service_catalog:
        return context
    #We are using the Neutron Client since they having Caching logic
    nclient = neutronv2_api.get_client(context, True)
    hclient = nclient.httpclient
    #Since the Neutron Client is cached, the token may already be
    #populated, so only need to authenticate if it isn't set yet
    if neutronv2_api._ADMIN_AUTH is None:
        hclient.authenticate()
        nclient.get_attr_metadata()
    #Determine whether the Token is Expiring and the Service Catalog
    expiring, service_catalog = _get_token_info(hclient)
    #Make sure the token isn't going to expire while executing
    if expiring:
        hclient.authenticate()
        nclient.get_attr_metadata()
    context.auth_token = neutronv2_api.\
        _ADMIN_AUTH.get_token(neutronv2_api._SESSION)
    #Cache the Service Catalog so that we can use it later
    if context_service_catalog is None:
        context_service_catalog = service_catalog
    context.service_catalog = context_service_catalog
    if context_service_catalog is None:
        LOG.info(_('No service_catalog found on network client: %s') %
                 nclient)
        return None
    return context
