#
#
# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# Copyright 2011 Piston Cloud Computing, Inc.
# Copyright 2012-2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Extends the Compute Manager to provide additional Capabilities for PowerVC
   and to override any necessary behavioral changes from the Community code.
   This code is organized into the following categories:
    1. Overriding Compute Manager Initialization
    2. Compute-related Manager Extensions
    3. Discovery-related Manager Extensions
    4. Image-related Manager Extensions
    5. Migration-related Manager Extensions
    6. Monitor-related Manager Extensions
    7. Network-related Manager Extensions
    8. Storage-related Manager Extensions
    9. Overriding the ResourceTracker and VirtAPI
"""

import os
import sys
if os.name == 'posix':
    import fcntl
else:
    # This leg is only for UT runs on Windows
    import mock
    sys.modules['fcntl'] = mock.MagicMock()

import copy
import errno
import inspect
import uuid
import functools
import greenlet
from eventlet import greenthread, GreenPool
from oslo_config import cfg
import six
import time
import traceback
import types
import re

from cinderclient import exceptions as cinder_exception

import nova.context
from nova import exception
from nova import network
from nova import rpc
from nova import utils
from nova.compute import manager
from nova.compute import resources
from nova.compute import resource_tracker
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import utils as compute_utils
from nova.compute import vm_states
from nova.i18n import _LE
from nova.i18n import _LI
from nova.image import glance
from nova.network import model as network_model
from nova.network import neutronv2
from nova import block_device
from nova import objects
from nova.objects import base as obj_base
from nova.objects import block_device as blkdev_obj
from nova.objects import compute_node as compute_obj
from nova.objects import instance as instance_obj
from nova.objects import quotas as quotas_obj
from oslo_utils import excutils
from oslo_utils import importutils, encodeutils
from oslo_serialization import jsonutils
from oslo_concurrency import lockutils
from oslo_log import log as logging
from nova.openstack.common import periodic_task
from oslo_utils import timeutils
from nova.openstack.common import eventlet_backdoor as backdoor
from nova.virt import block_device as blk_dev
from nova.virt import driver
from nova.volume import cinder as cinderv2
from nova.virt.libvirt import utils as libvirt_utils
from nova import safe_utils

from powervc_nova.common import block_device_utils
from powervc_nova.common import volume_type_utils
from powervc_nova.compute import send_ui_notification, \
    check_host_maintenance_mode, populate_admin_context
from powervc_nova.compute import notify_messages as notify_msg

from powervc_nova import cache_driver_data
from powervc_nova import logcall
from powervc_nova.compute import api as pvc_compute_rpcapi
from powervc_nova.compute import exception as pwrvc_excep
from powervc_nova.compute import common_constants
from powervc_nova.conductor import api as conductor
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.virt import discovery_driver
from powervc_nova.virt import netutils
from powervc_nova.virt.ibmpowervm.common import exception as pvm_exc
from powervc_nova.virt.ibmpowervm.common import migrate_utils
from powervc_nova.virt.ibmpowervm.common.constants \
    import CONNECTION_TYPE_NPIV, CONNECTION_TYPE_SSP, CONNECTION_TYPE_VSCSI
from powervc_nova.virt.ibmpowervm.hmc import event as hmcevent
from powervc_nova.volume import cinder as volume_cinder
from powervc_nova.objects.network import dom_kvm
from powervc_nova.network.powerkvm import agent

from powervc_nova import _
from powervc_nova import translate_message


LOG = logging.getLogger(__name__)


ibmpowervm_interval_opts = [
    cfg.IntOpt('ibmpowervm_activating_timeout', default=1200,
               help='Timeout (seconds) for VM in activating after deploy.'),
    cfg.IntOpt('ibmpowervm_iso_detach_timeout',
               default=3600,
               help='Timeout in seconds for removal of optical post-deploy.'),
    cfg.IntOpt('ibmpowervm_optical_cleanup_interval',
               default=1800,
               help='Interval to check for virtual optical needing removal '
               'post-deploy.'),
    cfg.IntOpt('inventory_host_interval', default=300,
               help='Interval to Refresh the Inventory for the Host'),
    cfg.IntOpt('quick_inventory_host_interval', default=60,
               help='Interval to do a quick check of host\'s VIOSes\' status'),
    cfg.IntOpt('inventory_instances_interval', default=120,
               help='Interval to Refresh the Inventory for Instances'),
    cfg.IntOpt('monitor_instances_interval', default=300,
               help='Interval to Refresh the Metrics for Instances'),
    cfg.IntOpt('metrics_message_size', default=20,
               help='The size limit of a compute.instance.metrics message. '
                    'For each message, the number of instances that is '
                    'reported.  For example if there are 40 instances on a '
                    'host, there will be 2 messages sent each time the '
                    'utilization data is collected (40 / 20 = 2)'),
    cfg.StrOpt('host_display_name', default=None,
               help='The display name set by the user for the Host'),
    cfg.IntOpt('volume_status_check_max_retries', default=300,
               help='The maximum number of volume status checks'
               ' during volume creation.'),
    cfg.IntOpt('volume_status_check_interval', default=6,
               help='Interval between status checks'
               ' during volume creation.'),
    cfg.IntOpt('cleanup_net_assn_interval', default=1200,
               help='Timeout (seconds) for cleanup network associations.'),
    cfg.StrOpt('boot_volume_name_prefix', default='boot-',
               help='Prefix for instance boot set volumes.'),
    cfg.StrOpt('data_volume_name_prefix', default='data-',
               help='Prefix for instance data volumes.'),
    cfg.BoolOpt('ibmpowervm_handle_out_of_band_migration', default=True,
                help='Handle out of band migration of instances'),
    cfg.BoolOpt('has_first_boot_occurred', default=False,
                help='Indicates if the agent has run through the first'
                     ' boot setup'),
    cfg.IntOpt('periodic_task_runtime_msg_threshold', default=300,
               help='The threshold (seconds) before messages are'
                    ' logged for periodic tasks running longer than the'
                    ' threshold.'),
    cfg.IntOpt('periodic_task_completion_msg_threshold', default=90,
               help='The threshold (seconds) before messages are'
                    ' logged for periodic tasks that ran longer to complete'
                    ' than the threshold.'),
    cfg.IntOpt('verify_mgmt_sys_max_retry_count', default=5,
               help='Retry count for verify management system'),
    cfg.IntOpt('verify_mgmt_sys_sleep_time_during_retry', default=20,
               help='Sleep time while retrying verify'
               'management system')
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_interval_opts)
CONF.import_opt('powervm_mgr_type', 'powervc_nova.virt.ibmpowervm.ivm')
CONF.import_opt('powervm_mgr', 'powervc_nova.virt.ibmpowervm.ivm')
CONF.import_opt('host_storage_type', 'powervc_nova.virt.ibmpowervm.ivm')

get_notifier = functools.partial(rpc.get_notifier, service='compute')
wrap_exception = functools.partial(exception.wrap_exception,
                                   get_notifier=get_notifier)

# Constant for Activating state used post-deploy
ACTIVATING = 'activating'
# waiting on host task state
WAITING_ON_HOST = 'waiting_on_host'
# time when the periodic task canary last ran
CANARY_RUN = None
# List of periodic tasks to remove. This will remove all occurrences
# of the name.  It's possible to have periodic tasks with the same name but
# it should not be done.  If overriding a periodic task, prefix the name with
# 'pvc' or something similar to create a unique name.
## This is not needed at the moment since we got a fix from OpenStack but
## good to keep around incase we need it in the future.
PERIODIC_TASKS_TO_REMOVE = ['update_available_resource', '_sync_power_states']
RUN_PERIODIC_TASK_ASYNC = ['pvc_update_available_resource',
                           'inventory_instances',
                           'monitor_hosts',
                           '_check_instance_vopt_removal',
                           '_pvc_sync_power_states'
                           '_poll_unconfirmed_resizes']

NOVA_COMP_MGR_FILE = inspect.getsourcefile(manager.ComputeManager)


@utils.expects_func_args('instance')
def pvc_comp_mgr_reverts_task_state(function):
    """
    IMPORTANT: This can only be used by PowerVCComputeManager.
    It has specific checking for whether an exception has
    raised from super.function to determine task_state has
    been reset already.
    Decorator to revert task_state on failure.
    """

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except exception.UnexpectedTaskStateError as e:
            # Note(maoy): unexpected task state means the current
            # task is preempted. Do not clear task state in this
            # case.
            with excutils.save_and_reraise_exception():
                LOG.info(_("Task possibly preempted: %s"),
                         e.format_message())
        except Exception:
            with excutils.save_and_reraise_exception():
                exc_raised_from_super = False
                try:
                    # check whether the exception is raise from super
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    tb_list = traceback.extract_tb(exc_traceback, 30)
                    for (filename, line, func_name, text) in tb_list:
                        if (function.__name__ == func_name and
                                NOVA_COMP_MGR_FILE == filename):
                            exc_raised_from_super = True
                            break
                except Exception as ex:
                    LOG.exception(ex)
                    pass
                try:
                    if not exc_raised_from_super:
                        wrapped_func = utils.get_wrapped_function(function)
                        keyed_args = safe_utils.getcallargs(wrapped_func,
                                                            context,
                                                            *args, **kwargs)
                        # NOTE(mriedem): 'instance' must be in keyed_args
                        # because we
                        # have utils.expects_func_args('instance') decorating
                        # this method.
                        instance_uuid = keyed_args['instance']['uuid']
                        self._instance_update(context, instance_uuid,
                                              task_state=None)
                    else:
                        LOG.info(_('The task_state value has been reverted in '
                                   'the parent method %s.')
                                 % function.__name__)
                except Exception as save_exc:
                    LOG.exception(save_exc)
                    pass

    return decorated_function


class PowerVCComputeManager(manager.ComputeManager):
    """Extends the base Compute Manager class with PowerVC capabilities"""

    def __init__(self, compute_driver=None, *args, **kwargs):
        """Constructor for the PowerVC extension to the Compute Manager"""
        try:
            # Remove any unwanted periodic tasks before initializing our
            # parent class, since it will start running tasks
            self._undefine_periodic_tasks()
            # Wrap the periodic tasks so we can log and time them
            self._init_ptask_nanny()

            super(PowerVCComputeManager, self).__init__(
                compute_driver='nova.virt.fake.FakeDriver', *args, **kwargs)
            # Quickly replace parents pool size.  Default is 1000 which could
            # overwhelm an HMC.
            self._sync_power_pool = GreenPool(size=10)
            # Since we need to add methods to the VirtAPI so the driver can
            # call other DB methods, we first construct a temporary driver so
            # that we can override the VirtAPI before we construct the
            # real driver
            self.compute_rpcapi = pvc_compute_rpcapi.PowerVCComputeRPCAPI()
            self.virtapi = PowerVCComputeVirtAPI(self)
            self.conductor_api = conductor.PowerVCConductorAPI()
            self.volume_api = volume_cinder.PowerVCVolumeAPI()
            self.driver = driver.\
                load_compute_driver(self.virtapi, compute_driver)

            # calculate the following flag
            # after actual driver is loaded
            self.use_legacy_block_device_info = \
                self.driver.need_legacy_block_device_info

            self.instance_chunk_map = dict()
            self.instance_chunk_counter = 0
            self.context_service_catalog = None

            self._periodic_task_pool = GreenPool(size=10)
            self._periodic_tasks_running = {}

            # Monkey patches:
            netutils.monkey_patch_netutils_get_injected_network_template()

        except Exception as exc:
            LOG.exception(_('error in ComputeMgr init, ending nova-compute'))
            raise exc

    def _undefine_periodic_tasks(self):
        for task_name, task in self._periodic_tasks:
            LOG.debug("Defined periodic task: %s" % task_name)
            if task_name in PERIODIC_TASKS_TO_REMOVE:
                LOG.debug("Removing defined periodic task: %s" % task_name)
                self._periodic_tasks.remove((task_name, task))

    def _init_ptask_nanny(self):
        nanny_wrapped = []
        for task_name, task in self._periodic_tasks:
            LOG.debug("Defined periodic task nanny for: %s" % task_name)
            nanny = functools.partial(self._periodic_task_nanny,
                                      task_=task, task_name=task_name)
            # Copy the settings for the task
            nanny._periodic_spacing = task._periodic_spacing
            nanny._periodic_immediate = task._periodic_immediate
            nanny._periodic_last_run = task._periodic_last_run
            # Add the new task to the list
            nanny_wrapped.append((task_name, nanny))

        # Replace the list of tasks with our wrapped ones
        self._periodic_tasks_orig = self._periodic_tasks
        self._periodic_tasks = nanny_wrapped

    def _periodic_task_nanny(self, *args, **kwargs):
        # Grab parameters
        context = args[1]
        task = kwargs['task_']
        task_name = kwargs['task_name']

        def run_task(task, task_name, context):
            LOG.debug('Starting periodic task: %s' % task_name)
            start_time = time.time()
            try:
                task(self, context)
                end_time = time.time()
                LOG.debug('Completed periodic task: %s' % task_name)
                delta = end_time - start_time
                if delta > CONF.periodic_task_completion_msg_threshold:
                    LOG.info(_('Periodic task %(task_name)s ran '
                               'for %(time)s seconds.')
                             % dict(task_name=task_name, time=int(delta)))
            except Exception:
                LOG.exception(_('Exception during periodic task %(task_name)s '
                                ' execution') % dict(task_name=task_name))
            finally:
                self._periodic_tasks_running.pop(task_name)

        # Only run the task if it's not already running
        if task_name not in self._periodic_tasks_running:
            self._periodic_tasks_running[task_name] = time.time()
            if task_name in RUN_PERIODIC_TASK_ASYNC:
                self._periodic_task_pool.spawn_n(run_task, task,
                                                 task_name, context)
            else:
                run_task(task, task_name, context)
        else:
            # The periodic task is still running a prior invocation
            # check if we should flag it.
            try:
                dispatch_time = self._periodic_tasks_running[task_name]
                delta = time.time() - dispatch_time
                if delta > CONF.periodic_task_runtime_msg_threshold:
                    LOG.warn(_('Periodic task %(task_name)s has been running '
                               'for %(time)s seconds.')
                             % dict(task_name=task_name, time=int(delta)))
            except KeyError:
                # It must have finished...
                pass

    #######################################################
    ##########  Override Initialization Methods  ##########
    #######################################################
    def pre_start_hook(self, **kwargs):
        """Override the Pre-Start Hook to fix Base Compute Manager issues"""
        # Call the Parent pre-start hook to create the ComputeNode in the DB
        super(PowerVCComputeManager, self).pre_start_hook(**kwargs)
        # After the Compute Node is inserted in the DB, then Inventory the Host
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            self.inventory_host(nova.context.get_admin_context())
            self.inventory_instances(nova.context.get_admin_context())

    @cache_driver_data(['get_lpars_feed'])
    def init_host(self):
        """Override the Init Host to fix Base Compute Manager issues"""
        try:
            super(PowerVCComputeManager, self).init_host()
            self._verify_mgmt_system_access(nova.context.get_admin_context())
            self.check_and_run_first_boot(nova.context.get_admin_context())
        except Exception:
            LOG.exception(_('ComputeMgr init_host error, ending nova-compute'))
            raise

    def _init_instance(self, context, instance):
        """Override Init-Instance method to fix Base Compute Manager issues"""
        try:
            super(PowerVCComputeManager,
                  self)._init_instance(context, instance)
        except Exception:
            LOG.exception(_('exception initializing the Instance, moving on'))

    @send_ui_notification(notify_msg.SOURCE_EVACUATED_INST_CLEANUP_ERROR,
                          None,
                          log_exception=True)
    def _destroy_evacuated_instances(self, context):
        """
            Override Base Compute Manager to update evacuated instances.
            If the hypervisor is KVM, read the instance metadata and
            delete the instance from the source if the instance host
            is different than the source host and instance system
            metadata contains the 'evacuated_to'.

            In case of PowerVM hypervisor, call back the driver method to
            handle evacuated instance.

            The method is overriden for sending notifications as well to make
            sure user is notified with the proper messages in case of cleanup
            failures from the source host or in case of successful cleanup.
        """
        is_kvm = False
        # Check whether KVM is the hypervisor as in case of KVM
        # we do not need to worry about calling driver and default
        # manager layer destroy instance should be called.
        if CONF.powervm_mgr_type is None:
            is_kvm = (CONF.libvirt.virt_type == 'kvm')
        # If the manager is not powervm manager type call the default
        # implementation for OpenStack destroy evacuated instance.
        if is_kvm:
            self._destroy_evacuated_instances_kvm(context)
        driver = self.driver
        try:
            LOG.debug('Update instead of delete evacuated instances.')
            # Get all LPARs on this host that are in PowerVC DB
            drv_instances = self.driver.list_instances()
            db_instances = self._query_db_instances(context)
            # If the Driver implemented, ask it to handle Evacuated
            # Instances
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                driver.handle_evacuated_instances(
                    context, db_instances, drv_instances)
        # We don't want to take down the process just because cleanup
        # failed
        except Exception:
            LOG.exception(_('exception in Evacuation Cleanup, moving on'))

    def _destroy_evacuated_instances_kvm(self, context):
        '''
            Overriden method as of the compute manager code as we needed
            to handle the condition in which we want to delete the VM.
            The method is same as of the OpenStack implemented method and
            we are changing the condition to handle out of band migration
            case.
        '''
        our_host = self.host
        filters = {'deleted': False}
        local_instances = self._get_instances_on_driver(context, filters)
        for instance in local_instances:
            # Get the system metadata for instance to see if the instance is
            # actually evacuated or not.
            sys_meta = instance.system_metadata
            evacuated_to = None
            if sys_meta and 'evacuated_to' in sys_meta:
                evacuated_to = sys_meta['evacuated_to']
            # Delete only when the instance has the metadata for dest host
            # and the current host is different than of evacuated to host.
            if evacuated_to:
                if our_host != evacuated_to and instance.host != our_host:
                    if (instance.task_state in [task_states.MIGRATING,
                                                task_states.RESIZE_MIGRATING,
                                                task_states.RESIZE_MIGRATED,
                                                task_states.RESIZE_FINISH]
                            or instance.vm_state in [vm_states.RESIZED]):
                        LOG.debug('Will not delete instance as its host ('
                                  '%(instance_host)s) is not equal to our '
                                  'host (%(our_host)s) but its task state is '
                                  '(%(task_state)s) and vm state is '
                                  '(%(vm_state)s)',
                                  {'instance_host': instance.host,
                                   'our_host': our_host,
                                   'task_state': instance.task_state,
                                   'vm_state': instance.vm_state},
                                  instance=instance)
                        continue
                    LOG.info(_("Deleting instance %(instance_name)s as its "
                               "host %(instance_host)s is not equal "
                               "to source host %(our_host)s") %
                             {'instance_name': instance.name,
                              'instance_host': instance.host,
                              'our_host': our_host},
                             instance=instance)
                    try:
                        network_info = self._get_instance_nw_info(context,
                                                                  instance)
                        bdi = self._get_instance_block_device_info(context,
                                                                   instance)
                        destroy_disks = not (self._is_instance_storage_shared(
                                             context, instance))
                    except exception.InstanceNotFound:
                        LOG.info(_("Instance has been marked deleted, "
                                   "skipping the deletion of instance "
                                   "%(instance_name)s") %
                                 {'instance_name': instance.name})
                        destroy_disks = True
                    if not destroy_disks:
                        self.driver.destroy(context, instance,
                                            network_info,
                                            bdi, destroy_disks)
                        info = {'msg': _("Instance {instance_name} "
                                         "evacuated to host {dest_host}, "
                                         "Deleted the instance from host "
                                         "{source_host}"),
                                'instance_name': instance.name,
                                'dest_host': evacuated_to,
                                'source_host': our_host}
                        notifier = rpc.get_notifier(service='compute',
                                                    host=self.host)
                        notifier.info(context, 'compute.instance.log', info)

    def _get_resource_tracker(self, nodename):
        """Override to return our own Resource Tracker implementation"""
        rt = self._resource_tracker_dict.get(nodename)
        if not rt:
            if nodename not in self.driver.get_available_nodes():
                msg = nodename + ' is not a valid node managed by the host.'
                raise exception.NovaException(msg)
            rt = PowerVCResourceTracker(self.host, self.driver, nodename)
            self._resource_tracker_dict[nodename] = rt
        return rt

    #######################################################
    ########  Compute-related Overrides/Extensions  #######
    #######################################################
    def check_and_run_first_boot(self, context):
        """
        Method to be run on first boot post registration
        """
        # check if first boot has already occurred, if yes return
        if CONF.has_first_boot_occurred:
            return
        #if not, check that driver is instance of ComputeDiscoveryDriver
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            # run the post registration against the driver
            LOG.debug('Running post_registration against ' + str(self.driver))
            self.driver.post_registration(context)
            # run openstack-config to set has_first_boot_occurred as True
            LOG.debug('Running openstack-config to set '
                      'has_first_boot_occurred')
            utils.execute('openstack-config', '--set',
                          CONF.config_file[0], 'DEFAULT',
                          'has_first_boot_occurred', True,
                          run_as_root=True, check_exit_code=[0])
            # update in-memory has_first_boot_occurred
            LOG.debug('Updating in-memory conf option has_first_boot_occurred')
            CONF.has_first_boot_occurred = True

    def end_compute_with_hmc(self, context, data):
        LOG.debug('Received end_compute_with_hmc cast %s' % data)
        hmc_ip = data['hmc_ip']
        if hmc_ip == CONF.powervm_mgr:
            # this is needed to trigger the health monitoring
            LOG.error(_('Received 401 unauthorized error from HMC %s')
                      % hmc_ip)
            sys.exit()

    @send_ui_notification(notify_msg.STOP_ERROR,
                          notify_msg.STOP_SUCCESS,
                          log_exception=True)
    def stop_instance(self, context, instance, clean_shutdown=True):
        """Stopping an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).\
            stop_instance(context, instance=instance,
                          clean_shutdown=clean_shutdown)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(notify_msg.START_ERROR,
                          notify_msg.START_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_START)
    def start_instance(self, context, instance):
        """Starting an instance on this host."""
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).start_instance(context,
                                                          instance=instance)

    @send_ui_notification(notify_msg.PAUSE_ERROR,
                          notify_msg.PAUSE_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_PAUSE)
    def pause_instance(self, context, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).pause_instance(context,
                                                          instance=instance)

    @send_ui_notification(notify_msg.RESUME_ERROR,
                          notify_msg.RESUME_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_UNPAUSE)
    def unpause_instance(self, context, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).unpause_instance(context,
                                                            instance=instance)

    @send_ui_notification(notify_msg.SUSPEND_ERROR,
                          notify_msg.SUSPEND_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_SUSPEND)
    def suspend_instance(self, context, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).suspend_instance(context,
                                                            instance=instance)

    @send_ui_notification(notify_msg.RESUME_ERROR,
                          notify_msg.RESUME_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_RESUME)
    def resume_instance(self, context, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).resume_instance(
            context, instance=instance)

    @manager.delete_image_on_error
    @send_ui_notification(notify_msg.CAPTURE_ERROR,
                          notify_msg.CAPTURE_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_CAPTURE)
    def snapshot_instance(self, context, image_id, instance):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).snapshot_instance(context,
                                                             image_id,
                                                             instance=instance)

    @send_ui_notification(notify_msg.RESTART_ERROR,
                          notify_msg.RESTART_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_REBOOT)
    def reboot_instance(self, context, instance, block_device_info,
                        reboot_type):
        # Instance must be passed as a keyword argument in order for
        # revert_task_state decorator to function properly
        super(PowerVCComputeManager, self).reboot_instance(
            context, instance, block_device_info, reboot_type)

    @pvc_comp_mgr_reverts_task_state
    @send_ui_notification(notify_msg.EVACUATE_ERROR,
                          notify_msg.EVACUATE_SUCCESS,
                          log_exception=True)
    def rebuild_instance(self, context, instance, orig_image_ref, image_ref,
                         injected_files, new_pass, orig_sys_metadata,
                         bdms, recreate, on_shared_storage,
                         preserve_ephemeral=False):
        """Destroy and re-make this instance.

        A 'rebuild' effectively purges all existing data from the system and
        remakes the VM with given 'metadata' and 'personalities'.

        :param context: `nova.RequestContext` object
        :param instance: Instance object
        :param orig_image_ref: Original image_ref before rebuild
        :param image_ref: New image_ref for rebuild
        :param injected_files: Files to inject
        :param new_pass: password to set on rebuilt instance
        :param orig_sys_metadata: instance system metadata from pre-rebuild
        :param bdms: block-device-mappings to use for rebuild
        :param recreate: True if the instance is being recreated (e.g. the
            hypervisor it was on failed) - cleanup of old state will be
            skipped.
        :param on_shared_storage: True if instance files on shared storage
        :param preserve_ephemeral: True if the default ephemeral storage
                                   partition must be preserved on rebuild
        """
        is_kvm = False
        if CONF.powervm_mgr_type is None:
            is_kvm = (CONF.libvirt.virt_type == 'kvm')

        if is_kvm:
            if not on_shared_storage:
                msg = (_("Evacuation is not supported if storage "
                         "is not shared."))
                LOG.error(msg)
                raise Exception(msg)
            if self._instance_on_disk(instance):
                # check if image-backing file path meets
                # the storage requirements.
                instance_base_path = libvirt_utils.get_instance_path(instance)
                disk_file = instance_base_path + '/disk'
                backing_file_path = libvirt_utils.\
                    get_disk_backing_file(disk_file, basename=False)
                LOG.debug("Backing file path: %s" % backing_file_path)
                # there isn't always a backing file with PowerKVM VM.
                # We need to handle cases where there is no backing file
                if (backing_file_path and
                        not re.match("^%s" % CONF.instances_path,
                                     backing_file_path)):
                    msg = (_("Backing file %(backing_file)s of image "
                             "%(disk_file)s is missing. "
                             "Ensure that the source and target hosts "
                             "have the same mount path.") %
                           {'backing_file': backing_file_path,
                            'disk_file': disk_file})
                    LOG.error(msg)
                    raise Exception(msg)

            super(PowerVCComputeManager, self).\
                rebuild_instance(context, instance=instance,
                                 orig_image_ref=orig_image_ref,
                                 image_ref=image_ref,
                                 injected_files=injected_files,
                                 new_pass=new_pass,
                                 orig_sys_metadata=orig_sys_metadata,
                                 bdms=bdms, recreate=recreate,
                                 on_shared_storage=on_shared_storage,
                                 preserve_ephemeral=preserve_ephemeral)

            # Store the dest host for the instance to which the instance
            # evacuate
            dest_host = instance.host
            # Update the instance system metadata so that the flag is
            # used for the destroy purpose of the VM's which are evacuated.
            sys_meta = instance.system_metadata
            sys_meta['evacuated_to'] = dest_host
            self._instance_update(context, instance.uuid,
                                  system_metadata=sys_meta)
            # Log in info about the instance which is evacuated.
            LOG.info(_("Instance %(instance_name)s was remote restarted to"
                       " host %(dest_host)s") %
                     {'instance_name': instance.name,
                      'dest_host': dest_host})

        else:
            self.rebuild_instance_pvm(context, instance, orig_image_ref,
                                      image_ref, injected_files, new_pass,
                                      orig_sys_metadata, bdms, recreate,
                                      on_shared_storage, preserve_ephemeral)

    @wrap_exception()
    @manager.wrap_instance_event
    @manager.wrap_instance_fault
    def rebuild_instance_pvm(self, context, instance, orig_image_ref,
                             image_ref, injected_files, new_pass,
                             orig_sys_metadata, bdms, recreate,
                             on_shared_storage, preserve_ephemeral=False):
        """Destroy and re-make this instance.

        A 'rebuild' effectively purges all existing data from the system and
        remakes the VM with given 'metadata' and 'personalities'.

        :param context: `nova.RequestContext` object
        :param instance: Instance object
        :param orig_image_ref: Original image_ref before rebuild
        :param image_ref: New image_ref for rebuild
        :param injected_files: Files to inject
        :param new_pass: password to set on rebuilt instance
        :param orig_sys_metadata: instance system metadata from pre-rebuild
        :param bdms: block-device-mappings to use for rebuild
        :param recreate: True if the instance is being recreated (e.g. the
            hypervisor it was on failed) - cleanup of old state will be
            skipped.
        :param on_shared_storage: True if instance files on shared storage
        :param preserve_ephemeral: True if the default ephemeral storage
                                   partition must be preserved on rebuild
        """
        context = context.elevated()
        # NOTE (ndipanov): If we get non-object BDMs, just get them from the
        # db again, as this means they are sent in the old format and we want
        # to avoid converting them back when we can just get them.
        # Remove this on the next major RPC version bump
        if (bdms and
            any(not isinstance(bdm, obj_base.NovaObject)
                for bdm in bdms)):
            bdms = None

        orig_vm_state = instance.vm_state
        with self._error_out_instance_on_exception(context, instance):
            LOG.info(_("Rebuilding instance"), context=context,
                     instance=instance)

            self._check_instance_exists(context, instance)

            # NOTE(mriedem): On a recreate (evacuate), we need to update
            # the instance's host and node properties to reflect it's
            # destination node for the recreate.
            node_name = None
            source_host = instance.host
            source_node = instance.node
            try:
                compute_node = self._get_compute_info(context, self.host)
                node_name = compute_node.hypervisor_hostname
            except exception.ComputeHostNotFound:
                LOG.exception(_LE('Failed to get compute_info for %s'),
                              self.host)
            finally:
                instance.host = self.host
                instance.node = node_name
                instance.save()

            if image_ref:
                image_meta = self.image_api.get(context, image_ref)
            else:
                image_meta = {}

            instance.power_state = self._get_power_state(context, instance)
            instance.task_state = task_states.REBUILDING
            instance.save(expected_task_state=[task_states.REBUILDING])

            self.network_api.setup_networks_on_host(context, instance,
                                                    self.host)
            self.network_api.setup_instance_network_on_host(context,
                                                            instance,
                                                            self.host)

            network_info = compute_utils.get_nw_info_for_instance(instance)
            if bdms is None:
                bdms = objects.BlockDeviceMappingList.\
                    get_by_instance_uuid(context, instance.uuid)

            block_device_info = \
                self._get_instance_block_device_info(context, instance,
                                                     bdms=bdms)

            def detach_block_devices(context, bdms):
                for bdm in bdms:
                    if bdm.is_volume:
                        volume = self.volume_api.get(context,
                                                     bdm.volume_id)
                        attachment = self._get_volume_attachment(
                            volume, instance['uuid'])
                        self.volume_api.detach(context, bdm.volume_id,
                                               attachment)

            files = self._decode_files(injected_files)

            kwargs = dict(
                context=context,
                instance=instance,
                image_meta=image_meta,
                injected_files=files,
                admin_password=new_pass,
                bdms=bdms,
                detach_block_devices=detach_block_devices,
                attach_block_devices=self._prep_block_device,
                block_device_info=block_device_info,
                network_info=network_info,
                preserve_ephemeral=preserve_ephemeral,
                recreate=recreate,
                source_host=source_host)
            try:
                self.driver.rebuild(**kwargs)
            except Exception:
                with excutils.save_and_reraise_exception():
                    # Set the instance source host back to original as the
                    # evacuation operation failed.
                    instance.host = source_host
                    instance.node = source_node
                    instance.save()

            self._update_instance_after_spawn(context, instance)
            instance.save(expected_task_state=[task_states.REBUILDING])

            if orig_vm_state == vm_states.STOPPED:
                LOG.info(_LI("bringing vm to original state: '%s'"),
                         orig_vm_state, instance=instance)
                instance.vm_state = vm_states.ACTIVE
                instance.task_state = task_states.POWERING_OFF
                instance.progress = 0
                instance.save()
                self.stop_instance(context, instance)

            self._update_scheduler_instance_info(context, instance)
            self._notify_about_instance_usage(context, instance, "rebuild.end",
                                              network_info=network_info)

    def _rebuild_default_impl(self, context, instance, image_meta,
                              injected_files, admin_password, bdms,
                              detach_block_devices, attach_block_devices,
                              network_info=None,
                              recreate=False, block_device_info=None,
                              preserve_ephemeral=False):
        '''
            The method is overriden for LP bug 1430042 and for
            RTC defect 39255.
            Once LP bug is fixed we will remove the method.
        '''
        if preserve_ephemeral:
            # The default code path does not support preserving ephemeral
            # partitions.
            raise exception.PreserveEphemeralNotSupported()

        detach_block_devices(context, bdms)

        if not recreate:
            self.driver.destroy(context, instance, network_info,
                                block_device_info=block_device_info)
        else:
            self.network_api.setup_instance_network_on_host(context, instance,
                                                            instance.host)

        instance.task_state = task_states.REBUILD_BLOCK_DEVICE_MAPPING
        instance.save(expected_task_state=[task_states.REBUILDING])

        new_block_device_info = attach_block_devices(context, instance, bdms)

        instance.task_state = task_states.REBUILD_SPAWNING
        instance.save(
            expected_task_state=[task_states.REBUILD_BLOCK_DEVICE_MAPPING])

        self.driver.spawn(context, instance, image_meta, injected_files,
                          admin_password, network_info=network_info,
                          block_device_info=new_block_device_info)

    def _instance_on_disk(self, instance):
        # ensure directories exist and are writable
        instance_path = libvirt_utils.get_instance_path(instance)
        LOG.debug('Checking instance files accessibility %s', instance_path)
        shared_instance_path = os.access(instance_path, os.W_OK)
        return shared_instance_path

    @send_ui_notification(notify_msg.DEPLOY_ERROR, notify_msg.DEPLOY_SUCCESS,
                          log_exception=True)
    def _build_and_run_instance(self, context, instance, image, injected_files,
                                admin_password, requested_networks,
                                security_groups, block_device_mapping, node,
                                limits, filter_properties):

        # Monkey patch method to intercept instance.save() calls
        def proxy_save(self, *args, **kwargs):
            # If super() is marking it active, let's change that!
            if self.vm_state == vm_states.ACTIVE and self.task_state is None:
                is_kvm = False
                if CONF.powervm_mgr_type is None:
                    is_kvm = (CONF.libvirt.virt_type == 'kvm')

                if not is_kvm:
                    # For PowerVM, set task states differently
                    # ISO deploys will go to stopped with no task state
                    # Deploys with do not boot option set to true will go
                    # to stopped with no task state
                    if instance['metadata'].get(
                            'do_not_boot') is not None and (
                            instance['metadata'].get(
                            'do_not_boot').upper() == 'TRUE'):
                        instance.vm_state = vm_states.STOPPED
                        instance.task_state = None
                    else:
                        # Non-ISO PowerVM deploys go to the ACTIVATING
                        # task state
                        instance.vm_state = vm_states.ACTIVE
                        instance.task_state = ACTIVATING

            # Call the original instance method
            self.orig_save(*args, **kwargs)

        # Monkey patch the object to call our version of instance.save()
        # Copy the Class save() and bind it to the instance object
        instance.orig_save = types.MethodType(instance_obj.Instance.save,
                                              instance)

        # Bind our version of the save() to the instance object
        instance.save = types.MethodType(proxy_save, instance)
        """Spawn an instance with error logging and update its power state.
           Override to send notification.
        """
        super(PowerVCComputeManager, self)._build_and_run_instance(
            context, instance, image,
            injected_files, admin_password,
            requested_networks, security_groups,
            block_device_mapping, node, limits,
            filter_properties)

        # Undo the monkey patch.  The methods are already bound to the instance
        instance.save = instance.orig_save

    def _do_build_and_run_instance(self, context, instance, image,
                                   request_spec, filter_properties,
                                   admin_password, injected_files,
                                   requested_networks, security_groups,
                                   block_device_mapping,
                                   node=None, limits=None):
        # OSEE 192018
        # Until it's fixed in Juno, we will override this method to store
        # context in strong store.
        context.update_store()
        super(PowerVCComputeManager, self)._do_build_and_run_instance(
            context, instance, image,
            request_spec, filter_properties, admin_password, injected_files,
            requested_networks, security_groups, block_device_mapping,
            node=node, limits=limits)

    def handle_lifecycle_event(self, event):
        '''
        Override to handle new events used in HMC environments.
        For PowerKVM, since state of the VM is overridden in the
        case of SUSPEND, we need to call get_info() to get
        the real state rather than relying on the event data.
        '''
        context = nova.context.get_admin_context()
        instance = instance_obj.Instance.get_by_uuid(
            context, event.get_instance_uuid())

        # Handle the event issued when the host of an instance goes down.
        if event.get_transition() == hmcevent.EVENT_HOST_DOWN:
            host_power_state = event.get_host_state()
            # Update states only if there is not a current task on the instance
            if instance.task_state is None:
                # If the host was powered down, then the instance will have
                # been powered off, so update the vm_state in the database
                if (host_power_state is not None and
                    'power off' in host_power_state and
                        instance.vm_state == vm_states.ACTIVE):
                    LOG.info(_("Host down event on VM %(uuid)s, because state"
                               " of the host is not operational. Updating VM"
                               " state to %(vm_state)s and task state to"
                               " %(task_state)s.") %
                             {'uuid': event.get_instance_uuid(),
                              'vm_state': vm_states.STOPPED,
                              'task_state': WAITING_ON_HOST})
                    self._instance_update(context, instance['uuid'],
                                          vm_state=vm_states.STOPPED,
                                          task_state=WAITING_ON_HOST)
                else:
                    LOG.info(_("Host down event on VM %(uuid)s, because state"
                               "of the host is not operational. Updating task"
                               " state of VM to %(task_state)s.") %
                             {'uuid': event.get_instance_uuid(),
                              'task_state': WAITING_ON_HOST})
                    self._instance_update(context, instance['uuid'],
                                          task_state=WAITING_ON_HOST)

            return
        # Handle the event to clear the waiting on host task state
        elif event.get_transition() == hmcevent.EVENT_CLEAR_TASK:
            if instance.task_state == WAITING_ON_HOST:
                self._instance_update(context, instance['uuid'],
                                      task_state=None)
            return

        inst_info = self.driver.get_info(instance)
        vm_power_state = inst_info.state

        LOG.info(_("Life cycle event %(state)d on VM %(uuid)s. State is"
                   " %(vm_state)s.") %
                 {'state': event.get_transition(),
                  'uuid': event.get_instance_uuid(),
                  'vm_state': vm_power_state})

        self._sync_instance_power_state(context,
                                        instance,
                                        vm_power_state)

    def _sync_instance_power_state(self, context, db_instance, vm_power_state,
                                   use_slave=False):
        """Align instance power state between the database and hypervisor.

        If the instance is not found on the hypervisor, but is in the database,
        then a stop() API will be called on the instance.
        """

        # We re-query the DB to get the latest instance info to minimize
        # (not eliminate) race condition.
        db_instance.refresh()
        db_power_state = db_instance.power_state
        vm_state = db_instance.vm_state

        if self.host != db_instance.host:
            # on the sending end of nova-compute _sync_power_state
            # may have yielded to the greenthread performing a live
            # migration; this in turn has changed the resident-host
            # for the VM; However, the instance is still active, it
            # is just in the process of migrating to another host.
            # This implies that the compute source must relinquish
            # control to the compute destination.
            LOG.info(_("During the sync_power process the "
                       "instance has moved from "
                       "host %(src)s to host %(dst)s") %
                     {'src': self.host,
                      'dst': db_instance.host},
                     instance=db_instance)
            return
        elif db_instance.task_state is not None:
            # on the receiving end of nova-compute, it could happen
            # that the DB instance already report the new resident
            # but the actual VM has not showed up on the hypervisor
            # yet. In this case, let's allow the loop to continue
            # and run the state sync in a later round
            LOG.info(_("During sync_power_state the instance has a "
                       "pending task. Skip."), instance=db_instance)
            return

        if vm_power_state != db_power_state:
            # power_state is always updated from hypervisor to db
            db_instance.power_state = vm_power_state
            db_instance.save()
            db_power_state = vm_power_state

        # Note(maoy): Now resolve the discrepancy between vm_state and
        # vm_power_state. We go through all possible vm_states.
        if vm_state in (vm_states.BUILDING,
                        vm_states.RESCUED,
                        vm_states.RESIZED,
                        vm_states.ERROR):
            # TODO(maoy): we ignore these vm_state for now.
            pass
        elif vm_state == vm_states.ACTIVE:
            # The only rational power state should be RUNNING
            if vm_power_state in (power_state.SHUTDOWN,
                                  power_state.CRASHED):
                # double check on vm_power_state before we running the stop api
                # if the vm_power_state is the same, just return and wait
                # for the next sync
                new_vm_pwr_state = self._get_power_state(context, db_instance)
                if vm_power_state != new_vm_pwr_state:
                    LOG.warn(_('During sync_power_state the instance '
                               'state is not up to date. Skip.'),
                             instance=db_instance)
                    return
                LOG.warn(_("Instance shutdown by itself. Calling "
                           "the stop API."), instance=db_instance)
                try:
                    # Note(maoy): here we call the API instead of
                    # brutally updating the vm_state in the database
                    # to allow all the hooks and checks to be performed.
                    self.compute_api.stop(context, db_instance)
                except Exception:
                    # Note(maoy): there is no need to propagate the error
                    # because the same power_state will be retrieved next
                    # time and retried.
                    # For example, there might be another task scheduled.
                    LOG.exception(_("error during stop for shutdown/crashed"),
                                  instance=db_instance)
            elif vm_power_state == power_state.SUSPENDED:
                # Instance is suspended. Setting state of
                # instance to suspended in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.SUSPENDED)
            elif vm_power_state == power_state.PAUSED:
                # Instance is paused. Setting state of
                # instance to paused in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.PAUSED)
            elif vm_power_state == power_state.NOSTATE:
                # Occasionally, depending on the status of the hypervisor,
                # which could be restarting for example, an instance may
                # not be found.  Therefore just log the condidtion.
                LOG.warn(_("Instance is unexpectedly not found. Ignore."),
                         instance=db_instance)
        elif vm_state in (vm_states.STOPPED,
                          vm_states.PAUSED,
                          vm_states.SUSPENDED):
            if vm_power_state == power_state.RUNNING:
                # Instance is running. Setting state of
                # instance to active in DB accordingly.
                self.set_vm_state(db_instance, vm_power_state,
                                  vm_states.ACTIVE)
            elif vm_power_state == power_state.SHUTDOWN:
                if vm_state != vm_states.STOPPED:
                    # Instance is stopped. Setting
                    # state of instance to stopped in DB.
                    self.set_vm_state(db_instance, vm_power_state,
                                      vm_states.STOPPED)
            elif vm_power_state == power_state.SUSPENDED:
                if vm_state != vm_states.SUSPENDED:
                    # If an instance is suspended out of band, its vm_state
                    # could become 'stopped' or 'paused' first. Sync up
                    # vm_state with power_state here.
                    self.set_vm_state(db_instance, vm_power_state,
                                      vm_states.SUSPENDED)
            else:
                # Do nothing if instance is in following state
                # as reported by hypervisor
                # NOSTATE
                # CRASHED
                # BUILDING
                LOG.warn(_("Instance state is %(s1)s as reported by "
                           "hypervisor, but its state in PowerVC is "
                           "%(s2)s. Ignore.")
                         % {'s1': vm_power_state, 's2': vm_state})
        elif vm_state in (vm_states.SOFT_DELETED,
                          vm_states.DELETED):
            if vm_power_state not in (power_state.NOSTATE,
                                      power_state.SHUTDOWN):
                # Note(maoy): this should be taken care of periodically in
                # _cleanup_running_deleted_instances().
                LOG.warn(_("Instance is not (soft-)deleted."),
                         instance=db_instance)

    def set_vm_state(self, db_instance, vm_power_state, new_state):
        """
        Set the vm_state of an instance to new_state
        """
        LOG.warn(_('Current instance state of %(name)s is %(old_state)s. '
                   'Instance power state is %(power_state)s. '
                   'Setting instance state to %(new_state)s '
                   'in database.') %
                 {'name': db_instance['display_name'],
                  'old_state': db_instance['vm_state'],
                  'power_state': vm_power_state, 'new_state': new_state})
        db_instance.vm_state = new_state
        db_instance.save()

    def _error_notify(self, instance, context, severity, details):
        info = {'msg': _('{message} Virtual Server: instance UUID:'
                         '{instance_id}, instance name: {instance_name} '
                         'on host: {host_name}'),
                'message': details,
                'instance_id': instance['uuid'],
                'instance_name': instance['display_name'],
                'host_name': self.host}
        notifier = rpc.get_notifier(service='compute', host=self.host)
        if severity == 'error':
            notifier.error(context, 'compute.instance.log', info)
        elif severity == 'info':
            notifier.info(context, 'compute.instance.log', info)
        elif severity == 'critical':
            notifier.critical(context, 'compute.instance.log', info)
        elif severity == 'warn':
            notifier.warn(context, 'compute.instance.log', info)
        elif severity == 'debug':
            notifier.debug(context, 'compute.instance.log', info)

    def _running_deleted_instances(self, context):
        """Override the running deleted to filter unmanaged VM's"""
        insts = super(PowerVCComputeManager,
                      self)._running_deleted_instances(context)
        #VM's that are unmanaged by PowerVC are being marked as soft deleted
        return [i for i in insts if i['vm_state'] != vm_states.SOFT_DELETED]

    def _query_db_instances(self, context, filter_instances=None):
        """Helper method to filter out the subset of the DB Instances"""
        filtered_instances = list()
        context = context.elevated()
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        # Query all of the Instance for the Host from the Database
        db_instances = instfact.find_all_instances_by_host(context, self.host)
        # If they didn't provide any Filter criteria, just return all
        if not filter_instances:
            return db_instances
        # Build a Map of UUID's for the Instances, for quicker lookup
        filter_map = dict([(instance['uuid'],
                            instance) for instance in filter_instances])
        # Loop through each of the DB Instances, seeing if there is a match
        for db_instance in db_instances:
            if db_instance['uuid'] in filter_map:
                filtered_instances.append(db_instance)
        return filtered_instances

    def _populate_admin_context(self, context):
        """Helper method to populate the Token/ServiceCatalog on the Context"""
        tmp_context = populate_admin_context(
            context, self.context_service_catalog)
        if tmp_context:
            self.context_service_catalog = tmp_context.service_catalog

    def _check_neutron_service(self, context):
        neutronv2.api.get_client(context, True).list_networks()

    def _check_cinder_service(self, context):
        cinderv2.cinderclient(context).client.get('/os-hosts')

    def _check_glance_service(self, context):
        glance.GlanceClientWrapper().call(context, 1, 'list')

    def _retry_operation(self, function, context,
                         max_retry_count, sleep_time):
        for x in range(max_retry_count):
            try:
                function(context)
                # if call succeeds then
                # break the loop
                break
            #Log the exception and re-throw
            #one we can parse in the Registration
            except Exception as e:
                with excutils.save_and_reraise_exception() as ctxt:
                    if x < max_retry_count - 1:
                        ctxt.reraise = False
                        greenthread.sleep(sleep_time)
                        LOG.info(_("Exception %s occurred. "
                                   "Retry in progress...") %
                                 six.text_type(e))
                    else:
                        LOG.exception(_('Error trying to connect to the '
                                        'Management System %s')
                                      % six.text_type(e))
                        raise Exception(_('Failed to connect to the '
                                        'Management System'))

    def _verify_mgmt_system_access(self, context):
        """Helper method to verify the Management System can be accessed"""
        max_retry_count = CONF.verify_mgmt_sys_max_retry_count
        sleep_time = CONF.verify_mgmt_sys_sleep_time_during_retry
        self._retry_operation(self._populate_admin_context, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_cinder_service, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_glance_service, context,
                              max_retry_count, sleep_time)
        self._retry_operation(self._check_neutron_service, context,
                              max_retry_count, sleep_time)

    #######################################################
    #######  Discovery-related Overrides/Extensions  ######
    #######################################################
    def discover_instances(self, context):
        """Returns a list of all of the VM's that exist on the Host"""
        driver = self.driver
        # Currently we won't throw an exception if it isn't a Discover Driver
        if not isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
            drvclass = driver.__class__.__name__
            LOG.warn(_('Driver %s does not implement Discover Driver')
                     % drvclass)
            return {'identifier': None, 'servers': [], 'chunking': False}
        # Call the Compute Discovery Driver to get a list of existing VM's
        all_instances = self._discover_instances(context)
        # We need to modify the Instances returned from the Driver slightly
        self._manipulate_driver_instances(context, all_instances, False)
        # Break up the list of VM's to a set of Chunks to be returned
        identifier = self._persist_instances_chunks(all_instances)
        # Return the first chunk of persisted volumes to the caller
        return self.get_next_instances_chunk(context, identifier)

    def query_instances(self, context, instance_ids, allow_unsupported=False):
        """Returns details about the VM's that were requested"""
        return_all = '*all' in instance_ids
        match_instances, query_instances = (list(), list())
        # First we need to figure out what VM's really exist on the Host
        all_instances = self._discover_instances(context)
        uuid_map = dict([(inst_id, '') for inst_id in instance_ids])
        # Loop through each VM's on the Host, seeing if we should query
        for instance in all_instances:
            # See if they requested all instances or this specific one
            if return_all or instance['uuid'] in uuid_map:
                support = instance.get('support', {})
                supported = support.get('status', 'supported')
                # If this is managed, then no need to do a query, just return
                if instance.get('managed') is True:
                    match_instances.append(instance)
                # If this isn't supported and no override, just return it
                elif not allow_unsupported and supported == 'not_supported':
                    match_instances.append(instance)
                # Otherwise it is supported/un-managed and a match so query
                else:
                    instance.pop('support', instance.pop('managed', None))
                    query_instances.append(instance)
        # Only worth calling the Driver if some VM's exist on the System
        if len(query_instances) > 0:
            instances = self.driver.query_instances(context, query_instances)
            match_instances.extend(instances)
        # We need to modify the Instances returned from the Driver slightly
        self._manipulate_driver_instances(
            context, match_instances, True, all_instances)
        # Break up the list of VM's to a set of Chunks to be returned
        identifier = self._persist_instances_chunks(match_instances)
        # Return the first chunk of persisted VM's to the caller
        return self.get_next_instances_chunk(context, identifier)

    @periodic_task.periodic_task(spacing=CONF.inventory_instances_interval)
    def inventory_instances(self, context, instances=None):
        """Periodic Task to Refresh the Inventory for the Instances"""
        driver, context = (self.driver, context.elevated())
        LOG.debug('Invoking Method to Refresh Instance Inventory - Begin')
        try:
            drv_insts = self.driver.list_instances()
            name_map = dict([(inst, '') for inst in drv_insts])
            # Retrieve All of the Instances from the Nova DB for the Given Host
            db_insts = self._query_db_instances(context, instances)
            db_insts = [inst for inst in db_insts if inst['name'] in name_map]
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                # We will also allow the Driver to handle Evacuated Instances
                driver.handle_evacuated_instances(context, db_insts, drv_insts)
                # Now we can actually collect Inventory on the Instances
                driver.inventory_instances(context, db_insts)
        except Exception as exc:
            LOG.warn(_('Error refreshing Instance Inventory'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Instance Inventory - End')

    def verify_host_running(self, context):
        """Verifies the nova-compute service for the Host is running"""
        return True

    @logcall
    @periodic_task.periodic_task(spacing=CONF.inventory_host_interval)
    @lockutils.synchronized('powervc-inventory', 'host-inventory-')
    def inventory_host(self, context):
        """Periodic Task to Refresh the Inventory for the Host"""
        driver, context = (self.driver, context.elevated())
        try:
            #The Periodic Task Context may not have the Token/ServiceCatalog
            self._populate_admin_context(context)
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                driver.inventory_host(context)
        except Exception as exc:
            LOG.warn(_('Error refreshing Host Inventory'))
            LOG.exception(exc)

    @logcall
    @periodic_task.periodic_task(spacing=CONF.quick_inventory_host_interval)
    @lockutils.synchronized('powervc-inventory', 'host-inventory-')
    def quick_inventory_host(self, context):
        '''
        Periodic Task to do a quick check of whether a full inventory is needed
        or not.  If so, it will initiate the full inventory.
        '''
        driver, context = (self.driver, context.elevated())

        try:
            #The Periodic Task Context may not have the Token/ServiceCatalog
            self._populate_admin_context(context)
            # If this is a Discovery Driver, then let it Gather Inventory
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                if driver.is_inventory_needed(context):
                    driver.inventory_host(context)
        except Exception as exc:
            LOG.exception(exc)

    @logcall
    def get_host_ovs(self, context):
        retVal = None
        try:
            if hasattr(self.driver, 'get_host_ovs'):
                LOG.debug('Retrieving host ovs data using Driver...')
                retVal = self.driver.get_host_ovs(context)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.exception(exc)

            error_obj = dom_kvm.Error("%s" % exc)
            ovs_dom = dom_kvm.HostOVSNetworkConfig(CONF.get("host"))

            if retVal is None:
                ovs_dom.error_list = [error_obj]

            else:
                ovs_dom.from_dict(retVal)
                ovs_dom.error_list.append(error_obj)

            retVal = ovs_dom.to_dict()

        return retVal

    @logcall
    def update_host_ovs(self, context, dom, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'update_host_ovs'):
                LOG.debug('Updating host ovs data using Driver...')
                retVal = self.driver.update_host_ovs(context,
                                                     dom,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    @logcall
    def create_host_ovs(self, context, dom, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'create_host_ovs'):
                LOG.debug('Creating host ovs using Driver...')
                retVal = self.driver.create_host_ovs(context,
                                                     dom,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    @logcall
    def delete_host_ovs(self, context, ovs_list, force_flag, rollback):
        retVal = {}
        try:
            if hasattr(self.driver, 'delete_host_ovs'):
                LOG.debug('Deleting host ovs using Driver...')
                retVal = self.driver.delete_host_ovs(context,
                                                     ovs_list,
                                                     force_flag,
                                                     rollback)
        except Exception as exc:
            """
            NOTE: Since this is the interface boundary between the agent
            and server we will not raise the exception and will just
            log it on the host side
            """
            LOG.error(exc)

            if agent.ERRORS_KEY not in retVal:
                retVal[agent.ERRORS_KEY] = [{'message': six.text_type(exc)}]
            else:
                retVal[agent.ERRORS_KEY].append(
                    {'message': six.text_type(exc)})

        return retVal

    def unmanage_host(self, context):
        """Allows the Driver to do any necessary cleanup on Host Removal"""
        # We only want to call unmanage on the Driver if it implements it
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('Unmanaging Host, allowing Driver to do cleanup...'))
            self.driver.unmanage_host(context)

    def prepare_instance(self, context, instance, preparation):
        """Prepare a VM for onboarding by changing the VM configuration"""
        # We only want to call prepare on the Driver if it implements it
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('prepare_instance called in manager'))
            instance = self.driver.prepare_instance(context, instance,
                                                    preparation)
            if instance:
                self._manipulate_driver_instances(context, [instance], True)
            return instance

    def finish_onboard(self, context, instance, preparation=None):
        """finish a VM onboarding by recreating the domain"""

        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            LOG.info(_('finish_onboard for instance %s') % instance['name'])
            instance = instance_obj.Instance.get_by_uuid(
                context, instance['uuid'])

            network_info = self._get_instance_nw_info(context, instance)

            block_device_info = self._get_instance_block_device_info(
                context, instance)

            self.driver.finish_onboard(context,
                                       instance,
                                       preparation,
                                       network_info,
                                       block_device_info)

    @lockutils.synchronized('instance_chunking', 'nova-')
    def get_next_instances_chunk(self, context, identifier):
        """Provides Chunking of VM's Lists to avoid QPID Limits"""
        instance_chunks = self.instance_chunk_map.get(identifier)
        # If the Identifier doesn't exist, we will just return an empty list
        if instance_chunks is None:
            return dict(identifier=identifier, servers=[], chunking=False)
        # If this is the last chunk (or no chunking), just return that list
        if len(instance_chunks) == 1:
            self.instance_chunk_map.pop(identifier, None)
            return dict(identifier=identifier,
                        servers=instance_chunks[0], chunking=False)
        # Otherwise return the first chunk and say that there are more left
        self.instance_chunk_map[identifier] = instance_chunks[1:]
        return dict(identifier=identifier,
                    servers=instance_chunks[0], chunking=True)

    def _discover_instances(self, context):
        """Internal Method to list of all of the VM's that exist on the Host"""
        # Call the Compute Discovery Driver to get a list of existing VM's
        drv_instances = self.driver.discover_instances(context)
        # Generate the UUID's for the VM's and determine which are Managed
        self._generate_instance_uuids(context, drv_instances)
        return drv_instances

    @lockutils.synchronized('instance_chunking', 'nova-')
    def _persist_instances_chunks(self, instances):
        """Internal Helper method to generate the UUID's for the VM's"""
        current_len = 0
        current_list, instance_chunks = (list(), list())
        # First get an identifier to be used to reference this chunking
        identifier = str(self.instance_chunk_counter)
        self.instance_chunk_counter = self.instance_chunk_counter + 1
        # Loop through each instance, breaking it up into chunks based on size
        while len(instances) > 0:
            instance = instances.pop(0)
            instance_len = len(str(instance))
            # If we could possibly go over the 64K Message size, break up
            if len(current_list) > 0 and (current_len + instance_len) > 40000:
                instance_chunks.append(current_list)
                current_list = []
                current_len = 0
            # Add this instance to the current chunk that is being returned
            current_len = current_len + instance_len
            current_list.append(instance)
        # Add the final chunk to the overall set of chunks for the instance
        # list
        instance_chunks.append(current_list)
        self.instance_chunk_map[identifier] = instance_chunks
        return identifier

    def _manipulate_driver_instances(self, context,
                                     drv_instances, query, discinsts=[]):
        """Internal Helper method to modify attributes on the Instances"""
        state_map = {'1': 'active', '3': 'paused', '4': 'stopped',
                     '6': 'error', '7': 'suspended', '9': 'building'}
        name_map = dict([(inst['name'], inst) for inst in discinsts])
        # Loop through each of the Driver Instances, modifying the data
        for inst in drv_instances:
            # Get some things off the discover call if query didn't return them
            disc_inst = name_map.get(inst['name'], {})
            state = inst.pop('state', disc_inst.get('state', 0))
            inst['id'] = inst.pop('uuid', disc_inst.get('uuid'))
            # If this is the result of a Driver Query Instances call
            if query:
                inst['power_state'] = state
                inst['vm_state'] = state_map.get(str(state), 'error')
                inst['memory_mb'] = int(inst.get('memory_mb', 0))
                inst['root_gb'] = int(inst.get('root_gb', 0))
                inst['ephemeral_gb'] = int(inst.get('ephemeral_gb', 0))
                vcpus = str(inst.get('vcpus', 0))
                if vcpus.find('.') > 0:
                    vcpus = vcpus[:vcpus.find('.')]
                inst['vcpus'] = int(vcpus)
            # If this is the result of a Driver Discover Instances call
            else:
                inst['status'] = state_map.get(str(state), 'error')
                inst.pop('power_specs', None)
            #We need to go through the reasons as convert to the Browser Locale
            if inst.get('support') is not None:
                if inst['support'].get('reasons') is not None:
                    reasons = inst['support']['reasons']
                    reasons = [translate_message(context, r) for r in reasons]
                    inst['support']['reasons'] = reasons
                #Also need to convert the Preparation Actions over
                if inst['support'].get('steps') is not None:
                    for step in inst['support']['steps']:
                        msgs = step.get('actions', [])
                        msgs = [translate_message(context, m) for m in msgs]
                        if step.get('actions') is not None:
                            step['actions'] = msgs

    def _generate_instance_uuids(self, context, drv_instances):
        """Internal Helper method to generate the UUID's for the VM's"""
        name_map, uuid_map = (dict(), dict())
        vm_name_map, vm_uuid_map = (dict(), dict())
        # Query the Database to get the Instance info for this Host
        db_instances = self._query_db_instances(context)
        # Loop through each of the Instances, creating the Maps for Lookup
        for db_instance in db_instances:
            power_specs = db_instance.get('power_specs', {})
            vm_uuid = power_specs.get('vm_uuid')
            vm_name = power_specs.get('vm_name')
            vm_uuid = '!' if vm_uuid is None else vm_uuid
            vm_name = '!' if vm_name is None else vm_name
            uuid_map[db_instance['uuid']] = db_instance
            name_map[db_instance.get('name', '!')] = db_instance
            vm_uuid_map[vm_uuid.lower()] = db_instance
            vm_name_map[vm_name] = db_instance
        # Cleanup any default invalid entries we added to the Maps
        name_map.pop('!', None)
        vm_name_map.pop('!', None)
        vm_uuid_map.pop('!', None)
        # Loop through the instances generating UUID's for any that need one
        for drv_vm in drv_instances:
            managed = False
            db_instance = uuid_map.get(drv_vm.get('uuid', '!'))
            power_specs = drv_vm.get('power_specs', {})
            vm_uuid = power_specs.get('vm_uuid')
            vm_name = power_specs.get('vm_name')
            # See if there was a match on the Name with what is in the DB
            if db_instance is None and vm_name is not None:
                db_instance = name_map.get(vm_name)
                if db_instance is None:
                    db_instance = vm_name_map.get(vm_name)
            # See if there was a match on the UUID with what is in the DB
            if db_instance is None and vm_uuid is not None:
                db_instance = uuid_map.get(vm_uuid.lower())
                if db_instance is None:
                    db_instance = vm_uuid_map.get(vm_uuid.lower())
            # If the UUID isn't set, do a secondary lookup to see if it exists
            if db_instance is not None:
                drv_vm['uuid'] = db_instance['uuid']
                drv_vm['name'] = db_instance['display_name']
                managed = True
            # Otherwise it isn't managed, so we need to generate one
            elif not drv_vm.get('uuid'):
                namesp = uuid.UUID('a0dd4880-6115-39d6-b26b-77df18fe749f')
                namestr = self._get_unique_instance_str(drv_vm)
                drv_vm['uuid'] = str(uuid.uuid3(namesp, namestr))
            # Set the Managed attribute based on what we determined earlier
            drv_vm['managed'] = managed

    def _get_unique_instance_str(self, instance):
        """Internal Helper method to create a unique string for the instance"""
        instance_name = instance.get('name', '')
        instance_id = instance.get('hostuuid', '')
        return "ibm-powervc://server/host='%s',id='%s',name='%s'" % \
            (self.host, str(instance_name), str(instance_id))

    #######################################################
    #########  Image-related Overrides/Extensions  ########
    #######################################################
    @periodic_task.periodic_task(spacing=600)
    def _check_instance_activating_time(self, context):
        """Ensure that instances are not stuck in activating."""
        timeout = CONF.ibmpowervm_activating_timeout
        if timeout == 0:
            return
        filters = {'task_state': ACTIVATING, 'host': self.host}
        activating_insts = instance_obj.InstanceList.get_by_filters(context,
                                                                    filters)
        # Loop through all the Activating Instance to do the time outs
        for instance in activating_insts:
            if timeutils.is_older_than(instance['created_at'], timeout):
                self._instance_update(context, instance['uuid'],
                                      task_state=None)
                LOG.warn(_("Instance activating timed out. "
                           "Resetting task state."), instance=instance)

    @periodic_task.periodic_task(
        spacing=CONF.ibmpowervm_optical_cleanup_interval)
    def _check_instance_vopt_removal(self, context):
        """Clean up virtual optical devices after 1 hour."""
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc':
            # only done for HMC
            timeout = CONF.ibmpowervm_iso_detach_timeout
            if timeout < 0:
                return

            insts = self.conductor_api.instance_get_all_by_host(
                context, self.host, columns_to_join=[])
            # Loop through all the Instances to find ones older than 1 hour
            old_instances = []
            for instance in insts:
                if timeutils.is_older_than(instance['created_at'],
                                           timeout):
                    old_instances.append(instance['uuid'])
            self.driver.vopt_cleanup(old_instances)

    def _await_volume_create(self, context, vol_id):
        """
        Wait for volume create to complete. This routine needs to
        wait until the driver moves the volume out of the creating
        state. Eventually it will move to available or error.
        For SSP Thick or EMC clones this can be double digit mins.
        :param context: operation context
        :param vol_id: volume id
        """
        while True:
            volume = self.volume_api.get(context, vol_id)
            volume_status = volume['status']
            # Would we every need downloading? Or other status here?
            if volume_status not in ['creating']:
                if volume_status != 'available':
                    LOG.warn(_("Volume id: %(vol)s finished being created"
                               " with status: '%(status)s'") %
                             {'vol': vol_id,
                              'status': volume_status})
                return
            # for now sleep based on the bd retry interval
            greenthread.sleep(CONF.block_device_allocate_retries_interval)

    def _convert_block_devices(self, bdms):
        """
        Convert a list of Nova object block device mappings into
        nova.virt.block_device wrappered instances.
        """
        driver_bdm_list = blk_dev.convert_volumes(bdms) +\
            blk_dev.convert_snapshots(bdms) +\
            blk_dev.convert_images(bdms) +\
            blk_dev.convert_blanks(bdms)
        return driver_bdm_list

    def _clone_bdm_source_volume(self, context, volume_api, bdm,
                                 instance, instance_flavor_vol_type):
        src_volume = volume_api.get(context, bdm.volume_id)
        LOG.debug('Creating volume from volume %s' % bdm.volume_id)

        source_metadata = src_volume['volume_metadata']

        target_metadata = {'instance_uuid': instance['uuid']}
        prefix = CONF.data_volume_name_prefix
        if 'is_boot_volume' in source_metadata and \
                source_metadata['is_boot_volume'] == 'True':
            target_metadata['is_boot_volume'] = 'True'
            prefix = CONF.boot_volume_name_prefix

        volume_name = self._generate_instance_volume_name(instance,
                                                          bdm,
                                                          prefix)

        return volume_api.create(context, bdm.volume_size, volume_name, '',
                                 source_vol=src_volume,
                                 metadata=target_metadata,
                                 volume_type=instance_flavor_vol_type)

    def _get_image_bdm_volumes_to_clone(self, context, image_ref,
                                        image_metadata=None):
        """
        Get the volumes from the image that we will clone.  We only clone
        volumes in image BDMs if they have delete_on_termination=true.
        """
        if not image_metadata:
            image_metadata = self._get_image_meta(context, image_ref)

        image_properties = image_metadata.get('properties', {})

        # Get the block device mappings defined by the image.
        image_defined_bdms = image_properties.get('block_device_mapping', [])
        image_defined_bdms = map(block_device.BlockDeviceDict,
                                 image_defined_bdms)
        vols_to_clone = []
        for bdm in image_defined_bdms:
            vol_id = bdm.get('volume_id')
            delete_on_termination = bdm.get('delete_on_termination')
            if vol_id and delete_on_termination:
                vols_to_clone.append(vol_id)
        return vols_to_clone

    def _validate_supported_bdms(self, bdms):
        for bdm in bdms:
            if isinstance(bdm, blk_dev.DriverBlankBlockDevice):
                continue
            elif isinstance(bdm, blk_dev.DriverSnapshotBlockDevice):
                exc_args = {'bdm_type': 'snapshot'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverEphemeralBlockDevice):
                exc_args = {'bdm_type': 'ephemeral'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverSwapBlockDevice):
                exc_args = {'bdm_type': 'swap'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverImageBlockDevice):
                exc_args = {'bdm_type': 'image'}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)
            elif isinstance(bdm, blk_dev.DriverVolumeBlockDevice):
                continue
            else:
                exc_args = {'bdm_type': type(bdm).__name__}
                raise pwrvc_excep.IBMPowerVCUnsupportedBDM(**exc_args)

    def _check_for_duplicate_attach(self, instance_uuid, bdms, image_volumes):
        """Check for duplicate volume attach BDMs and raise an exception if
        they are found. Do not prevent duplicate volume_ids in BDMs if the
        volume ID is from an image since the image volumes will be cloned
        before attach.

        @param bdms: The instance BDMs, expected to be a list of nova object
        BDMs
        @param image_volumes: The list of image volume IDs that will be cloned
        """
        found_volumes = []
        for bdm in bdms:
            if bdm.destination_type == 'volume' and bdm.volume_id:
                if bdm.volume_id in image_volumes:
                    continue
                if bdm.volume_id in found_volumes:
                    ex_ags = {'volume_id': bdm.volume_id,
                              'instuuid': instance_uuid}
                    raise pwrvc_excep.IBMPowerVCDuplicateVolumeAttach(**ex_ags)
                found_volumes.append(bdm.volume_id)

    def _prepare_volumes(self, context, volume_api, instance, bdms,
                         image_meta):
        """
        Prepare the volumes from the block device mappings.
        """
        LOG.debug('Preparing volumes for BDMs %s' % bdms)
        for bdm in bdms:
            sanitized_bdm = self._clean_bdm(bdm)
            LOG.debug('BDM: %s' % sanitized_bdm)
        virt_bdms = self._convert_block_devices(bdms)

        self._validate_supported_bdms(virt_bdms)

        image_ref = instance['image_ref']
        volumes_to_clone = self.\
            _get_image_bdm_volumes_to_clone(context, image_ref,
                                            image_metadata=image_meta)
        LOG.debug('Image volumes that will be cloned: %s' % volumes_to_clone)
        self._check_for_duplicate_attach(instance['uuid'],
                                         bdms, volumes_to_clone)
        source_vol_to_volume_type = volume_type_utils.\
            get_volume_types_from_flavor(context, volumes_to_clone,
                                         instance=instance)
        boot_index_to_vol = {}
        for bdm in virt_bdms:
            vol = None
            if isinstance(bdm, blk_dev.DriverBlankBlockDevice):
                if not bdm.volume_id:
                    prefix = CONF.data_volume_name_prefix
                    vol_name = self._generate_instance_volume_name(instance,
                                                                   bdm,
                                                                   prefix)
                    vol = volume_api.create(context, bdm.volume_size,
                                            vol_name, '')
                    LOG.debug('Blank volume created: %s' % vol)
            elif isinstance(bdm, blk_dev.DriverVolumeBlockDevice):
                if bdm.volume_id in volumes_to_clone:
                    flv_vol_type = source_vol_to_volume_type.get(bdm.volume_id)
                    vol_type_id = None
                    if flv_vol_type and 'id' in flv_vol_type:
                        vol_type_id = flv_vol_type['id']
                    vol = self._clone_bdm_source_volume(context, volume_api,
                                                        bdm, instance,
                                                        vol_type_id)
                    # Here we change the delete_on_termination value to True
                    # since we have created a cloned volume from an image
                    # volume and this volume will need to be deleted when the
                    # VM is deleted. We set delete_on_termination on the
                    # BDM object because the nova.virt.block_device class
                    # does not proxy delete_on_termination to the actual BDM
                    # object.
                    bdm._bdm_obj.delete_on_termination = True
                    bdm['delete_on_termination'] = True
                    LOG.debug('Image volume %s cloned to volume %s' %
                              (bdm.volume_id, vol))
            if vol:
                bdm.volume_id = vol['id']
                bdm.save()
                boot_index_to_vol[bdm['boot_index']] = vol

        # Check the volumes before leaving to catch volume creation errors
        # early, before the deploy proceeds to LPAR and network creation.
        bdms = [bdm for bdm in bdms if bdm.is_volume]
        for bdm in bdms:
            volume = volume_api.get(context, bdm['volume_id'])
            blank_volume = (bdm.source_type == 'blank')
            self._validate_for_error_status(context, instance, volume,
                                            blank_volume)

        # Update root_gb on the instance with the size of the volume at
        # boot_index = 0 to allow resize to work as it has in past releases.
        boot_disk_size = boot_index_to_vol[0]['size']
        args = {'root_gb': boot_disk_size}
        self.conductor_api.instance_update(context, instance['uuid'],
                                           **args)
        # update the memory object too.
        instance.update(args)

        return virt_bdms

    def _generate_instance_volume_name(self, instance, virtbdm, boot_or_data):
        """
        Generate volume names for the instance's volumes.
        """
        boot_index = virtbdm.get('boot_index')
        bdm_id = virtbdm._bdm_obj.id
        unique_id = (six.text_type(boot_index)
                     if boot_index is not None else six.text_type(bdm_id))
        if instance.get('host') and instance.get('name'):
            prefix = instance['name'] + "-" + instance['host']
        elif instance.get('uuid'):
            prefix = instance['uuid']
        else:
            prefix = ''
        return prefix + '-' + boot_or_data + unique_id

    def _validate_image_meta(self, image_meta):
        image_properties = image_meta.get('properties', {})
        required_props = ['bdm_v2', 'block_device_mapping']
        for prop in required_props:
            if not image_properties.get(prop):
                exc_args = {'image_name': image_meta['name'],
                            'property_name': prop}
                raise pwrvc_excep.\
                    IBMPowerVCMissingRequiredImageProperty(**exc_args)

    def _prep_block_device(self, context, instance, bdms,
                           do_check_attach=True):
        """
        Override nova ComputeManager's _prep_block_device
        in order to consume the new image meta data for PowerVC to
        prepare the VM's volumes.  Volumes are allocated but not attached here.
        :param context: operation context
        :param instance: compute instance
        :param bdms: The block device mapping based on image mapping
                     and block_device_mapping properties or the
                     block_device_mapping passed as parameter during
                     deploy.
        :return block_device_info:
        """
        if self._is_managing_PowerKVM():
            s = super(PowerVCComputeManager, self)
            return s._prep_block_device(context, instance, bdms)

        # this is the path for PowerVM volume backed VMs
        image_ref = instance['image_ref']
        if not image_ref:
            LOG.warn(_("ComputeMgr _prep_block_device() exception"
                       " since missing image_ref from instance"))
            raise exception.InvalidImageRef(image_href=image_ref)

        if not bdms:
            exc_args = {'instuuid': instance['uuid']}
            raise pwrvc_excep.IBMPowerVCMissingBDMs(**exc_args)

        image_meta = self._get_image_meta(context, image_ref)
        self._validate_image_meta(image_meta)

        self._prepare_volumes(context, self.volume_api, instance, bdms,
                              image_meta)

        blkdev_info = self._get_instance_block_device_info(context,
                                                           instance)
        # extend the block_device_info with call method for HMC driver
        # to call back boot volume attach during spawn.
        blkdev_info['powervc_blockdev_callbacks'] = {
            'validate_volumes': self._validate_volumes,
            'attach_volumes': self._attach_volumes_and_boot,
            'validate_attachment': self._validate_attachment}
        return blkdev_info

    @staticmethod
    def _get_image_meta(context, image_ref):
        image_service, image_id = glance.get_remote_image_service(context,
                                                                  image_ref)
        return image_service.show(context, image_id)

    #######################################################
    #######  Migration-related Overrides/Extensions  ######
    #######################################################
    @wrap_exception()
    @manager.reverts_task_state
    @manager.wrap_instance_event
    @manager.errors_out_migration
    @manager.wrap_instance_fault
    @send_ui_notification(notify_msg.RESIZE_ERROR, None,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_RESIZE)
    def resize_instance(self, context, instance, image,
                        reservations, migration, instance_type,
                        clean_shutdown=True):
        """Starts the migration of a running instance to another host."""

        quotas = quotas_obj.Quotas.from_reservations(context,
                                                     reservations,
                                                     instance=instance)
        with self._error_out_instance_on_exception(context, instance,
                                                   quotas=quotas):
            if not instance_type:
                instance_type = objects.Flavor.get_by_id(
                    context, migration['new_instance_type_id'])

            network_info = self._get_instance_nw_info(context, instance)

            migration.status = 'migrating'
            with migration.obj_as_admin():
                migration.save()
            instance.task_state = task_states.RESIZE_MIGRATING
            instance.save(expected_task_state=task_states.RESIZE_PREP)

            self._notify_about_instance_usage(
                context, instance, "resize.start", network_info=network_info)

            bdms = objects.BlockDeviceMappingList.get_by_instance_uuid(
                context, instance.uuid)
            block_device_info = self._get_instance_block_device_info(
                context, instance, bdms=bdms)

            timeout, retry_interval = self.\
                _get_power_off_values(context, instance, clean_shutdown)
            disk_info = self.driver.migrate_disk_and_power_off(
                context, instance, migration.dest_host,
                instance_type, network_info,
                block_device_info,
                timeout, retry_interval)

            migration_p = obj_base.obj_to_primitive(migration)
            instance_p = obj_base.obj_to_primitive(instance)
            self.network_api.migrate_instance_start(context,
                                                    instance_p,
                                                    migration_p)

            migration.status = 'post-migrating'
            with migration.obj_as_admin():
                migration.save()

            instance.host = migration.dest_compute
            instance.node = migration.dest_node
            instance.task_state = task_states.RESIZE_MIGRATED
            instance.save(expected_task_state=task_states.RESIZE_MIGRATING)

            self.compute_rpcapi.finish_resize(context, instance,
                                              migration, image, disk_info,
                                              migration.dest_compute,
                                              reservations=quotas.reservations)

            self._notify_about_instance_usage(context, instance, "resize.end",
                                              network_info=network_info)
            self.instance_events.clear_events_for_instance(instance)

    @wrap_exception()
    @send_ui_notification(notify_msg.RESIZE_ERROR, notify_msg.RESIZE_SUCCESS,
                          log_exception=True)
    def confirm_resize(self, context, instance, reservations, migration):

        super(PowerVCComputeManager, self). \
            confirm_resize(context, instance, reservations, migration)
        try:
            LOG.info(_("Refreshing statistics for the  %s instance.")
                     % instance.name)
            db_inst = self._query_db_instances(context, [instance])
            self.driver.inventory_instances(context, db_inst)
        except Exception as exc:
            LOG.warn(_('Failed to refresh instance invetory '))
            LOG.exception(exc)

    @wrap_exception()
    @manager.reverts_task_state
    @check_host_maintenance_mode(common_constants.OPER_MIGRATE_AS_TGT)
    def check_can_live_migrate_destination(self, ctxt, instance,
                                           block_migration=False,
                                           disk_over_commit=False):
        """
        When source host validation fails we need to clean up the driver's
        migration object on the target system.  We can't do this via the
        dictionaries passed between the methods.
        NOTE: No need to use @pvc_comp_mgr_reverts_task_state since
        super doesn't have @reverts_task_state
        """

        # Dictionary to pass between all the migration methods
        # on both the source and target host threads
        migrate_data = {}

        try:
            # Because the scheduler filters aren't invoked for targeted
            # migrations, we need to ensure that we aren't migrating outside
            # of the host aggregate domain or violating a collocation rule.
            migrate_utils.verify_aggregate_domain(ctxt, instance.display_name,
                                                  instance.host, CONF.host)
            migrate_utils.verify_collocation_rules(ctxt, instance, CONF.host)
            migrate_data = super(PowerVCComputeManager, self). \
                check_can_live_migrate_destination(
                    ctxt, instance=instance, block_migration=block_migration,
                    disk_over_commit=disk_over_commit)
        except pvm_exc.IBMPowerVMMigrationInProgress:
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(instance)

        return migrate_data

    @wrap_exception()
    def pre_live_migration(self, context, instance, block_migration=False,
                           disk=None, migrate_data=None):

        # Issue 5649: Onboarded lpars do not have a fixed IP assigned causing
        # pre_live_migration to fail.  Rather than wait for a community fix
        # we'll reimplement the manager level pre_live_migration without the
        # fixed IP check.
        try:
            block_device_info = self._get_instance_block_device_info(
                context, instance,
                refresh_conn_info=migrate_data['refresh_conn_info'])
            network_info = self._get_instance_nw_info(context, instance)
        except Exception:
            with excutils.save_and_reraise_exception():
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.get('dest_sys_name')
                migrate_utils.send_migration_failure_notification(context,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)
        self._notify_about_instance_usage(context, instance,
                                          "live_migration.pre.start",
                                          network_info=network_info)

        pre_live_migration_data = self.driver.pre_live_migration(
            context, instance, block_device_info,
            network_info, disk, migrate_data)

        # NOTE(tr3buchet): setup networks on destination host
        self.network_api.setup_networks_on_host(context, instance,
                                                self.host)

        self._notify_about_instance_usage(context, instance,
                                          "live_migration.pre.end",
                                          network_info=network_info)
        return pre_live_migration_data

    @wrap_exception()
    def live_migration(self, context, dest, instance,
                       block_migration=False, migrate_data=None):
        # BDMS will be updated by pre_live_migration() if there are
        # attached cinder volumes. The info in the BDMS may be
        # changed. Save it for roll-back during exception.
        migrate_data = dict(migrate_data or {})
        old_bdms_src = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        migrate_data['pre_migrate_bdms'] = old_bdms_src

        super(PowerVCComputeManager, self).live_migration(
            context, dest, instance, block_migration, migrate_data)

    @wrap_exception()
    @manager.reverts_task_state
    def post_live_migration_at_destination(self, context, instance,
                                           block_migration=False,
                                           migrate_data=None):
        """Post operations for live migration .

        :param context: security context
        :param instance: Instance dict
        :param block_migration: if true, prepare for block migration
        :param migrate_data: implementation specific data dictionary
        """
        # Notify post live at destination has started
        LOG.info(_('Post operation of migration started on destination host'),
                 instance=instance)
        network_info = self._get_instance_nw_info(context, instance)
        self._notify_about_instance_usage(context, instance,
                                          "live_migration.post.dest.start",
                                          network_info=network_info)

        # NOTE(tr3buchet): setup networks on destination host
        #                  this is called a second time because
        #                  multi_host does not create the bridge in
        #                  plug_vifs
        self.network_api.setup_networks_on_host(context, instance, self.host)

        migration = {'source_compute': instance['host'],
                     'dest_compute': self.host, }
        self.network_api.migrate_instance_finish(context, instance, migration)

        try:
            # Call driver
            block_device_info = self._get_instance_block_device_info(
                context, instance)
        except Exception:
            with excutils.save_and_reraise_exception():
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.get('dest_sys_name')
                migrate_utils.send_migration_failure_notification(context,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)

        self.driver.post_live_migration_at_destination(context, instance,
                                                       network_info,
                                                       block_migration,
                                                       block_device_info,
                                                       migrate_data)

        # Restore instance state
        current_power_state = self._get_power_state(context, instance)
        node_name = None
        try:
            compute_node = self._get_compute_info(context, self.host)
            node_name = compute_node['hypervisor_hostname']
        except exception.NotFound:
            LOG.exception(_('Failed to get compute_info for %s') % self.host)
        finally:
            # Do not check for expected task state following a successful
            # migration.  This leads to the health state following to go
            # critical when something other than migration causes the task
            # state to clear. Also, don't clear the task state here, as
            # it will be cleared by a thread kicked off by
            # driver.post_live_migration_at_destination when RMC becomes active
            # use instance.save instead since _instance_update return dict,
            # it will cause error in _notify_about_instance_usage
            instance.host = self.host
            instance.power_state = current_power_state
            instance.vm_state = vm_states.ACTIVE
            instance.task_state = None
            instance.node = node_name
            instance.save()

        # NOTE(vish): this is necessary to update dhcp
        self.network_api.setup_networks_on_host(context, instance, self.host)
        self._notify_about_instance_usage(context, instance,
                                          "live_migration.post.dest.end",
                                          network_info=network_info)

    @wrap_exception()
    @manager.reverts_task_state
    def rollback_live_migration_at_destination(self, context, instance,
                                               destroy_disks=True,
                                               migrate_data=None):

        # We're not calling 'super' so we still need to send notifications
        network_info = self._get_instance_nw_info(context, instance)
        self._notify_about_instance_usage(
            context, instance, "live_migration.rollback.dest.start",
            network_info=network_info)

        # NOTE(tr3buchet): tear down networks on destination host
        self.network_api.setup_networks_on_host(context, instance,
                                                self.host, teardown=True)

        # NOTE(vish): The mapping is passed in so the driver can disconnect
        #             from remote volumes if necessary
        block_device_info = self._get_instance_block_device_info(
            context, instance)

        # Don't call the parent roll-back, we define our own driver level
        self.driver.rollback_live_migration_at_destination(context, instance,
                                                           network_info,
                                                           block_device_info,
                                                           destroy_disks,
                                                           migrate_data)

        # unplug VLAN at Destination during rollback
        self.driver.unplug_vifs(instance, network_info)

        self._notify_about_instance_usage(
            context, instance, "live_migration.rollback.dest.end",
            network_info=network_info)

    @wrap_exception()
    @manager.reverts_task_state
    @manager.wrap_instance_fault
    def _post_live_migration(self, ctxt, instance,
                             dest, block_migration=False, migrate_data=None):
        """Post operations for live migration on source host.
        We override the manager level code as the behavior for
        IVM and HMC needs to be different

        This method is called from live_migration
        and mainly updating database record.

        :param ctxt: security context
        :param instance: Instance object
        :param dest: destination host
        :param block_migration: if true, prepare for block migration
        :param migrate_data: if not None, it is a dict which has data
        required for live migration without shared storage

        """
        try:
            # Notify the post migration has started
            network_info = self._get_instance_nw_info(ctxt, instance)
            self._notify_about_instance_usage(ctxt, instance,
                                              "live_migration._post.start",
                                              network_info=network_info)
            LOG.info(_('_post_live_migration() is started..'),
                     instance=instance)
            LOG.debug('Migration data = %s' % migrate_data)

            # Get the block device info
            bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                ctxt, instance['uuid'])
            block_device_info = self._get_instance_block_device_info(
                ctxt, instance, bdms=bdms)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.driver.migration_cleanup(instance, source=True)
                message = _("An internal error occurred. For more information"
                            ", see log file for host %s." % CONF.host)
                tgt_name = migrate_data.get('dest_sys_name')
                migrate_utils.send_migration_failure_notification(ctxt,
                                                                  instance,
                                                                  tgt_name,
                                                                  message)

        self.driver.post_live_migration(ctxt, instance,
                                        block_device_info, migrate_data)
        # Detach volumes for IVM only.
        # For NPIV/SSP storage, supported only by HMC managed systems,
        # the disk attachment is maintained by powervm during LPM.
        # From openstack's point of view, the attachment is the same no matter
        # which host it is running on. For example, in NPIV scenario, the disk
        # is attached to a virtualized fibre channel adapter assigned to the
        # instance. The attached disk is only visible to the instance.
        # This level of virtualization hides all the implementation details for
        # LPM, and it stays unchanged during the life cycle of the instance
        # regardless which host it is running on.
        refresh_connections = migrate_data['refresh_conn_info']
        if refresh_connections:
            LOG.debug("Terminating the storage connections for migration.")
            connector = self.driver.get_volume_connector(instance)
            for bdm in bdms:
                # NOTE(vish): We don't want to actually mark the volume
                #             detached, or delete the bdm, just remove the
                #             connection from this host.

                # remove the volume connection without detaching
                # from hypervisor because the instance is not
                # running anymore on the current host
                if bdm.is_volume:
                    self.volume_api.terminate_connection(ctxt, bdm.volume_id,
                                                         connector)
        else:
            LOG.debug("Skipping storage operations for migration.")

        # Releasing vlan.
        # (not necessary in current implementation?)
        migration = {'source_compute': self.host,
                     'dest_compute': dest, }
        self.network_api.migrate_instance_start(ctxt, instance, migration)

        # Define domain at destination host, without doing it,
        # pause/suspend/terminate do not work.
        self.compute_rpcapi.post_live_migration_at_destination(
            ctxt, instance, block_migration, dest, migrate_data)

        # NOTE(tr3buchet): tear down networks on source host
        self.driver.unplug_vifs(instance, network_info)
        self.network_api.setup_networks_on_host(ctxt, instance,
                                                self.host, teardown=True)
        self.instance_events.clear_events_for_instance(instance)
        self._update_scheduler_instance_info(ctxt, instance)
        self._notify_about_instance_usage(ctxt, instance,
                                          "live_migration._post.end",
                                          network_info=network_info)
        LOG.info(_('Migrating instance to %s finished successfully.'),
                 dest, instance=instance)

    @wrap_exception()
    @manager.reverts_task_state
    def _rollback_live_migration(self, context, instance, dest,
                                 block_migration, migrate_data=None):
        """Rollback operations for live migration on source host.
        We override the manager level code as the behavior for
        IVM and HMC needs to be different
        """

        # If pre_live_migration or live_migration fails
        # we need to clean up the migration dict on the source.
        self.driver.migration_cleanup(instance, source=True)

        # Notify rollback has started
        LOG.debug('Rollback starting on source host for instance %s. '
                  'Migration data = %s' % (instance['display_name'],
                                           migrate_data))
        self._notify_about_instance_usage(context, instance,
                                          "live_migration._rollback.start")
        # Set the state back to Active
        instance.vm_state = vm_states.ACTIVE
        instance.task_state = None
        instance.save(expected_task_state=[task_states.MIGRATING])

        # NOTE(tr3buchet): setup networks on source host (really it's re-setup)
        self.network_api.setup_networks_on_host(context, instance, self.host)

        # Detach volumes for IVM only.
        refresh_connections = migrate_data['refresh_conn_info']
        if refresh_connections:
            LOG.debug("Removing volume connections for migration.")
            bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                context, instance['uuid'])
            for bdm in bdms:
                if bdm.is_volume:
                    volume_id = bdm.volume_id
                    try:
                        self.compute_rpcapi.remove_volume_connection(context,
                                                                     instance,
                                                                     volume_id,
                                                                     dest)
                    except Exception as ex:
                        LOG.error(_("Failed to remove volume %(volume_id)s "
                                    "connection  due to error: %(error)s") %
                                  {'volume_id': volume_id, 'error': ex})
        else:
            LOG.debug("Skipping removing volume connection for migration.")

        # Rollback on destination
        self.compute_rpcapi.rollback_live_migration_at_destination(
            context, instance, dest, migrate_data=migrate_data)

        # Block device mapping for attached volumes may have been changed
        # by pre_live_migration(). Need to restore it after roll-back.
        self._restore_pre_migrate_bdms(context, migrate_data)

        # Notify we're done
        LOG.debug('Rollback ending on source host for instance %s.' %
                  instance['display_name'])
        self._notify_about_instance_usage(context, instance,
                                          "live_migration._rollback.end")

    #######################################################
    ########  Monitor-related Overrides/Extensions  #######
    #######################################################
    @periodic_task.periodic_task(spacing=CONF.monitor_instances_interval)
    def monitor_instances(self, context, instances=None):
        """Periodic Task to Refresh Metrics for the Instances"""
        driver, context = (self.driver, context.elevated())
        LOG.debug('Invoking Method to Refresh Instance Metrics - Begin')
        try:
            drv_insts = self.driver.list_instances()
            name_map = dict([(inst, '') for inst in drv_insts])
            # Retrieve All of the Instances from the Nova DB for the Given Host
            db_insts = self._query_db_instances(context, instances)
            db_insts = [inst for inst in db_insts if inst['name'] in name_map]
            # If this is a Discovery Driver, then let it Gather Metrics
            if isinstance(driver, discovery_driver.ComputeDiscoveryDriver):
                # It is only worth Gathering Metrics if there are Instances
                if len(db_insts) > 0:
                    metric_data = driver.monitor_instances(context, db_insts)
                    self._send_metric_notifications(context, metric_data)
        except Exception as exc:
            LOG.warn(_('Error refreshing Instance Metrics'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Instance Metrics - End')

    @periodic_task.periodic_task(spacing=CONF.monitor_instances_interval)
    def monitor_hosts(self, context, instances=None):
        """Periodic Task to Refresh Metrics for the hosts"""
        driver, context = (self.driver, context.elevated())
        LOG.debug('Invoking Method to Refresh Hosts Metric - Begin')
        try:
            metric_data = driver.monitor_hosts(context)
            if metric_data is None:
                return
            self._send_metric_notifications_host(context, metric_data)
        except Exception as exc:
            LOG.warn(_('Error refreshing Hosts Metric'))
            LOG.exception(exc)
        LOG.debug('Invoking Method to Refresh Hosts Metric - End')

    def _send_metric_notifications_host(self, context, metric_data):
        """Collects with host monitor data (i.e. CPU utilization)"""
        payload, payloads = (list(), list())
        # If there wasn't any Metric data, then log a debug statement
        if len(metric_data) <= 0:
            LOG.debug('No CPU data found: ' + str(metric_data))
        # Loop through the Metric Data on a per host level
        for i in range(len(metric_data)):
            data = metric_data[i]
            payload.append({'host_id': data['uuid'],
                            'cpu_utilization': data['cpu_utilization']})
            # If we have hit the Message Size limit, add a new chunk
            if (i > 1) and (i % CONF.metrics_message_size == 0):
                payloads.append(payload)
                payload = []
            # If this is the last one, then we want to add the last chunk
            elif i == (len(metric_data) - 1):
                payloads.append(payload)
        # Loop through each of the Payload Chunks, sending a Notification
        for p in payloads:
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.host.metrics', {'data': p})

    def _send_metric_notifications(self, context, metric_data):
        """Collects with LPAR monitor data (i.e. CPU utilization)"""
        payload, payloads = (list(), list())
        # If there wasn't any Metric data, then log a debug statement
        if len(metric_data) <= 0:
            LOG.debug('No CPU data found: ' + str(metric_data))
        # Loop through the Metric Data on a per-Instance level
        for i in range(len(metric_data)):
            inst_data = metric_data[i]
            payload.append({'instance_id': inst_data['uuid'],
                            'cpu_utilization': inst_data['cpu_utilization']})
            # If we have hit the Message Size limit, add a new chunk
            if (i > 1) and (i % CONF.metrics_message_size == 0):
                payloads.append(payload)
                payload = []
            # If this is the last one, then we want to add the last chunk
            elif i == (len(metric_data) - 1):
                payloads.append(payload)
        # Loop through each of the Payload Chunks, sending a Notification
        for p in payloads:
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.instance.metrics', {'data': p})

    #######################################################
    ########  Network-related Overrides/Extensions  #######
    #######################################################

    def _get_super(self):
        """
        Helper for UT since the mock library can't properly return the super
        class.
        """
        return super(PowerVCComputeManager, self)

    @wrap_exception()
    def external_instance_event(self, context, instances, events):
        """
        Wrapper until 29618 can be updated in the community
        """
        # Do the original...
        self._get_super().external_instance_event(context, instances, events)

        # Fire off an instance update for the appropriate objects
        for event in events:
            instance = [inst for inst in instances
                        if inst.uuid == event.instance_uuid][0]
            if event.name == 'network-changed':
                self._instance_update(context, instance['uuid'],
                                      updated_at=timeutils.utcnow())

    def attach_interface(self, context, instance, network_id, port_id,
                         requested_ip):
        """
        Subclassing simply to throw the instance update for Bumblebee
        """
        # call the parent
        ret = self._get_super().attach_interface(context, instance,
                                                 network_id, port_id,
                                                 requested_ip)

        self._instance_update(context, instance['uuid'],
                              updated_at=timeutils.utcnow())
        return ret

    def detach_interface(self, context, instance, port_id):
        """
        Subclassing simply to throw the instance update for Bumblebee
        """
        # call the parent
        ret = self._get_super().detach_interface(context, instance, port_id)

        self._instance_update(context, instance['uuid'],
                              updated_at=timeutils.utcnow())
        return ret

    def change_vif_sea(self, context, data):
        """
        This API calls the plug method on the compute driver to plug a VLAN
        on a VEA.
        :param context: Context Object
        :param data: The dictionary object containing information to call
        the plug  operation, sent by the API service.
        :return: Calls the compute driver and returns the plug call back.
        """
        greenthread.spawn(self._change_vif_sea, context, data)
        return True

    @logcall
    @lockutils.synchronized('change_powervm_vifs', 'nova-change-vif-powervm-')
    def _change_vif_sea(self, context, data):
        """
        This API calls the plug method on the compute driver to plug a VLAN
        on a VEA.  This is the internal call that should be done on the
        greenthread.

        :param context: Context Object
        :param data: The dictionary object containing information to call
        the plug  operation, sent by the API service.
        :return: Calls the compute driver and returns the plug call back.
        """
        # Obtain data from the input.  Start with common data.
        vlan = data['vlan']

        # Move on to the 'new' data
        # Note that the force=True is only used in the rollback, but
        # we set it here so that we don't have to rebuild.
        new_sea = data['new_sea']
        new_lpar = data['new_lpar_id']
        net_id = {'id': data['net_id']}
        new_vif = network_model.VIF('fake_id', 'fake_address', net_id,
                                    'fake_dev_name', sea_lpar_id=new_lpar,
                                    sea_name=new_sea, force=True,
                                    vlan=vlan)

        # Build a list of network ids that will be updated when complete
        net_ids_to_update = [data['net_id']]
        net_ids_to_update.extend(data.get('peers', []))

        # Build the 'old' data...what is 'currently' plugged in and find all
        # other vifs that will be affected by this change.
        old_sea = ''
        old_lpar = ''
        old_vif_list = []
        na_and_peers_tuple_list = []
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        for net_id_to_update in net_ids_to_update:
            net_assn = netafact.find_network_assoc(
                context, self.host, net_id_to_update)
            if(not net_assn or net_assn is None or not net_assn.sea or
               net_assn.sea is None):
                continue
            if net_id['id'] == net_id_to_update:
                old_sea = net_assn.sea.name
                old_lpar = net_assn.sea.vioserver.lpar_id
            na_and_peers_tuple_list.append((net_assn.sea.name,
                                            [net_id_to_update]))

            # It's possible to find duplicates, so make sure we aren't already
            # replugging this vif.
            net_assn_lpar = net_assn.sea.vioserver.lpar_id
            already_listed = False
            for old_vif_entry in old_vif_list:
                if(old_vif_entry['meta']['sea_name'] == net_assn.sea.name or
                   old_vif_entry['network']['id'] == net_id_to_update):
                    # We've already got this vif on the list, skip it
                    already_listed = True
            if not already_listed:
                cur_vif = network_model.VIF('fake_id', 'fake_address',
                                            {'id': net_id_to_update},
                                            'fake_dev_name',
                                            sea_lpar_id=net_assn_lpar,
                                            sea_name=net_assn.sea.name,
                                            force=True,
                                            vlan=vlan)
                old_vif_list.append(cur_vif)

        # Build the driver (IVM or HMC) that will perform the
        # plug/unplug for us
        vif_conf = CONF.powervm_vif_driver
        vif_drv = importutils.import_object(vif_conf, host_name=self.host)

        # For logging purposes, get the network name.
        net_api = network.API()
        neutron_obj = net_api.get(context, data['net_id'])
        if neutron_obj is not None:
            net_name = neutron_obj.get('name', data['net_id'])
        else:
            net_name = data['net_id']

        try:
            # Now start to plug/unplug
            msg = (_("Network %(net)s is being modified to use Shared "
                     "Ethernet Adapter (SEA) %(new_sea)s instead of SEA "
                     "%(old_sea)s on host %(host)s.") %
                   {'host': self.host, 'net': net_name,
                    'old_sea': old_sea, 'new_sea': new_sea})
            LOG.info(msg)

            # Happy path should just be 'unplug and plug'
            vif_drv.unplug(None, old_vif_list)

            # We update the data base now with the new SEA and network_ids.
            # This is because the corresponding plug call will check the
            # database for the network association to find the SEA to work
            # against.
            self.__update_na_and_peers_in_db(context, new_lpar,
                                             [(new_sea, net_ids_to_update)])
            vif_drv.plug(None, [new_vif])

            # Lastly notify the user that the change operation has completed
            msg = (_("Network %(net)s was successfully modified to use Shared "
                     "Ethernet Adapter (SEA) %(new_sea)s instead of SEA "
                     "%(old_sea)s on host %(host)s. Connectivity was restored "
                     "on all virtual machines that use the network") %
                   {'host': self.host, 'net': net_name,
                    'old_sea': old_sea, 'new_sea': new_sea})
            info = {'msg': msg, 'sea_name': new_sea, 'host_name': self.host}
            LOG.info(msg)
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.instance.log', info)

        except Exception as ex:
            # Something went wrong...we need to log and then attempt to replug
            # to the original state.
            LOG.exception(ex)

            # Update the VIFs, note that we now need to set the force to true
            try:
                vif_drv.unplug(None, [new_vif])

                # Update the database to set the Network Mapping back to
                # the original value. This would mean reverting the peers to
                # the older SEA.
                self.__update_na_and_peers_in_db(context, old_lpar,
                                                 na_and_peers_tuple_list)

                vif_drv.plug(None, old_vif_list)
            except Exception as ex2:
                LOG.exception(ex2)

            # Lastly notify the user that something has gone wrong.
            msg = (_("The adapter change on Host %(host)s for Network "
                     "%(net)s failed to finish successfully.  The system has "
                     "returned the following error message: %(msg)s.  "
                     "Please review the issue and retry the command.") %
                   {'host': self.host, 'msg': ex.message,
                    'net': net_name})
            info = {'msg': msg, 'sea_name': new_sea, 'host_name': self.host}
            LOG.warn(msg)
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.warn(context, 'compute.instance.log', info)

    @periodic_task.periodic_task(spacing=CONF.cleanup_net_assn_interval)
    def _cleanup_network_associations(self, context):
        """
        This task periodically cleans up network associations
        of deleted networks
        """
        LOG.info(_('pvc_nova.compute.manager.PowerVCComputeManager '
                   '_cleanup_network_associations: updating network'
                   ' associations...'))
        if isinstance(self.driver, discovery_driver.ComputeDiscoveryDriver):
            self.driver.cleanup_network_associations(
                context, self.host)

    def __update_na_and_peers_in_db(self, context, lpar_id, sea_na_tuple_list):
        """
        Updates the network associations for the net_ids with a new SEA.

        :param context: The user context making the request
        :param lpar_id: The lpar housing the sea
        :param sea_na_tuple_list: A list of SEA/NA tuples to update.
        """
        netafact = network_dom.NetworkAssociationFactory.get_factory()
        hostfact = network_dom.ManagedHostFactory.get_factory()
        transaction = netafact.get_resource_transaction()
        with transaction.begin(subtransactions=True):
            host_dom = hostfact.find_host_by_name(
                context, self.host, transaction)

            for ttuple in sea_na_tuple_list:
                sea = ttuple[0]
                net_ids = ttuple[1]
                sea_dom = host_dom.find_adapter(lpar_id, sea)
                for net_id in net_ids:
                    netafact.update_network_assoc(
                        context, self.host, net_id, sea_dom, transaction)

    def _cleanup_imported_networks(self, context, network_info):
        """Cleanup Networks if they were for Imported VM's and not used"""
        network_ids = dict()
        NETNM1, NETNM2 = ('Imported Flat Network', 'Imported Network for VLAN')
        try:
            client = neutronv2.api.get_client(context)
            # Loop through each of the VIF's, check if any are
            for vif in network_info:
                if vif.get('network') and vif['network'].get('label'):
                    name = vif['network']['label']
                    # We only want to delete it if it is an Import created one
                    if name.startswith(NETNM1) or name.startswith(NETNM2):
                        network_ids[vif['network']['id']] = ''
            # Loop through the Network ID's seeing if they have no Ports left
            for net_id in network_ids:
                name = client.show_network(net_id)['network']['name']
                # Check the Name again to make sure user didn't rename later
                if not name.startswith(NETNM1) and not name.startswith(NETNM2):
                    continue
                # If no ports were left, then we want to delete the Network
                if len(client.list_ports(network_id=net_id)['ports']) <= 0:
                    LOG.info(_("Removing unused Imported Network %s") % net_id)
                    client.delete_network(net_id)
        # Since the cleanup isn't essential, just log errors and don't fail
        except Exception as exc:
            LOG.warn(_("Error trying to Cleanup Imported Networks"))
            LOG.exception(exc)

    #######################################################
    # Moved check_host_maintance_mode here so we can catch
    # the exeption in attach_volume() and reset the state
    # of the volume to available when this check fails.
    #######################################################
    @check_host_maintenance_mode(common_constants.OPER_AT_VOL)
    def attach_volume_check_VM_state(self, context):
        """
        checks that the VM is not in maint mode.
        :param context: operation context
        :return: None
        """
        LOG.debug("attach_volume_check")

    #######################################################
    ########  Storage-related Overrides/Extensions  #######
    #######################################################
    @manager.object_compat
    @wrap_exception()
    @pvc_comp_mgr_reverts_task_state
    @manager.wrap_instance_fault
    def reserve_block_device_name(self, context, instance, device,
                                  volume_id, disk_bus=None, device_type=None,
                                  return_bdm_object=False):
        # copied from openstack nova manager.py to restrict attaching same
        # volume to same instance twice

        @utils.synchronized(instance.uuid)
        def do_reserve():
            if volume_id:
                bdm = self._get_bdm_for_instance_and_vol(context, instance,
                                                         volume_id)
                if bdm:
                    reason = "Volume %s already attached to instance %s"\
                        % (volume_id, instance.uuid)
                    raise exception.InvalidVolume(reason=reason)
            bdms = (
                objects.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance.uuid))

            device_name = compute_utils.get_device_name_for_instance(
                context, instance, bdms, device)

            # NOTE(vish): create bdm here to avoid race condition
            bdm = objects.BlockDeviceMapping(
                context=context,
                source_type='volume', destination_type='volume',
                instance_uuid=instance.uuid,
                volume_id=volume_id or 'reserved',
                device_name=device_name,
                disk_bus=disk_bus, device_type=device_type)
            bdm.create()

            if return_bdm_object:
                return bdm
            else:
                return device_name

        return do_reserve()

    @wrap_exception()
    @pvc_comp_mgr_reverts_task_state
    @manager.wrap_instance_fault
    @send_ui_notification(notify_msg.ATTACH_ERROR, notify_msg.ATTACH_SUCCESS,
                          log_exception=True)
    def attach_volume(self, context, volume_id, mountpoint, instance,
                      bdm=None):
        """
        attach a volume to an instance, prevent any image volume
        from being attached.
        :param context: operation context
        :param volume_id: volume id for attaching
        :param mountpoint: mount point
        :param instance: compute instance for attaching.
        :param bdm: block device mapping object.
        :return: None
        """
        if not volume_id and bdm:
            volume_id = bdm.volume_id
        # check that the vm is in a state that allows attach
        try:
            self.attach_volume_check_VM_state(context)
        except Exception:
            # reset the state of the volume to available
            # we don't want to leave the state as attaching
            with excutils.save_and_reraise_exception():
                self.volume_api.unreserve_volume(context, volume_id)
                LOG.info("Cannot attach image because of the state of the VM")

        volume = self.volume_api.get(context, volume_id)
        metadata = volume.get("volume_metadata")

        # For image-volume, is_image_volume will be set
        # in the volume metadata.
        if metadata and ('is_image_volume' in metadata and
                         bool(metadata['is_image_volume'])):
            # roll back attaching state and log a warning message
            # if it is an image volume
            LOG.warn(_("Cannot attach image volume: %(volume_id)s") % locals())
            self.volume_api.unreserve_volume(context, volume_id)
            bdm.destroy(context)
            raise exception.InvalidVolume(reason='Cannot attach image volume')
        else:
            # Cleanup bdms for this specific volume_id before
            # attachment. Retreive a list of all bdms, run through the list
            # and destroy only bdms that match given volume_id

            s = super(PowerVCComputeManager, self)
            s.attach_volume(context, volume_id, mountpoint, instance, bdm=bdm)

    def _get_all_bdms_for_instance_and_vol(self, context, instance, volume_id):
        """ Helper function to retrieve all bdms for
        instance"""
        # go get the bdm for this instance and volume id
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        return bdms

    def _get_volume_attachment(self, volume, instance_uuid):
        # returns the attachment id else return none
        attachment = self.volume_api.get_volume_attachment(
            volume, instance_uuid)
        if attachment:
            return attachment['id']
        else:
            return None

    def detach_volume(self, context, volume_id, instance):
        """This is done due to changes in OSEE code for rpc 4.0
        in order to support notifications arguments needs to be
        passed as key value pair
        """
        return self.pvc_detach_volume(context, volume_id=volume_id,
                                      instance=instance)

    @wrap_exception()
    @manager.reverts_task_state
    @manager.wrap_instance_fault
    @send_ui_notification(notify_msg.DETACH_ERROR, notify_msg.DETACH_SUCCESS,
                          log_exception=True)
    @check_host_maintenance_mode(common_constants.OPER_DT_VOL)
    def pvc_detach_volume(self, context, volume_id, instance):
        """
        detach a volume from an instance, prevent any boot volume
        from being detached.
        :param context: operation context
        :param volume_id: volume id for detaching
        :param instance: compute instance for detaching.
        :return: None
        """
        volume = self.volume_api.get(context, volume_id)
        is_ivm = CONF.powervm_mgr_type == 'ivm'

        # OSEE detach_volume part. PowerVC need to retrieve
        # the saved connector per BDM during detach for NPIV.
        bdm = self._get_bdm_for_instance_and_vol(context, instance, volume_id)
        if CONF.volume_usage_poll_interval > 0:
            vol_stats = []
            mp = bdm['device_name']
            # Handle bootable volumes which will not contain /dev/
            if '/dev/' in mp:
                mp = mp[5:]
            try:
                vol_stats = self.driver.block_stats(instance, mp)
            except NotImplementedError:
                pass

            if vol_stats:
                LOG.debug(_("Updating volume usage cache with totals"))
                rd_req, rd_bytes, wr_req, wr_bytes, flush_ops = vol_stats
                self.conductor_api.vol_usage_update(context, volume_id,
                                                    rd_req, rd_bytes,
                                                    wr_req, wr_bytes,
                                                    instance,
                                                    update_totals=True)

        self._detach_volume(context, instance, bdm)

        if is_ivm:
            # call the driver to get the connector.
            # this will be the IVM FC attachment since the connector
            # could be different after migration.
            connector = self.driver.get_volume_connector(instance)
        else:
            # get a default SPP connector
            connector = self._get_SSP_volume_connector(instance, bdm)

            # Get a real FC connector
            if not connector:
                connector = self._get_bdm_volume_connector(context,
                                                           instance,
                                                           bdm,
                                                           volume_id)
            # HMC managed system and has no connector, try to generate
            # it from the saved connection_info.
            if not connector:
                connector = self._construct_connector_from_conninfo(instance,
                                                                    bdm)
            # Last try by called the virt driver. Could be expensive and
            # have exception.
            if not connector:
                try:
                    connector = self.driver.get_volume_connector(instance)
                except Exception as ex:
                    LOG.error(_("Failed to detach volume %(volid)s from "
                                "instance %(inst)s due to null "
                                "volume connector. Error: %(error)s") %
                              {'volid': volume_id,
                               'inst': instance['uuid'],
                               'error': ex})
                    err_msg = {'volid': volume_id,
                               'inst': instance['uuid']}
                    raise pwrvc_excep.IBMPowerVCVInvalidVolumeConnector(
                        **err_msg)
        LOG.info(_("Instance %(inst)s detach volume %(volid)s with connector: "
                   "%(conn)s ") %
                 {'inst': instance['uuid'],
                  'volid': volume_id,
                  'conn': connector})
        try:
            # For Onboarded VM: the connector returned from the
            # driver.get_volume_connector does not have connection_type
            # set. Need to check if wwpn_vscsi is present
            attachment = self._get_volume_attachment(
                volume, instance['uuid'])
            self.volume_api.detach(context.elevated(), volume_id,
                                   attachment)

            conninfo = jsonutils.loads(bdm['connection_info'])
            self.driver.call_terminate_connection(context, instance,
                                                  bdm['volume_id'],
                                                  connector, conninfo)
        except Exception:
            with excutils.save_and_reraise_exception():
                # Roll back detaching state.
                # Otherwise, volume stucks in detaching state forever.
                self.volume_api.roll_detaching(context, volume_id)

        # destroy all bdms matching this volume_id
        self._get_all_bdm_for_instance_and_vol_and_destroy(context, instance,
                                                           volume_id)

        self._instance_update(context, instance['uuid'],
                              updated_at=timeutils.utcnow())
        info = dict(volume_id=volume_id)
        self._notify_about_instance_usage(
            context, instance, "volume.detach", extra_usage_info=info)

    def _get_all_bdm_for_instance_and_vol_and_destroy(self, context, instance,
                                                      volume_id):
        """ Helper function to retrieve the volume_id for
        instance"""
        # go get the bdm for this instance and volume id
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        bdm = [bdm for bdm in bdms if bdm['volume_id'] and
               bdm['volume_id'] == volume_id]

        for specific_bdm in bdm:
            specific_bdm.destroy()

    def remove_volume_connection(self, context, volume_id, instance):
        """Override Remove Volume Connection to Clean-up Stale Disks"""
        bdm = self._get_bdm_for_instance_and_vol(context, instance, volume_id)
        cinfo = jsonutils.loads(bdm['connection_info'])
        super(PowerVCComputeManager,
              self).remove_volume_connection(context, volume_id, instance)
        self.driver.remove_volume_device_at_destination(
            context, instance, cinfo)

    def _delete_instance(self, context, instance, bdms,
                         reservations=None):
        """Overridden to handle the HypervisorUnavailable exception
        in the case of PowerKVM. When the libvirt service is down Delete
        task hang indefinitely."""
        s = super(PowerVCComputeManager, self)
        try:
            s._delete_instance(context, instance, bdms,
                               reservations)
        except exception.HypervisorUnavailable:
            with excutils.save_and_reraise_exception():
                if (instance.task_state == task_states.DELETING):
                    instance.task_state = None
                    instance.expected_task_state = None
                    instance.save()

    @send_ui_notification(notify_msg.DELETE_ERROR,
                          notify_msg.DELETE_SUCCESS,
                          log_exception=True)
    def terminate_instance(self, context, instance, bdms, reservations):
        """Terminate an instance on this host.
           Override to send ui notification. Notify in the compute manager
           instead of the driver's destroy method, so notifications are only
           sent when a user issues a delete on the instance.
        """
        super(PowerVCComputeManager, self).terminate_instance(
            context, instance, bdms, reservations)

    def _shutdown_instance(self, context, instance,
                           bdms, requested_networks=None, notify=True,
                           try_deallocate_networks=True):
        """ override openstack _shutdown_instance in order to
        save the volume connector before destroy the partition.
        For NPIV/SSP attached volumes, they are not assigned to host.
        The connector doesn't exist after destroy the instance.
        For NPIV, this connector is important to remove the
        hostmap and delete the fabric zone. For SSP, termiante_connection
        is a no-op. For IVM, it is the same as openstack."""

        """Shutdown an instance on this host."""
        context = context.elevated()
        LOG.info(_('%(action_str)s instance: %(instuuid)s') %
                 {'action_str': 'Terminating', 'instuuid': instance['uuid']},
                 context=context, instance=instance)

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.start")

        # get network info before tearing down
        try:
            network_info = compute_utils.get_nw_info_for_instance(instance)
        except (exception.NetworkNotFound, exception.NoMoreFixedIps):
            network_info = network_model.NetworkInfo()

        # refresh bdms from database to sync up the in memory copy
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])

        # get bdms and volume connector before destroying the instance
        vol_bdms = [bdm for bdm in bdms if bdm.is_volume]

        # Defect 24039 : Added the check to handle the case of KVM
        # for Cleaning up the iSCSI volume connectors.
        is_kvm = False
        if CONF.powervm_mgr_type is None:
            is_kvm = (CONF.libvirt.virt_type == 'kvm')

        if is_kvm:
            vol_connector = self.driver.get_volume_connector(instance)
        else:
            # for NPIV, the connector has been saved in for each BDM. And
            # they might be different for each volume.
            vol_connector = None
        # LPAR get_info
        # Handling the HypervisorUnavailable exception forPowerKVM
        # If VM is Error: Send notification and delete it from DB
        # else send notification and mark it to Error.

        try:
            lpar_state = self.driver.get_info(instance).state
        except exception.InstanceNotFound:
            lpar_state = power_state.NOSTATE
        except exception.HypervisorUnavailable as e:
            LOG.exception(_("Failed to retrieve LPAR, Delete VM Failed"
                            "%s") % e)
            vm_state = instance['vm_state']
            if (vm_state != vm_states.ERROR):
                reason = (_("Unable to Delete the virtual machine,"
                            "Virtual machine will be changed to Error "
                            "State. Reason: %s. For more information "
                            "see the host log files.") % e)
                LOG.exception(_("Unable to Delete the virtual machine, "
                                "Virtual machine will be changed to Error "
                                "State. Reason: %s. For more information,"
                                "see the host log files.") % e)
                self._error_notify(instance, context, 'error', reason)
                raise e
            reason = (_("Unable to delete the virtual machine. The virtual "
                        "machine will be removed from PowerVC management. "
                        "Reason: %s. For more information, see the host log"
                        "files.") % e)
            LOG.exception(_("Unable to delete the virtual machine. The "
                            "virtual machine will be removed from PowerVC "
                            "management. Reason: %s. For more information,"
                            "see the host log files.") % e)
            self._error_notify(instance, context, 'error', reason)
            return

        # If LPAR has been deleted out-of-band, the attachment information
        # exists while connector is empty. Need to be able to destroy
        # the instance for cleanup purpose.
        LOG.info(_("Destroy Virtual Server %(lpar_name)s, current LPAR state:"
                   " %(lpar_state)s") %
                 {'lpar_name': instance['name'],
                  'lpar_state': lpar_state})

        block_device_info = self._get_instance_block_device_info(
            context, instance)
        LOG.debug("block_device_info: %s" %
                  str(self._clean_bdm_info(block_device_info)))
        # Attempt driver destroy before releasing ip, may
        # want to keep ip allocated for certain failures
        # The volume mapping for the instance will be removed
        # by driver as well. _shutdown_instance should handle
        # the storage side hostmap and SAN fabric zoning removal.
        try:
            self.driver.destroy(context, instance, network_info,
                                block_device_info)

        except exception.InstancePowerOffFailure:
            # if the instance can't power off, don't release the ip
            with excutils.save_and_reraise_exception():
                pass
        except Exception:
            with excutils.save_and_reraise_exception():
                # deallocate ip and fail without proceeding to
                # volume api calls, preserving current behavior
                try:
                    if try_deallocate_networks:
                        self._try_deallocate_network(context, instance,
                                                     requested_networks)
                        #If this is an imported network, we want to clean it up
                        if requested_networks is None:
                            self._cleanup_imported_networks(context,
                                                            network_info)
                except Exception as ex:
                    # LPAR may have been destroyed and there is no way
                    # back. Try to clean up as much as we can.
                    # The error has been logged in _try_deallocate_network.
                    reason = (_("Failed to remove the VLAN associated with "
                                "the LPAR virtual ethernet. Please check "
                                "host log for further details. Error: %(ex)s")
                              % dict(ex=ex))
                    self._error_notify(instance, context, 'warn', reason)
                    LOG.warn(_("Failed to remove the VLAN associated with "
                               "LPAR virtual ethernet. Instance: %(inst)s. "
                               "Error:%(error)s") %
                             {'lpar_name': instance['name'],
                              'inst': instance['uuid'],
                              'error': ex})
        try:
            if try_deallocate_networks:
                self._try_deallocate_network(context, instance,
                                             requested_networks)
                #If this is an imported network, we want to clean it up
                if requested_networks is None:
                    self._cleanup_imported_networks(context, network_info)
        except Exception as ex:
            # LPAR may have been destroyed and there is no way
            # back. Try to clean up as much as we can.
            # The error has been logged in _try_deallocate_network.
            reason = (_("Deleted LPAR successfully. But failed to remove "
                        "the VLAN associated with the LPAR virtual ethernet. "
                        "Please check host log for further "
                        "details. Error: %(ex)s") % dict(ex=ex))
            self._error_notify(instance, context, 'warn', reason)
            LOG.warn(_("Deleted LPAR %(lpar_name)s successfully. But failed "
                       "to remove the VLAN associated with LPAR virtual "
                       "ethernet. Instance: %(inst)s. Error:%(error)s") %
                     {'lpar_name': instance['name'],
                      'inst': instance['uuid'],
                      'error': ex})

        # Get the volumes from the VM's image to ensure we do not delete
        # image volumes when deploys fail before the image volume in the BDM
        # is cloned and BDM updated.
        img_vols_to_clone = []
        try:
            image_ref = instance['image_ref']
            if image_ref != 'ONBOARDED_NO_IMAGE':
                img_vols_to_clone = self.\
                    _get_image_bdm_volumes_to_clone(context, image_ref)
        except Exception:
            msg_args = {'image': image_ref,
                        'instance': instance['uuid']}
            LOG.warn(_('The image %(image)s for instance %(instance)s could '
                       'not be found.') % msg_args)
        last_connector = None
        # Cleanup the zones on the 'first' disk rather than the last disk
        # We keep trying to cleanup until we get a valid connector back
        # from the cleanup helper with clean_zones property.
        # Handles the case where we have an invalid or partial connector.
        # Random order of SSP VSCSI and NPIV is handled because
        # clean_zones only shows up in the connector if the volume type
        # is npiv.
        clean_zones = True
        for bdm in vol_bdms:
            last_connector = self._bdm_cleanup_helper(
                context, instance, bdm, vol_connector, clean_zones,
                img_vols_to_clone)
            if last_connector is not None and \
               'clean_zones' in last_connector:
                clean_zones = False
        vol_connector = last_connector if last_connector else vol_connector

        # cleanup instance after block device cleanup.
        args = {'root_gb': 0}
        instance.update(args)
        self.conductor_api.instance_update(context, instance['uuid'],
                                           **args)

        if notify:
            self._notify_about_instance_usage(context, instance,
                                              "shutdown.end")

    def _is_volume_boot_set(self, context, volume_id):
        volume = self.volume_api.get(context, volume_id)
        if 'is_boot_volume' in volume['volume_metadata'] and \
                volume['volume_metadata']['is_boot_volume'] == 'True':
            return True
        return False

    @logcall
    def _attach_volume(self, context, instance, bdm):
        """
        Extend this method to make sure instance has the required virtual
        adapter and maps to physical fibre channel adapter prior to
        cinder's initialize_connection.
        """
        volume_id = bdm.volume_id
        is_boot_set = self._is_volume_boot_set(context, volume_id)
        mountpoint = bdm['mount_device']

        try:
            stg_type = self.volume_api.get_storage_type_for_volume(
                context, volume_id)
            # since vSCSI and NPIV both has the storage type as 'fc'
            # we need to identify if the volume attachment to be done
            # for vSCSI or NPIV before we pass to prep_instance_for_attach
            # in order to create the proper SCSI(vSCSI) or FC (NPIV) Adapters
            if stg_type == 'fc':
                connector = self.driver.get_volume_connector(instance)
                if not connector:
                    raise Exception(_('Cannot get volume connector to '
                                      'virtual machine %s') % instance['name'])
                if is_boot_set:
                    connector['connection_type'] = \
                        connector.pop('boot_conn_types', None)
                else:
                    connector['connection_type'] = \
                        connector.pop('data_conn_types', None)

                if ((connector['connection_type'] and
                     CONNECTION_TYPE_VSCSI in connector['connection_type'])):
                    stg_type = CONNECTION_TYPE_VSCSI

            self.driver.prep_instance_for_attach(context, instance, stg_type)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                self.volume_api.unreserve_volume(context, volume_id)

                # We send an asynchronous notification that attachment failed,
                # but do not attempt to gather more useful info for the error
                # report (e.g. storage provider metainfo since that is likely
                # how we got here (volume_api.get_storage_type_for_volume())
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                if (isinstance(ex.value, dict) and
                   'kwargs' in ex.value and
                   'reason' in ex.value['kwargs']):
                    error = six.text_type(ex.value['kwargs']['reason'])
                else:
                    error = six.text_type(ex.value)
                log_msg = (_("Failed to attach volume %(volume_id)s to "
                             "%(name)s, likely due to lack of connectivity "
                             "to storage provider(s). Error: %(error)s") %
                           {'name': instance['name'],
                            'volume_id': volume_id,
                            'error': error})
                LOG.error(log_msg)

                msg = _("Could not attach volume {volume_id} to virtual "
                        "server {instance_name}.  Check that host "
                        "{host_name} has connectivity to all storage "
                        "providers.  Check the host's log files for "
                        "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='volume.%s' % volume_id)
                notifier.error(context, 'volume.update.errors', notify_info)
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

        try:
            info = self._attach_volume_osee(context, volume_id, mountpoint,
                                            instance, is_boot_set)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                # We send an asynchronous notification that attachment failed.
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                error = ex.value

                try:
                    # Attempt to gather some more useful information for the
                    # error report.

                    # Volume details
                    vol = self.volume_api.get(context, volume_id)
                    notify_info['volume_display_name'] = vol['display_name']

                    # Storage provider details
                    provider_info = self.volume_api.get_stg_provider_info(
                        context, volume_id=volume_id)
                    update = {'storage_provider_display_name':
                              provider_info['host_display_name'],
                              'storage_provider_name': provider_info['host'],
                              'storage_provider_id': provider_info['id'],
                              'error': error}
                    notify_info.update(update)

                    # UI replaces volume_id with volume_display_name hyperlink.
                    msg = _("Could not attach volume {volume_id} to "
                            "virtual server {instance_name}.  Check that "
                            "host {host_name} has connectivity to storage "
                            "provider {storage_provider_id}. If the provider "
                            "is a Fibre Channel SAN controller, then check "
                            "that the host has connectivity to the FC switch "
                            "for each registered fabric (ensure the FC "
                            "switches are operational). Also, check the "
                            "host's log files for further details of this "
                            "error: %(error)s") % {'error': error}

                except Exception as e2:
                    LOG.exception(e2)
                    # If anything goes wrong, fall back to just reporting the
                    # information that we have to hand.  We'd end up here if we
                    # failed due to Cinder being down, for example.
                    msg = _("Could not attach volume {volume_id} to virtual "
                            "server {instance_name}.  Check that host "
                            "{host_name} has connectivity to all storage "
                            "providers.  Check the host's log files for "
                            "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='volume.%s' % volume_id)
                notifier.error(context, 'volume.update.errors', notify_info)
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

        return info

    def _detach_volume(self, context, instance, bdm):
        """Do the actual driver detach using block device mapping."""
        volume_id = bdm['volume_id']

        s = super(PowerVCComputeManager, self)
        try:
            s._detach_volume(context, instance, bdm)
        except Exception:
            with excutils.save_and_reraise_exception() as ex:
                if (isinstance(ex.value, dict) and
                   'kwargs' in ex.value and
                   'reason' in ex.value['kwargs']):
                    error = six.text_type(ex.value['kwargs']['reason'])
                else:
                    error = six.text_type(ex.value)
                # Attempt to gather some more useful information for the
                # error report.
                notify_info = {'instance_id': instance['uuid'],
                               'instance_name': instance['name'],
                               'volume_id': volume_id,
                               'host_name': self.host}
                try:
                    # Volume details
                    vol = self.volume_api.get(context, volume_id)
                    notify_info['volume_display_name'] = vol['display_name']

                    # Storage provider details
                    provider_info = self.volume_api.get_stg_provider_info(
                        context, volume_id=volume_id)
                    update = {'storage_provider_display_name':
                              provider_info['host_display_name'],
                              'storage_provider_name': provider_info['host'],
                              'storage_provider_id': provider_info['id'],
                              'error': error}
                    notify_info.update(update)

                    # UI replaces volume_id with volume_display_name hyperlink.
                    msg = _("Could not detach volume {volume_id} from "
                            "virtual server {instance_name}.  Check that "
                            "host {host_name} has connectivity to storage "
                            "provider {storage_provider_id}. If the provider "
                            "is a Fibre Channel SAN controller, then check "
                            "that the host has connectivity to the FC switch "
                            "for each registered fabric (ensure the FC "
                            "switches are operational). Also check the "
                            "host's log files for further details of this "
                            "error: {error}")

                except Exception as e2:
                    LOG.exception(e2)
                    # If anything goes wrong, fall back to just reporting the
                    # information that we have to hand.  We'd end up here if we
                    # failed due to Cinder being down, for example.
                    msg = _("Could not detach volume {volume_id} from virtual "
                            "server {instance_name}.  Check that host "
                            "{host_name} has connectivity to all storage "
                            "providers.  Check the host's log files for "
                            "further details of this error.")

                notify_info['msg'] = msg
                notifier = rpc.get_notifier(service='compute', host=self.host)
                notifier.error(context, 'compute.instance.log', notify_info)

    def _cleanup_volumes(self, context, instance_uuid, bdms, raise_exc=True):
        """ Override nova compute manager's method. All the volumes
        that have been marked with 'delete_on_termination' are
        handled by _shutdown_instance already.
        NO-OP the parent's _cleanup_volumes otherwise
        it will generate an exception in the compute log.
        """
        pass

    def _restore_pre_migrate_bdms(self, context, migrate_data):
        """Override restore pre-migrate BDMS data for attached volumes"""
        org_bdms = migrate_data.get("pre_migrate_bdms")
        if not org_bdms:
            return
        for bdm in org_bdms:
            connection_info = {'connection_info': bdm['connection_info']}
            self.conductor_api.block_device_mapping_update(
                context, bdm['id'], connection_info)

    def _get_instance_block_device_info(self, context, instance,
                                        refresh_conn_info=False, bdms=None):
        """Override Nova compute manager's method in order to control
        refresh_conn_info. For NPIV VFC attached volume(HMC only), the
        connection stays unchanged during the instance's life cycle.
        There is no need to refresh.
        """
        s = super(PowerVCComputeManager, self)
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc' \
                and refresh_conn_info:
            # refresh connection is true under resize. There is no need
            # to refresh for NPIV/SSP attached volume during resize.
            return s._get_instance_block_device_info(
                context, instance, refresh_conn_info=False, bdms=bdms)
        else:
            return s._get_instance_block_device_info(
                context, instance, refresh_conn_info=refresh_conn_info,
                bdms=bdms)

    def _attach_volumes_and_boot(self, context, instance):
        """
        This is callback function to attach volumes before boot
        up lpar.
        :param context: operation context
        :param instance: compute instance
        :return: block_device_info
        """
        LOG.debug("Begin _attach_volumes_and_boot(), %s" % instance['uuid'])

        self._validate_state_before_attach_and_boot(
            context, instance['uuid'],
            method=self._attach_volumes_and_boot.func_name
        )

        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        original_volume_bdms = [bdm for bdm in bdms if bdm.is_volume]

        attached_bdms = []

        # loop through all the volumes ready to attach to the instance
        # during boot and handle the volumes attach.
        # Sort the volumes by 'boot_index' so they get attached
        # in the proper boot order.
        sorted_bdms = block_device_utils.sort_instance_bdms(bdms)
        for bdm in sorted_bdms:
            LOG.debug("Processing BDM %s" % bdm)
            volume_id = bdm['volume_id']

            mountpoint = bdm['device_name']
            # save delete_on_termination since nova compute manage
            # _attach_volume will set delete_on_termination to False
            # regardless.

            delete_on_termination = bdm['delete_on_termination']

            if not mountpoint:
                # now generate a unique mountpoint. Since the instance
                # is not in the state the volume can be attached to,
                # no serialization is needed here. The volume has been
                #  reserved. It cannot be attached to other instance
                # at the same time.

                # bdms is updated every time a volume is attached. Get new copy
                bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
                    context, instance['uuid'])

                # Take the bdms out of the list that don't have a device_name,
                # because nova/compute/utils/get_next_device_name() expects all
                # device_names to have values.
                bdms = [b for b in bdms if b.get('device_name')]

                mountpoint = compute_utils.get_device_name_for_instance(
                    context, instance, bdms, None)

                # Set the device_name in the bdm so attach doesn't try to
                # create a new bdm.
                self._update_bdm_device_name(context, bdm['id'], mountpoint)
                # set the device name in the bdm we already have.
                bdm['device_name'] = mountpoint

            data = {}
            try:
                driver_bdm_list = self._convert_block_devices([bdm])
                self._attach_volume(context, instance, driver_bdm_list[0])
                # _attach_volume is from parent class, which overwrite the
                # delete_on_termination to False regardless. Restore the proper
                # delete_on_termination after attach_volume.
                data = {'delete_on_termination': delete_on_termination}
                self.conductor_api.block_device_mapping_update(
                    context, bdm['id'], data)

            except Exception:
                with excutils.save_and_reraise_exception() as ex:
                    ex_info = ex.value

                    reason = (_("Could not deploy %(name)s due to "
                                "volume: %(volume_id)s attach failure. "
                                "Error: %(error)s") %
                              {'name': instance['name'],
                               'volume_id': volume_id,
                               'error': ex_info})
                    # The notification has been sent by _attach_volume
                    # don't send it again. Just log the error.
                    LOG.error(reason)
                    # also need to terminate and detach all the attached
                    # volumes if we fail to attach boot volume
                    # For the 'delete_on_termination' volumes,
                    # _delete_instance will clean them up during "delete"
                    # Just clean up attachment here. _delete_instnace will
                    # also clean up the instance BDMS.
                    attached_vol_ids = []
                    for tmpbdm in attached_bdms:
                        volid = tmpbdm['volume_id']
                        attached_vol_ids.append(volid)
                        self._detach_volume(context, instance, tmpbdm)
                        connector = \
                            self.driver.get_volume_connector(instance)
                        self.driver.call_terminate_connection(context,
                                                              instance,
                                                              volid,
                                                              connector)
                        volume = self.volume_api.get(context, volid)
                        attachment = self._get_volume_attachment(
                            volume, instance['uuid'])
                        self.volume_api.detach(context, volid,
                                               attachment)

                    # All volumes have their state set to 'attaching' by the
                    # _validate_volumes callback before this method is called.
                    # We need to unreserve any volumes that were not
                    # successfully attached to free-up any volumes that the
                    # user added to the deploy as 'attach existing volume'.
                    for bdm in original_volume_bdms:
                        if bdm['volume_id'] not in attached_vol_ids:
                            self.volume_api.unreserve_volume(context,
                                                             bdm['volume_id'])

            attached_bdms.append(bdm)

        # return block_device_info
        return self._get_instance_block_device_info(context, instance)

    def _clean_conn_info(self, connection_info):
        # cleanup the password out of an iscsi connector
        if (connection_info['driver_volume_type'] != 'iscsi'):
            return connection_info
        if 'data' not in connection_info:
            return connection_info
        if 'auth_password' not in connection_info['data']:
            return connection_info
        conn_info = copy.deepcopy(connection_info)
        conn_info['data']['auth_password'] = "**********"
        return conn_info

    def _clean_bdm(self, bdm):
        # sanitize the password out of an iscsi bdm
        c_info = bdm.get("connection_info")
        if c_info:
            if isinstance(c_info, basestring):
                c_info_d = jsonutils.loads(c_info)
            elif isinstance(c_info, dict):
                c_info_d = c_info
            else:
                return bdm
            sanitized = self._clean_conn_info(c_info_d)
            if c_info_d == sanitized:
                return bdm
            bdm_copy = copy.deepcopy(bdm)
            bdm_copy["connection_info"] = sanitized
            return bdm_copy
        return bdm

    def _clean_bdm_info(self, bdm_info):
        # sanitize the passwords out of a bdm_info
        bdm_list = []
        found_diff = False
        if bdm_info and 'block_device_mapping' in bdm_info:
            for bdm in bdm_info['block_device_mapping']:
                bdm_sanitized = self._clean_bdm(bdm)
                bdm_list.append(bdm_sanitized)
                if bdm != bdm_sanitized:
                    found_diff = True
            if found_diff:
                bdm_info_copy = copy.deepcopy(bdm_info)
                bdm_info_copy['block_device_mapping'] = bdm_list
                return bdm_info_copy
        return bdm_info

    def _rollback_initialize_connection(self, context, instance, volume_id,
                                        connector, conninfo=None):
        """
        Calls the terminate connection to remove the mapping from the
        SAN, It handles the case of Dual VIOS mapping in case of
        vSCSI attachments
        """
        LOG.info(_("_rollback_initialize_connection %(con)s"
                   "and %(con_info)s"), {'con': connector,
                                         'con_info': conninfo})
        self.driver.call_terminate_connection(context, instance, volume_id,
                                              connector, conninfo)

    def _get_connection_type(self, connector):
        if (connector is not None and 'connection_type' in connector
                and (connector['connection_type'] is not None)):
            return connector['connection_type'][0]

    @logcall
    def _attach_volume_osee(self, context, volume_id, mountpoint,
                            instance, is_boot_set):
        context = context.elevated()
        LOG.info(_('Attaching volume %(volume_id)s to %(mountpoint)s'),
                 {'volume_id': volume_id, 'mountpoint': mountpoint},
                 context=context, instance=instance)
        try:
            connector = self.driver.get_volume_connector(instance)
            msg = (_("Attaching volume %(volume_id)s to instance %(inst)s with"
                     " connector: %(connector)s") %
                   {'volume_id': volume_id,
                    'inst': instance['uuid'],
                    'connector': connector})
            LOG.info(msg)
            #for boot volumes the boot_connectivity value needs to be passed
            #in connector info
            c_info = None
            if connector:
                if is_boot_set:
                    connector['connection_type'] = connector.pop(
                        'boot_conn_types')\
                        if 'boot_conn_types' in connector else None
                    connector.pop('data_conn_types', None)
                else:
                    connector['connection_type'] = connector.pop(
                        'data_conn_types')\
                        if 'data_conn_types' in connector else None
                    connector.pop('boot_conn_types', None)

                if connector['connection_type'] and\
                   CONNECTION_TYPE_VSCSI in connector['connection_type']:
                    # for vSCSI we do not need this property so pop it
                    if 'phy_to_virt_initiators' in connector.keys():
                        connector.pop('phy_to_virt_initiators')

                    wwpn_vscsi = connector.pop('wwpn_vscsi')
                    vscsi_lun_list = []
                    for wwpns in wwpn_vscsi:
                        connector['wwpns'] = wwpns['wwpns']
                        connector['host'] = wwpns['vios_host']
                        connector['ip'] = wwpns['vios_ip']
                        c_info = self.volume_api.\
                            initialize_connection(context,
                                                  volume_id, connector)
                        vios_target_lun = {
                            'vios_name': wwpns['vios_name'],
                            'target_lun': c_info['data']['target_lun'],
                            'vios_wwpn': wwpns['wwpns'],
                            'target_wwpn': c_info['data']['target_wwn']
                        }
                        vscsi_lun_list.append(vios_target_lun)
                    if c_info:
                        c_info['vscsi_target_lun'] = vscsi_lun_list
                    else:
                        raise Exception("Failed to intialize vSCSI Volume "
                                        "connection")
                else:
                    # for NPIV we do not need this property so pop it
                    if 'wwpn_vscsi' in connector.keys():
                        connector.pop('wwpn_vscsi')
            if not c_info:
                c_info = self.volume_api.\
                    initialize_connection(context, volume_id, connector)
            #[shyama]connection_info should now have
            # {'driver_volume_type':'ssp'/'fibre_channel',
            #  'data': {..'connector_type':'npiv'/'pv_vscsi'...} }
            msg = (_("Successfully established volume %(volid)s "
                     "connectivity, connection_info: %(conn_info)s") %
                   dict(volid=volume_id,
                        conn_info=self._clean_conn_info(c_info)))
            LOG.info(msg)
        except Exception:  # pylint: disable=W0702
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Failed to connect to volume %(volume_id)s "
                                "while attaching at %(mountpoint)s"),
                              {'volume_id': volume_id,
                               'mountpoint': mountpoint},
                              context=context, instance=instance)
                self.volume_api.unreserve_volume(context, volume_id)
        if 'serial' not in c_info:
            c_info['serial'] = volume_id

        try:
            self.driver.attach_volume(context,
                                      c_info,
                                      instance,
                                      mountpoint,
                                      encryption=None)
            msg = (_("Successfully attached volume: %(volid)s to instance:"
                     "%(inst)s.") % dict(volid=volume_id,
                                         inst=instance['uuid']))
            LOG.info(msg)
        except Exception:  # pylint: disable=W0702
            with excutils.save_and_reraise_exception():
                LOG.exception(_("Failed to attach volume %(volume_id)s "
                                "at %(mountpoint)s") %
                              {'volume_id': volume_id,
                               'mountpoint': mountpoint},
                              context=context, instance=instance)
                LOG.info(_("Clean up failed volume %(volid)s attachment "
                           "to instance %(inst)s with connector: "
                           "%(connector)s") %
                         {'volid': volume_id,
                          'inst': instance['uuid'],
                          'connector': connector})
                self._rollback_initialize_connection(context, instance,
                                                     volume_id,
                                                     connector,
                                                     c_info)

        self.volume_api.attach(context,
                               volume_id,
                               instance['uuid'],
                               mountpoint)
        # validate connector and connection_info and merge.
        # the updated connection info will have connector saved it
        # if a valid NPIV or SSP connector. Each volume attached
        # will its own connector saved.
        conn_info = self._update_conninfo_with_connector(
            context, instance, volume_id, c_info, connector)
        values = {
            'instance_uuid': instance['uuid'],
            'connection_info': jsonutils.dumps(conn_info),
            'device_name': mountpoint,
            'delete_on_termination': False,
            'virtual_name': None,
            'snapshot_id': None,
            'volume_id': volume_id,
            'volume_size': None,
            'no_device': None}
        self.conductor_api.block_device_mapping_update_or_create(context,
                                                                 values)
        self._instance_update(context, instance['uuid'],
                              updated_at=timeutils.utcnow())
        info = dict(volume_id=volume_id)
        self._notify_about_instance_usage(
            context, instance, "volume.attach", extra_usage_info=info)

    def _validate_state_before_attach_and_boot(self, context,
                                               instance_uuid, method=None):
        """
        validate instance state before attach volumes and boot

        :param context: operation context
        :param instance_uuid: compute instance
        :param method: compute manager method that calls validate
        :return: instance if validate or None
        """

        instance = instance_obj.Instance.get_by_uuid(context, instance_uuid)
        vm_state = instance['vm_state']
        task_state = instance['task_state']

        if (vm_state != vm_states.BUILDING or
                task_state != task_states.SPAWNING):
            msg = (_("_validate_volumes() is called when instance is "
                     "in the wrong state: vm_state=%(vm_state)s "
                     "task_state=%(task_state)s") % locals())
            LOG.error(msg)
            exc_args = {'instance_uuid': instance_uuid,
                        'attr': vm_state,
                        'state': task_state,
                        'method': method}
            raise exception.InstanceInvalidState(**exc_args)

        return instance

    def _validate_attachment(self, context, instance_uuid):
        pass

    def _notify_deploy_volume_create_failure(self, context, instance, volume):
        create_fail_msg = volume['volume_metadata'].get(common_constants.
                                                        CREATE_FAILURE_KEY)
        if not create_fail_msg:
            create_fail_msg = _('No detailed volume error message.')
        reason = (_("Instance %(instance_name)s volume creation "
                    "failure. Health status: %(hstatus)s. "
                    " Volume error: %(volume_error)s.") %
                  {'instance_name': instance['display_name'],
                   'hstatus': volume.get('health_status'),
                   'volume_error': create_fail_msg})
        LOG.error(reason)
        self._error_notify(instance, context, 'error', reason)

    def _validate_for_error_status(self, context, instance, volume, blank_vol):
        if 'error' in volume['status']:
            vol_msg = self._notify_deploy_volume_create_failure(context,
                                                                instance,
                                                                volume)
            src_vol_id = volume['source_volid']
            exc_args = {'instname': instance['display_name'],
                        'instuuid': instance['uuid'],
                        'vol_error': vol_msg}
            # Throw different exceptions if this is a volume create
            # from volume vs a blank volume create vs an existing
            # volume that is being attached which happens to be in error.
            if src_vol_id:
                exc_args['srcvolid'] = src_vol_id
                raise pwrvc_excep.IBMPowerVCVolumeCloneFailure(**exc_args)
            elif blank_vol:
                raise pwrvc_excep.IBMPowerVCVolumeCreateFailure(**exc_args)
            else:
                exc_args['vol_id'] = volume['id']
                exc_args['vol_name'] = volume['display_name']
                raise pwrvc_excep.IBMPowerVCExistingVolumeInError(**exc_args)

    def _validate_volumes(self, context, instance_uuid):
        """
        callback function used by virtual driver during spawn to
        validate all the volumes state before attach those
        volumes and boot.
        :param context: operation context
        :param instance_uuid: compute instance uuid
        :return: None
        """
        LOG.debug("Begin _validate_volumes():")
        instance = self._validate_state_before_attach_and_boot(
            context, instance_uuid, method=self._validate_volumes.func_name
        )
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        bdms = [bdm for bdm in bdms if bdm.is_volume]
        for bdm in bdms:
            volume = self.volume_api.get(context, bdm.volume_id)
            blank_volume = (bdm.source_type == 'blank')
            self._validate_for_error_status(context, instance, volume,
                                            blank_volume)
            if volume['status'] not in ['available', 'in-use']:
                stg_type = 'fc'
                try:
                    stg_type = self.volume_api.get_storage_type_for_volume(
                        context, bdm.volume_id)
                except Exception:
                    stg_type = 'fc'

                try:
                    if stg_type != 'cluster':
                        self._await_block_device_map_created(context,
                                                             bdm.volume_id)
                    else:
                        # Wait forever for SSP cluster volume clone creations
                        self._await_volume_create(context, bdm.volume_id)
                except exception.VolumeNotCreated:
                    with excutils.save_and_reraise_exception():
                        volume = self.volume_api.get(context, bdm.volume_id)
                        self._notify_deploy_volume_create_failure(context,
                                                                  instance,
                                                                  volume)

                volume = self.volume_api.get(context, bdm.volume_id)
            # make sure the volume is in the right state to attach
            self.volume_api.check_attach(context, volume, instance=instance)
            # change the volume to "attaching" state
            self.volume_api.reserve_volume(context, volume['id'])

    def _reset_volume_bdm(self, bdm, source_volume_id, img_vols_to_clone):
        if not bdm:
            return
        if bdm['delete_on_termination']:
            if (bdm['source_type'] == 'volume' and
                    source_volume_id in img_vols_to_clone):
                bdm['volume_id'] = source_volume_id
                bdm['delete_on_termination'] = False
                LOG.info(_('Resetting BDM %(bdm_id)s volume_id to %(vol_id)s'
                           'and delete_on_termination value to False.')
                         % {'bdm_id': bdm['id'],
                            'vol_id': source_volume_id})
            elif bdm['source_type'] == 'blank':
                bdm['volume_id'] = None
                LOG.info(_('Resetting BDM %(bdm_id)s volume_id to %(vol_id)s')
                         % {'bdm_id': bdm['id'],
                            'vol_id': 'None'})
        bdm['connection_info'] = None
        bdm.save()

    def _bdm_cleanup_helper(self, context, instance, bdm, connector=None,
                            clean_zones=False, img_vols_to_clone=[]):
        """
        The helper function to be used by _shutdown_instance(). The
        purpose is to consolidate all the volume detach and terminate
        connection function needed during VM destroy.
        """
        if not bdm:
            return

        volume_id = None
        source_volume_id = None
        try:
            volume_id = bdm['volume_id']
            # BDMs with source type of blank may not have a volume_id
            # if the deploy failed very early.
            if not volume_id:
                LOG.debug("BDM has the volume_id set to None")
                return

            # We do not process volumes that are in the image and will be
            # cloned as all operations would be against a clone and not
            # the image volume.
            if volume_id in img_vols_to_clone:
                LOG.debug("Volume %(vol)s is a part of the image %(img)s" %
                          {'vol': volume_id, 'img': img_vols_to_clone})
                return

            volume = self.volume_api.get(context, volume_id)
            source_volume_id = volume['source_volid']
            stg_type = self.volume_api.get_storage_type_for_volume(context,
                                                                   volume_id)
            # For the volume that is not fully attached yet.
            if not self._bdm_has_attachment(bdm):
                # Volumes are 'reserved' and placed into 'attaching' state
                # early in the deploy process before the LPAR is created and
                # before attachment really begins.
                # We need to unreserve the volumes so that volumes marked with
                # delete_on_termination=true can be deleted and volumes
                # marked with delete_on_termination=false can be freed up
                # so they are available for VM attach to other VMs on deploy
                # failure or this same VM on deploy-fail-reschedule.
                if volume['status'] == 'attaching':
                    self.volume_api.unreserve_volume(
                        context, bdm['volume_id'])
                if bdm['delete_on_termination']:
                    LOG.debug("Terminating bdm: %(bdm)s" % dict(bdm=bdm))
                    # if the volume is creating we are deleting a failed
                    # deploy. So we need to wait for the volume to
                    # finish creating before we delete it.
                    # If we leave it as 'creating' delete will fail.
                    # The volume create will eventually finish or fail
                    # so wait here.
                    # If the volume is attached we would not get here
                    # and we would
                    if volume['status'] == 'creating':
                        self._await_volume_create(context,
                                                  volume_id)
                    self.volume_api.delete(context, bdm['volume_id'])
                return

            if not connector:
                # retrieve the connector from bdm connection_info
                connector = self._get_bdm_volume_connector(
                    context, instance, bdm, bdm['volume_id'])
                # for SSP connector will be None
                if not connector and stg_type == 'fc':
                    # the connector is lost, but there is an
                    # attachment.
                    if volume['attach_status'] == 'attached':
                        connector = self._construct_connector_from_conninfo(
                            instance, bdm)

                        # for onboarded vm's vscsi volumes
                        # wwpns are missing so get connector again
                        if not connector:
                            try:
                                c_info = bdm.get("connection_info")
                                connection_info = jsonutils.loads(c_info)
                                vol_conn = connection_info.get(
                                    "volume_connector")
                                connection_type = vol_conn.get(
                                    'connection-type')
                                if CONNECTION_TYPE_VSCSI == connection_type:
                                    connector = self.driver. \
                                        get_volume_connector(instance,
                                                             isVscsi=True)
                                    LOG.info(_("Retrieved connector from"
                                             " driver : %(connector)s" % dict(
                                                 connector=connector)))
                            except Exception as ex:
                                LOG.error(_("Failed to retrieve connector for "
                                            "instance %(inst)s due to ,"
                                            "%(error)s") %
                                          {'inst': instance['uuid'],
                                           'error': ex})
                        if not connector:
                            # This is critical condition. Send UI notification.
                            message = (_("Could not terminate storage host "
                                         "mapping and SAN zoning for volume: "
                                         "%(volid)s when deleting "
                                         "the virtual server. Please check the"
                                         " host logs for further details.") %
                                       dict(volid=volume_id))
                            self._error_notify(instance, context,
                                               'critical',
                                               message)
                            # No need to try terminate_connection which will
                            # raise exception. The nova "detached" volume
                            # will be left in 'in-use' state. This needs
                            # some out of band clean up.
                            # TODO: Need to have cinder volume driver to
                            # rebuild the connector and send it back.
                            return
                        # Built a connnector based on the connection_info.
                        LOG.debug("connector created from connection_info: "
                                  "%(connector)s" % dict(connector=connector))

            LOG.info(_("Detach volume %(volume_id)s for instance %(inst)s "
                       "with connector: %(connector)s") %
                     {'volume_id': volume_id,
                      'inst': instance['uuid'],
                      'connector': connector})

            # detach then terminate
            attachment = self._get_volume_attachment(
                volume, instance['uuid'])
            if attachment:
                self.volume_api.detach(context, volume_id, attachment)

            if (connector is not None and 'connection_type' in connector and
                CONNECTION_TYPE_VSCSI in connector['connection_type']) or\
               (connector is not None and 'wwpn_vscsi' in connector):
                if 'phy_to_virt_initiators' in connector and\
                   not connector['phy_to_virt_initiators']:
                    connector.pop('phy_to_virt_initiators', None)
                conninfo = jsonutils.loads(bdm['connection_info'])
                self.driver.call_terminate_connection(context, instance,
                                                      bdm['volume_id'],
                                                      connector, conninfo)
            else:
                # clean_zones only applies to NPIV (stg_type fc)
                # no zoning for iscsi or vscsi
                if stg_type == 'fc' and clean_zones:
                    # It is the last attached volume for VM. Tell cinder that
                    # it is time to remove all the zones.
                    connector['clean_zones'] = 'all'
                self.volume_api.terminate_connection(context,
                                                     bdm['volume_id'],
                                                     connector)
            # make sure the volume in attaching state changes back to
            # available, so it could be handled by cleanup_volumes().
            self.volume_api.unreserve_volume(context, bdm['volume_id'])
            if volume_id and bdm['delete_on_termination']:
                LOG.debug("Terminating bdm: %(bdm)s" % dict(bdm=bdm))
                self.volume_api.delete(context, bdm['volume_id'])
        except cinder_exception.EndpointNotFound as exc:
            # if the failure is due to endpoint not found,
            # it's likely that context is out of date.
            # We should fail the operation and ask for retry
            with excutils.save_and_reraise_exception():
                raise Exception(_('Failed to access cinder service to '
                                  'remove volumes. Try the operation again.'))
        except exception.DiskNotFound as exc:
            LOG.warn(_('Ignoring DiskNotFound: %(exc)s, instance: %(inst)s,'
                       'volume id: %(volid)s') %
                     dict(exc=exc, inst=instance['uuid'], volid=volume_id))
        except exception.VolumeNotFound as exc:
            LOG.warn(_('Ignoring VolumeNotFound: %(exc)s, instance: %(inst)s,'
                       'volume id: %(volid)s') %
                     dict(exc=exc, inst=instance['uuid'], volid=volume_id))
        except Exception as exc:
            LOG.exception(exc)
            # differentiate based on SSP or not
            if not connector or ('connection_type' in connector and
               CONNECTION_TYPE_SSP in connector['connection_type']):
                reason = (_("The virtual server was deleted. "
                            "However, a virtual server cleanup "
                            "problem occurred on one of the "
                            "following operations: "
                            "The host mapping was unable to be "
                            "removed for one or more volumes on the "
                            "storage controllers and the host "
                            "mapping may require manual "
                            "cleanup, or one or more volumes on the "
                            "storage controllers failed removal and "
                            "may require manual cleanup. "
                            "Error: %(exc)s")
                          % dict(exc=exc))
            else:
                reason = (_("The virtual server was deleted. "
                            "However, a virtual server cleanup "
                            "problem occurred on one of the "
                            "following operations: "
                            "The host mapping was unable to be "
                            "removed for one or more volumes on the "
                            "storage controllers and the host "
                            "mapping may require manual "
                            "cleanup, or one or more volumes on the "
                            "storage controllers failed removal and "
                            "may require manual cleanup, "
                            "or the NPIV zone information was "
                            "unable to be removed from the switch "
                            "fabric and the zone may require manual "
                            "cleanup. Error: %(exc)s")
                          % dict(exc=exc))
            self._error_notify(instance, context, 'info', reason)
            LOG.debug("Unable to terminate and detach volume for instance "
                      "%(inst)s %(volid)s with connector: %(conn)s, BDM: "
                      "%(bdm)s" %
                      {'volid': bdm['volume_id'],
                       'inst': instance['uuid'],
                       'conn': connector,
                       'bdm': self._clean_bdm(bdm)})
        finally:
            try:
                context.read_deleted = 'yes'
                self._reset_volume_bdm(bdm, source_volume_id,
                                       img_vols_to_clone)
            finally:
                context.read_deleted = 'no'
        return connector

    @logcall
    def _get_bdm_volume_connector(self, context, instance, bdm, volume_id):
        """ For each volume that has been attached, it will be attached
        with a certain connector which will not change during the life
        cycle of the attachment. Save the connector to the connection_info.
        And it will be retrieved during terminate_connection. For NPIV,
        get_volume_connector could change since the SCG could be different.
        By saving the connector, the terminate_connection will be successful.
        Otherwise, the terminate_connector may not work if a different
        connector is given that is different from the one that used
        to initialize_connection().
        """

        if not bdm:
            bdm = self._get_bdm_for_instance_and_vol(context, instance,
                                                     volume_id)
        if not bdm or bdm['volume_id'] != volume_id:
            return None

        connector = None
        c_info = bdm.get("connection_info")

        if c_info and isinstance(c_info, basestring):
            try:
                connection_info = jsonutils.loads(c_info)
                sanitized_bdm = self._clean_bdm(bdm)
                LOG.debug("BDM: %(bdm)s" % {'bdm': sanitized_bdm})
                vol_conn = connection_info.get("volume_connector")
                connector = self._volume_connector_validator(vol_conn)
            except:
                # invalid connector
                connector = None
        if not connector:
            LOG.info(_("Volume_connector is missing from block device "
                       "mapping for instance %(vm_id)s and volume "
                       "%(vol_id)s. A connector may be generated.") %
                     dict(vm_id=instance['id'], vol_id=volume_id))
        return connector

    @logcall
    def _get_SSP_volume_connector(self, instance, bdm):
        """ Look into the BDM at the connection_info contained
        within the BDM and determine if this is BDM is of
        """
        if not bdm or not bdm.get("connection_info"):
            return None
        try:
            c_info = bdm.get("connection_info")
            connection_info = jsonutils.loads(c_info)
            vol_conn = connection_info.get("volume_connector")
            c_type = vol_conn['connection-type']
            if c_type == CONNECTION_TYPE_SSP:
                volume_connector = {'connection-type': CONNECTION_TYPE_SSP,
                                    'connector': None}
                return volume_connector
        except Exception:
            # We could have just returned None here
            # but we pass here so we return None below
            # We get here when we have an incomplete connector
            # None causes us to keep processing
            pass

        return None

    def _construct_connector_from_conninfo(self, instance, bdm):
        """ The connection_info for NPIV attached volume has an
        attribute called "initiator_target_map". It contains the
        vfc wwpn to target wwpn mapping. This could also be used
        to build a potential "good enough" connector for
        terminate_connection to use.
        """
        connector = None
        if not bdm or not bdm.get("connection_info"):
            return connector
        try:
            conninfo = jsonutils.loads(bdm['connection_info'])
            if not conninfo.get('driver_volume_type'):
                return connector
            conn_type = conninfo['driver_volume_type']
            if conn_type == 'fibre_channel' and conninfo.get('data'):
                conn_data = conninfo['data']
                if conn_data.get('initiator_target_map'):
                    # this is the NPIV connection. Let's construct the
                    # connector from the map.
                    # initiator_target_map looks like this:
                    # {"c0507606529b01a3":
                    #         ["500507680510041a", "500507680510041b"],
                    #  "c0507606529b01a2":
                    #         ["500507680510041a", "500507680510041b"],
                    #  ....}
                    i_t_map = conn_data['initiator_target_map']
                    if i_t_map and isinstance(i_t_map, dict):
                        i_wwpns = set(i_t_map.keys())
                        connector = {'connection_type': [CONNECTION_TYPE_NPIV],
                                     'host': instance['name'],
                                     'wwpns': list(i_wwpns),
                                     'phy_to_virt_initiators': None}
                if(conn_data.get('wwpns')):
                    connector = {'connection_type': [CONNECTION_TYPE_VSCSI],
                                 'host': instance['name'],
                                 'wwpns': conn_data.get('wwpns')}
            elif conn_type == CONNECTION_TYPE_SSP:
                connector = {'connection_type': [CONNECTION_TYPE_SSP],
                             'connector': None}
        except Exception as ex:
            LOG.exception(ex)
            connector = None
        LOG.debug("_construct_connector_from_conninfo: %(connector)s" %
                  locals())
        return connector

    @logcall
    def _update_conninfo_with_connector(self, context, instance, volume_id,
                                        connection_info, connector):
        """
        This is helper function to update the connection_info with a new
        volume_connector field if the attached volume connector is VFC(NPIV)
        or SSP based. For HMC managed PowerVC, the volume connector will
        not change during the life cycle of the volume attachment even
        after migration. Save the connector can help speed up the
        detach_volume patch as well destroy/cleanup path with may not
        have a valid volume connector since LPAR may be in some invalid
        state.
        """
        if not volume_id or not connection_info or not connector:
            return connection_info
        volume_connector = None

        try:
            volume = self.volume_api.get(context, volume_id)

            if 'error' in volume['status']:
                # The volume is not healthy after initialize_connection.
                # something went wrong and we don't want to update
                # the BDM with the connection info.
                raise ValueError
            if connection_info['driver_volume_type'] == 'ibm_ssp':
                volume_connector = {'connection-type': CONNECTION_TYPE_SSP,
                                    'connector': None}
            elif (connection_info['driver_volume_type'] == 'fibre_channel'
                  and connection_info['data']['connector_type'] ==
                  CONNECTION_TYPE_NPIV
                  and connector['phy_to_virt_initiators']
                  and connector['host']
                  and connector['wwpns']
                  and connection_info['data']['target_wwn']
                  and int(connection_info['data']['target_lun']) >= 0
                  and connection_info['data']['volume_id'] == volume_id):
                # if there is anything missing, it will cause exception.
                volume_connector = {'connection-type': CONNECTION_TYPE_NPIV,
                                    'connector': connector}
            elif (connection_info['driver_volume_type'] == 'fibre_channel'
                  and connection_info['data']['connector_type'] ==
                  CONNECTION_TYPE_VSCSI
                  and connector['host']
                  and connector['wwpns']
                  and connection_info['data']['target_wwn']
                  and int(connection_info['data']['target_lun']) >= 0
                  and connection_info['data']['volume_id'] == volume_id):
                # if there is anything missing, it will cause exception.
                volume_connector = {'connection-type': CONNECTION_TYPE_VSCSI,
                                    'connector': connector}
            else:
                volume_connector = None
        except:
            volume_connector = None
        # for IVM san FC connection, it will not be updated with the
        # connector
        if volume_connector:
            connection_info.update({'volume_connector': volume_connector})
        return connection_info

    @logcall
    def _volume_connector_validator(self, volume_connector):
        """
        Validate the volume connector format saved into the. Not the content.
        This is the volume_connector saved in the BDM connection_info.
        Will return the real volume connector if it is valid.
        For IVM FC connector, we don't save it since it could be
        attached to both source/target host during migration.
        volume_connector:
            { 'connection-type': 'npiv'| 'ibm_ssp',
              'connector': {...} actual connector returned from
              get_volume_connector
            }
            fc is the legacy FibreChannel connector
            npiv is the NPIV VFC connector
            ibm_ssp is the ssp vscsi connector
        """

        if not volume_connector:
            return None

        conn_info = None
        try:
            c_type = volume_connector['connection-type']
            conn_info = volume_connector['connector']
            if c_type == CONNECTION_TYPE_NPIV:
                if (not conn_info['host'] or
                    not conn_info['wwpns'] or
                        not conn_info['phy_to_virt_initiators']):
                    conn_info = None
            elif c_type == CONNECTION_TYPE_VSCSI:
                if (not conn_info['host'] or
                        not conn_info['wwpns']):
                    conn_info = None
            elif c_type == CONNECTION_TYPE_SSP:
                conn_info = None
            else:
                conn_info = None
        except:
            # validation failure
            conn_info = None
        return conn_info

    def _get_bdm_for_instance_and_vol(self, context, instance, volume_id):
        """ Helper function to retrieve the volume_id for
        instance"""
        # go get the bdm for this instance and volume id
        bdms = blkdev_obj.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance['uuid'])
        bdm = [bdm for bdm in bdms if bdm['volume_id'] and
               bdm['volume_id'] == volume_id and bdm['connection_info']]
        return bdm[0] if bdm else None

    def _update_bdm_device_name(self, context, bdm_id, device_name):
        value = {'device_name': device_name}
        self.conductor_api.block_device_mapping_update(context,
                                                       bdm_id,
                                                       value)

    def _bdm_has_attachment(self, bdm):
        if not bdm or not bdm.get("connection_info"):
            return False
        connection_info = bdm.get("connection_info")
        if (not connection_info or 'driver_volume_type' not in
                connection_info or 'data' not in connection_info):
            return False
        else:
            return True

    def _valid_volume_connector(self, connector):
        if not connector:
            return False
        if connector.get('host') and connector.get('wwpns'):
            return True
        else:
            return False

    def _is_managing_PowerKVM(self):
        if hasattr(self.driver, "_get_hypervisor_type"):
            hypervisor_type = self.driver._get_hypervisor_type()
            # For PowerKVM, use openstack block device path directly
            if hypervisor_type and hypervisor_type.lower() == 'qemu':
                return True
        return False

    def is_instance_on_host(self, context, instance_uuid, instance_name):
        """ Answer query from another host if instance
            is on this host.
        """
        ret = self.driver.is_instance_on_host(instance_uuid, instance_name)
        if ret:
            LOG.info(_('Virtual machine %(name)s (%(uuid)s) is being '
                       'managed by this host.') %
                     {'uuid': instance_uuid, 'name': instance_name})
        else:
            LOG.info(_('Virtual machine %(name)s (%(uuid)s) is not '
                       'being managed by this host.') %
                     {'uuid': instance_uuid, 'name': instance_name})
        return ret

    @periodic_task.periodic_task(spacing=60, run_immediately=True)
    def canary(self, context):
        """
        A simple periodic task to record when it ran last
        This will give us an indication if periodic tasks
        are running normally or if they may be hung.
        """
        global CANARY_RUN
        CANARY_RUN = time.time()

    @periodic_task.periodic_task
    @cache_driver_data(['get_available_resource'])
    def pvc_update_available_resource(self, context):
        return super(PowerVCComputeManager, self).update_available_resource(
            context)

    #@compute_utils.periodic_task_spacing_warn("sync_power_state_interval")
    @periodic_task.periodic_task(spacing=CONF.sync_power_state_interval,
                                 run_immediately=True)
    def _pvc_sync_power_states(self, context):
        """
        Override _sync_power_states for performance purpose.
        With HMC, each get_info would result a k2operator call of
        LogicalPartition/<uuid> GET. The data is cached in k2operator as
        feed or single entry unmarshalled. In most cases, we have to search
        via lpars feed. When there are hundreds of LPARs, the unmarshalling
        process can take a while.
        To improve performance, we will get lpars feed unmarshalled first
        and pass it for each get_info.
        """
        is_hmc = False
        if CONF.powervm_mgr_type and CONF.powervm_mgr_type.lower() == 'hmc':
            is_hmc = True
        if not is_hmc:
            return super(PowerVCComputeManager, self)._sync_power_states(
                context)

        db_instances = objects.InstanceList.get_by_host(context,
                                                        self.host,
                                                        use_slave=True)

        num_vm_instances = self.driver.get_num_instances()
        num_db_instances = len(db_instances)

        if num_vm_instances != num_db_instances:
            LOG.warn(_("While synchronizing instance power states, found "
                       "%(num_db_instances)s instances in the database and "
                       "%(num_vm_instances)s instances on the hypervisor."),
                     {'num_db_instances': num_db_instances,
                      'num_vm_instances': num_vm_instances})

        def _sync(db_instance):
            # NOTE(melwitt): This must be synchronized as we query state from
            #                two separate sources, the driver and the database.
            #                They are set (in stop_instance) and read, in sync.
            @utils.synchronized(db_instance.uuid)
            def query_driver_power_state_and_sync():
                self._pvc_query_driver_power_state_and_sync(context,
                                                            db_instance,
                                                            lpars_feed=lpars)

            try:
                # this if stanza is moved from
                # _query_driver_power_state_and_sync for performance purpose
                if db_instance.task_state is not None:
                    LOG.info(_LI("During sync_power_state the instance has a "
                                 "pending task (%(task)s). Skip."),
                             {'task': db_instance.task_state},
                             instance=db_instance)
                else:
                    query_driver_power_state_and_sync()
            except Exception:
                LOG.exception(_LE("Periodic sync_power_state task had an "
                                  "error while processing an instance."),
                              instance=db_instance)

            self._syncs_in_progress.pop(db_instance.uuid)

        lpars = self.driver.get_lpars_feed()
        for db_instance in db_instances:
            # process syncs asynchronously - don't want instance locking to
            # block entire periodic task thread
            uuid = db_instance.uuid
            if uuid in self._syncs_in_progress:
                LOG.debug('Sync already in progress for %s' % uuid)
            else:
                LOG.debug('Triggering sync for uuid %s' % uuid)
                self._syncs_in_progress[uuid] = True
                self._sync_power_pool.spawn_n(_sync, db_instance)

    def _pvc_query_driver_power_state_and_sync(self, context, db_instance,
                                               lpars_feed=None):
        # No pending tasks. Now try to figure out the real vm_power_state.
        try:
            vm_instance = self.driver.get_info(db_instance,
                                               lpars_feed=lpars_feed)
            vm_power_state = vm_instance.state
        except exception.InstanceNotFound:
            vm_power_state = power_state.NOSTATE
        # Note(maoy): the above get_info call might take a long time,
        # for example, because of a broken libvirt driver.
        try:
            self._sync_instance_power_state(context,
                                            db_instance,
                                            vm_power_state,
                                            use_slave=True)
        except exception.InstanceNotFound:
            # NOTE(hanlind): If the instance gets deleted during sync,
            # silently ignore.
            pass

    def set_host_dead(self, context, dead=True, target_host=None):
        if (self._is_managing_PowerKVM()):
            # for KVM, we can only enable remote restart if the host
            # service is dead
            raise exception.ComputeServiceInUse(host=self.host)
        else:
            self.driver.set_host_dead(context, self.host,
                                      dead, target_host)
            self.update_available_resource(context)

    def _get_instances_on_driver(self, context, filters=None):
        """Override to include vm_names in filters for instance queries.
        """
        if not filters:
            filters = {}
        try:
            driver_uuids = self.driver.list_instance_uuids()
            if len(driver_uuids) == 0:
                # Short circuit, don't waste a DB call
                return instance_obj.InstanceList()
            filters['uuid'] = driver_uuids
            local_instances = instance_obj.InstanceList.get_by_filters(
                context, filters, use_slave=True)
            return local_instances
        except NotImplementedError:
            pass

        # WI 46938, performance improvement to query instances
        # based off of the vm_name in the instance_power_specs. The
        # community manager employs brute force iteration over all
        # deleted instances which can be expensive at scale.
        filters['vm_names'] = self.driver.list_instances()
        return instance_obj.InstanceList.get_by_filters(context, filters,
                                                        use_slave=True)


#######################################################
##########  Override ResourceTracker Class  ###########
#######################################################
class PowerVCResourceTracker(resource_tracker.ResourceTracker):
    """Extends the base ResourceTracker class for PowerVC to fix issues"""

    def __init__(self, host, driver, nodename):
        # Monkey patch the resource handler to load our resources
        resources.RESOURCE_NAMESPACE = 'powervc_nova.compute.resources'
        """Constructor for the PowerVC extension to the Resource Tracker."""
        super(PowerVCResourceTracker, self).__init__(host, driver, nodename)
        self.conductor_api = conductor.PowerVCConductorAPI()

    def _get_compute_node(self, context):
        """Override the Compute Node Query to not query by Node Name"""
        try:
            n = compute_obj.ComputeNodeList.get_all_by_host(context, self.host)
            return obj_base.obj_to_primitive(n[0])
        except exception.NotFound:
            return None

    def _resource_change(self, resources, save=True):
        """Override the Resource Changed method to handle calling twice"""
        new_rscs = copy.deepcopy(resources)
        old_rscs = copy.deepcopy(self.old_resources)
        # Convert the stats to a dictionary to be able to do comparisons
        if old_rscs.get('stats') and not isinstance(old_rscs['stats'], dict):
            old_rscs['stats'] = jsonutils.loads(old_rscs['stats'])
        if new_rscs.get('stats') and not isinstance(new_rscs['stats'], dict):
            new_rscs['stats'] = jsonutils.loads(new_rscs['stats'])
        has_changed = cmp(old_rscs, new_rscs) != 0

        # If the resources changed and they asked to save, update existing
        # and return result
        if has_changed and save:
            self.old_resources = copy.deepcopy(resources)
            return has_changed

        # If they haven't asked to save, compare resources without metrics
        # included. Since metrics are volatile, this will protect against too
        # many notifications being sent for compute node updates.
        if not save:
            new_rscs.pop('metrics', old_rscs.pop('metrics', None))
            return cmp(old_rscs, new_rscs) != 0

    def _verify_resources(self, resources):
        """Override Verify Resources to not wipe out States in Error case"""
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        # If the only thing in the stats is the state and is error, this means
        # we couldn't connect on startup, so bypass everything and set to error
        if len(stats) <= 4 and (state != 'operating' and state != 'standby'):
            resources = copy.deepcopy(self.old_resources)
            if resources.get('stats') and not isinstance(resources['stats'],
                                                         dict):
                resources['stats'] = jsonutils.loads(resources['stats'])
            resources.get('stats', {}).update(stats)
        else:
            # if the system is powered off, it will not have some of the
            # key values.
            # Now we can do the super class verification of the resource data
            super(PowerVCResourceTracker, self)._verify_resources(resources)

    def _write_ext_resources(self, resources):
        """Override Write Resources so that our Stats aren't wiped out"""
        stats = resources.get('stats', {})
        # Convert the stats to a dictionary to be able to merge in
        if stats and not isinstance(stats, dict):
            stats = jsonutils.loads(stats)
        # First call the super class to let it write out the resources
        super(PowerVCResourceTracker, self)._write_ext_resources(resources)
        # Now add back in our stats that were wiped out by the super class
        for key, val in stats.iteritems():
            if key not in resources['stats']:
                resources['stats'][key] = val

    def _update(self, context, values):
        """Override Update Compute Node to send additional notifications"""
        orig_values = copy.deepcopy(self.old_resources)
        old_hostname = self.compute_node.get('hypervisor_hostname')
        info = dict(host=self.host, hypervisor_hostname=self.nodename)
        info['service_id'] = self.compute_node.get('service_id')
        info['compute_node_id'] = self.compute_node.get('id')
        info['host_display_name'] = encodeutils.safe_decode(
            CONF.host_display_name, 'utf-8')
        # Now we can do the super class to update the Compute Node in the DB
        super(PowerVCResourceTracker, self)._update(context, values)
        # Send a notification that we are updating the Compute Node
        if self._resource_change(orig_values, False):
            LOG.info(_("Sending notifiation that the Compute Node changed..."))
            notifier = rpc.get_notifier(service='compute', host=self.host)
            notifier.info(context, 'compute.node.update.start', info)
            notifier.info(context, 'compute.node.update.end', info)
        # See if we need to update the VM's due to the Hostname changing
        new_hostname = self.compute_node.get('hypervisor_hostname')
        if old_hostname and new_hostname and old_hostname != new_hostname:
            self._update_vm_hostnames(context, new_hostname)

    def _report_hypervisor_resource_view(self, resources):
        """Log the hypervisor's view of free resources.

        This is just a snapshot of resource usage recorded by the
        virt driver.

        The following resources are logged:
            - free memory
            - free disk
            - free CPUs
            - assignable PCI devices

        Override, so that the driver can log the appropriate value
        for free disk. Free disk on a host is only applicable in
        local storage environments.
        """
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
        else:
            if hasattr(self.driver, "log_free_disk"):
                is_audit = False
                self.driver.log_free_disk(is_audit)

                free_ram_mb = (resources['memory_mb']
                               - resources['memory_mb_used'])
                LOG.debug(_("Hypervisor: free ram (MB): %s") % free_ram_mb)

                vcpus = resources['vcpus']
                if vcpus:
                    free_vcpus = vcpus - resources['vcpus_used']
                    LOG.debug(_("Hypervisor: free VCPUs: %s") % free_vcpus)
                else:
                    LOG.debug(_("Hypervisor: VCPU information unavailable"))

                if 'pci_passthrough_devices' in resources and \
                        resources['pci_passthrough_devices']:
                    LOG.debug(_("Hypervisor: assignable PCI devices: %s") %
                              resources['pci_passthrough_devices'])
                else:
                    LOG.debug(_("Hypervisor: no assignable PCI devices"))
            else:
                super(PowerVCResourceTracker, self). \
                    _report_hypervisor_resource_view(resources)

    def _report_final_resource_view(self, resources):
        """Report final calculate of free memory, disk, CPUs, and PCI devices,
        including instance calculations and in-progress resource claims. These
        values will be exposed via the compute node table to the scheduler.

        Override, so that the driver can log the appropriate value
        for free disk. Free disk on a host is only applicable in
        local storage environments.
        """
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
        else:
            if hasattr(self.driver, "log_free_disk"):
                is_audit = True
                self.driver.log_free_disk(is_audit)

                LOG.info(_("Free ram (MB): %s") % resources['free_ram_mb'])

                vcpus = resources['vcpus']
                if vcpus:
                    free_vcpus = vcpus - resources['vcpus_used']
                    LOG.info(_("Free VCPUS: %s") % free_vcpus)
                else:
                    LOG.info(_("Free VCPU information unavailable"))

                if 'pci_devices' in resources:
                    LOG.info(_("Free PCI devices: %s") %
                             resources['pci_devices'])
            else:
                super(PowerVCResourceTracker, self). \
                    _report_final_resource_view(resources)

    def _update_vm_hostnames(self, context, new_hostname):
        """Helper method to update the hostname on VM's to the new hostname"""
        try:
            LOG.info(_("Updating VM's to the new Hostname %s.") % new_hostname)
            # Retrieve all of the Instances for this Host so we can update them
            insts = instance_obj.InstanceList.get_by_host(context, self.host)
            # Loop through each of the Instances, updating the Hostname
            for inst in insts:
                inst.node = new_hostname
                inst.save(context)
        # If we got an exception just log it since don't want to fail
        except Exception as exc:
            LOG.warn(_("Error updating VM's to the new Hostname"))
            LOG.exception(exc)

    def _update_usage_from_instances(self, context, resources, instances):
        """Override the method so we don't try to update instance info
        when the system is down.
        """
        stats = resources.get('stats', {})
        state = stats.get('hypervisor_state')
        if (len(stats) <= 4 and
                (state != 'operating' and state != 'standby')):
            LOG.debug('Host is not up and cannot update resource view.')
            # merge in the stats into existing
            self.stats.digest_stats(resources.get('stats'))
        else:
            super(PowerVCResourceTracker, self). \
                _update_usage_from_instances(context, resources, instances)


#######################################################
###########  Override ComputeVirtAPI Class  ###########
#######################################################
class PowerVCComputeVirtAPI(manager.ComputeVirtAPI):
    """Extends the base ResourceTracker class for PowerVC to fix issues"""

    def __init__(self, compute):
        """Constructor for the PowerVC extension to the Virt API."""
        super(PowerVCComputeVirtAPI, self).__init__(compute)


#######################################################
###########  Periodic Task Watchdog         ###########
#######################################################
#
def dump_thread_stacks():
    for i, t in enumerate(backdoor._find_objects(greenlet.greenlet)):
        LOG.error('Greenthread %s %s' % (i, t))
        LOG.error(traceback.format_stack(t.gr_frame))


def periodic_task_watchdog():
    """
    This works by inserting a "canary" periodic task into the list of tasks
    to run, which records a timestamp each time it runs.  Then this method
    is run as a regular (non periodic) task to checkup on the canary and
    see if it's still alive.  If it hasn't reported a timestamp since the
    timeout then we know the periodic tasks are hung or starved.
    """
    LOG.info(_('Starting periodic task watch dog.'))
    delay_interval = 600
    last_print = None
    last_check = time.time()
    global CANARY_RUN
    while True:
        try:
            greenthread.sleep(delay_interval)
            curr_time = time.time()
            # Check if it appears the time was changed and if so
            # just skip this check and we'll get it next interval
            if curr_time - last_check > 2 * delay_interval:
                continue
            # Don't do anything unless the canary has run at least once
            if CANARY_RUN:
                # Check if the canary ran since the last time we printed
                # out a message saying they stopped and if so, they must
                # have resumed.
                if last_print and last_print != CANARY_RUN:
                    last_print = None
                    LOG.info(_('Detected periodic tasks have resumed.'))

                # if canary has run but earlier than our interval
                if (last_print != CANARY_RUN and
                        (CANARY_RUN + delay_interval < curr_time)):
                    # Seems the periodic tasks are stuck, dump thread stacks
                    LOG.error(_('Detected periodic tasks have stopped'
                                ' running.'))
                    last_print = CANARY_RUN
                    dump_thread_stacks()
        except Exception as ex:
            LOG.exception(six.text_type(ex))
        finally:
            last_check = time.time()

# Spawn a thread to be the periodic task watchdog
greenthread.spawn(periodic_task_watchdog)


#######################################################
###########  Monkey Patching                ###########
#######################################################
#
# Monkey patch file locking because it does spin locks
# and gives no indication of troubles.  The result is
# that periodic tasks can hang, if NFS lockd continues
# to return EAGAIN.  Usually restarting lockd helps.
class _Monkey_FcntlLock(lockutils._FcntlLock):

    def __init__(self, name):
        super(_Monkey_FcntlLock, self).__init__(name)
        self._max_retries = 4096
        self._max_time = 3600
        self._msg_interval = 2048

    def trylock(self):
        tries = 0
        start_time = time.time()
        err_msg = 'Unable to obtain lock for file: %s'
        while True:
            try:
                fcntl.lockf(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
                return self
            except IOError as e:
                tries += 1
                if e.errno in (errno.EACCES, errno.EAGAIN):
                    # Check if we've been trying for quite a while.  Use
                    # both time and tries, in order to avoid time change
                    # issues and yet give it enough tries.
                    if (time.time() - start_time > self._max_time and
                            tries > self._max_retries):
                        LOG.error(_(err_msg) % self.fname)
                        raise exception.NovaException(message=(_(err_msg)
                                                               % self.fname))
                    # If we think we might have an issue, send a warning
                    if tries % self._msg_interval == 0:
                        LOG.warn(_(err_msg) % self.fname)
                    time.sleep(0.05)
                else:
                    raise


# Inject our own overwritten lock function to avoid spinning forever
setattr(lockutils, 'InterProcessLock', _Monkey_FcntlLock)
