# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Overrides the Conductor API for additional PowerVC DB Access:
    1. Override compute_node_delete to not allow OS to erroneously delete
"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils

from nova import rpc
from nova import utils
from nova.conductor import api

from powervc_nova.conductor import manager
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCConductorAPI(api.API):
    """ Extends the base Conductor API class with PowerVC capabilities """

    def __init__(self):
        """Constructor for the PowerVC extension to the Conductor API"""
        super(PowerVCConductorAPI, self).__init__()
        if cfg.CONF.conductor.use_local:
            LOG.info(self.__class__.__name__ + ": Using local access to "
                     "PowerVCConductorManager")
            self._manager = utils.ExceptionHelper(
                manager.PowerVCConductorManager())

    def compute_node_create(self, context, values):
        """Override Compute Node create to send additional notifications"""
        info = dict(host=values['host'], service_id=values['service_id'])
        info['hypervisor_hostname'] = values['hypervisor_hostname']
        info['host_display_name'] = encodeutils.safe_decode(
            cfg.CONF.host_display_name, 'utf-8')
        # Send a notification that we are creating the Compute Node
        notifier = rpc.get_notifier(service='compute', host=values['host'])
        notifier.info(context, 'compute.node.create.start', info)
        # Call the super class to actually create the Compute Node
        node = super(PowerVCConductorAPI,
                     self).compute_node_create(context, values)
        info['compute_node_id'] = node['id']
        # Send a notification that we finished creating the Compute Node
        notifier.info(context, 'compute.node.create.end', info)
        return node

    def compute_node_delete(self, context, node):
        """Override Compute Node delete to not delete on Hostname change"""
        # Since OpenStack will try to remove Compute Nodes when the Hostname
        # changes and we can't override the code that does this change and
        # the delete call, we will override the delete call to not let them
        # delete it.  We have went with an approach to just update the
        # hypervisor_hostname property on the existing Compute Node instead.
        LOG.info(_("Ignoring Compute Node delete request on hostname change."))
