#
#
# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
#
#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Overrides the Conductor Manager for additional PowerVC DB Access:
    1. Override instance_update to break out the PowerSpecs
    2. Add-in notify_unmanage to allow cleanup of KVM Host when switching
"""

import sys
import copy
import itertools
import nova.context
import oslo_messaging as messaging
from oslo_config import cfg
from nova import exception
from nova import notifications
from nova import objects
from nova import rpc as rpc_api
from nova.compute import vm_states
from nova.conductor import manager
from nova.conductor.tasks import live_migrate
from oslo_serialization import jsonutils
from oslo_log import log as logging
from nova.objects import service as service_obj
from nova.openstack.common import periodic_task
from nova.scheduler import utils as scheduler_utils

from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.compute import populate_admin_context
from powervc_nova.compute import api as compute_api
from powervc_nova.db import api as db_api
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.objects.storage import dom as storage_dom
from powervc_nova import _

LOG = logging.getLogger(__name__)
HOST_REG = 'powervc_discovery.registration.compute.host_registrar'


class PowerVCConductorManager(manager.ConductorManager):
    """Extends the base Conductor Manager class with PowerVC capabilities"""

    def __init__(self, *args, **kwargs):
        """Constructor for the PowerVC extension to the Conductor Manager"""
        super(PowerVCConductorManager, self).__init__(*args, **kwargs)
        self.additional_endpoints.append(_PowerVCConductorManagerProxy(self))
        # Construct a Default Factory from each Module to allow initialization
        self.hmcfac = compute_dom.ManagementConsoleFactory.get_factory()
        self.seafac = network_dom.SharedEthernetAdapterFactory.get_factory()
        self.scgfac = storage_dom.StorageConnectivityGroupFactory.get_factory()

    #######################################################
    ##########  Override Initialization Methods  ##########
    #######################################################
    def pre_start_hook(self, **kwargs):
        """Override the Pre-Start Hook to perform some initialization"""
        # Call the Parent preStart hook if it ever implements anything
        super(PowerVCConductorManager, self).pre_start_hook(**kwargs)
        try:
            # Make sure we reconcile the Aggregates right away on startup
            self.reconcile_aggregates(nova.context.get_admin_context())
            # We need to create the default SCG the first time Conductor starts
            self.scgfac.create_default_fc_scg(nova.context.get_admin_context())
            # Force the DB API to initialize by doing a random query for HMC's
            self.hmcfac.find_all_hmcs(nova.context.get_admin_context())
        except:
            pass  # We don't care if it fails, since it is just initialization

    @periodic_task.periodic_task(
        spacing=cfg.CONF.reconcile_host_aggregates_interval)
    def reconcile_aggregates(self, context):
        try:
            agg_utils.reconcile_aggregates(
                cfg.CONF.default_host_aggregate_name)
        except Exception as exc:
            LOG.warn(_("Error reconciling host aggregates: %s") % exc)
            LOG.exception(exc)

    ###################################################
    ###### PowerSpec Conductor API Implementation #####
    ###################################################
    def instance_update(self, context, instance_uuid,
                        updates, service):
        """ Update the attributes for the Instance given in the Database """
        # STG Defect 29618 - monkey patch info_from_instance() method
        def patched_info_from_instance(context, instance_ref, network_info,
                                       system_metadata, **kw):
            instance_info = notifications.orig_info_from_instance(
                context, instance_ref, network_info, system_metadata)
            try:
                updated_at = instance_ref.get('updated_at', None)
                if updated_at:
                    instance_info['updated_at'] = updated_at
            except Exception:
                pass
            return instance_info
        if not hasattr(notifications, 'orig_info_from_instance'):
            notifications.orig_info_from_instance = \
                notifications.info_from_instance
            notifications.info_from_instance = patched_info_from_instance

        updates = copy.deepcopy(updates)
        pwr_specs = updates.pop('power_specs', None)
        # Call the Parent's method to actually update the Instance in the DB
        instance_ref = super(PowerVCConductorManager, self).\
            instance_update(context, instance_uuid, updates, service)
        # If there were any Power Specs given, update those in the database
        if pwr_specs is not None:
            pwr_specs = db_api.\
                instance_power_specs_update(context, instance_uuid, pwr_specs)
            instance_ref['power_specs'] = self._convert_power_specs(pwr_specs)
        return instance_ref

    ####################################################
    ##########  UnManage Host Implementation  ##########
    ####################################################
    def notify_unmanage(self, context, host, mgmt_ip):
        """Notifies this Management System to remove Host Management"""
        info = dict(hostname=host, ip=mgmt_ip)
        try:
            # Log a message for debug purposes that were are removing the Host
            LOG.info(_("Removing Host %(hostname)s, switching "
                       "to Management System %(ip)s...") % info)
            # Generate a Context with a Token to use for the Requests
            context = populate_admin_context()
            # If the Compute Node doesn't exist, we don't need to notify
            comp_node = self._get_compute_node(context, host)
            if comp_node is None:
                return
            # Notify the old Management System is no longer managing the host
            text = _("The PowerVC management system at %(ip)s is taking over "
                     "management of host %(hostname)s.  The host will be "
                     "removed from this management system.") % info
            anotifier = rpc_api.get_notifier('compute', host)
            anotifier.info(context, 'compute.instance.log', {'msg': text})
            try:
                __import__(HOST_REG)
                # Call the Host Registrar to do the full clean-up of the Host
                get_registrar = getattr(sys.modules[HOST_REG], 'get_registrar')
                registrar = get_registrar(context, host_name=host)
                registrar.skip_remote_commands = True
                registrar.deregister(force=True)
            except Exception as ex:
                LOG.warn(_("Exception trying to fully remove the Host."))
                LOG.exception(ex)
                # Send a notification that we are removing the Compute Node
                anotifier.info(context, 'compute.node.delete.start', comp_node)
                # Fall-back to just cleaning the DB, if the main flow failed
                hostfact = compute_dom.ManagedHostFactory.get_factory()
                hostfact.delete_host(context, host)
                # Send a notification that we removed the Compute Node
                anotifier.info(context, 'compute.node.delete.end', comp_node)
            # Log a message for debug purposes that we removed the Host
            LOG.info(_("Removed Host %(hostname)s, switching "
                       "to Management System %(ip)s.") % info)
        except Exception as exc:
            # Log the Exception that occurred while trying to Remove the Host
            LOG.warn(_("Failed to remove Host %(hostname)s while "
                       "switching to Management System %(ip)s") % info)
            LOG.exception(exc)

    ####################################################
    ########  Internal Conductor Helper Methods  #######
    ####################################################
    @staticmethod
    def _get_compute_node(context, host):
        """Helper method to query the Compute Node from the DB"""
        service = service_obj.Service.get_by_compute_host(context, host)
        # If we weren't able to find the Server or Compute Node, just return
        if service is None or service['compute_node'] is None:
            return None
        # Return the key info from the Compute Node as a dictionary
        compute_node = service['compute_node']
        return dict(compute_node_id=compute_node.id,
                    host=service.host, service_id=compute_node.service_id,
                    hypervisor_hostname=compute_node.hypervisor_hostname)

    @staticmethod
    def _convert_power_specs(power_specs):
        """Internal Helper Method to Convert PowerSpecs to a Dictionary"""
        lst = ['created_at', 'updated_at', 'deleted_at', 'deleted', 'instance']
        power_specs = jsonutils.to_primitive(power_specs)
        # There are certain properties on the PowerSpecs we don't want returned
        for attr in lst:
            if attr in power_specs:
                power_specs.pop(attr, None)
        return power_specs

""""
Monkey patch validate_filter to always return true if called with Nova Affinity
filters. PowerVC uses its own affinity filters, but still needs
validate_filter to return true during scheduler utils setup_instance_group.
"""


def patch_validate_filter(filter_name):
    if (filter_name == 'ServerGroupAffinityFilter' or
            filter_name == 'ServerGroupAntiAffinityFilter'):
        return True
    else:
        return filter in cfg.CONF.scheduler_default_filters

setattr(scheduler_utils, 'validate_filter', patch_validate_filter)


###################################################
#### Conductor Manager Adapter Extension Class ####
###################################################
class _PowerVCConductorManagerProxy(object):
    """Adapter Class to extend Remote Methods for the Conductor"""

    target = messaging.Target(version='2.0')

    def __init__(self, manager):
        """Construct a Class to extend Remote Methods for the Conductor"""
        self.manager = manager

    def notify_unmanage(self, context, host, mgmt_ip):
        """Notifies this Management System to remove Host Management"""
        return self.manager.notify_unmanage(context, host, mgmt_ip)


###################################################
#########   Override Live Migration Task   ########
###################################################
"""
Since the default RPC timeout in OpenStack is way too low in many cases for
migrations, PowerVC is extending a few of the RPC API call methods to set a
larger timeout (based on a configuration property). This works in most places
in the Compute Manager since we specify the PowerVC RPC API there, but for
the check_can_live_migrate_destination call in the live_migration module this
directly constructs its own RPC API, so it doesn't pick up the override.

We need to have a better solution for 2Q14 (larger default, OSCE increasing
timeout, etc), but temporarily for 1.2 FP1 are putting in a change to inject
our own extended LiveMigrationTask class that only overrides the RPC API
used. This will inject our own execute method that just constructs this Task.
"""


class _Extended_LiveMigrationTask(live_migrate.LiveMigrationTask):
    """Extends the default Live Migration Task to override the RPC API used"""

    def __init__(self, context, instance, destination,
                 block_migration, disk_over_commit):
        """Constructor for the PowerVC extension to Live Migration Task"""
        super(_Extended_LiveMigrationTask, self).\
            __init__(context, instance, destination,
                     block_migration, disk_over_commit)
        # Use our own extended RPC API instead of the default
        self.compute_rpcapi = compute_api.PowerVCComputeRPCAPI()


def _extended_live_migrate_execute(context, instance, destination,
                                   block_migration, disk_over_commit):
    """Overrides the default live_migrate execute to use our extended Task"""
    task = _Extended_LiveMigrationTask(context, instance, destination,
                                       block_migration, disk_over_commit)
    return task.execute()


# Inject our own overwritten live migrate execute function instead
setattr(live_migrate, 'execute', _extended_live_migrate_execute)


"""
Monkey patch ComputeTaskManager's build_instances method to pass the
requested networks to the scheduler's select_destinations method via
the filter_properties (this was introduced when OSCE changed the
scheduler to have a less "hands on" role in the deployment process
in Juno.

TODO(jwcroppe): Remove this monkey patch once OSCE defect
https://bugs.launchpad.net/nova/+bug/1363568 is fixed; undo the
RTC changeset associated with defect RTC 29891.
"""


def _extended_build_instances(self, context, instances, image,
                              filter_properties, admin_password,
                              injected_files, requested_networks,
                              security_groups, block_device_mapping=None,
                              legacy_bdm=True):
    # TODO(ndipanov): Remove block_device_mapping and legacy_bdm in version
    #                 2.0 of the RPC API.
    request_spec = scheduler_utils.build_request_spec(context, image,
                                                      instances)
    # TODO(danms): Remove this in version 2.0 of the RPC API
    if (requested_networks and
            not isinstance(requested_networks,
                           objects.NetworkRequestList)):
        requested_networks = objects.NetworkRequestList(
            objects=[objects.NetworkRequest.from_tuple(t)
                     for t in requested_networks])
    # NOTE(jwcroppe): This one line below is what the above monkey patch
    # is needed for.
    filter_properties['requested_networks'] = requested_networks
    # TODO(melwitt): Remove this in version 2.0 of the RPC API
    flavor = filter_properties.get('instance_type')
    if flavor and not isinstance(flavor, objects.Flavor):
        # Code downstream may expect extra_specs to be populated since it
        # is receiving an object, so lookup the flavor to ensure this.
        flavor = objects.Flavor.get_by_id(context, flavor['id'])
        filter_properties = dict(filter_properties, instance_type=flavor)

    try:
        scheduler_utils.setup_instance_group(context, request_spec,
                                             filter_properties)
        # check retry policy. Rather ugly use of instances[0]...
        # but if we've exceeded max retries... then we really only
        # have a single instance.
        scheduler_utils.populate_retry(filter_properties, instances[0].uuid)
        hosts = self.scheduler_client.select_destinations(
            context, request_spec, filter_properties)
    except Exception as exc:
        updates = {'vm_state': vm_states.ERROR, 'task_state': None}
        for instance in instances:
            self._set_vm_state_and_notify(
                context, instance.uuid, 'build_instances', updates,
                exc, request_spec)
        return

    for (instance, host) in itertools.izip(instances, hosts):
        try:
            instance.refresh()
        except (exception.InstanceNotFound,
                exception.InstanceInfoCacheNotFound):
            LOG.debug('Instance deleted during build', instance=instance)
            continue
        local_filter_props = copy.deepcopy(filter_properties)
        scheduler_utils.populate_filter_properties(local_filter_props, host)
        # The block_device_mapping passed from the api doesn't contain
        # instance specific information
        bdms = objects.BlockDeviceMappingList.get_by_instance_uuid(
            context, instance.uuid)

        self.compute_rpcapi.build_and_run_instance(
            context, instance=instance, host=host['host'],
            image=image,
            request_spec=request_spec,
            filter_properties=local_filter_props,
            admin_password=admin_password,
            injected_files=injected_files,
            requested_networks=requested_networks,
            security_groups=security_groups,
            block_device_mapping=bdms, node=host['nodename'],
            limits=host['limits'])


setattr(manager.ComputeTaskManager,
        'build_instances',
        _extended_build_instances)
