# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import uuid
from datetime import datetime
from sqlalchemy import MetaData, Table
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, V1232

NOVA_CONF = '/etc/nova/nova.conf'
PASTE_CONF = '/etc/nova/api-paste.ini'
CA_CERTS = '/etc/pki/tls/certs/powervc.crt'
FLAVOR_JSON = '/etc/nova/flavors-config.json'
POWERVC_MANAGE = '/opt/ibm/powervc/bin/powervc-manage'


######################################################
#########  Nova Data Manager Definition  #############
######################################################
class NovaDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'nova'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return True

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db2', 'rabbitmq', 'httpd',
                'keystone', 'cinder', 'glance', 'neutron']

    ###############################################
    ####  Data Manager - Configure Interface  #####
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(NovaDataManager, self).configure_component()
        #Do the KVM Host-specific Nova Configuration
        if not utils.is_management_system():
            config = utils.get_config_data()
            #For the Ephemeral Disk Path we need to also create the directory
            self.__create_ephemeral_disk_path(config)
            #Configure libvirt to be able to work properly for migrations
            self.__configure_libvirt_for_migration(config)
        #On the Management System we need to insert the PowerVM Flavors
        elif utils.get_offering() == 'powervm':
            self.__create_powervm_flavors()
        #Change the permissions for the Nova configuration files
        self.__update_nova_conf_permissions()

    ###############################################
    #####  Data Manager - Upgrade Interface  ######
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        #First we need to wipe out all of the extraneous notification_drivers
        if utils.compare_version(version, V123) < 0:
            files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
            for afile in (files + ['nova.conf']):
                utils.run_command("/bin/sed -i -e '/notification_"
                                  "driver = /d' /etc/nova/%s" % afile)
        #Now call the Super-class to Update the Configuration Files
        super(NovaDataManager, self).upgrade_conf_files(version)
        #Although not related to the CONF files, we need to delete
        #the duplicate Compute Nodes before the Nova DBSync is run
        if utils.is_management_system():
            self.__cleanup_deleted_compute_nodes()

    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        #If coming from before 1.2.3, need to insert the new flavors
        if utils.compare_version(version, V123) < 0:
            if utils.get_offering() == 'powervm':
                self.__create_powervm_flavors()
        #Change the permissions for the Nova configuration files
        self.__update_nova_conf_permissions()

    ###############################################
    ######  Data Manager - Backup Interface  ######
    ###############################################
    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config = utils.get_config_data()
        #Generate a unique time-stamp to put in the existing nova.conf files
        timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
        config['nova_restore_timestamp'] = timestamp
        #Loop through each CONF file adding a time-stamp to determine changes
        for host in self.__get_compute_hosts():
            conf_file = '/etc/nova/nova-%s.conf' % host
            utils.set_openstack_config_value(
                conf_file, 'restore_timestamp', timestamp)

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""
        config = utils.get_config_data()
        #Loop through each CONF file seeing if the host is new or old
        for host in self.__get_compute_hosts():
            conf_file = '/etc/nova/nova-%s.conf' % host
            val = utils.get_openstack_config_value(
                conf_file, 'restore_timestamp')
            #If file came from the restore, we need to verify service exists
            if val is None or val != config['nova_restore_timestamp']:
                self.__update_new_host(host)
            #If file is left from before the restore, need to delete it
            else:
                self.__remove_old_host(host)

    ###############################################
    ##  Data Manager - Subclass Implementation  ###
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'nova'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        #If this is the Management System, we need different services
        if utils.is_management_system():
            services = ['api', 'conductor', 'scheduler']
            services.extend(self.__get_compute_services())
            return ['openstack-nova-%s' % svc for svc in services]
        return ['openstack-nova-compute']

    def _get_disabled_services(self):
        """Provides a list of the Disabled Services for the Component"""
        #We want to disable a bunch of services if this is the Mgmt System
        services = ['cells', 'cert', 'compute', 'console', 'consoleauth',
                    'metadata-api', 'objectstore', 'network', 'xvpvncproxy']
        return ['openstack-nova-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['nova-api', 'nova-conductor', 'nova-scheduler',
                'nova-manage', 'httpd-error', 'powervc-dbsync']
        return ['/var/log/nova/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('NOVA', 'nova')

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/nova-manage db sync',
                '/opt/ibm/powervc/bin/powervc-dbsync nova']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {NOVA_CONF: dict()}
        #If this is the Mgmt System, also need to update the Host Conf Files
        if utils.is_management_system():
            for host in self.__get_compute_hosts():
                config['/etc/nova/nova-%s.conf' % host] = dict()
        #Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._added_auth_conf_properties(config[key])
            self._added_messaging_conf_properties(config[key])
            self.__added_nova_common_conf_properties(config[key])
            #Add the Specific Properties for Mgmt System vs KVM Host
            if utils.is_management_system():
                self.__added_nova_mgmtsystem_conf_properties(config[key])
            else:
                self.__added_nova_kvmhost_conf_properties(config[key])
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {NOVA_CONF: dict()}
        #If this is the Mgmt System, also need to update the Host Conf Files
        if utils.is_management_system():
            for host in self.__get_compute_hosts():
                config['/etc/nova/nova-%s.conf' % host] = dict()
        #Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._deleted_auth_conf_properties(config[key])
            self._deleted_messaging_conf_properties(config[key])
            self.__deleted_nova_conf_properties(config[key])
        return config

    def _get_audit_conf_properties(self):
        """Provides a tuple of configuration properties for auditing"""
        sects = ['composite:openstack_compute_api_v2',
                 'composite:openstack_compute_api_v21',
                 'composite:openstack_compute_api_v3']
        return (PASTE_CONF, sects, ['keystone', 'keystone_nolimit'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/nova']

    def _get_certificate_authority_info(self):
        """Provides the info to create the Component's CA Certificate"""
        ca_info = dict(component='libvirt', altnamexts=[],
                       extensions=['ca', 'client', 'server'])
        ca_info['privkey'] = '/etc/pki/libvirt/private/ca.key'
        ca_info['cacert'] = '/etc/pki/libvirt/certs/ca.crt'
        return ca_info if utils.get_offering() != 'powervm' else None

    def _get_kvmhost_certificate_info(self):
        """Provides the info to create the KVM Host Signed Certificates"""
        cacert = ('/etc/pki/libvirt/certs/ca.crt', '/etc/pki/CA/cacert.pem')
        ccsr = dict(certfile='clientcert.pem',
                    keyfile='clientkey.pem', exttype='client')
        scsr = dict(certfile='servercert.pem',
                    keyfile='serverkey.pem', exttype='server')
        return dict(cacert=cacert, certdir='/etc/pki/libvirt',
                    group='root', csrs=[ccsr, scsr])

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('nova', NOVA_CONF, 'keystone_authtoken', 'admin_password')]

    def _temporary_openstack_patches(self):
        """Perform any temporary openstack patching for the Component"""
        #Temporarily patch the nova.objects.base class for the
        #issue with converting Message objects to str-type strings
        objbefore = 'str(value_arg)'
        objafter = 'unicode(value_arg) if isinstance('
        objafter += 'value_arg, Exception) else str(value_arg)'
        objfile = '/usr/lib/python2.7/site-packages/nova/objects/base.py'
        utils.sed_replace(objfile, objbefore, objafter)

    ###############################################
    #### Internal Data Manager Helper Methods  ####
    ###############################################
    def __get_compute_hosts(self):
        """Helper method to get a list of registered compute hosts"""
        files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
        return [f[5:-5] for f in files]

    def __get_compute_services(self):
        """Helper method to get a list of nova-compute-xxx services"""
        files1 = utils.get_matching_files(
            '/etc/init.d', 'openstack-nova-compute-', '')
        files2 = utils.get_matching_files(
            '/usr/lib/systemd/system', 'openstack-nova-compute-', '')
        if files1:
            return [f[15:] for f in files1]
        return [f[15:-8] for f in files2]

    def __added_nova_common_conf_properties(self, config):
        """Add in the Common nova.conf Configuration Properties"""
        drivers = ['ceilometer.compute.nova_notifier',
                   'messaging', 'nova.ibm.ego.notifier.rpc_notifier']
        for sect in ['cinder', 'glance', 'neutron']:
            config.setdefault(sect, dict())
        #Add in the properties for the Cinder Section
        config['cinder'].update(dict(
            timeout=('1800', V123),
            catalog_info=('volume:cinder:publicURL', V123)))
        #Add in the properties for the Glance Section
        uri = '%(local_ip)s' if utils.is_management_system() else '%(mgmt_ip)s'
        config['glance'].update(dict(
            host=(uri, V122), port=('9292', V122)))
        #Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            admin_auth_url=('https://%s:5000/v2.0/' % uri, V123),
            timeout=('60', V123), admin_tenant_name=('service', V122),
            admin_username=('neutron', V122), auth_strategy=('keystone', V122),
            admin_password=('aes:%(neutron_password)s', V123)))
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            notification_driver=(drivers, V123),
            resize_to_same_host=('require', V120),
            instance_build_timeout=('7200', V120),
            default_log_levels=('urllib3=WARN', V120),
            running_deleted_instance_action=('log', V123),
            heal_instance_info_cache_interval=('180', V123),
            ibm_notification_topics=('ibm_notifications', V120),
            use_ipv6=(str(utils.get_config_data()['ip_version'] == '6'), V122),
            flat_injected=('True', V121), instance_delete_interval=('0', V123),
            service_down_time=('120', V122), resize_confirm_window=('1', V120),
            report_interval=('30', V122), management_ip=('%(mgmt_ip)s', V120),
            instance_name_template=(
                '<%(display_name).13s-%(uuid).8s-%(id)08x>', V120),
            compute_manager=(
                'powervc_nova.compute.manager.PowerVCComputeManager', V120),
            injected_network_template=(
                '<$pybasedir/powervc_nova/virt/interfaces.template>', V121)))

    def __added_nova_mgmtsystem_conf_properties(self, config):
        """Add in the Management System nova.conf Configuration Properties"""
        for sect in ['cinder', 'glance', 'neutron', 'database', 'conductor']:
            config.setdefault(sect, dict())
        config.setdefault('ibm_prs_ceilometer', dict())
        #Add in the properties for the Cinder/DB/Conductor Sections
        config['cinder']['insecure'] = ('True', V123)
        config['database'].update(dict(
            pool_timeout=('60', V123),
            max_overflow=('100', V123), max_pool_size=('30', V123),
            connection=('aes:ibm_db_sa://nova:%(nova_'
                        'password)s@%(local_ip)s:50110/nova', V123)))
        config['conductor'].update(dict(
            use_local=('True', V120), workers=('4', V121),
            manager=('powervc_nova.conductor.'
                     'manager.PowerVCConductorManager', V120)))
        #Add in the properties for the Glance Section
        config['glance'].update(dict(
            protocol=('http', V122), api_insecure=('True', V122),
            api_servers=('http://%(local_ip)s:9292', V122)))
        #Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            insecure=('True', V123),
            url=('http://%(local_ip)s:9696', V123)))
        #Add in the properties for the PRS Ceilometer Section
        config['ibm_prs_ceilometer'].update(dict(
            user_domain_name=('Service', V123),
            password=('aes:%(ceilometer_password)s', V123),
            auth_url=('https://%(local_ip)s:5000/v2.0/', V123),
            project_name=('service', V123), username=('ceilometer', V123),
            project_domain_name=('Service', V123), insecure=('True', V123)))
        #Add in the properties for the Oslo Currency Section
        config['oslo_concurrency'] = \
            dict(lock_path=('/var/lib/nova/tmp', V123))
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            osapi_compute_listen_port=('8774', V121),
            quota_server_group_members=('100', V123),
            install_ip_option=('%(ip_option)s', V122),
            install_ip_version=('%(ip_version)s', V122),
            osapi_compute_listen=('%(local_ip)s', V121),
            api_paste_config=('/etc/nova/api-paste.ini', V120),
            osapi_glance_link_prefix=('https://%(mgmt_ip)s:9292', V121),
            osapi_compute_link_prefix=('https://%(mgmt_ip)s:8774', V121),
            quota_volumes=('8192', V123),  quota_instances=('3500', V123),
            metadata_workers=('4', V121), osapi_compute_workers=('4', V121),
            quota_ram=('24000000', V123), quota_gigabytes=('3000000', V123),
            quota_cores=('35000', V123), enabled_apis=('osapi_compute', V122),
            osapi_max_limit=('4000', V123), san_disk_prefix=('powervc', V120),
            osapi_compute_extension=(
                'powervc_nova.api.plugins.api_extensions_v2', V121),
            volume_api_class=(
                'powervc_nova.volume.cinder.PowerVCVolumeAPI', V122),
            non_inheritable_image_properties=(
                'cache_in_nova, bittorrent, block_device_mapping', V123)))
        #Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_nova_powervm_conf_properties(config)
        else:
            self.__added_nova_powerkvm_conf_properties(config)

    def __added_nova_kvmhost_conf_properties(self, config):
        """Add in the KVM Host nova.conf Configuration Properties"""
        for sect in ['cinder', 'glance', 'neutron', 'libvirt', 'ssl']:
            config.setdefault(sect, dict())
        #Add in the properties for the SSL/Cinder Section
        config['ssl']['ca_file'] = (CA_CERTS, V122)
        config['cinder']['cafile'] = (CA_CERTS, V123)
        #Add in the properties for the Glance Section
        config['glance'].update(dict(
            protocol=('https', V122),
            api_servers=('https://%(mgmt_ip)s:9292', V122)))
        #Add in the properties for the Neutron Section
        config['neutron'].update(dict(
            ca_certificates_file=(CA_CERTS, V123),
            url=('https://%(mgmt_ip)s:9696/', V122)))
        #Add in the properties for the Libvirt Section
        config['libvirt'].update(dict(
            virt_type=('kvm', V121), libvirt_type=('kvm', V121),
            cpu_mode=('none', V121), inject_partition=('-2', V121),
            live_migration_uri=('<qemu+tls://%s/system>', V121),
            volume_drivers=('iscsi=powervc_nova.virt.powerkvm.'
                            'volume.PowerKVMSCSIVolumeDriver', V122),
            live_migration_flag=('VIR_MIGRATE_UNDEFINE_SOURCE,VIR_MIGRATE_'
                                 'PEER2PEER,VIR_MIGRATE_TUNNELLED', V121)))
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            integration_bridge=('br-int', V121),
            dhcp_domain=('%(dhcp_domain)s', V121),
            instances_path=('%(instances_path)s', V121),
            state_path=('/var/lib/libvirt/images/nova', V121),
            host_display_name=('%(host_display_name)s', V121),
            compute_driver=(
                'powervc_nova.virt.powerkvm.driver.PowerKVMDriver', V121),
            network_api_class=('powervc_nova.network.neutronv2.'
                               'powerkvm_api.PowerKVMAPI', V121),
            compute_monitors=('ComputeDriverCPUMonitor', V122),
            host_type=('kvm', V121), enable_kvm_ovs_auto_config=('True', V121),
            vnc_enabled=('False', V121), ca_certificates_file=(CA_CERTS, V121),
            firewall_driver=('nova.virt.firewall.NoopFirewallDriver', V122),
            network_allocate_retries=('5', V123)))

    def __added_nova_powervm_conf_properties(self, config):
        """Add in the PowerVM-specific nova.conf Configuration Properties"""
        netclass = 'powervc_nova.network.neutronv2.ibmpowervm_api.PowerVMAPI'
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            compute_monitors=('PowerVMDriverCPUMonitor', V1232),
            compute_available_monitors=(
                'powervc_nova.compute.monitors.virt.powervm_'
                'cpu_monitor.PowerVMDriverCPUMonitor', V1232),
            network_api_class=(netclass, V121),
            force_config_drive=('always', V121),
            send_rmc_busy_message=('True', V122),
            supported_host_types=('powervm', V122),
            block_device_allocate_retries=('300', V122),
            automated_powervm_vlan_cleanup=('True', V122),
            powervm_img_remote_path=('/home/padmin', V120),
            block_device_allocate_retries_interval=('6', V122),
            powervm_img_local_path=('/var/lib/nova/images', V120),
            powervm_mgr_type=('hmc', V122),  host_type=('powervm', V120)))

    def __added_nova_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific nova.conf Configuration Properties"""
        netclass = 'powervc_nova.network.neutronv2.powerkvm_api.PowerKVMAPI'
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            network_api_class=(netclass, V121),
            enable_kvm_ovs_auto_config=('True', V122),
            host_type=('kvm', V121), supported_host_types=('kvm', V122)))

    def __deleted_nova_conf_properties(self, config):
        """Remove the Common nova.conf Configuration Properties"""
        #Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            sql_connection=V123, lock_path=V123,
            neutron_admin_password=V123, neutron_url=V122,
            cinder_api_insecure=V123, cinder_http_timeout=V123,
            glance_api_servers=V122, glance_api_insecure=V122,
            glance_host=V122, glance_port=V122, glance_protocol=V122,
            neutron_admin_username=V122, neutron_api_insecure=V122,
            neutron_admin_auth_url=V122, neutron_auth_strategy=V122,
            neutron_admin_tenant_name=V122, cinder_ca_certificates_file=V123,
            backdoor_port=V123))

    def __cleanup_deleted_compute_nodes(self):
        """Helper method to remove the deleted Compute Node entries"""
        #Since for some inexplicable reason, OpenStack decided to put a
        #Unique Constraint on the Compute Nodes DB Table that didn't include
        #the deleted flag, we need to remove the deleted entries to avoid
        #having a unique constraint error occurring when the DBSync is run
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        table = Table('compute_nodes', metadata, autoload=True)
        table.delete().where(table.c.deleted != 0).execute()

    def __update_new_host(self, host):
        """Helper method to make sure the service exists for the host"""
        service = 'openstack-nova-compute-%s' % host
        conf_file = '/etc/nova/nova-%s.conf' % host
        #We only want to create the service if this is a PowerVM Mgmt System
        if utils.get_offering() != 'powerkvm':
            if not self.__service_file_exists(host):
                utils.move_file(conf_file, conf_file + '_backup')
                utils.run_command(POWERVC_MANAGE + ' -o addhost -s ' +
                                  service + ' -f ' + conf_file + '_backup')
                utils.remove_file(conf_file + '_backup')

    def __remove_old_host(self, host):
        """Helper method to make sure the service is removed for old host"""
        service = 'openstack-nova-compute-%s' % host
        conf_file = '/etc/nova/nova-%s.conf' % host
        #We only want to delete the service if this is a PowerVM Mgmt System
        if utils.get_offering() != 'powerkvm':
            if self.__service_file_exists(host):
                utils.run_command(
                    POWERVC_MANAGE + ' -o removehost -s ' + service)
        #Make sure the CONF file is gone even it is a PowerKVM Mgmt System
        utils.remove_file(conf_file)

    def __service_file_exists(self, host):
        """Helper method to see if the service file exists for the host"""
        svc = 'openstack-nova-compute-%s' % host
        svc1 = '/etc/rc.d/init.d/%s' % svc
        svc2 = '/usr/lib/systemd/system/%s.service' % svc
        return utils.file_exists(svc1) or utils.file_exists(svc2)

    def __update_nova_conf_permissions(self):
        """Helper method to update the permissions for the nova conf files"""
        if utils.is_management_system():
            utils.chown('/etc/nova', 'nova', 'nova')
            utils.chown('/etc/nova/nova.conf', 'nova', 'nova')

    def __configure_libvirt_for_migration(self, config):
        """Helper method to configure libvirt to work with migration"""
        #Open up the libvirt-tls service through firewalld
        utils.run_command('firewall-cmd --permanent --add-service=libvirt-tls')
        utils.run_command('firewall-cmd --reload')
        #If a NFS Ephemeral Disk Path, we need to open up SELinux for it
        if config.get('host_storage_type') == 'nfs':
            utils.run_command('/sbin/setsebool -P virt_use_nfs 1')
        #Update the libvirtd config file to listen for incoming connections
        utils.run_command('/bin/sed -i \'/#LIBVIRTD_ARGS="--listen"/c\\'
                          'LIBVIRTD_ARGS="--listen"\' /etc/sysconfig/libvirtd')
        #Restart the libvirtd service when we are done to take affect
        utils.run_command('/usr/sbin/service libvirtd restart')

    def __create_ephemeral_disk_path(self, config):
        """Helper method create the State/Instances Path if doesn't exist"""
        state_path = '/var/lib/libvirt/images/nova'
        subdirs = ['buckets', 'images', 'instances', 'keys', 'networks', 'tmp']
        self.__prep_nova_path_directory(state_path, subdirs)
        #If the Instances Path is provided, make sure nova has access to them
        if config.get('instances_path'):
            self.__prep_nova_path_directory(config['instances_path'], [''])

    def __prep_nova_path_directory(self, adir, subdirs):
        """Helper method to make sure nova has access to a directory"""
        for subdir in ([''] + subdirs):
            full_path = adir + '/' + subdir
            #If the Path doesn't exist yet, then we need to create it
            if not utils.file_exists(full_path):
                su = '/usr/bin/su -s /bin/bash nova -c "%s"'
                utils.run_command(su % ('/bin/mkdir -p -m 755 ' + full_path))
            #As a fall-back, try creating as root if nova doesn't have access
            if not utils.file_exists(full_path):
                utils.makedirs(full_path)
            #Make sure that Nova is the Owner and has permission to it
            utils.chown(full_path, 'nova', 'nova')
            utils.chmod(full_path, 755)
        #We now want to make sure Nova has permission all the way up the chain
        self.__grant_search_to_parents(adir)

    def __grant_search_to_parents(self, adir):
        """Helper method to grant public excute to the directory and parents"""
        utils.run_command('/usr/bin/chmod o+x ' + adir)
        if adir.rfind('/') > 0:
            self.__grant_search_to_parents(adir[:adir.rfind('/')])

    def __create_powervm_flavors(self):
        """Helper method to create the Flavors specific to PowerVM"""
        flavor_data = json.loads(utils.read_from_file(FLAVOR_JSON))
        global_specs = flavor_data.pop('pvm_global_extra_specs')
        #Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        flv_table = Table('INSTANCE_TYPES', metadata, autoload=True)
        spc_table = Table('INSTANCE_TYPE_EXTRA_SPECS', metadata, autoload=True)
        flav_insert, spec_insert = flv_table.insert(), spc_table.insert()
        connection = engine.connect()
        #Loop through each of the flavor info, creating the database entries
        for flavor_key in flavor_data:
            #If this is the extra specs, it is just an auxiliary entry
            if flavor_key.endswith('_extra_specs'):
                continue
            #Get the entry and extra specs from the overall json dictionary
            flavor = flavor_data.get(flavor_key)
            flavor_specs = flavor_data.get(flavor_key + '_extra_specs')
            flavor_specs.update(global_specs)
            #Insert the Flavor in the instance_types database table
            rs = connection.execute(
                flav_insert, name=flavor['name'], created_at=datetime.utcnow(),
                deleted=0, disabled=0, is_public=1, memory_mb=flavor['ram'],
                root_gb=flavor['disk'], vcpus=flavor['vcpus'], swap=0,
                ephemeral_gb=0, rxtx_factor=1.0, flavorid=str(uuid.uuid4()))
            flavor_id = rs.inserted_primary_key[0]
            #Insert the Extra Specs into the instance_type_extra_specs table
            for key, val in flavor_specs.iteritems():
                spec_row = dict(instance_type_id=flavor_id, deleted=0, key=key,
                                value=val, created_at=datetime.utcnow())
                connection.execute(spec_insert, spec_row)
        connection.close()
