# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import re
from datetime import datetime
from sqlalchemy import MetaData, Table
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, V1232, V999

from powervc_oslo import _

NOVA_CONF = '/etc/nova/nova.conf'
RTPOLICY_FILE = '/etc/nova/rtpolicy_conf.xml'


######################################################
##########  EGO Data Manager Definition  #############
######################################################
class EgoDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'ego'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db2', 'rabbitmq', 'httpd', 'keystone', 'nova']

    ###############################################
    ##  Data Manager - Configure Implementation  ##
    ###############################################
    def configure_component(self):
        """Performs Configuration specific to the given Component"""
        super(EgoDataManager, self).configure_component()
        #Need to replace the RTPolicy file with the PowerVC one
        afile = '/etc/nova/rtpolicy_conf_%s.xml.pvc' % utils.get_offering()
        utils.copy_file(afile, '/etc/nova/rtpolicy_conf.xml')
        utils.remove_file(afile)
        #Need to remove the SetUID from the EGOSH File
        egosh_path = utils.run_command_output(
            '/usr/bin/find /opt/ibm/ego/ -name egosh', False)[1].strip()
        utils.run_command('/usr/bin/chmod -s ' + egosh_path)

    ###############################################
    #####  Data Manager - Upgrade Interface  ######
    ###############################################
    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        super(EgoDataManager, self).upgrade_component(version)
        #Extract the directory/name of the Backup Archive from the Global Data
        config = utils.get_config_data()
        tardir, tarfile = config.get('restore_dir'), config.get('restore_file')
        #If the Mgmt System and the Archive Info, extract the rtcpolicy.xml
        if utils.is_management_system() and tardir and tarfile:
            data = utils.extract_tar_file_content(
                tardir, tarfile, [RTPOLICY_FILE])
            #Upgrade the Aggregate Metadata if we are coming from 1.2.1/1.2.2
            if data and utils.compare_version(version, V123) < 0:
                self.__upgrade_aggregate_metadata(tardir, tarfile, data[0])

    ###############################################
    ##  Data Manager - System Changed Interface  ##
    ###############################################
    def management_ip_changed(self, upgrading):
        """Updates the Management IP in the CONF files when they change"""
        utils.set_openstack_config_value(
            NOVA_CONF, 'ego_master_list', utils.get_local_hostname())

    ###############################################
    ##  Data Manager - Subclass Implementation  ###
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'nova'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        services = ['ibm-notification', 'ibm-ego-ha-service',
                    'ibm-ego-resource-optimization']
        return ['ego'] + ['openstack-nova-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['nova-ibm-notification', 'nova-ibm-ego-ha-service',
                'nova-ibm-ego-resource-optimization']
        return ['/var/log/nova/%s.log' % log for log in logs]

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {NOVA_CONF: dict()}
        #Add in all of the Common EGO Configuration Properties
        self.__added_ego_common_conf_properties(config[NOVA_CONF])
        #Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_ego_powervm_conf_properties(config[NOVA_CONF])
        else:
            self.__added_ego_powerkvm_conf_properties(config[NOVA_CONF])
        return config

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/opt/ibm/ego/kernel/conf', '/opt/ibm/ego/kernel/work']

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return ['/etc/nova/rtpolicy_conf.xml',
                '/opt/ibm/ego/kernel/conf/ego.conf',
                '/opt/ibm/ego/kernel/conf/profile.ego',
                '/opt/ibm/ego/kernel/conf/ego.cluster.cluster1']

    ###############################################
    #### Internal Data Manager Helper Methods  ####
    ###############################################
    def __added_ego_common_conf_properties(self, config):
        """Add in the Management System nova.conf Configuration Properties"""
        config.setdefault('DEFAULT', dict())
        config.setdefault('ibm_prs_ceilometer', dict())
        #Find the Path to the EGO Shared Library on this System
        ego_path = utils.run_command_output(
            '/usr/bin/find /opt/ibm/ego/ -name "ego.so"', False)[1].strip()
        ego_path = ego_path[:ego_path.rfind('/')+1]
        hostname = utils.get_local_hostname()
        #Add in the Common EGO properties for the Default Section
        config['DEFAULT'].update(dict(
            ego_lib_path=(ego_path, V999),
            prs_ut_data_sample_num=('30', V123),
            ego_master_list=('%s' % hostname, V120),
            prs_ha_timeout_seconds=('3600', V123),
            policy_per_host_aggregate=('True', V123),
            prs_sync_aggregate_reserveon=('True', V123),
            scheduler_driver=('nova.scheduler.ibm.ego.ego_'
                              'scheduler.EGOScheduler', V120),
            scheduler_manager=('nova.scheduler.ibm.ego.ego_scheduler_'
                               'manager.SchedulerManager', V120),
            prs_mt_migration_preferred_order=('memory_mb.most', V122),
            ha_policy_instance_driver=('powervc_nova.prs.instance_driver.'
                                       'PowerVCHAInstanceDriver', V123),
            ram_allocation_ratio=('1.0', V120), ego_kd_port=('7870', V120),
            ha_policy_hypervisor_driver=('powervc_nova.prs.hypervisor_driver.'
                                         'PowerVCHAHypervisorDriver', V123),
            prs_driver=('powervc_nova.prs.prs_driver.PowerVCPRSDriver', V123)))
        #Add in the properties for the IBM PRS Ceilometer Section
        config['ibm_prs_ceilometer'].update(dict(
            auth_url=('https://%(local_ip)s:5000', V123),
            insecure=('True', V123),
            project_domain_name=('Service', V123),
            project_name=('service', V123),
            user_domain_name=('Service', V123),
            username=('ceilometer', V123),
            password=('aes:%(ceilometer_password)s', V123)))

    def __added_ego_powervm_conf_properties(self, config):
        """Add in the PowerVM-specific nova.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            cpu_allocation_ratio=('1.0', V120),
            reserved_host_memory_mb=('0', V120),
            scheduler_available_filters=(self.__get_scheduler_filters(),
                                         V1232),
            scheduler_default_filters=(
                'RamFilter,CoreFilter,ComputeFilter,RetryFilter,Availability'
                'ZoneFilter,ImagePropertiesFilter,ComputeCapabilitiesFilter,'
                'MaintenanceFilter,PowerVCServerGroupAffinityFilter,PowerVC'
                'ServerGroupAntiAffinityFilter,PowerVCHostAggregateFilter,'
                'PowerVMNetworkFilter,PowerVMProcCompatModeFilter,PowerLMB'
                'SizeFilter,PowerMigrationLicenseFilter,PowerVMMigrationCount'
                'Filter,PowerVMHMCStorageFilter,PowerVMIBMiMobilityFilter,'
                'PowerVMRemoteRestartFilter,PowerVMRemoteRestartSameHMCFilter,'
                'PowerVMEndianFilter,PowerVMGuestCapableFilter,'
                'PowerVMSharedProcPoolFilter', V1232)))

    def __added_ego_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific nova.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            cpu_allocation_ratio=('4.0', V121),
            scheduler_available_filters=(self.__get_scheduler_filters(), V123),
            scheduler_default_filters=(
                'RamFilter,CoreFilter,ComputeFilter,RetryFilter,Availability'
                'ZoneFilter,ImagePropertiesFilter,ComputeCapabilitiesFilter,'
                'MaintenanceFilter,DiskFilter,PowerVCServerGroupAffinityFilter'
                ',PowerVCServerGroupAntiAffinityFilter,PowerVCHostAggregate'
                'Filter,PowerKVMNetworkFilter,PowerKVMSMTFilter,PowerKVM'
                'StorageFilter', V123)))

    def __get_scheduler_filters(self):
        """Helper method to build a list of scheduler filters"""
        #Make sure to update the Version number above when changing lists
        common_filters = [
            'host_aggregate_filter.PowerVCHostAggregateFilter',
            'affinity_filter.PowerVCServerGroupAffinityFilter',
            'affinity_filter.PowerVCServerGroupAntiAffinityFilter']
        #Make sure to update the Version number above when changing lists
        powerkvm_filters = [
            'smt_filter.PowerKVMSMTFilter',
            'network_filter.PowerKVMNetworkFilter',
            'storage_filter.PowerKVMStorageFilter']
        #Make sure to update the Version number above when changing lists
        powervm_filters = [
            'endian_filter.PowerVMEndianFilter',
            'network_filter.PowerVMNetworkFilter',
            'hmc_storage_filter.PowerVMHMCStorageFilter',
            'guest_capable_filter.PowerVMGuestCapableFilter',
            'ibmi_mobility_filter.PowerVMIBMiMobilityFilter',
            'remote_restart_filter.PowerVMRemoteRestartFilter',
            'migration_count_filter.PowerVMMigrationCountFilter',
            'proc_compat_mode_filter.PowerVMProcCompatModeFilter',
            'remote_restart_filter.PowerVMRemoteRestartSameHMCFilter',
            'shared_proc_pool_filter.PowerVMSharedProcPoolFilter']
        prefix = 'powervc_nova.scheduler.filters.common.'
        filters = [(prefix + filt) for filt in common_filters]
        #Depending on if it is PowerVM or KVM, need to add different filters
        if utils.get_offering() == 'powervm':
            prefix = 'powervc_nova.scheduler.filters.powervm.'
            filters.extend([(prefix + filt) for filt in powervm_filters])
        else:
            prefix = 'powervc_nova.scheduler.filters.kvm.'
            filters.extend([(prefix + filt) for filt in powerkvm_filters])
        return filters

    def __upgrade_aggregate_metadata(self, tardir, tarfile, policy_data):
        """Helper method to upgrade the Aggregate MetaData to the DB"""
        initial_policy_id = ''
        state = ''
        for line in policy_data.split():
            line = line.strip()
            if line.startswith('<id>'):
                initial_policy_id = re.compile('<id>(.*?)</id>', re.DOTALL |
                                               re.IGNORECASE).findall(line)
            elif line.startswith('<state>'):
                state = re.compile('<state>(.*?)</state>', re.DOTALL |
                                   re.IGNORECASE).findall(line)
                if str(state) == "['enabled']":
                    break
        initial_policy_id = str(initial_policy_id).strip("['']")
        mgr.LOG.info(_("Restoring initial policy id: %(initial_policy_id)s; "
                       "state: %(state)s") % {'initial_policy_id':
                                              initial_policy_id,
                                              'state': state})
        engine = utils.get_database_engine(NOVA_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='nova')
        aggr_table = Table('AGGREGATES', metadata, autoload=True)
        meta_table = Table('AGGREGATE_METADATA', metadata, autoload=True)
        #Insert the Default Aggregate into the DB
        aggr_table.insert().values(
            id=1, name='Default Group', created_at=datetime.utcnow(),
            deleted=0).execute()
        #Insert each of the Aggregate MetaData Items into the DB
        self.__insert_aggregate_metadata(meta_table, 1, 'hapolicy-id', '1')
        self.__insert_aggregate_metadata(meta_table, 1,
                                         'hapolicy-run_interval', '1')
        self.__insert_aggregate_metadata(meta_table, 1,
                                         'hapolicy-stabilization', '1')

        if initial_policy_id:
            self.__insert_aggregate_metadata(meta_table, 1, 'initialpolicy-id',
                                             str(initial_policy_id))

    def __insert_aggregate_metadata(self, table, aggr_id, key, value):
        """Helper method to insert an individual Aggregate MetaData Item"""
        table.insert().values(
            aggregate_id=aggr_id, key=key, value=value,
            created_at=datetime.utcnow(), deleted=0).execute()
