# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
This is the object-relational mapping (ORM) for PowerVC extensions to
the OpenStack Nova Database Schema.  These extensions include new database
tables to represent HMC's, VIOS's, SCG's, Network Adapters, FC Ports,
Health Status, On-boarding Task Progress, etc.
"""

from sqlalchemy import Column, ForeignKey, event
from sqlalchemy import Integer, String, Boolean, DateTime, Float, Text
from sqlalchemy.orm import relationship, backref

import nova.db.sqlalchemy.api as base_dbapi
import nova.db.sqlalchemy.models as base_dbmodels
from nova.objects import base as base_obj
from nova.objects import instance as inst_obj
from nova.objects import fields as fields_util
from oslo_log import log as logging
from oslo_utils import timeutils

NAME_LEN, DESC_LEN = (255, 1023)
STATUS_LEN, MODE_LEN = (31, 31)
USER_LEN, CRED_LEN = (63, 255)
UUID_LEN, IP_LEN = (36, 39)

LOG = logging.getLogger(__name__)


######################################################
#############    HMC Model Definition    #############
######################################################
class HmcDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """ The HmcDTO class provides the ORM for the ibm_hmcs table """
    __tablename__ = 'ibm_hmcs'
    __table_args__ = ()
    #Hardware Management Console DTO Model Attributes
    _pk_id = Column('id', Integer, primary_key=True)
    hmc_uuid = Column('uuid', String(UUID_LEN), nullable=False)
    hmc_display_name = Column('display_name', String(NAME_LEN), nullable=False)
    access_ip = Column('access_ip', String(NAME_LEN), nullable=False)
    user_id = Column('user_id', String(USER_LEN), nullable=False)
    password = Column('user_credentials', String(CRED_LEN), nullable=False)
    registered_at = Column('registered_at', DateTime, default=timeutils.utcnow)


class HmcHostsDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """ The HmcHostsDTO class provides the ORM for the ibm_hmc_hosts table """
    __tablename__ = 'ibm_hmc_hosts'
    __table_args__ = ()
    #HMC Hosts DTO Model Attributes
    _pk_id = Column('id', Integer, primary_key=True)
    hmc_uuid = Column('hmc_uuid', String(UUID_LEN), nullable=False)
    host_name = Column('host_name', String(NAME_LEN), nullable=False)


class HmcHostClustersDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The HostClusterDTO class provides the ORM for the host_clusters table"""
    __tablename__ = 'ibm_hmc_host_clusters'
    __table_args__ = ()
    #HMC Host Clusters DTO Model Attributes
    _pk_id = Column('id', Integer, primary_key=True)
    hmc_uuid = Column('hmc_uuid', String(UUID_LEN), nullable=False)
    host_name = Column('host_name', String(NAME_LEN), nullable=False)
    cluster_id = Column('cluster_id', String(UUID_LEN), nullable=False)


######################################################
#############  Instance Model Definition  ############
######################################################
class InstancePowerSpecsDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The InstPowerSpecDTO class provides the ORM for the power_specs Table"""
    __tablename__ = 'instance_power_specs'
    __table_args__ = ()
    #Instance Power Specification DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    instance_uuid = Column('instance_uuid', String(UUID_LEN), nullable=False)
    vm_id = Column('vm_id', String(UUID_LEN))
    vm_uuid = Column('vm_uuid', String(UUID_LEN))
    vm_name = Column('vm_name', String(NAME_LEN))
    vios_ids = Column('vios_ids', String(MODE_LEN))
    vcpu_mode = Column('vcpu_mode', String(MODE_LEN))
    memory_mode = Column('memory_mode', String(MODE_LEN))
    vcpus = Column('vcpus', Integer)
    min_vcpus = Column('min_vcpus', Integer)
    max_vcpus = Column('max_vcpus', Integer)
    proc_units = Column('proc_units', Float)
    min_proc_units = Column('min_proc_units', Float)
    max_proc_units = Column('max_proc_units', Float)
    memory_mb = Column('memory_mb', Integer)
    min_memory_mb = Column('min_memory_mb', Integer)
    max_memory_mb = Column('max_memory_mb', Integer)
    root_gb = Column('root_gb', Integer)
    ephemeral_gb = Column('ephemeral_gb', Integer)
    pending_action = Column('pending_action', String(STATUS_LEN))
    pending_onboard_actions = Column(
        'pending_onboard_actions', String(NAME_LEN))
    processor_compatibility = Column(
        'processor_compatibility', String(MODE_LEN))
    current_compatibility_mode = Column(
        'current_compatibility_mode', String(MODE_LEN))
    desired_compatibility_mode = Column(
        'desired_compatibility_mode', String(MODE_LEN))
    dedicated_sharing_mode = Column(
        'dedicated_sharing_mode', String(MODE_LEN))
    dedicated_proc = Column('dedicated_proc', String(MODE_LEN))
    avail_priority = Column('avail_priority', Integer)
    shared_weight = Column('shared_weight', Integer)
    rmc_state = Column('rmc_state', String(STATUS_LEN))
    uncapped = Column('uncapped', Boolean)
    operating_system = Column('operating_system', String(NAME_LEN))
    cpu_topology = Column('cpu_topology', Text)
    vnc_info = Column('vnc_info', Text)
    srr_capability = Column('srr_capability', Boolean)
    srr_state = Column('srr_state', Text)
    shared_proc_pool_id = Column('shared_proc_pool_id', Integer)
    shared_proc_pool_name = Column('shared_proc_pool_name', String(32))
    _instance = relationship(
        base_dbmodels.Instance,
        backref=backref('_power_specs', uselist=False,
                        lazy='joined', innerjoin=True),
        foreign_keys=instance_uuid,
        primaryjoin='InstancePowerSpecsDTO.instance_uuid == Instance.uuid')


######################################################
##############   SPP Model Definition    #############
######################################################
class SharedProcessorPoolDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The SPPDTO class provides the ORM for the shared_proc_pool Table"""
    __tablename__ = 'shared_processor_pool'
    __table_args__ = ()
    # Shared Processor Pool DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    pool_id = Column('pool_id', Integer, nullable=False)
    pool_name = Column('pool_name', String(32), nullable=False)
    host_name = Column('host_name', String(NAME_LEN), nullable=False)
    pool_uuid = Column('pool_uuid', String(UUID_LEN), nullable=False)
    maximum_proc_units = Column('maximum_proc_units', Float)
    reserved_proc_units = Column('reserved_proc_units', Float)
    available_proc_units = Column('available_proc_units', Float)


######################################################
#############   VIOS Model Definition    #############
######################################################
class VioServerDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The VioServerDTO class provides the ORM for the server_vio Table"""
    __tablename__ = 'server_vio'
    __table_args__ = ()
    # VIO Server DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    id = Column('id', String(UUID_LEN), nullable=False)
    lpar_id = Column('lpar_id', Integer, nullable=False)
    lpar_name = Column('lpar_name', String(NAME_LEN), nullable=False)
    state = Column('state', String(STATUS_LEN))
    rmc_state = Column('rmc_state', String(STATUS_LEN))
    cluster_provider_name = Column('cluster_provider_name', String(NAME_LEN))
    backend_uuid = Column('backend_uuid', String(UUID_LEN))
    _compute_node_id = Column('compute_node_id', Integer,
                              ForeignKey('compute_nodes.id'), nullable=False)


######################################################
#############  Network Model Definition  #############
######################################################
class SharedEthernetAdapterDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The SEA DTO class provides the ORM for the adap_ethernet_shared Table"""
    __tablename__ = 'adapter_ethernet_shared'
    __table_args__ = ()
    #Shared Ethernet Adapter DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    id = Column('id', String(UUID_LEN), nullable=False)
    name = Column('adapter_name', String(NAME_LEN), nullable=False)
    slot = Column('slot', Integer)
    state = Column('state', String(STATUS_LEN))
    _primary_vea_pk_id = Column('primary_vea_pk_id', Integer, nullable=False)
    _control_channel_vea_pk_id = Column('control_channel_vea_pk_id', Integer)
    _vios_pk_id = Column('vios_pk_id', Integer,
                         ForeignKey('server_vio.pk_id'), nullable=False)


class VirtualEthernetAdapterDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The VEA DTO class provides the ORM for the adapt_ethernet_virt Table"""
    __tablename__ = 'adapter_ethernet_virtual'
    __table_args__ = ()
    #Virtual Ethernet Adapter DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    id = Column('id', String(UUID_LEN), nullable=False)
    name = Column('adapter_name', String(NAME_LEN), nullable=False)
    slot = Column('slot', Integer)
    state = Column('state', String(STATUS_LEN))
    pvid = Column('pvid', Integer)
    is_trunk = Column('is_trunk', Boolean)
    trunk_pri = Column('trunk_priority', Integer)
    ieee_eth = Column('is_ieee_ethernet', Boolean)
    vswitch_name = Column('vswitch_name', String(NAME_LEN))
    _vlan_ids_string = Column('vlan_ids_string', String(NAME_LEN))
    _vios_pk_id = Column('vios_pk_id', Integer,
                         ForeignKey('server_vio.pk_id'), nullable=False)
    _sea_pk_id = Column('sea_pk_id', Integer,
                        ForeignKey('adapter_ethernet_shared.pk_id'))


class NetworkAssociationDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The Network Assoc DTO class provides the ORM for the net_assoc Table"""
    __tablename__ = 'host_network_association'
    __table_args__ = ()
    #Network Association DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    id = Column('id', String(UUID_LEN), nullable=False)
    neutron_net_id = Column('neutron_network_id',
                            String(UUID_LEN), nullable=False)
    _sea_pk_id = Column('sea_pk_id', Integer,
                        ForeignKey('adapter_ethernet_shared.pk_id'))
    _compute_node_id = Column('compute_node_id', Integer,
                              ForeignKey('compute_nodes.id'), nullable=False)


######################################################
############  Storage Model Definition   #############
######################################################
class StorageConnectivityGroupDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The SCG DTO class provides the ORM for the storage_conn_group Table"""
    __tablename__ = 'storage_connectivity_group'
    __table_args__ = ()
    #Storage Connectivity Group DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    id = Column('id', String(UUID_LEN), nullable=False)
    display_name = Column('display_name', String(NAME_LEN), nullable=False,
                          unique=True)
    enabled = Column('enabled', Boolean)
    auto_defined = Column('auto_defined', Boolean)
    auto_add_vios = Column('auto_add_vios', Boolean)
    fc_storage_access = Column('fc_storage_access', Boolean)
    port_tag = Column('port_tag', String(NAME_LEN))
    cluster_provider_name = \
        Column('cluster_provider_name', String(NAME_LEN))
    priority_cluster = Column('priority_cluster', Integer)
    priority_external = Column('priority_external', Integer)
    vios_redundancy = Column('vios_redundancy', Integer)
    exact_redundancy = Column('exact_redundancy', Boolean)
    vios_redundancy = Column('vios_redundancy', Integer)
    exact_redundancy = Column('exact_redundancy', Boolean)
    npiv_fabric = Column('npiv_fabric', String(NAME_LEN))
    boot_connectivity = Column('boot_connectivity', String(NAME_LEN))
    data_connectivity = Column('data_connectivity', String(NAME_LEN))
    backend_host_group = Column('backend_host_group', String(NAME_LEN))
    include_untagged = Column('include_untagged', Boolean)


class FcPortDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The FCPort DTO class provides the ORM for the port_fc Table"""
    __tablename__ = 'port_fc'
    __table_args__ = ()
    #FC Port DTO Model Attributes
    id = Column('id', String(NAME_LEN), nullable=False)
    name = Column('name', String(NAME_LEN))
    status = Column('status', String(STATUS_LEN))
    enabled = Column('enabled', Boolean)
    wwpn = Column('wwpn', String(UUID_LEN))
    adapter_id = Column('adapter_id', String(UUID_LEN), nullable=False)
    port_tag = Column('port_tag', String(NAME_LEN))
    fabric = Column('fabric', String(NAME_LEN))
    fc_protocol = Column('fc_protocol', String(STATUS_LEN))
    max_vfcs = Column('max_vfcs', Integer)
    mapped_vfcs = Column('mapped_vfcs', Integer)
    reserved_vfcs = Column('reserved_vfcs', Integer)
    avail_active_vfcs = Column('avail_active_vfcs', Integer)
    _pk_id = Column('pk_id', Integer, primary_key=True)
    _vios_pk_id = Column('vios_pk_id', Integer,
                         ForeignKey('server_vio.pk_id'), nullable=False)


class ScgViosAssociationDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The ScgViosAssocDTO class provides the ORM for the scg_vios Table"""
    __tablename__ = 'scg_vios_association'
    __table_args__ = ()
    #SCG to VIO Server Association DTO Model Attributes
    _pk_id = Column('pk_id', Integer, primary_key=True)
    _scg_pk_id = Column('scg_pk_id', Integer, ForeignKey(
                        'storage_connectivity_group.pk_id'), nullable=False)
    _vios_pk_id = Column('vios_pk_id', Integer,
                         ForeignKey('server_vio.pk_id'), nullable=False)


######################################################
##########  Health Status Model Definition   #########
######################################################
class ComputeNodeHealthStatusDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The Compute Health DTO class provides the ORM for the health Table"""
    __tablename__ = 'compute_node_health_status'
    __table_args__ = ()
    #Compute Health Status DTO Model Attributes
    id = Column(String(UUID_LEN), primary_key=True)
    health_state = Column(String(255))
    reason = Column(Text)
    unknown_reason_details = Column(Text)


class InstanceHealthStatusDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The Instance Health DTO class provides the ORM for the health Table"""
    __tablename__ = 'instance_health_status'
    __table_args__ = ()
    #Instance Health Status DTO Model Attributes
    id = Column(String(UUID_LEN), primary_key=True)
    health_state = Column(String(255))
    reason = Column(Text)
    unknown_reason_details = Column(Text)


class HMCHealthStatusDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The HMC Health DTO class provides the ORM for the health Table"""
    __tablename__ = 'ibm_hmc_health_status'
    __table_args__ = ()
    #HMC Health Status DTO Model Attributes
    id = Column(String(36), primary_key=True)
    health_state = Column(String(255))
    reason = Column(Text)
    unknown_reason_details = Column(Text)


######################################################
##########  On-board Task Model Definition   #########
######################################################
class OnboardTaskDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The OnboardTaskDTO class provides the ORM for the onboard_tasks table"""
    __tablename__ = 'onboard_tasks'
    __table_args__ = ()
    #On-board Task DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    host = Column('host', String(NAME_LEN), nullable=False)
    status = Column('status', String(STATUS_LEN), default='waiting',
                    nullable=False)
    started = Column('started_at', DateTime,
                     default=timeutils.utcnow, nullable=False)
    ended = Column('ended_at', DateTime)
    progress = Column('progress', Integer, default=0, nullable=False)


class OnboardTaskServerDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The OnboardTaskServerDTO class provides the ORM for the tasks table"""
    __tablename__ = 'onboard_task_servers'
    __table_args__ = ()
    #On-board Task Server DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    task_id = Column('task_id', Integer, nullable=False)
    server_uuid = Column('server_uuid', String(UUID_LEN), nullable=False)
    server_name = Column('server_name', String(NAME_LEN), nullable=False)
    status = Column('status', String(STATUS_LEN), nullable=False)
    fault_message = Column('fault_message', String(NAME_LEN))


######################################################
########  Console Messages Model Definition   ########
######################################################
class ConsoleMessagesDTO(base_dbmodels.BASE, base_dbmodels.NovaBase):
    """The ConsoleMessagesDTO class provides the ORM for the cons_msgs table"""
    __tablename__ = 'ibm_console_messages'
    __table_args__ = ()
    #Console Messages DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    msg_time = Column('msg_time', DateTime,
                      default=timeutils.utcnow, nullable=False)
    msg_type = Column('msg_type', String(63), nullable=False)
    msg_user = Column('msg_user', String(255), nullable=True)
    priority = Column('priority', String(31), nullable=False)
    publisher = Column('publisher', String(127), nullable=True)
    payload = Column('payload', Text, nullable=False)


###################################################
##########   List of PowerVC 1.2 DTO's   ##########
###################################################
POWERVC_V1R2_DTOS = [
    HmcDTO,  HmcHostsDTO, HmcHostClustersDTO, InstancePowerSpecsDTO,
    VioServerDTO, SharedEthernetAdapterDTO, VirtualEthernetAdapterDTO,
    NetworkAssociationDTO, StorageConnectivityGroupDTO, FcPortDTO,
    ScgViosAssociationDTO, ComputeNodeHealthStatusDTO,
    InstanceHealthStatusDTO, HMCHealthStatusDTO,
    OnboardTaskDTO, OnboardTaskServerDTO
]

POWERVC_V1R2M2_DTOS = [ConsoleMessagesDTO]

POWERVC_V1R2M3MF2_DTOS = [SharedProcessorPoolDTO]


###################################################
#########  Override PowerSpecs on Instance  #######
###################################################
#The way OpenStack is setup each Instance/VM basically has 2 names, the
#"display name" that the user specifies and the "name" that is internally
#generated based on a template and what is used to actually create the VM
#on the Hyper-visor.  This template name isn't actually persisted and is
#generated on the fly when someone asks for it.  This works well for deployed
#instances, but for any vm's that are changed outside of PowerVC or are
#on-boarded, the template name won't match the actual name of the VM.
#
#This code below is to inject method overrides for the name and power_specs
#properties on the OpenStack SQLAlchemy and Nova.Objects Instance classes
#so that the PowerSpecs attributes are available on the Instance object and
#also so that that the name property returns the vm_name value from the
#PowerSpecs if it is set rather than the generated template name.  Both the
#on-boarding and out-of-band changes will cause the vm_name to be set to
#the actual VM's name on the Hyper-visor, so will always reflect the actual
#name rather than some internally generated name based on the id by OpenStack.
class PowerSpecsField(fields_util.FieldType):
    def coerce(self, obj, attr, value):
        if value is not None and not isinstance(value, dict):
            ValueError(_('The PowerSpecs must be a Dictionary'))
        return dict(value.iteritems()) if value is not None else dict()


def _instance_extra_specs(self):
    """Override the Instance ExtraSpecs to return PowerSpecs"""
    return ['name', 'power_specs']


def _instance_power_specs_get(self):
    """Override the Instance PowerSpecs Property Get Method"""
    adict = dict()
    attrs = ['created_at', 'updated_at', 'deleted_at', 'deleted', '_instance']
    try:
        power_specs = self['_power_specs']
        #Loop through each of the Columns in the Table, adding to Dictionary
        for column in power_specs.__table__.columns:
            if column.name not in attrs:
                adict[column.name] = getattr(power_specs, column.name)
        return adict
    #To work around OpenStack not returning the full retrieved
    #Instance from the instance_create method, allow an exception
    except Exception:
        return adict


def _instance_power_specs_set(self, value):
    """Override the Instance PowerSpecs Property Set Method"""
    pass


def _instance_name(self):
    """Override the SQLAlchemy Instance Name method"""
    power_specs = self.get('power_specs')
    power_specs = power_specs if power_specs is not None else dict()
    vm_name = power_specs.get('vm_name')
    if vm_name is not None and len(vm_name) > 0:
        return vm_name
    vm_name = self._orig_name()
    #We need to sanitize the template name and substitute invalid chars
    return _sanitize_instance_name(vm_name)

_INVALID_NAME_CHARS = '()\<>*$&?|[]\'"`,/#=:;'
_INVALID_NAME_CHAR_MAP = dict([(c, '') for c in list(_INVALID_NAME_CHARS)])


def _sanitize_instance_name(vm_name):
    """Replace invalid characters in the template name for the VM"""
    replaced_char = False
    #If the VM Name is None or an empty string, just return it
    if vm_name:
        new_name = list(vm_name)
        #Loop through each character in the string, substituting if invalid
        for index in range(len(new_name)):
            ordinal = ord(new_name[index])
            out_of_range = ordinal < 32 or ordinal > 126
            #Substitute any spaces with underscore in the VM name
            if ordinal == 32:
                new_name[index] = '_'
                replaced_char = True
            #Replace reserved or non-ASCII characters with the caret symbol
            elif out_of_range or new_name[index] in _INVALID_NAME_CHAR_MAP:
                new_name[index] = '^'
                replaced_char = True
        #For performance only convert new name to a string if it changed
        if replaced_char:
            return "".join(new_name)
    return vm_name


def _instance_create_listener(mapper, connection, target):
    """Helper method to listen for Instance Creations to create PowerSpecs"""
    try:
        session = base_dbapi.get_session()
        #Need to create a default PowerSpecs when the Instance is created
        with session.begin():
            specs_ref = InstancePowerSpecsDTO()
            specs_ref.update(dict(instance_uuid=target['uuid']))
            specs_ref.save(session=session)
    except Exception as exc:
        LOG.warn('Error inserting PowerSpecs in the database')
        LOG.exception(exc)


#Inject the Name and PowerSpecs into the SQLAlchemy Instance
_orig_dbinst_name = getattr(base_dbmodels.Instance, 'name')
_db_power_spec_methods = property(fget=_instance_power_specs_get,
                                  fset=_instance_power_specs_set)
setattr(base_dbmodels.Instance, '_extra_keys',
        property(fget=_instance_extra_specs))
setattr(base_dbmodels.Instance, '_orig_name', _orig_dbinst_name.fget)
setattr(base_dbmodels.Instance, 'power_specs', _db_power_spec_methods)
setattr(base_dbmodels.Instance, 'name', property(fget=_instance_name))

#Inject the Name and PowerSpecs into the Instance NovaObject
inst_obj.Instance.fields['power_specs'] = \
    fields_util.Field(PowerSpecsField(), nullable=True)
base_obj.make_class_properties(inst_obj.Instance)
_orig_instobj_name = getattr(inst_obj.Instance, 'name')
setattr(inst_obj.Instance, '_orig_name', _orig_instobj_name.fget)
setattr(inst_obj.Instance, 'name', property(fget=_instance_name))

#Listen for notifications after an Instance is created in the DB
event.listen(base_dbmodels.Instance,
             'after_insert', _instance_create_listener)
