# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import objects
from nova.compute import vm_states

from oslo_log import log as logging

from nova.ibm.ego.hapolicy.instance_driver import driver

from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVCHAInstanceDriver(driver.HAInstanceDriver):
    """PowerVC's HA service driver class to extend PRS base logic."""

    @staticmethod
    def _get_compute_node(context, svc_host):
        """
        This returns the respective compute node given a service host.

        :param context: A context object that is used to authorize DB access
        :param svc_host: The service host whose compute node to locate
        :returns: The corresponding compute node or None if not found
        """
        context = context.elevated()
        compute_nodes = objects.ComputeNodeList.get_all(context) or []
        compute_node = None
        for node in compute_nodes:
            if node.host == svc_host:
                compute_node = node
                break

        return compute_node

    @staticmethod
    def _include_instance(context, inst, svc_host, hv_is_up):
        """Determines whether the instance should be rebuilt."""
        node = PowerVCHAInstanceDriver._get_compute_node(context, svc_host)
        valid_states = [vm_states.ACTIVE, vm_states.STOPPED]
        if not hv_is_up:
            # VMs that are in ERROR state can also be 'rebuilt'
            valid_states.append(vm_states.ERROR)

        if inst.vm_state not in valid_states:
            # If the VM state doesn't allow it to be rebuild (i.e., per
            # OpenStack's rules), then we need to filter it.
            LOG.debug("Bypassing instance %(name)s (%(uuid)s) because it is "
                      "not in a valid state to be rebuilt; its state is "
                      "currently %(state)s and it needs to be one of "
                      "%(states)s."
                      % dict(name=inst.display_name, uuid=inst.uuid,
                             states=', '.join(valid_states),
                             state=inst.vm_state))
            return False

        if (node.hypervisor_type == pvmc.IBM_POWERVM_HYPERVISOR_TYPE and
                not hv_is_up):
            # If the hypervisor is down and it's PowerVM, then we need to make
            # sure the VM is remote restartable before passing it to the
            # HA engine.
            srr_state = inst.system_metadata.get('srr_state')
            if srr_state != 'Remote_Restartable':
                LOG.debug("Bypassing instance %(name)s (%(uuid)s) because "
                          "it is not configured for 'remote restart' "
                          "operations on the hypervisor."
                          % dict(name=inst.display_name, uuid=inst.uuid))
                return False

        return True

    def select_instances(self, context, svc_host, hv_is_up):
        """
        Select all instances from this host that should be remote restarted.

        :param context: nova context
        :param svc_host: service host
        :param hv_is_up: True if the hypervisor is active and False otherwise
        :return: [(instance, action)]
        """
        if hv_is_up:
            # If the hypervisor is up, we shouldn't migrate anything because
            # PowerVC doesn't support pro-active evacuation at this time.
            LOG.error(_("The instance driver's select_instances() was called "
                        "when the hypervisor was deemed active.  This should "
                        "not have happened because PowerVC does not support "
                        "predictive host evacuations at this time."))
            return []

        insts = [inst for inst in
                 objects.InstanceList.get_by_host(
                     context, svc_host, expected_attrs=['uuid', 'vm_state',
                                                        'display_name',
                                                        'system_metadata'])
                 if PowerVCHAInstanceDriver._include_instance(context, inst,
                                                              svc_host,
                                                              hv_is_up)]
        return [(inst, None) for inst in insts]
