# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from nova import objects
from nova import utils
from oslo_log import log as logging

from nova.ibm.ego import ego_constants
from nova.ibm.ego.hapolicy import constants as ha_const

from powervc_nova.db import api as pvcdb_api
from powervc_nova.scheduler.filters import PowerVCBaseFilter
from powervc_nova.virt.ibmpowervm.common import constants as pvmc
from powervc_nova import logcall
from powervc_nova import _

LOG = logging.getLogger(__name__)


class PowerVMRemoteRestartFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the PowerVM host supports remote
    restart operations.  The simplified remote restart operations are only
    supported on POWER8 or newer hosts.
    """

    def __init__(self):
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_CREATE,
             ego_constants.FILTER_WORKING_SCOPE_RESIZE,
             ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
             ego_constants.FILTER_WORKING_SCOPE_EVACUATE,
             ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        if not self._allow_oper_for_powervm(filter_properties, all_host_attrs):
            LOG.debug("Ignoring PowerVMRemoteRestartFilter...")
            return

        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        operation_type = request_spec.get('operation_type')
        is_deploy = (operation_type ==
                     ego_constants.FILTER_WORKING_SCOPE_CREATE)
        is_other = (operation_type in
                    [ego_constants.FILTER_WORKING_SCOPE_RESIZE,
                     ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
                     ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
                     ego_constants.FILTER_WORKING_SCOPE_UNSHELVE])
        apply_filter = False

        if is_deploy:
            # For deploys, check the flavor's extra_specs.
            instance_type = filter_properties.get('instance_type', {})
            extra_spec = instance_type.get('extra_specs', {})
            flavor_srr_cap = str(extra_spec.get('powervm:srr_capability',
                                                False)).lower()
            if flavor_srr_cap == 'true':
                apply_filter = True
        elif is_other:
            # For migrations/unshelves, check the instance system metadata.
            system_metadata = utils.instance_sys_meta(instance_props) or {}
            srr_cap = str(system_metadata.get('srr_capability', False)).lower()
            if srr_cap == 'true':
                apply_filter = True
        else:
            # For evacuates, always apply the filter.
            apply_filter = True

        if not apply_filter:
            LOG.debug("Ignoring PowerVMRemoteRestartFilter because the "
                      "virtual machine does not require PowerVM's "
                      "simplified remote restart capabilities.")
            return

        # In non-evacuate cases, don't apply the check for the number of remote
        # restart operations in progress.
        if (operation_type in
                [ego_constants.FILTER_WORKING_SCOPE_CREATE,
                 ego_constants.FILTER_WORKING_SCOPE_RESIZE,
                 ego_constants.FILTER_WORKING_SCOPE_LIVE_MIGRATION,
                 ego_constants.FILTER_WORKING_SCOPE_COLD_MIGRATION,
                 ego_constants.FILTER_WORKING_SCOPE_UNSHELVE]):
            req_str = ("$%(rr_cap_name)s == True" %
                       dict(rr_cap_name=pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE))
        # If the HA engine invoked us, then we shouldn't filter based on count
        # because the HA engine needs to 'throttle' the requests appropriately
        # (and if we filtered, the HA engine would get NoValidHost, which would
        # send it down an error path).  Note the operation type is also
        # implicitly 'evacuate' in the below cases as well.
        elif ego_constants.HA_MAINTENANCE_STATUS in filter_properties:
            LOG.info(_("PowerVMRemoteRestartFilter invoked via a rebuild "
                       "operation from the HA engine; ignoring progress "
                       "counts as the HA engine will address those."))
            req_str = ("$%(rr_cap_name)s == True" %
                       dict(rr_cap_name=pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE))
        else:
            req_str = ("($%(rr_cap_name)s == True) && ($%(rrip_cap_name)s "
                       "< $'%(rrs_cap_name)s')" %
                       dict(rr_cap_name=pvmc.PVM_SIMP_REMOTE_RESTART_CAPABLE,
                            rrip_cap_name=pvmc.NUM_RR_IN_PROGRESS,
                            rrs_cap_name=ha_const.PARALLEL_DEST_REBUILD))
        self._select(req_str)
        LOG.debug("After passing the PowerVMRemoteRestartFilter, the resource "
                  "requirement expression is (%s)." % req_str)


class PowerVMRemoteRestartSameHMCFilter(PowerVCBaseFilter):
    """
    This EGO scheduler filter ensures that the remote restart operations stay
    within the domain of a single HMC as cross-HMC remote restart operations
    are not allowed at this time.
    """

    def __init__(self):
        self._set_filter_scope(
            [ego_constants.FILTER_WORKING_SCOPE_EVACUATE])
        self._set_error_message(_("No suitable host could be found for the "
                                  "remote restart operation because the "
                                  "source host uses an HMC that is not shared "
                                  "with other hosts in the environment."))

    @staticmethod
    def _get_hmc_uuids(context, host_name):
        """Retrieve a list of HMC UUIDs managing the host."""
        return [str(hmc['hmc_uuid']) for hmc in
                pvcdb_api.hmc_find_all(context, dict(host_name=host_name))]

    @logcall
    def parse(self, filter_properties, all_host_attrs):
        context = filter_properties.get('context')
        context = context.elevated()
        request_spec = filter_properties.get('request_spec', {})
        instance_props = request_spec.get('instance_properties', {})
        host = instance_props.get('host')
        src_hmc_uuids = PowerVMRemoteRestartSameHMCFilter._get_hmc_uuids(
            context, host)
        valid_hosts = []

        for node in objects.ComputeNodeList.get_all(context):
            dest_host = node.host
            dst_hmc_uuids = PowerVMRemoteRestartSameHMCFilter._get_hmc_uuids(
                context, dest_host)
            # Perform the set intersection to find common management HMCs.
            intersection = list(set(src_hmc_uuids) & set(dst_hmc_uuids))

            if not intersection:
                # No common management HMCs; not a valid destination host.
                continue
            valid_hosts.append(dest_host)

        host_to_hyper_map = self._get_hosts_to_hypervisors_map(context)
        ret = [host_to_hyper_map[host] for host in valid_hosts
               if host in host_to_hyper_map]
        LOG.debug(("Setting included hosts in "
                   "PowerVMRemoteRestartSameHMCFilter (%s).") % ret)
        self._include_hosts(ret)
