# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from powervc_nova import logcall
import six
import time
import exception
from eventlet import greenthread

from nova.compute import vm_states

from oslo_config import cfg
CONF = cfg.CONF
from nova import conductor
from nova.objects import instance as instance_obj
from oslo_log import log as logging
from powervc_nova.virt.ibmpowervm.common.constants import IBMI_OS_DISTRO
from powervc_nova import _
import powervc_nova.virt.ibmpowervm.hmc.k2.constants as k2const
from powervc_nova.virt.ibmpowervm.hmc.k2oprguard import K2OperGuard
from powervc_nova.virt.ibmpowervm.hmc.k2.lparentry_wrapper \
    import LPAREntryWrapper

LOG = logging.getLogger(__name__)
ACTIVATING = 'activating'
ibmpowervm_opts = [
    cfg.IntOpt('ibmpowervm_rmc_active_timeout',
               default=1200,
               help='Timeout in seconds for rmc to be active after deploy.')
]
CONF.register_opts(ibmpowervm_opts)


@logcall
def check_task_state(context, instance, host, lpar_uuid):
    """Set task state to activating while waiting for RMC to be ready
    :param context: nova context for this operation
    :params instance:
        nova.db.sqlalchemy.models.Instance object
        instance object that was deployed.
    """
    def reset_task_state():
        try:
            conductor.API().instance_update(context,
                                            instance['uuid'],
                                            task_state=None,
                                            expected_task_state=ACTIVATING)
        except Exception:
            LOG.warning(_('Unable to reset task state to None for instance %s')
                        % instance['name'])

    def get_lpar_activation_info():
        """Gets the refcode, dlpar connectivity, and rmc state from K2"""
        with K2OperGuard(host=host) as k2oper:
            k2resp = k2oper.read(k2const.MGT_SYS, host.uuid,
                                 k2const.LPAR, lpar_uuid)
            lpar_entry = k2resp.entry
            lpar_wrapper = LPAREntryWrapper(lpar_entry)
        refcode = lpar_wrapper.get_ref_code()
        dlpar, rmc_state = lpar_wrapper.check_dlpar_connectivity()
        return refcode, dlpar, rmc_state

    inst_name = instance['name']

    start_time = time.time()
    timeout = CONF.ibmpowervm_rmc_active_timeout
    refcode, dlpar, rmc_state = get_lpar_activation_info()
    # poll instance for reference code
    ref_code_fail = False
    ref_code_unsupported_os_ver = False
    system_metadata = instance.get('system_metadata')
    os_distro = system_metadata.get('image_os_distro')
    ref400 = '00000000'
    wait_inc = 15
    while rmc_state is None or rmc_state != 'active' or dlpar is False:
        greenthread.sleep(wait_inc)
        curr_time = time.time()
        # wait up to (timeout) seconds for start up
        if (curr_time - start_time) > timeout:
            if IBMI_OS_DISTRO == os_distro:
                LOG.debug("Timed out waiting for ok or failed reference code "
                          "for %s." % inst_name)
            else:
                LOG.debug("Timed out waiting for RMC to be active "
                          "for %s." % inst_name)
            reset_task_state()
            return
        if refcode == 'No license':
            ref_code_fail = True
            break
        if refcode == 'B700F120':
            ref_code_unsupported_os_ver = True
            break
        # There is no need to wait for RMC active for IBM i
        if IBMI_OS_DISTRO == os_distro:
            if refcode == ref400:
                reset_task_state()
                return
        instance_existance = instance_obj.\
            Instance.get_by_uuid(context, instance['uuid'])
        # Check to make sure VM wasn't deleted
        if instance_existance:
            refcode, dlpar, rmc_state = get_lpar_activation_info()
        else:
            break

    else:
        reset_task_state()

    if ref_code_fail:
        # set to error state
        no_license_args = {'instance_name': inst_name}
        err_msg = exception.IBMPowerVMNoLicenseSRC(
            **no_license_args).message
        values = {'instance_uuid': instance['uuid'],
                  'code': 500,
                  'message': six.text_type(err_msg),
                  'details': six.text_type(err_msg),
                  'host': CONF.host}
        conductor.API().instance_fault_create(context, values)
        conductor.API().instance_update(context,
                                        instance['uuid'],
                                        vm_state=vm_states.ERROR,
                                        task_state=None)
        return

    if ref_code_unsupported_os_ver:
        # set to error state
        unsupported_os_ver_args = {'instance_name': inst_name}
        err_msg = exception.IBMPowerVMUnSupportedOSVersion(
            **unsupported_os_ver_args).message
        values = {'instance_uuid': instance['uuid'],
                  'code': 500,
                  'message': unicode(err_msg),
                  'details': unicode(err_msg),
                  'host': CONF.host}
        conductor.API().instance_fault_create(context, values)
        conductor.API().instance_update(context,
                                        instance['uuid'],
                                        vm_state=vm_states.ERROR,
                                        task_state=None)
        return
