# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import six

MANAGED_SYSTEM_TYPE = 'ManagedSystem'
RAW_METRICS_TYPE = 'RawMetrics'
LONG_TERM_MONITOR_TYPE = 'LongTermMonitor'
PCM_SERVICE = 'pcm'

from oslo_log import log as logging
from powervc_nova import _

LOG = logging.getLogger(__name__)


class CpuUtilProcessor(object):
    """
    This class is used to collect and calculate the CPU utilization
    of all LPARs (VMs) on a given CEC (host).
    """
    def __init__(self):
        pass

    def get_cpu_utilization_data(self, oper, cec_uuid, is_host=False):
        cpu_util_data = {}

        try:
            ltm_response = CpuUtilProcessor._get_managed_system_ltm(oper,
                                                                    cec_uuid)

            if ltm_response is None:
                return cpu_util_data

            phyp_links = []
            if ltm_response.status == 200:
                for entry in ltm_response.feed.entries:
                    category = entry.properties.get('category', '')

                    if category == 'phyp':
                        phyp_links.append(entry.properties.get('link', ''))
            else:
                LOG.info(_('Response status was not 200; instead, it was: %s.')
                         % str(ltm_response.status))
                return cpu_util_data

            # sort by time stamp
            phyp_links.sort()

            # We need at least two records to figure out the CPU utilization
            num_links = len(phyp_links)
            if num_links > 1:
                data_t0_link = phyp_links[num_links-2]
                data_t1_link = phyp_links[num_links-1]

                data_t0_json = json.loads(oper.readbyhref(data_t0_link).body)
                data_t1_json = json.loads(oper.readbyhref(data_t1_link).body)

                util_data = data_t0_json['systemUtil']['utilSample']
                # get total "activated" cores
                total_cores = util_data['processor']['configurableProcUnits']

                data_lpar_t0 = CpuUtilProcessor.\
                    _parse_ltm_raw_data(data_t0_json)
                data_lpar_t1 = CpuUtilProcessor.\
                    _parse_ltm_raw_data(data_t1_json)

                cpu_util_data_subResult = CpuUtilProcessor.\
                    _calculate_result_pair(data_lpar_t0, data_lpar_t1)
                cpu_util_data.update(cpu_util_data_subResult)

                # If host is passed as an argument, calculate util for VIOS
                # and system firmware.
                if is_host:
                    data_vios_t0 = CpuUtilProcessor.\
                        _parse_ltm_raw_data(data_t0_json, is_vios=True)
                    data_vios_t1 = CpuUtilProcessor.\
                        _parse_ltm_raw_data(data_t1_json, is_vios=True)
                    cpu_util_data_subResult = CpuUtilProcessor.\
                        _calculate_result_pair(data_vios_t0, data_vios_t1)
                    cpu_util_data.update(cpu_util_data_subResult)
                    cpu_util_data.update({'TotalCores': total_cores})
                    # TODO: calculate util for system firmware
            else:
                LOG.info(_('The number of PHYP entries is not > 1; '
                         'instead, it was: %s.') % str(num_links))
        except Exception:
            LOG.error(_('Error processing cpu_util data.'), exc_info=True)

        return cpu_util_data

    @staticmethod
    def _calculate_result_pair(data_lpar_t0, data_lpar_t1):
        cpu_util_data = {}
        for key, value_t0 in six.iteritems(data_lpar_t0):
            value_t1 = data_lpar_t1.get(key, None)
            # data for the LPAR must be at both collection points
            if value_t1 is not None:
                cpu_util = \
                    CpuUtilProcessor._calc_cpu_utilization(value_t0,
                                                           value_t1)
                # avoid negative values for CPU util
                if cpu_util < 0:
                    cpu_util = 0
                cpu_util_data[value_t0['uuid']] = float(cpu_util)
        return cpu_util_data

    @staticmethod
    def _get_managed_system_ltm(operator, cec_uuid):
        try:
            raw_data_response = operator.read(MANAGED_SYSTEM_TYPE,
                                              cec_uuid,
                                              RAW_METRICS_TYPE,
                                              LONG_TERM_MONITOR_TYPE,
                                              service=PCM_SERVICE)
            return raw_data_response
        except Exception:
            # Removing exc_info flag so that Logger does not print the
            # complete exception stack. For RTC defect 10175
            LOG.error(_("Unable to get managed system LTM property for "
                      "uuid: %s") % str(cec_uuid))
            return None

    @staticmethod
    def _parse_ltm_raw_data(data, is_vios=False):
        data_timestamp = data['systemUtil']['utilSample']['timeStamp']
        cycles = data['systemUtil']['utilSample']['timeBasedCycles']
        if is_vios:
            utils = data['systemUtil']['utilSample']['viosUtil']
        else:
            utils = data['systemUtil']['utilSample']['lparsUtil']
        results = {}
        for util in utils:
            result = {}
            result['timestamp'] = data_timestamp
            result['uuid'] = util['uuid']
            result['name'] = util['name']
            result['Cycles'] = cycles
            result['utilizedCappedProcCycles'] = \
                util['processor']['utilizedCappedProcCycles']
            result['utilizedUnCappedProcCycles'] = \
                util['processor']['utilizedUnCappedProcCycles']
            result['idleProcCycles'] = \
                util['processor']['idleProcCycles']

            results[result['uuid']] = result

        return results

    @staticmethod
    def _calc_cpu_utilization(data_t0, data_t1):
        cs_t0 = data_t0['Cycles']
        cs_t1 = data_t1['Cycles']

        entitled = float(cs_t1 - cs_t0)
        if entitled == 0:
            return 0

        cap_cs_t0 = data_t0['utilizedCappedProcCycles']
        uncap_cs_t0 = data_t0['utilizedUnCappedProcCycles']
        idle_cs_t0 = data_t0['idleProcCycles']

        cap_cs_t1 = data_t1['utilizedCappedProcCycles']
        uncap_cs_t1 = data_t1['utilizedUnCappedProcCycles']
        idle_cs_t1 = data_t1['idleProcCycles']

        # This formula is for both dedicated and shared CPUs
        #
        # For shared CPUs, the idle CPU is always 0
        #
        # For dedicated CPU, idle_cpu is reported, so that will
        # be greater than zero and we have to deduct that when
        # determining CPU Utilization.
        #
        # When the CPU mode is dedicated and share_idle_procs is set,
        # donated CPU could be greater than 0.  If it is set to
        # not shared, then it will be 0.
        cpu_cycles_used = float((cap_cs_t1 - cap_cs_t0) +
                                (uncap_cs_t1 - uncap_cs_t0) -
                                (idle_cs_t1 - idle_cs_t0))
        cpu_util = cpu_cycles_used / entitled
        return cpu_util
